// BudaDoc.cpp : implementation of the CBudaDoc class
//
/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "BudaMaster.h"
#include "Buda.h"
#include "BudaDoc.h"
#include "BudaView.h"
#include "MITableView.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "TimingDialog.h"
#include "DisplayOptions.h"
#include "MITableSetup.h"
#include "SaveFileOpsDlg.h"
#include "CalibrateDlg.h"
#include "SetupFFT.h"
#include "CSDetParmsDlg.h"
#include "Diagnostics.h"
#include "SineWaveDlg.h"
#include "FFTWorkerThread1.h"
#include "SelectOldSourceDlg.h"
#include "ChildFrm.h"
#include "ChangeMemo.h"
#include "AcquireScansDlg.h"
#include "AdjustDacInitDlg.h"
#include "DeconOptionsDlg.h"
#include "FilterDlg.h"
#include "ChefDlg.h"
#include "ChargeStateDetermination.h"
//#include "ExportDlg.h"
#include "DebugDialog.h"

// intel's digital signalling processing library
#define nsp_UsesAll
#include "nsp.h"

//#include "TestDialog.h"
#include <errno.h>

//#define DEBUG_ESTIMATENOISELEVEL_FLAG
//#define DEBUG_PEAKFITFLAG
//#define DEBUG_SERIALIZE_PS

#ifdef _DEBUG
	#define new DEBUG_NEW
	#undef THIS_FILE
	static char THIS_FILE[] = __FILE__;
#endif

// Buda's debug defines
#ifdef _DEBUG
	#define DEBUG_FREQTOMASS
	//#define DEBUG_VIEWRANGE
	//#define _CRTDBG_MAP_ALLOC
	//#include <CRTDBG.H>
	//#define CHKMEM _CrtCheckMemory()
#endif

/////////////////////////////////////////////////////////////////////////////
// CBudaDoc

IMPLEMENT_DYNCREATE(CBudaDoc, CDocument)

BEGIN_MESSAGE_MAP(CBudaDoc, CDocument)
	//{{AFX_MSG_MAP(CBudaDoc)
	ON_COMMAND(IDD_TIMINGINFO, OnTiminginfo)
	ON_COMMAND(ID_VIEW_TESTAPODIZATION, OnTestApodization)
	ON_COMMAND(ID_CHARGESTATES, OnChargeStates)
	ON_COMMAND(ID_CALIBRATE, OnCalibrate)
	ON_COMMAND(ID_PEAK_MATCH, OnPeakMatch)
	ON_UPDATE_COMMAND_UI(ID_PEAK_MATCH, OnUpdatePeakMatch)
	ON_UPDATE_COMMAND_UI(ID_CALIBRATE, OnUpdateCalibrate)
	ON_COMMAND(ID_FILE_SAVE_AS, OnFileSaveAs)
	ON_COMMAND(ID_FILE_SAVE, OnFileSave)
	ON_COMMAND(ID_DEBUG_TOOLS, OnDebugTools)
	ON_COMMAND(ID_DEBUG_SINE, OnDebugSine)
	ON_COMMAND(ID_READ_TEST_FFT_FILE, OnReadTestFftFile)
	ON_COMMAND(ID_DEBUG_MAKE_SINE_WAVE, OnDebugMakeSineWave)
	ON_COMMAND(ID_STOPFFTBTN, OnStopfftbtn)
	ON_UPDATE_COMMAND_UI(ID_STOPFFTBTN, OnUpdateStopfftbtn)
	ON_COMMAND(ID_RUN_EXP_BUTTON, OnRunExpButton)
	ON_UPDATE_COMMAND_UI(ID_RUN_EXP_BUTTON, OnUpdateRunExpButton)
	ON_COMMAND(ID_TUNE_MODE, OnTuneMode)
	ON_UPDATE_COMMAND_UI(ID_TUNE_MODE, OnUpdateTuneMode)
	ON_UPDATE_COMMAND_UI(ID_SETUP_BUTTON, OnUpdateSetupButton)
	ON_COMMAND(ID_TOOLS_CHANGEMEMO, OnChangeMemo)
	ON_COMMAND(ID_ADJUST_VOLTAGES, OnAdjustVoltages)
	ON_COMMAND(ID_TOOLS_CHEFCALCULATOR, OnToolsChefcalculator)
//	ON_COMMAND(ID_FILE_EXPORT, OnFileExport)
	ON_MESSAGE(WM_NEWDATAREADY, OnExperimentDone)
	ON_COMMAND(ID_SETUP_OPTIONS, OnSetupOptions)
	ON_COMMAND(ID_DEBUG_DIALOG_MASSTOFREQ, OnDebugDialogMasstofreq)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CBudaDoc construction/destruction

CBudaDoc::CBudaDoc()
{
	// initiallize all data pointers = 0.
	asserted = FALSE;
	pSetupView = 0;
	FreqArray = 0;
	TimeArray = 0;
	FITable = 0;
	StatusBar = 0;
	pApp = 0;
	pWnd = 0;
	Instrument = 0;
	sys = 0;
	SetupViewOpen = FALSE;
	CalculatingFFT = FALSE;
	FFTIsDirty = FALSE;
	CurrentSpectrumFocus = 0;

	// define default values for CBudaDoc Variables
	pApp = (CBudaApp*)AfxGetApp();
	pFFT = pApp->pFFT;
	pWnd = (CMainFrame*)AfxGetApp()->m_pMainWnd;
	DumpToDiagFile("Creating new CBudaDoc");

	Instrument = GetInstrument();

	EnableStopFFTBtn = FALSE;

	InitializeParameters();
	sys = GetSys();
}

CBudaDoc::~CBudaDoc()
{
}

BOOL CBudaDoc::OnNewDocument()
{
	DumpToDiagFile("entering CBudaDoc::OnNewDocument");
	if (!CDocument::OnNewDocument())
		return FALSE;

	CString DefaultLoadSpectrum = sys[0].DefaultSetupFile;
	if (!OnOpenDocument(DefaultLoadSpectrum))
		return FALSE;

	SetTitle(DefaultLoadSpectrum);
	DumpToDiagFile("leaving CBudaDoc::OnNewDocument   -  returned true");
	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////
// CBudaDoc serialization

void CBudaDoc::Serialize(CArchive& ar)
{
	DumpToDiagFile("entering CBudaDoc::Serialize");

	long MajorVersion, MinorVersion, FileType, inpts=0; 
	CFile *arf;
	pFFT->Lock();
	FileWasRead = TRUE;

	arf = ar.GetFile();
	if (ar.IsStoring())  
	{  // this file ONLY stores Version 8 files.
		StoreVersion8File(arf,0);
	} else {
		// first read in version info
		inpts+=arf->Read(&FileType,4);
		inpts+=arf->Read(&MajorVersion,4);
		inpts+=arf->Read(&MinorVersion,4);

		if(inpts<12) FileError("Error Reading Version Number");
		// then reset file to point to beginning of file
		arf->SeekToBegin();
		
		// then read in the particular version
		//if(FileType == SETUP_FILE) {
		//	MessageBox("This file is a setup file.  You cannot load it here.\nGo to 'Setup|Load Setup File...' to open this file","Error");
		//}

		switch(MajorVersion) {
		case 6:
			ReadVersion6File(arf,0);
			break;
			// end of loading version 6
		case 7:
			ReadVersion6File(arf,0);
			break;
			// end of loading version 6
		case 8:
			ReadVersion8File(arf,0);
			break;
			// end of loading version 8
		default:
			if(AfxMessageBox("Is this a Bruker Data File?",MB_YESNO,0) == IDYES)
				ReadBrukerFIDFile(arf,0);
			else
				FileError("Cannot Recognize File");
		}

		if(!FileWasRead) {
			AfxThrowArchiveException(CArchiveException::generic,NULL);
		}

		// if OLD file and hiresesi and hiresmaldi not set, ask for them
		if(!PS.HiResESI && !PS.HiResMALDI && PS.MajorVersion<=6 && PS.MinorVersion<50) {
			CSelectOldSourceDlg dlg;
			dlg.DoModal();
			if(dlg.m_SourceType == 1) PS.HiResMALDI = 1;
			if(dlg.m_SourceType == 2) PS.HiResESI = 1;
		}

		sys = GetSys(); 
		TDisDirty = TRUE;
		PSisDirty = TRUE;
	} // end of if storing/loading
	//arf->Close();

	pFFT->Unlock();

	ResetMenus();
	UpdateSystemPS();

	DumpToDiagFile("leaving CBudaDoc::Serialize");
}

/////////////////////////////////////////////////////////////////////////////
// CBudaDoc diagnostics

#ifdef _DEBUG

void CBudaDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CBudaDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}

//void CBudaDoc::OnDebugTestdialog() 
//{
//	CTestDialog dlg;
//	dlg.DoModal();
//}

void CBudaDoc::OnDebugWritecalfile() 
{
	DumpToDiagFile("entering CBudaDoc::OnDebugWritecalfile");

	CFile calfile;

	char* pathname;
	pathname = new char[256];
	strcpy(pathname,pApp->ProgramDirectory);
	strcat(pathname,"MALDI.cal");

	calfile.Open(pathname,  CFile::modeWrite | CFile::modeCreate, NULL);

	CArchive ar(&calfile,CArchive::store,4096,NULL);

	long calhdr = 0xcafd;
	ar.Write(&calhdr,4);
	long calvers = 10;
	double A = 71456456.45645;
	double B = -1.234567;
	double hms = 0;
	double trap = 13.33;
	double ints = 1.0;
	long calpts = 1;
	char caldt[40] = "13-FEB-1998 12:34:56                   ";
	char caldes[64]= "preliminary test cal file for win cal version 10               ";


	ar.Write(&calvers,4);
	ar.Write(&A,8);
	ar.Write(&B,8);
	ar.Write(&hms,8);
	ar.Write(&trap,8);
	ar.Write(&ints,8);
	ar.Write(&calpts,4);
	ar.Write(&caldt,40);
	ar.Write(&caldes,64);

	double m=1000.000;
	double f=65000.000;
	float v = 0.5;
	float amp = 100.0;
	long Incl = 0;
	char date[40] = "13-FEB-1998 12:34:56                   ";

	ar.Write(&m,8);
	ar.Write(&f,8);
	ar.Write(&v,4);
	ar.Write(&amp,4);
	ar.Write(&Incl,4);
	ar.Write(date,40);

	ar.Close();
	delete pathname;

	DumpToDiagFile("leaving CBudaDoc::OnDebugWritecalfile");
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CBudaDoc commands
/*
void CBudaDoc::FFT()
{
	DumpToDiagFile("entering CBudaDoc::FFT");

	// this function performs the forward direction (time -> freq) 
	// FFT on a transient by creating a frequency array, copying the
	// time domain data into it, and calling the Numerical Recipes
	// in place real FFT code.  It then calculates the magnitude mode
	// intensities
	// this function also times the FFT and stores its duration in 
	// the class's 'duration' variable
	long s;
	long i;

	if(TestForBigFFTs()) return;

	HaltFFTFlag = FALSE;
	EnableStopFFTBtn = TRUE;
	// turn on the hourglass cursor
	pApp->DoWaitCursor(1);

	UpdateAllViews(NULL);

	// main FFT loop, one fft for each spectrum.
	for(s=0; s<PS.SpectraAcquired; s++) {	
			// set the transient to the correct one
		SetCurrentSpectrumFocus(s);

			// do the FFT
		GetCurrentFreqArray(s);
	} 
	
	if(HaltFFTFlag) {
		IsPeakPicked=FALSE;
		SetIsFFTed(FALSE);
		for(i=0; i<s; i++) {
			//delete the MITables and such up to this point
			delete [] DS[i].FITable;
			DS[i].TotalIonInt=0;
			PS.PeakCount[i]=0;
			DS[i].NoiseLevel=0;
		}
	} else {
		IsPeakPicked=TRUE;
		IsFFTed = TRUE;
	}
	EnableStopFFTBtn = FALSE;
	ResetMenus();

	DumpToDiagFile("leaving CBudaDoc::FFT");
}
*/

void CBudaDoc::OnTiminginfo() 
{
	DumpToDiagFile("entering CBudaDoc::OnTiminginfo");

	CTimingDialog dlg;
	// initialize dialog data
	dlg.numpts = PS.FFTSize;
	dlg.duration = duration;
	dlg.DoModal();

	DumpToDiagFile("leaving CBudaDoc::OnTiminginfo");
}

void CBudaDoc::NullTerminate(char *mystring, int length)
{
	DumpToDiagFile("entering CBudaDoc::NullTerminate(..)");

	int i;

	for(i=length-1; i>0; i--) {
		if(mystring[i]==' ') mystring[i]='\0';
		else i=0;
	}	

	DumpToDiagFile("leaving CBudaDoc::NullTerminate(..)");
}

/*
void CBudaDoc::Fourier(double *data, long nn, int isign)
{
	DumpToDiagFile("entering CBudaDoc::Fourier(..)");

	long n,mmax,m,j,istep,i;
	double wtemp,wr,wpr,wpi,wi,theta;
	double tempr,tempi;

	n=nn << 1;
	j=1;
	for (i=1;i<n;i+=2) { // this is the bit rev section
		if (j > i) {
			tempr=data[j];
			data[j]=data[i];
			data[i]=tempr;

			tempr=data[j+1];
			data[j+1]=data[i+1];
			data[i+1]=tempr;
		}
		m=n >> 1;
		while (m >= 2 && j > m) {
			j -= m;
			m >>= 1;
		}
		j += m;
	}
	mmax=2;
	while (n > mmax) {	// outer log2(nn) loop
		istep=mmax << 1;
		theta=isign*(6.28318530717959/mmax);
		wtemp=sin(0.5*theta);
		wpr = -2.0*wtemp*wtemp;
		wpi=sin(theta);
		wr=1.0;
		wi=0.0;
		for (m=1;m<mmax;m+=2) {  // the 2 nested inner loops
			for (i=m;i<=n;i+=istep) {
				j=i+mmax;
				tempr=wr*data[j]-wi*data[j+1];
				tempi=wr*data[j+1]+wi*data[j];
				data[j]=data[i]-tempr;
				data[j+1]=data[i+1]-tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
		}
		mmax=istep;
	}	

	DumpToDiagFile("leaving CBudaDoc::Fourier(..)");

}
*/

/*
void CBudaDoc::RealFFT(double *data, long nn, int isign)
{
	DumpToDiagFile("entering CBudaDoc::RealFFT(..)");

	// this function calls the Fourier function and then de-scrambles
	// (or the reverse for IFFT's) the data to do the in place real
	// fft.  This function is a slightly modified version of Numerical
	// Recipe's realft(...) function pg 417 of Numerical Recipes in C
	long i, i1, i2, i3, i4, n2p3;
	double c1=0.5, c2, h1r, h1i, h2r, h2i;
	double wr, wi, wpr, wpi, wtemp, theta;
	long n;

	n=nn/2;
	theta = 3.141592653589793/(double) n;
	if (isign == 1) {
		c2 = -0.5;
		Fourier(data, n, 1);  // foward direction fft
	
	} else {
		c2=0.5;
		theta = -theta;
	}

	wtemp = sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi = sin(theta);
	wr = 1.0+wpr;
	wi = wpi;
	n2p3 = 2*n+3;

	// de-scrambling (or scrambling) section
	for (i=2;i<=n/2;i++) {
		i4=1+(i3=n2p3-(i2=1+(i1=i+i-1)));
		//ASSERT(i1>0 && i1<=nn);
		//ASSERT(i2>0 && i2<=nn);
		//ASSERT(i3>0 && i3<=nn);
		//ASSERT(i4>0 && i4<=nn);
		h1r=c1*(data[i1]+data[i3]);
		h1i=c1*(data[i2]-data[i4]);
		h2r=-c2*(data[i2]+data[i4]);
		h2i=c2*(data[i1]-data[i3]);
		data[i1]=h1r+wr*h2r-wi*h2i;
		data[i2]=h1i+wr*h2i+wi*h2r;
		data[i3]=h1r-wr*h2r+wi*h2i;
		data[i4]=-h1i+wr*h2i+wi*h2r;
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	if (isign==1) {
		data[1]=(h1r=data[1])+data[2];
		data[2]=h1r-data[2];
	} else {
		data[1]=c1*((h1r=data[1])+data[2]);
		data[2]=c1*(h1r-data[2]);
		Fourier(data,n,-1);	// inverse FFT
	}
	DumpToDiagFile("leaving CBudaDoc::RealFFT(..)");

}
*/

/*
void CBudaDoc::Magnitude()
{
//	DumpToDiagFile("entering CBudaDoc::Magnitude()");

	long i,n;
	//double* magn;
	//double* data;
	double *swap;
	double scale;
	n=PS.SpectrumPoints*2;

	//data=FreqArray;
	//magn=new double [n/2];
	scale=4.0/n;

	for (i=0; i<n; i+=2) {
		temp[i/2]=sqrt(FreqArray[i]*FreqArray[i]+FreqArray[i+1]*FreqArray[i+1])*scale;
		ScaleFactor = ( ScaleFactor < temp[i/2] ) ? temp[i/2] : ScaleFactor;
	}
	//delete [] FreqArray;
		//delete data;
	ScaleFactor=100/ScaleFactor;  // Scalefactor = 100/max point
	swap = temp;
	temp = FreqArray;
	FreqArray=swap;

//	DumpToDiagFile("leaving CBudaDoc::Magnitude()");
}
*/
/*
void CBudaDoc::Zerofill(int numzerofills, long n)
{
//	DumpToDiagFile("entering CBudaDoc::Zerofill(..)");

	numzerofills = numzerofills; // just so I don't get an "unreferenced formal parameter" warning
	long i;
	int j=1;
	double *swap;

	//for (i=numzerofills; i>0; i--) j=j*2;
	//PS.FFTSize = j*n;
	//temp=new double [PS.FFTSize+2];
	for(i=0;i<PS.FFTSize+2;i++) {
		if (i<n) temp[i]=FreqArray[i];
		else temp[i]=0.0;
	}

	swap = FreqArray;
	//delete [] FreqArray;
	FreqArray=temp;
	temp = swap;
//	DumpToDiagFile("leaving CBudaDoc::Zerofill(..)");
}
*/

double CBudaDoc::GetTrapVoltsOnDetect()
{
	// we need to search the GenPS structure in PS to find the
	// voltage that is at the same time as the detect pulse.  
	// the detectevent index in GenPS is defined by PS.DetectEvent
	// but then we need to search the trap voltages for the volts
	// at the same time.

	float start;	// the times of the start and 
	float finish;	// finish of the detect event
	float DACvolts;

	GetEventEdges(PS.DetectEvent, &start, &finish);
	//
	// now, return the DacVoltage on Trap2 at the start time
	DACvolts = DacValueAtTime(DAC_TRAP2, start);
	return DACvolts;
}

double CBudaDoc::CalculateIntShift()
{
	/*double IntShift;
	if (PS.ScansPerSpectrum<1 || PS.ADCGain<1) PS.IntensityShift=0;
	else
		IntShift=PS.IntensityShift*TrapVoltsOnDetect*TotalIonInt/double(PS.ScansPerSpectrum)/double(PS.ADCGain);
	return IntShift;
	*/
	return PS.IntensityShift;
}

/*
void CBudaDoc::Apodize(long n)
{
	long i;
	double factor;

	switch (PS.WindowFunction) {
	case 0:
		// no apod (or rectangular window function)
		break;
	case 1:
		// full Hanning apod
		// TimeArray[i] *= (1.0 - cos(2*pi*i/numpts))/2
		for(i=0;i<n/2;i++) {
			factor=(1.0-cos(2.0*pi*i/n))/2;
			FreqArray[i]=FreqArray[i]*factor;
			if (i>0) FreqArray[n-i]=FreqArray[n-i]*factor;
		}
		break; // end case 1
	case 2:
		// partial Hanning apod
		// Hanning apod as above but just on first and last
		// 25% of spectrum
		for(i=0;i<n/4;i++) {
			factor=(1.0-cos(pi*i/n))/2;
			FreqArray[i]=FreqArray[i]*factor;
			FreqArray[n-i]=FreqArray[n-i]*factor;
		} // end for
		break;

	
	} // end switch

	//DumpToDiagFile("leaving CBudaDoc::Apodize(..)");
} // end Apodize()
*/

void CBudaDoc::OnTestApodization() 
{
	//Apodize(TimeArray);	
}

/*
void CBudaDoc::PeakPick()
{
//	DumpToDiagFile("entering CBudaDoc::PeakPick()");

	int ptr = 0;			// peakstack pointer
	long i;		// cycle pointer to loop through FreqArray
	BOOL Upside = TRUE;		// flag if on side of increasing slope
	BOOL FitPeak = FALSE;	// flag if ready to fit the peak
	BOOL SplitPeak = FALSE; // flag if found split peak
	long minpt,maxpt;
	double PeakStack[100];		// intensity values of peaks
	double PeakXStack[100];		// frequency values of peaks

	// define a default FITable
	//FITable = new struct FITableElement[MAXNUMPEAKS];

	minpt=MassToPt(PS.DetectHighMass);
	maxpt=MassToPt(PS.DetectLowMass);
	NumPeaks = 0;
	// peak requirements:
	//		must have 2 or more points above noise
	//		must have "smooth" surface where smooth is
	//			defined as:
	//				before summit, no negative deviations 
	//				in intensity greater than noise
	//				after summit, no positive deviations 
	//				in intensity greater than noise
	//			
	//
	// pseudocode:
	// double buff[100];
	// ptr=0;
	// noise = EstimateNoiseLevel()
	//
	// for(i=0; i<numpts; i++)
	//	 if (freqarray(i)>noise)
	//		buff[ptr]=freqarray[i]; ptr++;
	//		if ptr>1
	//			if upside
	//				if split peaks 
	//					remove last point
	//					peakfit(buff, ptr)
	//					ptr=0
	//					buff[ptr]=freqarray[i]; ptr++;
	//				else, check for summit
	//					set downside
	//				endif 
	//			else downside
	//				if split peaks
	//					as above
	//				endif
	//			endif
	//		else
	//			continue to next i // only one point so far
	//		endif
	//	else
	//		if ptr>1
	//			remove last point
	//			peakfit(buff, ptr)
	//			ptr=0
	//			buff[ptr]=freqarray[i]; ptr++;
	//		else
	//			continue to next i // noise
	//	endif
	//endfor

	ptr = -1;
	for (i=minpt;i<maxpt;i++) {
		if (FreqArray[i]>NoiseLevel) {
			ASSERT(ptr<100);
			if(ptr<0) { 
				//PushPeakStack(i-1);
				ptr++;
				PeakStack[ptr]=FreqArray[i-1];
				PeakXStack[ptr]=PtToFreq(i-1);
			}
			// PushPeakStack(i);
			if (ptr<98) {
				ptr++;
				PeakStack[ptr]=FreqArray[i];
				PeakXStack[ptr]=PtToFreq(i);
			}
			// in the next two if's ignore all peaks that are not 
			// significant slope changes
			if(Upside && (PeakStack[ptr-1]>PeakStack[ptr]) && (FreqArray[i+1]<PeakStack[ptr])) {
				// two points in a row with a negative slope
				// so we are over the summit
				Upside = FALSE;
			}
			if(Upside && (PeakStack[ptr-1]-PeakStack[ptr]>3*NoiseLevel) && (FreqArray[i+1]>PeakStack[ptr])) {
				// significant slope change and split peak
				// split peak
				FitPeak=TRUE;
				SplitPeak=TRUE;
				ptr--;  // PopPeakStack();
			}
			if(!Upside && (PeakStack[ptr]-PeakStack[ptr-1]>3*NoiseLevel) && (FreqArray[i+1]<PeakStack[ptr])) {
				// significant slope change and split peak
				Upside=TRUE;  // going back up the side
				FitPeak=TRUE;
				SplitPeak=TRUE;
				ptr--;  // PopPeakStack();
			}
		} else {// else if point<noise
			if (ptr>0) { // have the whole peak now 
				ptr++;
				PeakStack[ptr]=FreqArray[i+1];
				PeakXStack[ptr]=PtToFreq(i+1);
				FitPeak = TRUE;
			}
		}
		ASSERT(i<PS.SpectrumPoints);
		if (FitPeak) {
			PeakFit(PeakStack, PeakXStack, ptr);
			ptr=-1; // ClearPeakStack();
			FitPeak=FALSE;
		}
		if (SplitPeak) {
			// Push first point back onto stack
			ptr++;
			PeakStack[ptr]=FreqArray[i];
			PeakXStack[ptr]=PtToFreq(i);
			SplitPeak=FALSE;
		}
	} //end for loop
	IsMIed = TRUE;
//	DumpToDiagFile("leaving CBudaDoc::PeakPick()");

} // end PeakPick()
*/

/*
void CBudaDoc::GetTotalIonInt() 
{
//	DumpToDiagFile("entering CBudaDoc::GetTotalIonInt(..)");
	long i;

	if(!IsMIed) {
		return;
	}

	TotalIonInt = 0;
	for(i=0; i<NumPeaks; i++) {
		if(sys->ExcludeNoResPeaks && FITable[i].RP>0) TotalIonInt += FITable[i].I;
		if(!sys->ExcludeNoResPeaks) TotalIonInt += FITable[i].I;
	}
//	DumpToDiagFile("leaving CBudaDoc::GetTotalIonInt(..)");
}
*/

/*
void CBudaDoc::PeakFit(double *PeakStack, double *PeakXStack, int numpts)
// PeakStack is the array of intensity values
// PeakXStack is the corresponding array of mass values
// numpts is the number of points in these arrays
{
//	DumpToDiagFile("entering CBudaDoc::PeakFit(..)");
	int i, PeakFitPts;
	double maxy = 0;
	double maxx=0, a[3], *x, *y;
	double deltafreq;
	double sumxy,sumy;

	BOOL PeakWasFit=FALSE;

#ifdef DEBUG_PEAKFITFLAG
	DumpToDiagFile("PeakFit(): ++++++++++++++++++++++++++++++++++++++\n");
#endif

	switch (PeakFitType) {
	case PARABOLIC_LEAST_SQUARES:
			a[0]=a[1]=a[2]=0.0;  // initialize parabolic fit constants
				// we want to do the fit with x=the index value and afterwards
				// calulate the matching frequency by interpolation as this 
				// decreases the roundoff errors in the least squares
				//
				// also, we want to set PeakWasFit = FALSE if one of the following
				// conditions is true:
				//		1) if <3 points above noise threshold
				//		2) if <3 points above the half height of the peak

			// first get new y and x arrays that only include the good points
			// good points are those above half height of peak and those above
			// threshold.
			x=new double [numpts];
			y=new double [numpts];
			maxy = 0;
			for(i=0; i<numpts; i++) {
				if (PeakStack[i]>maxy) maxy=PeakStack[i];
			}
			PeakFitPts = 0;
			for(i=0; i<numpts; i++) {
				if (PeakStack[i]>(maxy*PEAK_FIT_THRESHOLD)) {
					y[PeakFitPts]=PeakStack[i];
					x[PeakFitPts]=i;
					PeakFitPts++;
				}
			}

			#ifdef DEBUG_PEAKFITFLAG
			for(i=0; i<numpts; i++) {
				CString str;
				str.Format("PeakFit(): peak point %d int = %f freq = %f\n",i,PeakStack[i],PeakXStack[i]);
				DumpToDiagFile(str);
			}
			#endif

			PeakWasFit=ParabolicLeastSquares(y, x, PeakFitPts, a);
			if (PeakWasFit) {
				// calculate the center pt and intensity and verify that
				//		1) x[center pt] is in x[0]-x[numpt-1] range
				//		2) intensity is not > 2* maxintensity 
				maxx=(-1*a[1])/(2*a[0]);
				maxy=a[0]*maxx*maxx + a[1]*maxx + a[2];
				//if (maxy>2*maxpeak) PeakWasFit=FALSE;
				deltafreq=PeakXStack[1]-PeakXStack[0];
				maxx=PeakXStack[0]+deltafreq*maxx;
				// if maxx outside range of input points, not a good fit
				if (maxx>PeakXStack[numpts] || maxx<PeakXStack[0]) PeakWasFit = FALSE;
			#ifdef DEBUG_PEAKFITFLAG
				CString str;
				str.Format("peakfit constants a=%f, b=%f, c=%f\n", a[0], a[1], a[2]);
				DumpToDiagFile(str);
				str.Format("peakfit point maxx = %f and maxy = %f\n",maxx,maxy);
				DumpToDiagFile(str);
			#endif
			}
			delete [] x;
			delete [] y;
		break;
	case MAXIMUM_PT:
		for (i=0; i<numpts; i++) {
			if (maxy<PeakStack[i]) {
				maxy = PeakStack[i];
				maxx = PeakXStack[i];
			}
		}
		PeakWasFit=TRUE;
		break;
	case WEIGHTED_AVERAGE:
		// x is the intensity weighted average of the x's
		// y is the max intensity
		sumxy=0.0;
		sumy=0.0;
		maxy=0;
		for (i=0; i<numpts; i++) {
			sumxy=sumxy+PeakStack[i]*PeakXStack[i];
			sumy=sumy+PeakStack[i];
			maxy = (maxy<PeakStack[i]) ? PeakStack[i] : maxy;
		}
		maxx=sumxy/sumy;
		PeakWasFit=TRUE;
		break;
	}

	double RP;
	if (PeakWasFit) {
		if (NumPeaks<MAXNUMPEAKS) {
			FITable[NumPeaks].F=maxx;
			FITable[NumPeaks].I=maxy;
			RP = InterpolateForRes(PeakStack,PeakXStack,numpts);
			FITable[NumPeaks].RP = RP;
			FITable[NumPeaks].CS = 1;
			NumPeaks++;
		} else {
		// error message...
		}
	} else {
		// peak could not be fit using the current method
	}

//	DumpToDiagFile("leaving CBudaDoc::PeakFit(..)");
}//end PeakFit
*/

/*
void CBudaDoc::EstimateNoiseLevel(double *data, double *noise)
{
//	DumpToDiagFile("entering CBudaDoc::EstimateNoiseLevel(..)");
	long i,j;
	double sum=0.0;
	long highpt,lowpt,numpts;
	double scale;
	scale = 0;

	lowpt=MassToPt(PS.DetectHighMass);
	highpt=MassToPt(PS.DetectLowMass);
	if (highpt<lowpt) { // swap them
		numpts = highpt; // use numpts as a temp variable
		highpt = lowpt;
		lowpt = numpts;
	}

	if ((highpt-lowpt)<10) {
		*noise=0.0;
		return;
	}
	for (i=lowpt; i<highpt; i++) {
		sum+=data[i];
		if(data[i]>scale) scale = data[i]; // get highest point for scale
	}
	numpts=highpt-lowpt;
	*noise=10*sum/numpts; // first estimate is 10* average

	for (j=0; j<NUM_NOISE_LEVEL_ITERATION; j++) {
		sum=0.0;
		for (i=lowpt; i<highpt; i++) {
			if(data[i]<*noise) {
				sum+=data[i]*data[i];
			} else { 
				sum+=(*noise)*(*noise)*NOISE_LEVEL_OVERCORRECTION_FACTOR;
			}
		}
		*noise=NOISE_LEVEL_INCLUDE_FACTOR*sqrt(sum/numpts); // include all pts < 2*rms
#ifdef DEBUG_ESTIMATENOISELEVEL_FLAG
		CString str;
		str.Format("EstimateNoiseLevel(): iter %d, *noise %f\n",j,NoiseLevel);
		DumpToDiagFile(str);
#endif
	}
	CurrentScaleFactor = 100.0/scale;
	if (*noise<0.03*scale) *noise = 0.03*scale;
	NoiseLevel = *noise;

//	DumpToDiagFile("leaving CBudaDoc::EstimateNoiseLevel(..)");
}
*/

double CBudaDoc::PtToFreq(long pt)
{
	//DumpToDiagFile("entering CBudaDoc::PtToFreq(..)");
	double freq;
	double deltaf;
	
	deltaf = PS.AdcRate/(2.0*(double)PS.SpectrumPoints);

	freq = deltaf*(double)pt;

	if (PS.DetectMode == MixerMode) freq = PS.MixerFrequency - freq;

	//DumpToDiagFile("leaving CBudaDoc::PtToFreq(..)");
	return freq;
}

double CBudaDoc::FreqToMass(double freq, double A, double B)
{
	//DumpToDiagFile("entering CBudaDoc::FreqToMass(..)");

	double locA,locB;
	double mass, f1, f2, m1, m2;

	freq = CyclotronFrequency(freq,1);
	if (freq<=0) freq=1;

	if (PS.PeakMatching) {
		// calibration in PeakMatch Mode
		if (PS.RefPeak[1].ActualMass < 1.0) {
			// only one peak to match against
			f1 = CyclotronFrequency(PS.RefPeak[0].MeasuredFrequency,PS.RefPeak[0].TrapVoltage,PS.RefPeak[0].TotalIntensity,1,1,1);

			f2 = A + PS.RefPeak[0].ActualMass*(freq-f1);

			if (f2 > 0.0) mass = PS.RefPeak[0].ActualMass*A/f2;
			else mass = 99999.0;
		} else {
			// two peakmatch peaks
			m1=PS.RefPeak[0].ActualMass;
			m2=PS.RefPeak[1].ActualMass;
			f1 = CyclotronFrequency(PS.RefPeak[0].MeasuredFrequency,PS.RefPeak[0].TrapVoltage,PS.RefPeak[0].TotalIntensity,1,1,1);
			f2 = CyclotronFrequency(PS.RefPeak[1].MeasuredFrequency,PS.RefPeak[1].TrapVoltage,PS.RefPeak[1].TotalIntensity,1,1,1);
			locA = (f1-f2)*(m1*m2/(m2-m1));
			locB = f1 - locA/m1;

			if ((freq - locB) >= 500.0) 
				mass = locA/(freq - locB);
			else
				mass = locA/500.0;
		}
	} else {
		// normal calibration - not peakmatch mode
		//frequency corrections
	
		if ((freq-B) > 500.0) mass = A / (freq - B);
		else mass =  A / 500.0;
	}
	//DumpToDiagFile("leaving CBudaDoc::FreqToMass(..)");

#ifdef DEBUG_FREQTOMASS
	if(mass>601.92 && mass <602) {
		CString str,crlf = "\r\n";
		TRACE("======CBudaDoc::FreqToMass====="+crlf);
		str.Format("freq : %1.3f",freq);
		TRACE(str+crlf);
		str.Format("A,B : %1.3f,%1.3f",A,B);
		TRACE(str+crlf);
		str.Format("mass : %1.3f",mass);
		TRACE(str+crlf);
	}
#endif
	
	return mass;
}

double CBudaDoc::MassToFreq(double mass)
{
	return MassToFreq(mass, PS.A, PS.B);
}

double CBudaDoc::MassToFreq(double mass, double A, double B)
{
	//DumpToDiagFile("entering CBudaDoc::MassToFreq(..)");
	double freq=0;

	// debug test to see if this mass is reasonable.
	#ifdef _DEBUG
	if (mass < 100 || mass >10000) {
		TRACE("CBudaDoc::MassToFreq : Warning mass outside normal range\n	Should you be using FreqToMass instead?\n");
	}
	#endif

	if(PS.PeakMatching) {
		double f0,m0,f1,m1,locA,locB;
		f0=f1=m0=m1=locA=locB=0;

		/* derivation:
		
		fundamental equation : f = A/(m-B)	
							or m = A/f + B	(eqn. 0)
		f0 = A / (m0 - B)					(eqn. 1)
		f1 = A / (m1 - B)					(eqn. 2)

		solve for A and B 

		A = f1 * ( m1 - B)					(eqn. 2b)
		f0 = (f1 * ( m1 - B)/(m0 - B)		(eqn. 3)	//substitution in eqn 1

		only unknown in eqn 3 is B.  solve for it.
		f0*m0 - f0*B = f1*m1 - f1*B
		b * (f1 - f0) = f1*m1 - f0*m0
		b = (f1*m1 - f0*m0)/(f1 - f0)

		substitute back into eqn 2b to get A

		This works fine for two refpeaks, but if we only have one ref peak,
		then what?  We have to make an assumption.
		
		two possibilities; A = sys.A or B = sys.B

		*/
		m0 = PS.RefPeak[0].ActualMass;
		m1 = PS.RefPeak[1].ActualMass;
		if(m0 >= 1) f0 = CyclotronFrequency(PS.RefPeak[0].MeasuredFrequency,PS.RefPeak[0].TrapVoltage,PS.RefPeak[0].TotalIntensity,1,1,1);
		if(m1 >= 1) f1 = CyclotronFrequency(PS.RefPeak[1].MeasuredFrequency,PS.RefPeak[1].TrapVoltage,PS.RefPeak[1].TotalIntensity,1,1,1);

		// apply frequency corrections to f1 and f0
		if(m1 < 1 && m0 < 1) { // no ref peaks just use A and B
			locA = A;
			locB = B;
		} else if (m1 < 1) {  // only one refpeak
			// only one peak to match against
			// not needing f1, so use it as a temp variable
			f1 = (m0 - mass)/(mass*m0);
			freq = A * f1 + f0;
		} else {
			// two ref peaks
			// my derivation
			//locB = (f1*m1 - f0*m0)/(f1-f0);
			//locA = f1 * ( m1 - locB);
			// copied down from freqtomass - ricks derivation
			locA = (f0-f1)*(m1*m0/(m1-m0));
			locB = f0 - locA/m0;

			if(mass > locB) freq = locA / (mass - locB); 
		}
	} else {	// normal (not peakmatch) conversion
		if (mass > 1.0) freq = (A/mass) + B;
		else freq = A+B;
		// remove the frequency corrections
	}
	freq = CyclotronFrequency(freq, -1);

	if (freq<0) freq=0;
	if (PS.DetectMode == MixerMode && (PS.MixerFrequency-freq)<0)
		freq = PS.MixerFrequency;

#ifdef DEBUG_FREQTOMASS
	if(mass>601.92 && mass <602) {
		CString str,crlf = "\r\n";
		TRACE("======CBudaDoc::MassToFreq====="+crlf);
		str.Format("mass,A,B : %1.3f,%1.3f,%1.3f",mass,A,B);
		TRACE(str+crlf);
		str.Format("freq : %1.3f",freq);
		TRACE(str+crlf);
	}
#endif

	return freq;
}

long CBudaDoc::FreqToPt(double freq)
{
	//DumpToDiagFile("entering CBudaDoc::FreqToPt(..)");
	long pt;
	double dpt;
	double deltaf;
	
	deltaf = PS.AdcRate/(2.0*(double)PS.SpectrumPoints);

	dpt = freq/deltaf;
	pt=(long)dpt;
	if((dpt-(double)pt)>0.5) pt++;

	if (PS.DetectMode == MixerMode) {
		if(fabs(freq-PS.MixerFrequency)<PS.AdcRate/2) {
			// if mixer mode and freq is within nyquist of mixerfreq
			freq=PS.MixerFrequency-freq;
			dpt = freq/deltaf;
			pt = long(dpt);
			if((dpt-(double)pt)>0.5) pt++;
		} else {
			pt = 0;  // assign foldback frequencies to DC
		}
	}
	//DumpToDiagFile("leaving CBudaDoc::FreqToPt(..)");

	return pt;
}

long CBudaDoc::MassToPt(double mass)
{
	//DumpToDiagFile("entering CBudaDoc::MassToPt(..)");
	double freq;
	long pt;

	freq=MassToFreq(mass, PS.A, PS.B);
	pt=FreqToPt(freq);
	//DumpToDiagFile("leaving CBudaDoc::MassToPt(..)");
	return pt;
}

long CBudaDoc::TimeToPt(double time)
{
	return long(time*PS.AdcRate);
}

/*
BOOL CBudaDoc::ParabolicLeastSquares(double *y, double *x, int numpts, double *ret)
{  // this function performs a parabolic least squares fit on the x,y arrays each of length numpts
   // that are input.  ret[0], ret[1], ret[3] are the three fit constants

	DumpToDiagFile("entering CBudaDoc::ParabolicLeastSquares(..)");
	if (numpts<3) return FALSE; // can't parabolic least squares with less than 3 peaks

	int i;
	double sumx4, sumx3, sumx2, sumx1, sumx0; // Sum(x^4), Sum(x^3), ...
	double sumy2, sumy1, sumy0;				  // Sum(y*x^2), Sum(y*x), Sum(y)
	double a, b, c;		// parabola coefficients
	double dp;  // denominator in the determinants.

	sumx4=sumx3=sumx2=sumx1=sumx0=0.0;
	sumy2=sumy1=sumy0=0;

	for (i=0; i<numpts; i++) {
		sumx0++;
		sumx1=sumx1+x[i];
		sumx2=sumx2+x[i]*x[i];
		sumx3=sumx3+x[i]*x[i]*x[i];
		sumx4=sumx4+x[i]*x[i]*x[i]*x[i];
		sumy0=sumy0+y[i];
		sumy1=sumy1+y[i]*x[i];
		sumy2=sumy2+y[i]*x[i]*x[i];
	}

	#ifdef DEBUG_PEAKFITFLAG
	CString str;
	str.Format("PeakFit(): array constants: %f %f %f\n",sumx4, sumx3, sumx2);
	DumpToDiagFile(str);
	str.Format("PeakFit(): array constants: %f %f %f\n",sumx3, sumx2, sumx1);
	DumpToDiagFile(str);
	str.Format("PeakFit(): array constants: %f %f %f\n",sumx2, sumx1, sumx0);
	DumpToDiagFile(str);
	str.Format("PeakFit(): number of points in peak = %d\n",numpts);
	DumpToDiagFile(str);
	str.Format("PeakFit(): vector constants: %f %f %f\n",sumy2, sumy1, sumy0);
	DumpToDiagFile(str);
	#endif
	// now calculate a,b,c coefficients for the parabola
	dp = GetDeterminant(sumx4, sumx3, sumx2,
						sumx3, sumx2, sumx1,
						sumx2, sumx1, sumx0);
	if (dp == 0) 
	{
		DumpToDiagFile("leaving CBudaDoc::ParabolicLeastSquares(..) - FALSE");
		return FALSE;
	}
	a  = GetDeterminant(sumy2, sumx3, sumx2,
						sumy1, sumx2, sumx1,
						sumy0, sumx1, sumx0)/dp;
	b  = GetDeterminant(sumx4, sumy2, sumx2,
						sumx3, sumy1, sumx1,
						sumx2, sumy0, sumx0)/dp;
	c  = GetDeterminant(sumx4, sumx3, sumy2,
						sumx3, sumx2, sumy1,
						sumx2, sumx1, sumy0)/dp;

	// verify that a is negative so that found a peak not a valley 
	if (a>0) return FALSE;

	// if we get this far, the peak is good, so return it
	ret[0]=a;
	ret[1]=b;
	ret[2]=c;

	DumpToDiagFile("leaving CBudaDoc::ParabolicLeastSquares(..) - TRUE");
	return TRUE;
}
*/

/*
double CBudaDoc::GetDeterminant(double a1, double a2, double a3, double b1, double b2, double b3, double c1, double c2, double c3)
{
	return a1*b2*c3+a2*b3*c1+a3*b1*c2-a1*b3*c2-a2*b1*c3-a3*b2*c1;
}
*/

/*
double CBudaDoc::InterpolateForRes(double *y, double *x, int ptr)
{
	DumpToDiagFile("entering CBudaDoc::InterpolateForRes(..)");
	// find maximum point, then find two peaks closest to the FWHM points
	// on a model peak, interpolate to find the approximate x values of these
	// points and then calculate f/delta(f) and return it.
	int i,maxx=0;
	double maxy,b,m,retval;
	double fwhmx1, fwhmx2;
	int pt1,pt2,count;

	// find maximum point and count number of points above NoiseLevel
	maxy=0;
	count=0;
	for (i=0; i<ptr; i++) {
		if (y[i] > maxy) {
			maxy =  y[i];
			maxx=i;
		}
		if (y[i]>NoiseLevel) count++;
	}
	if (count<3) return -1;

	// now find left side two points straddling the half height
	pt1=pt2=0;
	for (i=0; i<ptr; i++) {
		pt1=i;
		pt2=i+1;
		if (y[pt1]>maxy/2) break;
	}
	if (pt2>ptr) {
		pt2=ptr;
		pt1=ptr-1;
	}
	// for y=mx+b, m=y2-y1/x2-x1, b=y2-mx2
	m=(y[pt2]-y[pt1])/(x[pt2]-x[pt1]);
	b=y[pt2]-m*x[pt2];
	// now, x=(y-b)/m;
	fwhmx1=(maxy/2-b)/m;
	// now find right side two points straddling the half height
	pt1=pt2=0;
	for (i=maxx; i<ptr ; i++) {
		pt1=i;
		pt2=i+1;
		if (y[pt2]<maxy/2) break;
	}
	if (pt2>ptr) {
		pt2=ptr;
		pt1=ptr-1;
	}
	// for y=mx+b, m=y2-y1/x2-x1, b=y2-mx2
	m=(y[pt2]-y[pt1])/(x[pt2]-x[pt1]);
	b=y[pt2]-m*x[pt2];
	// now, x=(y-b)/m;
	fwhmx2=(maxy/2-b)/m;

	retval= x[maxx]/(fwhmx2-fwhmx1);
	DumpToDiagFile("leaving CBudaDoc::InterpolateForRes(..)");
	return retval;
}
*/

double CBudaDoc::PtToMass(long pt)
{
	//DumpToDiagFile("entering CBudaDoc::PtToMass(..)");
	double freq, mass;

	//ASSERT(pt >=0 && pt<=PS.SpectrumPoints);
	if(!asserted) {
		if(pt < 0 || pt > PS.SpectrumPoints) {
			MessageBox("The Point-to-mass calculation has exceeded it's bounds,\n some data calculated now might be suspect.  \n Note: only the Charge State Determination routines are known to cause this bug so far.", "Warning!");
			asserted = TRUE;
			if(pt > PS.SpectrumPoints) pt = PS.SpectrumPoints;
			if(pt < 0) pt = 0;
		}
	}

	freq=PtToFreq(pt);
	mass=FreqToMass(freq, PS.A, PS.B);

	//DumpToDiagFile("leaving CBudaDoc::PtToMass(..)");
	return mass;
}

double CBudaDoc::PtToTime(long pt)
{
return pt/PS.AdcRate;
}

void CBudaDoc::OnChargeStates() 
{
//	if (AssignChargStates) ChargeStateDetermination();
}	

void CBudaDoc::SortFITableOnAbundance()
{
	DumpToDiagFile("entering CBudaDoc::SortFITableOnAbundance()");
	int i,j;
	FITableElement temp;
	for(i=1; i<PS.Spec[CurrentSpectrumFocus].NumPeaks; i++) {
		if (SortIncreasing) {
			if (FITable[i].I<FITable[i-1].I) {
				j=i;
				while(j>0 && FITable[j].I<FITable[j-1].I) {
					temp = FITable[j];
					FITable[j]=FITable[j-1];
					FITable[j-1]=temp;
					j--;
				}
			}
		} else {
			if (FITable[i].I>FITable[i-1].I) {
				j=i;
				while(j>0 && FITable[j].I>FITable[j-1].I) {
					temp = FITable[j];
					FITable[j]=FITable[j-1];
					FITable[j-1]=temp;
					j--;
				}
			}
		}
	}

	DumpToDiagFile("leaving CBudaDoc::SortFITableOnAbundance()");
}

void CBudaDoc::OnSetupOptions() 
{
	DumpToDiagFile("entering CBudaDoc::OnSetupOptions()");
	CPropertySheet props;
	CDisplayOptions	disp_ops;   // options page for display
	CSetupFFT setup_fft;		// options page for FFT
	CMITableSetup mi_ops;		// options page for MI Table
	//CSaveFileOpsDlg fileops;	// options page for save file
	CCSDetParmsDlg CSops;		// options page for charge state determination
	CDeconOptionsDlg DeconOps;	// options page for deconvolution settings
	// idea didn't work.  see fftworkerthread for details.
//	CFilterDlg Filter_Ops;		// options page for time domain filtering

	BOOL FFTOptionsChanged = FALSE;
	// apply current variables to FFT page
	setup_fft.m_apodfunc=PS.WindowFunction;
	setup_fft.m_IsPeakPicking = PeakPickFlag;
	if(sys->AutoLabelCharge) setup_fft.m_AutoCalcCS = TRUE;
	else setup_fft.m_AutoCalcCS = FALSE;
	setup_fft.m_NumZeroFills = PS.ZeroFills;
	setup_fft.DataDirty = FALSE;
	setup_fft.m_TruncatePoints = PS.TruncatePoints;
	setup_fft.m_TruncateOffset = PS.TruncateOffset;
	setup_fft.TruncateTransient = PS.TruncateTransient;
	setup_fft.PDoc = this;
	//setup_fft.m_IonIntensity = TotalIonInt;

	BOOL DisplayParmsChanged = FALSE;
	DisplayParmsChanged=DisplayParmsChanged;
	// apply current DisplayParms variables to Display Parms page
	disp_ops.pDoc = this;
	disp_ops.dp = DisplayParms;

	// apply current MITable parms to the MITable property page
	mi_ops.SortMode = SortOnProperty;
	mi_ops.SortIncreasing = SortIncreasing;
	//mi_ops.SICmode = SICmode;
	mi_ops.AutoDetectThreshold = pApp->AutoDetectThreshold;
//	mi_ops.ThresholdMethod = ThresholdMethod;
	sys->MinimumAutoThreshold = pApp->MinimumAutoThreshold;
	mi_ops.m_MinimumThreshold = sys->MinimumAutoThreshold;


	BOOL SaveFileOpsChanged=FALSE;
	SaveFileOpsChanged=SaveFileOpsChanged;

	// apply current save file parms to the CSaveFileOptionsDlg page
	/*fileops.SaveFileOpsChanged = FALSE;
	fileops.AutoSave = AutoSave;
	fileops.SaveMITable = PS.MITableIsStored;
	fileops.SaveSpectrum = PS.FrequencyIsStored;
	fileops.SaveTruncated = SaveTruncatedTransient;
	fileops.SaveTransient = PS.TransientIsStored;
	fileops.m_DefaultDir = DefaultSaveDirectory;
	fileops.m_DefFileNameType = DefaultFileNameType;
	*/

	// setup CSops property page
	CSops.pDoc=this;
	CSops.m_HighChargeLimit = CSHighChargeLimit;
	CSops.m_CSDetMethod = CSDetMethod;

	// setup CDeconOps property page
	DeconOps.pDoc = this;
	DeconOps.m_AdductMass = DeconAdductMass;
	DeconOps.m_DeconMethod = DeconMethod;
	DeconOps.m_HighMass = DeconHighMassLimit;
	DeconOps.m_LowMass = DeconLowMassLimit;
	DeconOps.m_MaxCharge = DeconHighChargeLimit;
	DeconOps.m_MinCharge = DeconLowChargeLimit;
	DeconOps.m_DeconRes = DeconResolution;

	// setup CFilterDlg property page
	// idea didn't work.  see fftworkerthread for details.
//	Filter_Ops.m_BoxcarAvg = BoxcarAvg;
//	Filter_Ops.m_BoxcarPoints = BoxcarPoints;

	// now load all the pages
	props.AddPage(&disp_ops);
	props.AddPage(&setup_fft);
	props.AddPage(&mi_ops);
	//props.AddPage(&fileops);
	props.AddPage(&CSops);
	props.AddPage(&DeconOps);
//	props.AddPage(&Filter_Ops);
	props.SetActivePage(SetupOptionsInitPage);

	if (!props.DoModal()) return;

	// retrieve current variables from FFT page
	if(PS.WindowFunction != setup_fft.m_apodfunc){
		PS.WindowFunction = setup_fft.m_apodfunc;
		sys->WindowFunction = PS.WindowFunction;
		FFTOptionsChanged=TRUE;
	}
	PeakPickFlag = setup_fft.m_IsPeakPicking;
	if(setup_fft.m_AutoCalcCS && sys->AutoLabelCharge == 0) {
		sys->AutoLabelCharge = 1;
		FFTOptionsChanged=TRUE;
	}
	else if (!setup_fft.m_AutoCalcCS && sys->AutoLabelCharge == 1) {
		sys->AutoLabelCharge = 0;
		FFTOptionsChanged=TRUE;
	}

	if(PS.ZeroFills != setup_fft.m_NumZeroFills) {
		PS.ZeroFills = setup_fft.m_NumZeroFills;
		sys->ZeroFills = PS.ZeroFills;
		FFTOptionsChanged=TRUE;
	}
	if(PS.TruncatePoints != setup_fft.m_TruncatePoints) {
		PS.TruncatePoints = setup_fft.m_TruncatePoints;
		FFTOptionsChanged=TRUE;
	}
	if(PS.TruncateOffset != setup_fft.m_TruncateOffset) {
		PS.TruncateOffset = setup_fft.m_TruncateOffset;
		FFTOptionsChanged=TRUE;
	}
	if(PS.TruncateTransient != setup_fft.TruncateTransient) {
		PS.TruncateTransient = setup_fft.TruncateTransient;
		FFTOptionsChanged=TRUE;
	}
	if (FFTOptionsChanged && IsFFTed()) {
		if(pWnd->MessageBox("Recalculate FFT?","FFT Options Changed",MB_YESNO | MB_ICONQUESTION) == IDYES) {
			// recalculate fft
			SetIsFFTed(FALSE);
			FFTIsDirty = TRUE;
			SetIsMIed(FALSE);
			delete [] FITable;
			PS.Spec[CurrentSpectrumFocus].FITable = 0;
			FITable = 0;
			RecalcAllViews();
		}
	}

	// retrieve current DisplayParms variables from Display Parms page
	DisplayParms = disp_ops.dp;

	/*DisplayParms.allparms = disp_ops.m_disp_parms;
	DisplayParms.massrange = disp_ops.m_disp_massrange;
	DisplayParms.acqrate = disp_ops.m_disp_acqrate;
	DisplayParms.apodfunc = disp_ops.m_disp_apodfunc;
	DisplayParms.calconst = disp_ops.m_disp_calconst;
	DisplayParms.exciteamp = disp_ops.m_disp_exciteamp;
	DisplayParms.filename = disp_ops.m_disp_filename;
	DisplayParms.notes = disp_ops.m_disp_notes;
	DisplayParms.numpts = disp_ops.m_disp_numpts;
	DisplayParms.numzerofills = disp_ops.m_disp_numzerofills;
	DisplayParms.timedate = disp_ops.m_disp_timedate;
	DisplayParms.trapvolts = disp_ops.m_disp_trapvolts;
	DisplayParms.labelpoints = disp_ops.m_disp_labelpoints;
	DisplayParms.numscans = disp_ops.m_disp_numscans;
	DisplayParms.LabelPeakPos = disp_ops.m_LabelPeakPos;
	DisplayParms.noisethreshold = disp_ops.m_disp_noisethreshold;
	DisplayParms.LabelCharge = disp_ops.m_disp_lablecharge;
	DisplayParms.timedate = disp_ops.m_disp_time;
	DisplayParms.PosNeg = disp_ops.m_disp_posneg;
	DisplayParms.ScaleFactor = disp_ops.m_DispScalefactor;
	*/

	// retrieve parms from the MITable property page
	SortOnProperty = mi_ops.SortMode;
	SortIncreasing = mi_ops.SortIncreasing;
	if(mi_ops.SortDirty) SortMITable();
			/// SIC vs TIC stuff
	//SICmode = mi_ops.SICmode;	
	//SICMass = mi_ops.m_SICMass;
	//SICWidth = mi_ops.m_SICMassWidth;
			/// manual entry of noise threshold stuff
	pApp->AutoDetectThreshold = mi_ops.AutoDetectThreshold;
//	ThresholdMethod = mi_ops.ThresholdMethod;
	sys->MinimumAutoThreshold = mi_ops.m_MinimumThreshold;
	pApp->MinimumAutoThreshold = sys->MinimumAutoThreshold;

	// retrieve fileops parms
	/*if(AutoSave != fileops.AutoSave) {
		AutoSave = fileops.AutoSave;
		SetModifiedFlag(TRUE);
	}
	if(PS.MITableIsStored != fileops.SaveMITable) {
		PS.MITableIsStored = fileops.SaveMITable;
		SetModifiedFlag(TRUE);
	}
	if(PS.FrequencyIsStored != fileops.SaveSpectrum) {
		PS.FrequencyIsStored = fileops.SaveSpectrum;
		SetModifiedFlag(TRUE);
	}
	if(SaveTruncatedTransient != fileops.SaveTruncated) {
		SaveTruncatedTransient = fileops.SaveTruncated;
		SetModifiedFlag(TRUE);
	}
	if(PS.TransientIsStored != fileops.SaveTransient) {
		PS.TransientIsStored = fileops.SaveTransient;
		SetModifiedFlag(TRUE);
	}
	if(DefaultSaveDirectory != fileops.m_DefaultDir) {
		DefaultSaveDirectory = fileops.m_DefaultDir;
		SetModifiedFlag(TRUE);
	}
	if(DefaultFileNameType != fileops.m_DefFileNameType) {
		DefaultFileNameType = fileops.m_DefFileNameType;
		SetModifiedFlag(TRUE);
	}
	if(fileops.SaveFileOpsChanged) SetModifiedFlag(TRUE);
	*/

	// retrieve CSops property page information
	CSHighChargeLimit = CSops.m_HighChargeLimit;
	CSDetMethod = CSops.m_CSDetMethod;
	pApp->CSHighChargeLimit = CSHighChargeLimit;
	sys->ChargeLimit = CSHighChargeLimit;
	pApp->CSDetMethod = CSDetMethod;

	// retrieve CDeconOps property page
	DeconAdductMass = DeconOps.m_AdductMass;
	DeconMethod = DeconOps.m_DeconMethod;
	DeconHighMassLimit = DeconOps.m_HighMass;
	DeconLowMassLimit = DeconOps.m_LowMass;
	DeconHighChargeLimit = DeconOps.m_MaxCharge;
	DeconLowChargeLimit = DeconOps.m_MinCharge;
	DeconResolution = DeconOps.m_DeconRes;
	pApp->DeconAdductMass = DeconAdductMass;
	pApp->DeconMethod = DeconMethod;
	pApp->DeconHighMassLimit = DeconHighMassLimit;
	pApp->DeconLowMassLimit = DeconLowMassLimit;
	pApp->DeconHighChargeLimit = DeconHighChargeLimit;
	pApp->DeconLowChargeLimit = DeconLowChargeLimit;
	pApp->DeconResolution = DeconResolution;

		// setup CFilterDlg property page
	// idea didn't work.  see fftworkerthread for details.
//	BoxcarAvg = Filter_Ops.m_BoxcarAvg;
//	BoxcarPoints = Filter_Ops.m_BoxcarPoints;
//	pApp->BoxcarAvg = BoxcarAvg;
//	pApp->BoxcarPoints = BoxcarPoints;

	UpdateAllViews(NULL,0,NULL);
	DumpToDiagFile("leaving CBudaDoc::OnSetupOptions()");
}


void CBudaDoc::SortFITableOnMass()
{
	DumpToDiagFile("entering CBudaDoc::SortFITableOnMass()");
	int i,j;
	FITableElement temp;
	for(i=1; i<PS.Spec[CurrentSpectrumFocus].NumPeaks; i++) {
		if (SortIncreasing) {
			if (GetMITableMass(i)<GetMITableMass(i-1)) {
				j=i;
				while(j>0 && GetMITableMass(j)<GetMITableMass(j-1)) {
					temp = FITable[j];
					FITable[j]=FITable[j-1];
					FITable[j-1]=temp;
					j--;
				}
			}
		} else {
			if (GetMITableMass(i)>GetMITableMass(i-1)) {
				j=i;
				while(j>0 && GetMITableMass(j)>GetMITableMass(j-1)) {
					temp = FITable[j];
					FITable[j]=FITable[j-1];
					FITable[j-1]=temp;
					j--;
				}
			}
		}
	}
	DumpToDiagFile("leaving CBudaDoc::SortFITableOnMass()");
}

void CBudaDoc::SortFITableOnFrequency()
{
	DumpToDiagFile("entering CBudaDoc::SortFITableOnFrequency()");
	int i,j;
	FITableElement temp;
	for(i=1; i<PS.Spec[CurrentSpectrumFocus].NumPeaks; i++) {
		if (SortIncreasing) {
			if (FITable[i].F<FITable[i-1].F) {
				j=i;
				while(j>0 && FITable[j].F<FITable[j-1].F) {
					temp = FITable[j];
					FITable[j]=FITable[j-1];
					FITable[j-1]=temp;
					j--;
				}
			}
		} else {
			if (FITable[i].F>FITable[i-1].F) {
				j=i;
				while(j>0 && FITable[j].F>FITable[j-1].F) {
					temp = FITable[j];
					FITable[j]=FITable[j-1];
					FITable[j-1]=temp;
					j--;
				}
			}
		}
	}
	DumpToDiagFile("leaving CBudaDoc::SortFITableOnFrequency()");
}

void CBudaDoc::SortMITable()
{
	DumpToDiagFile("entering CBudaDoc::SortMITable()");
	FITable = PS.Spec[CurrentSpectrumFocus].FITable;
	switch(SortOnProperty) {
	case 1: 
		SortFITableOnMass();
		break;
	case 2:
		SortFITableOnAbundance();
		break;
	case 3:
		SortFITableOnFrequency();
		break;
	}
	DumpToDiagFile("leaving CBudaDoc::SortMITable()");
}

void CBudaDoc::OnCalibrate() 
{
	DumpToDiagFile("entering CBudaDoc::OnCalibrate()");
	if (!IsMIed()) {
		MessageBox("Cannot calibrate without a peaklist\nCalculate an MI table first","Error: Cannot Calibrate");
		return;
		//PeakPickFlag = TRUE;
		//PeakPick(&PS, CurrentSpectrumFocus, MassToPt(PS.DetectHighMass), MassToPt(PS.DetectLowMass), FreqArray);
	}

	CCalibrateDlg dlg;

	// load CalibrateDlg with parms
	dlg.PS = &PS;
	dlg.sys = sys;
	dlg.FITable = FITable;
	dlg.pDoc = this;
	dlg.IsPeakMatching = PS.PeakMatching;

	// DoModal
	dlg.DoModal();

	DumpToDiagFile("leaving CBudaDoc::OnCalibrate()");
}

void CBudaDoc::OnPeakMatch() 
{
	DumpToDiagFile("entering CBudaDoc::OnPeakMatch()");
	//if (IsPeakMatching) IsPeakMatching = FALSE;
	//else IsPeakMatching = TRUE;

	//if (IsPeakMatching) 
	OnCalibrate();

	// OnCalibrate will update PS.PeakMatching if necessary

	DumpToDiagFile("leaving CBudaDoc::OnPeakMatch()");
}

void CBudaDoc::OnUpdatePeakMatch(CCmdUI* pCmdUI) 
{
	//DumpToDiagFile("entering CBudaDoc::OnUpdatePeakMatch(..)");
	if (IsFFTed()) pCmdUI->Enable(TRUE);
	else pCmdUI->Enable(FALSE);

	//if (PS.PeakMatching) pCmdUI->SetCheck(1);
	//else 
	pCmdUI->SetCheck(0);	
	//DumpToDiagFile("leaving CBudaDoc::OnUpdatePeakMatch(..)");
}

double CBudaDoc::CyclotronFrequency(double freq, int direction)
{
	if(PS.Spec == 0) {
		return CyclotronFrequency(freq,float(TrapVoltsOnDetect),0,PS.ScansPerSpectrum,PS.ADCGain,direction);
	} else
	return CyclotronFrequency(freq,float(TrapVoltsOnDetect),PS.Spec[CurrentSpectrumFocus].TotalIonInt,PS.ScansPerSpectrum,PS.ADCGain,direction);
}

inline CInstrument* CBudaDoc::GetInstrument()
{
	return &(pApp->Instrument);
}

struct sys* CBudaDoc::GetSys()
{
	return Instrument->GetSys();
}

void CBudaDoc::OnUpdateCalibrate(CCmdUI* pCmdUI) 
{
	if (IsFFTed()) pCmdUI->Enable(TRUE);
	else pCmdUI->Enable(FALSE);		
}

double CBudaDoc::GetExciteVoltage()
{
	return 50.0;
}

void CBudaDoc::InitializeParameters()
{
	PS.cfp = 0;
	PS.arb = 0;
	PS.dac = 0;
	PS.GenPS = 0;
	PS.Spec = 0;

	TDisDirty = TRUE;
	PSisDirty = TRUE;

	InitializePS();
	Setup2DDataMenu();
	CurrentFreqArrayPtr=0;
	order = 0;
	TransientIsDefined = FALSE;
	IsCSed = FALSE;
	AutoSave = 0;
	SaveTruncatedTransient = 0;
	DefaultSaveDirectory = "C:\\Buda\\FTData\\";
	DefaultFileNameType = 0;
	SetupOptionsInitPage = 0;
	//TimeArray = new double [2];
	PS.PeakFitType=PARABOLIC_LEAST_SQUARES;
	IsPeakMatching = FALSE;
	IsPeakPicked=FALSE;
	PeakPickFlag=FALSE;
	//TotalIonInt = 100;
	IsDefined = FALSE;
	SetIsFFTed(FALSE);
	SetIsMIed(FALSE);
	SortIncreasing = TRUE;
	SortOnProperty = 1;
	IFFTFlag = 1; // FFT's currently go forward = time->freq

	// SIC vs TIC
	SICmode = FALSE;
	SICMass = 1000;
	SICWidth = 1;
	AutoDetectThreshold = pApp->AutoDetectThreshold;
	ThresholdMethod = pApp->ThresholdMethod;
	CSHighChargeLimit = pApp->CSHighChargeLimit;
	CSDetMethod = pApp->CSDetMethod;
	
	// decon parameters
	DeconHighChargeLimit = pApp->DeconHighChargeLimit;
	DeconLowChargeLimit = pApp->DeconLowChargeLimit;
	DeconMethod = pApp->DeconMethod;
	DeconHighMassLimit = pApp->DeconHighMassLimit;
	DeconLowMassLimit = pApp->DeconLowMassLimit;
	DeconResolution = pApp->DeconResolution;
	DeconAdductMass = pApp->DeconAdductMass;

	// filter parameters
	BoxcarAvg = pApp->BoxcarAvg;
	BoxcarPoints = pApp->BoxcarPoints;

	GetDefaultDisplayParms();
}

void CBudaDoc::MessageBox(CString str1, CString str2)
{
	pWnd->MessageBox(str1,str2, MB_OK);
}

void CBudaDoc::ReadVersion6File(CFile* arf, int mode)
{
	DumpToDiagFile("entering CBudaDoc::ReadVersion6File(..)");
	// note : mode 0 is the normal read file mode.  mode 1
	// is the read setupfile mode
	CheckMemState("ReadVersion6File:1");

	long i,s;
	unsigned long inpts;
	long* ElapsedTime=0;
	char* junk=0;
	struct PSDesc6 event[TOTAL_PS_EVENTS];

	// read in the individual elements of PS that have not
	// changed from version 6 to version 8
	inpts=0;
	inpts+=arf->Read(&(PS.FileType),4);
	if(PS.FileType == SETUP_FILE) mode = 1;
	if(mode == 1) PS.FileType = SETUP_FILE;

	inpts+=arf->Read(&(PS.MajorVersion),4);
	inpts+=arf->Read(&(PS.MinorVersion),4);
	inpts+=arf->Read(&(PS.UserID),8);
	inpts+=arf->Read(&(PS.DataFileName),64);
	inpts+=arf->Read(&(PS.TimeAcquired),40);
	inpts+=arf->Read(&(PS.TimeProcessed),32);
	inpts+=arf->Read(&(PS.HiResMALDI),4);
	PS.HiResMALDI = PS.HiResMALDI & 0x000000ff;
	inpts+=arf->Read(&(PS.HiResESI),4);
	PS.HiResESI = PS.HiResESI & 0x000000ff;
	if(PS.HiResMALDI && PS.HiResESI) {
		PS.HiResESI = 0;
		PS.HiResMALDI = 0;
	}
	inpts+=arf->Read(&(PS.ExperimentMemo),60);
	PS.ExperimentMemo[59]='\0';
	inpts+=arf->Read(&(PS.ExperimentNotes),600);
	PS.ExperimentNotes[599]='\0';
	inpts+=arf->Read(&(PS.DetectMode),2);
	inpts+=arf->Read(&(PS.ADCGain),2);
	// old data files may have PS.ADCGain = 0 - which is bad for
	// the calibration in freqtomass()
	if(PS.ADCGain != 1 && PS.ADCGain != 2 && PS.ADCGain != 5 ) {
		PS.ADCGain = 1;
	}
	inpts+=arf->Read(&(PS.DetectEvent),4);
	inpts+=arf->Read(&(PS.DetectLowMass),8);
	inpts+=arf->Read(&(PS.DetectHighMass),8);
	inpts+=arf->Read(&(PS.MonitorMass),8);
	inpts+=arf->Read(&(PS.MixerFrequency),8);
	inpts+=arf->Read(&(PS.AdcRate),8);
	inpts+=arf->Read(&(PS.ScanMode),4);
	CheckMemState("ReadVersion6File:2");
	ElapsedTime = new long[MAX_SPECTRA];
	inpts+=arf->Read(ElapsedTime,4*MAX_SPECTRA);
	inpts+=arf->Read(&(PS.TransientPoints),4);
	inpts+=arf->Read(&(PS.SpectraRequested),4);
	inpts+=arf->Read(&(PS.SpectraAcquired),4);
	inpts+=arf->Read(&(PS.CurrentSpectrum),4);
	inpts+=arf->Read(&(PS.ScansPerSpectrum),4);
	inpts+=arf->Read(&(PS.ScansRequested),4);
	inpts+=arf->Read(&(PS.TriggerMode),4);
	inpts+=arf->Read(&(PS.CalibrationVersion),4);
	inpts+=arf->Read(&(PS.A),8);
	inpts+=arf->Read(&(PS.B),8);
	inpts+=arf->Read(&(PS.HighMassShift),8);
	inpts+=arf->Read(&(PS.TrapShift),8);
	inpts+=arf->Read(&(PS.IntensityShift),8);
	inpts+=arf->Read(&(PS.PeakMatching),4);
	inpts+=arf->Read(&(PS.PeakMatchFactor),8);
	inpts+=arf->Read(&(PS.RefPeak),sizeof(struct PeakMatchPoint)*MAX_REF_PEAK);
	inpts+=arf->Read(&(PS.SpectrometerType),4);
	inpts+=arf->Read(&(PS.DCRemoval),4);
	PS.DCRemoval = sys->DCRemoval;
	inpts+=arf->Read(&(PS.WindowFunction),4);
	inpts+=arf->Read(&(PS.TruncatePoints),4);
	inpts+=arf->Read(&(PS.ZeroFills),4);
	inpts+=arf->Read(&(PS.FFTSize),4);
	inpts+=arf->Read(&(PS.SpectrumPoints),4);
	inpts+=arf->Read(&(PS.FrequencyDataType),4);
	inpts+=arf->Read(&(PS.ZoomFactor),4);
	inpts+=arf->Read(&(PS.ZoomLowMass),8);
	inpts+=arf->Read(&(PS.ZoomHighMass),8);
	inpts+=arf->Read(&(PS.ZoomFrequencyShift),8);
	inpts+=arf->Read(&(PS.ZoomTransientPoints),4);
	inpts+=arf->Read(&(PS.ZoomPointShift),4);

	// create and fill in PeakCount
	long *PeakCount=0;
	CheckMemState("ReadVersion6File:3");
	PeakCount = new long[MAX_SPECTRA];
	inpts+=arf->Read(PeakCount,4*MAX_SPECTRA);

	// create and fill in cfp, dac, arb
	CheckMemState("ReadVersion6File:4");
	if(PS.cfp != 0) {
		delete [] PS.cfp;
		PS.cfp = 0;
	}
	if(PS.dac != 0) {
		delete [] PS.dac;
		PS.dac = 0;
	}
	if(PS.arb != 0) {
		delete [] PS.arb;
		PS.arb = 0;
	}
	CheckMemState("ReadVersion6File:5");
	PS.cfp = new struct ChirpPacket[MAX_SYN_TRIGGERS];
	PS.dac = new struct DacPacket[MAX_DACS];
	PS.arb = new struct ArbPacket[MAX_ARB_TRIGGERS];
	CheckMemState("ReadVersion6File:6");
	inpts+=arf->Read(PS.cfp,sizeof(struct ChirpPacket)*MAX_SYN_TRIGGERS);
	inpts+=arf->Read(PS.dac,sizeof(struct DacPacket)*MAX_DACS);
	inpts+=arf->Read(PS.arb,sizeof(struct ArbPacket)*MAX_ARB_TRIGGERS);
	PS.NumCFP = MAX_SYN_TRIGGERS;
	PS.NumDAC = MAX_DACS;
	PS.NumARB = MAX_ARB_TRIGGERS;

	// create and fill in GenPS
	inpts+=arf->Read(event,sizeof(struct PSDesc6)*TOTAL_PS_EVENTS);
	if(PS.GenPS != 0) {
		delete [] PS.GenPS;
		PS.GenPS = 0;
	}		
	CheckMemState("ReadVersion6File:7");
	PS.GenPS = new struct PSDesc[MAX_PS_EVENTS];
	for(i=0; i<TOTAL_PS_EVENTS; i++) {
		PS.GenPS[i].Channel=GetV6ChannelNumber(i);
		if(i == LAST_USER16_DAC+1) event[i].Type=NO_EVENT;
		PS.GenPS[i].Type = event[i].Type;
		PS.GenPS[i].Linking = event[i].Linking;
		PS.GenPS[i].Link2Event = event[i].Link2Event;
		PS.GenPS[i].Delay = event[i].Delay;
		PS.GenPS[i].Width = event[i].Width;
		PS.GenPS[i].Value = event[i].Value;
	}
	PS.NumEvents = TOTAL_PS_EVENTS;
	SortGenPS();
	//delete [] event;


	CheckMemState("ReadVersion6File:7a");
	if (inpts != PS6_NUM_BYTES) {
		FileError("WARNING!\nError 01:\nError Reading Version 6 file\nHeader Incorrect");
		junk = new char[PS6_NUM_BYTES-inpts];
		arf->Read(junk,PS6_NUM_BYTES-inpts);
		delete [] junk;
	}

	// now define the new data elements that weren't defined in version 6
	if(PS.FileType == SINGLE_TRANSIENT_FILE) PS.SpectraAcquired = 1;
	if(mode == 0) PS.TransientIsStored = 1;
	else PS.TransientIsStored = 0;
	PS.MITableIsStored = 0;
	PS.FrequencyIsStored = 0;
	PS.TruncateTransient = 0;
	PS.TruncatePoints = PS.TruncatePoints/2;
	PS.TruncateOffset = 0;
	char c[8] = "Unknown";
	for (i=0; i<128; i++) {
		if(i<7) {
			PS.SystemName[i]=c[i];
			PS.SystemSerialNumber[i]=c[i];
		}
		else if (i<32) {
			PS.SystemName[i]=0;
			PS.SystemSerialNumber[i]=0;
		}
		else PS.SystemName[i]=0;
	}
	if(PS.Spec != 0) {
		delete [] PS.Spec;
		PS.Spec = 0;
	}
	CheckMemState("ReadVersion6File:8");
	PS.Spec = new struct Spectrum[PS.SpectraAcquired];
	CheckMemState("ReadVersion6File:9");

	// if reading a setup file, skip all the read data crap
	if(mode == 0) {
		// read in the data element's size (either 16 bit or 32 bit)
		unsigned long HeaderSize, StepSize, ActualPosition=0, NextPosition=0;
		inpts += arf->Read(&DataPointSize,4);
		PS.TransientIsStored = DataPointSize;
		PS.MITableIsStored = 0;
		PS.FrequencyIsStored = 0;
		StepSize = PS.TransientPoints*DataPointSize/8;
		HeaderSize = arf->GetPosition();
		ASSERT(HeaderSize == inpts);

			// fill in remainder of PS.Spec stuff
		CheckMemState("ReadVersion6File:10");
		for(s=0; s<PS.SpectraAcquired; s++) {
			PS.Spec[s].ElapsedTime = ElapsedTime[s];
			PS.Spec[s].NumPeaks = PeakCount[s];
			PS.Spec[s].IsMIed = FALSE;
			PS.Spec[s].IsFFTed = FALSE;
			PS.Spec[s].NoiseLevel = 0;
			PS.Spec[s].ScaleFactor = 1;
			PS.Spec[s].TotalIonInt = 0;
			PS.Spec[s].FITable = 0;
			NextPosition = HeaderSize + s*StepSize;
			ActualPosition = arf->Seek(NextPosition,CFile::begin);
			if(ActualPosition != NextPosition) {
				CString str;
				str.Format("Cannot Access Data position (byte offset from beginning of file)\n at %ld in file %s",NextPosition,arf->GetFileName());
				MessageBox(str,"Read File Error");
				return;
			}
			PS.Spec[s].TArrayFileOffset = arf->GetPosition();
		} // end for s loop
		CheckMemState("ReadVersion6File:11");

		CurrentSpectrumFocus = -1;
		ReadTimeArray(0,arf);
		CurrentSpectrumFocus = 0;
		TransientIsDefined = TRUE;
		CheckMemState("ReadVersion6File:12");

	} else {	// mode == 1 (SETUP_FILE)
		for(s=0; s<PS.SpectraAcquired; s++) {
			// for setup files only,
			PS.Spec[s].ElapsedTime = ElapsedTime[0];
			PS.Spec[s].NumPeaks = PeakCount[0];
			PS.Spec[s].IsMIed = FALSE;
			PS.Spec[s].IsFFTed = FALSE;
			PS.Spec[s].NoiseLevel = 0;
			PS.Spec[s].ScaleFactor = 1;
			PS.Spec[s].TotalIonInt = 0;
			PS.Spec[s].FITable = 0;
			PS.Spec[s].TArrayFileOffset = 0;
		}
	}
	CheckMemState("ReadVersion6File:13");
	SetupParmsOnRead();

	CheckMemState("ReadVersion6File:14");
	if(ElapsedTime != 0) {
		delete [] ElapsedTime;
		ElapsedTime = 0;
	}
	if(PeakCount != 0) {
		delete [] PeakCount;
		PeakCount = 0;
	}

	// set FileWasRead to TRUE so that OnOpenDocument can decide not
	// to kill this document
	FileWasRead = TRUE;

	CheckMemState("ReadVersion6File:15");
	DumpToDiagFile("leaving CBudaDoc::ReadVersion6File(..)");
} // end ReadVersion6File

void CBudaDoc::StoreVersion8File(CFile *arf,int mode)
{
	DumpToDiagFile("entering CBudaDoc::StoreVersion8File(..)");
	long i,s,OldFileType;
	OldFileType = PS.FileType;
	PS.FileType = VERSION_8_COMPOSITE_FILE;
	if(mode == 1) PS.FileType = SETUP_FILE;
	PS.MajorVersion = CURRENT_MAJOR_VERSION_NUMBER;
	PS.MinorVersion = CURRENT_MINOR_VERSION_NUMBER;

	if (SaveTruncatedTransient != 0) {
		PS.TransientPoints = PS.TruncatePoints;
	} else {
		PS.TruncateOffset = 0;
		PS.TruncatePoints = PS.TransientPoints;
	}

	// first write the header
	arf->SeekToBegin();
	arf->Write(&PS, sizeof(PS));

	// then write the cfp array
	for(i=0; i<PS.NumCFP; i++) {
		arf->Write(&(PS.cfp[i]), sizeof(struct ChirpPacket));
	}

	// then write the dac array
	for(i=0; i<PS.NumDAC; i++) {
		arf->Write(&(PS.dac[i]), sizeof(struct DacPacket));
	}

	// then write the arb array
	for(i=0; i<PS.NumARB; i++) {
		arf->Write(&(PS.arb[i]), sizeof(struct ArbPacket));
	}

	// then write the GenPS array
	for(i=0; i<PS.NumEvents; i++) {
		arf->Write(&(PS.GenPS[i]), sizeof(struct PSDesc));
	}

	// if this is a setup file, stop here otherwise
	//	then write the Spec array 

	if( mode == 0) { // not a Setup File
		// then, for each spectrum in the Spec array (0 .. PS.SpectraAcquired)
		unsigned long Headersize, NextPosition, ActualPosition, StepSize;
		Headersize = arf->GetPosition() + sizeof(struct Spectrum);
		StepSize = PS.TransientPoints*PS.TransientIsStored/8;
		for(i=0; i<PS.SpectraAcquired; i++) {
			NextPosition = Headersize + i*StepSize;
			PS.Spec[i].TArrayFileOffset = NextPosition;
			arf->Write(&(PS.Spec[i]), sizeof(struct Spectrum));
		}
		ActualPosition = arf->GetPosition();
		ASSERT(ActualPosition == Headersize);
		
		// then write the data for each spectrum
		// do this simply by cycling through the DS values
		for (s=0; s<PS.SpectraAcquired; s++) {
			SetCurrentSpectrumFocus(s);
			// first write time domain data if it is stored
			//		note that PS.TransientIsStored = 0,8,16,32 indicating the size of the data elements
			NextPosition = Headersize + s*StepSize;
			DataPointSize = PS.TransientIsStored;
			switch (PS.TransientIsStored) {
			case 8:
				char *c;		
				c = new char[PS.TransientPoints];
				for(i=0; i<PS.TransientPoints; i++) {
					c[i]=char(TimeArray[i+PS.TruncateOffset]);
				}
				arf->Write(c,PS.TransientPoints);
				delete [] c;
				break;
			case 16:
				short *sh;
				sh = new short[PS.TransientPoints];
				for(i=0; i<PS.TransientPoints; i++) {
					sh[i]=short(TimeArray[i+PS.TruncateOffset]);
				}
				arf->Write(sh,2*PS.TransientPoints);
				delete [] sh;
				break;
			case 32:
				long *l;
				l = new long[PS.TransientPoints];
				for(i=0; i<PS.TransientPoints; i++) {
					l[i]=TimeArray[i+PS.TruncateOffset];
				}
				arf->Write(l,4*PS.TransientPoints);
				delete [] l;
				break;
			default:
				break;
			} 
			ActualPosition = arf->GetPosition();
			if(ActualPosition != NextPosition + StepSize) {
				MessageBox("Error on Write, could not write whole file","File Error");
				return;
			}
		}

		for(s=0; s<PS.SpectraAcquired; s++) {
			// write MITables if they are stored
			//		note that PS.MITableIsStored = 0,1 
			//
			FITable = PS.Spec[s].FITable;
			if(PS.MITableIsStored) {
				if (IsMIed()) {
					arf->Write(FITable,PS.Spec[s].NumPeaks*sizeof(FITableElement));
				} else
					pWnd->MessageBox("Error 60: \nTrying to save the MI Table\nbut no MI table Available","MI table not found");
					// todo: should now destroy the current file.
			}
		} // end for s loop

		//	finally write the (one and only) FreqArray (if needed)
		// second write frequency domain data if it is stored
		// note that frequency is always stored as doubles
		if (PS.FrequencyIsStored) {
			arf->Write(FreqArray,PS.SpectrumPoints);
		}
	} // if mode == 0
	if(mode == 1) PS.FileType = OldFileType;

	DumpToDiagFile("leaving CBudaDoc::StoreVersion8File(..)");
	
} // end StoreVersion8File

void CBudaDoc::ReadVersion8File(CFile *arf,int mode)
{
	DumpToDiagFile("entering CBudaDoc::ReadVersion8File(..)");
	long i,s;

	arf->Read(&(PS.FileType),4);
	arf->Read(&(PS.MajorVersion),4);
	arf->Read(&(PS.MinorVersion),4);

	if(PS.MajorVersion != 8) {
		MessageBox("Cannot recognize file version number.\nThis file is not version 8.x","Error");
		return;
	}
	if(PS.MinorVersion != 0) {
		MessageBox("Cannot recognize file version number.\nThis file is not version 8.0","Error");
		return;
	}

	InitializePS();	// initializes the PS arrays

	// store these pointers before reading PS and overwriting them.
	struct ChirpPacket*	cfp = PS.cfp;
	struct DacPacket*	dac = PS.dac;
	struct ArbPacket*	arb = PS.arb;
	struct PSDesc*		GenPS = PS.GenPS;
	struct Spectrum*	Spec = PS.Spec;

	arf->SeekToBegin();
	arf->Read(&PS, sizeof(PS));

	//restore the pointers
	PS.cfp = cfp;
	PS.dac = dac;
	PS.arb = arb;
	PS.GenPS = GenPS;
	PS.Spec = Spec;

	// then read the cfp array
	if(PS.NumCFP > MAX_SYN_TRIGGERS) PS.NumCFP = MAX_SYN_TRIGGERS;
	for(i=0; i<PS.NumCFP; i++) {
		arf->Read(&(PS.cfp[i]), sizeof(struct ChirpPacket));
	}

	// then read the dac array
	if(PS.NumDAC > MAX_DACS) PS.NumDAC = MAX_DACS;
	for(i=0; i<PS.NumDAC; i++) {
		arf->Read(&(PS.dac[i]), sizeof(struct DacPacket));
	}

	// then read the arb array
	if(PS.NumARB > MAX_ARB_TRIGGERS) PS.NumARB = MAX_ARB_TRIGGERS;
	for(i=0; i<PS.NumARB; i++) {
		arf->Read(&(PS.arb[i]), sizeof(struct ArbPacket));
	}

	// then read the GenPS array
	if(PS.NumEvents > MAX_PS_EVENTS) PS.NumEvents = MAX_PS_EVENTS;
	for(i=0; i<PS.NumEvents; i++) {
		arf->Read(&(PS.GenPS[i]), sizeof(struct PSDesc));
	}

	// if this is a setup file, stop here otherwise
	//	then read the Spec array 

	if(PS.FileType == SETUP_FILE) mode = 1;

	if( mode == 0) { // not a Setup File
		// then, for each spectrum in the Spec array (0 .. PS.SpectraAcquired)
		PS.Spec = new struct Spectrum[PS.SpectraAcquired];
		for(i=0; i<PS.SpectraAcquired; i++) {
			arf->Read(&(PS.Spec[i]), sizeof(struct Spectrum));
		}

		unsigned long HeaderSize, StepSize, ActualPosition=0, NextPosition=0;
		DataPointSize = PS.TransientIsStored;
		StepSize = PS.TransientPoints*DataPointSize/8;
		HeaderSize = arf->GetPosition();

		// read the first transient
		CurrentSpectrumFocus = -1;
		ReadTimeArray(0,arf);
		CurrentSpectrumFocus = 0;
	
		TransientIsDefined = TRUE;
		// jump to the end of the last transient
		NextPosition = HeaderSize + PS.SpectraAcquired*StepSize;
		ActualPosition = arf->Seek(NextPosition,CFile::begin);
		if(ActualPosition != NextPosition) {
			CString str;
			str.Format("Cannot Access Data position (byte offset from beginning of file)\n at %ld in file %s",NextPosition,arf->GetFileName());
			MessageBox(str,"Read File Error");
			return;
		} 

		for (s=0; s<PS.SpectraAcquired; s++) {
			// read MITables if they are stored
			//		note that PS.MITableIsStored = 0,1 
			if(PS.MITableIsStored) {
				PS.Spec[s].FITable = new struct FITableElement[PS.Spec[s].NumPeaks];
				FITable = PS.Spec[s].FITable;
				arf->Read(FITable,PS.Spec[s].NumPeaks*sizeof(FITableElement));
				//PS.IsMIed = TRUE;
				SetIsMIed(s,TRUE);
			} else {
				//PS.IsMIed = FALSE;
				SetIsMIed(s,FALSE);
				PS.Spec[s].FITable = 0;
				FITable = 0;
			}
		} // end for s loop

		//	finally read the (one and only) FreqArray (if needed)
		// second read frequency domain data if it is stored
		// note that frequency is always stored as doubles
		if (PS.FrequencyIsStored) {
			SetIsFFTed(TRUE);
			FreqArray = new float[PS.SpectrumPoints];
			arf->Read(FreqArray,PS.SpectrumPoints);
		} else {
			SetIsFFTed(FALSE);
			FreqArray = 0;
		}

	} // if mode == 0
	else {
		// for setup files only,
		SetIsFFTed(FALSE);
		SetIsMIed(FALSE);
		IsCSed = FALSE;
		IsDefined = FALSE;
		IsPeakPicked = FALSE;
	}

	SetupParmsOnRead();

	FileWasRead = TRUE;

	DumpToDiagFile("leaving CBudaDoc::ReadVersion8File(..)");

} // end ReadVersion8File(...)

void CBudaDoc::FileError(CString str)
{
	MessageBox(str,"File Error");
	FileWasRead = FALSE;
}


void CBudaDoc::SetupParmsOnRead()
{
	DumpToDiagFile("entering CBudaDoc::SetupParmsOnRead()");
	if (PS.PeakMatching) IsPeakMatching = TRUE;
	else IsPeakMatching = FALSE;

	// convert all the strings to '/0' terminated strings
	NullTerminate(PS.DataFileName,64);
	NullTerminate(PS.UserID,8);
	NullTerminate(PS.TimeAcquired,40);
	NullTerminate(PS.TimeProcessed,32);
	NullTerminate(PS.ExperimentMemo,60);
	NullTerminate(PS.ExperimentNotes,600);
		// Assign TrapVoltsDuringDetect
	if (PS.FileType != BRUKERFILE) TrapVoltsOnDetect=GetTrapVoltsOnDetect();
	else TrapVoltsOnDetect=0;
	// Calculate Intensity Shift Factor
	if (PS.FileType != BRUKERFILE) PS.IntensityShift=CalculateIntShift();
	else PS.IntensityShift=0;

	IsDefined = TRUE;
	SingleFFTFlag = FALSE;

	// check to see that the PS.DetectLowMass is valid
	if(PS.DetectMode!= MixerMode && PS.DetectLowMass < FreqToMass(PS.AdcRate/2))
		{
			PS.DetectLowMass = FreqToMass(PS.AdcRate/2);
		}
	else if(PS.DetectMode == MixerMode && PS.DetectLowMass <FreqToMass(PS.MixerFrequency + PS.AdcRate/2)) 
		PS.DetectLowMass = FreqToMass(PS.MixerFrequency + PS.AdcRate/2);

	// create FreqArray
	//FreqArray = new float [2];
	FreqArray = 0;

	DumpToDiagFile("leaving CBudaDoc::SetupParmsOnRead()");
}

inline void CBudaDoc::DumpToDiagFile(CString str)
{
//	CHKMEM;
#ifdef DEBUG_DUMP_DIAGS_TO_FILE
	str=str+"\r\n";
	pApp->diagfile.Write(str,str.GetLength());
	//TRACE(str);
#endif
}

/*
void CBudaDoc::Truncate()
{
//	DumpToDiagFile("entering CBudaDoc::Truncate()");
	//double* temp;
	//temp = new double[PS.TruncatePoints+2];
	double *swap;

	long i;
	for (i=PS.TruncateOffset; i<PS.TruncateOffset+PS.TruncatePoints; i++) 
		temp[i-PS.TruncateOffset] = FreqArray[i];

	for (i=PS.TruncatePoints; i<PS.TransientPoints; i++) temp[i]=0;

	//PS.FFTSize = PS.TruncatePoints;

	swap = FreqArray;
	FreqArray = temp;
	temp = swap;
//	DumpToDiagFile("leaving CBudaDoc::Truncate()");
}
*/

CString CBudaDoc::GetDefaultFileName() 
{
	DumpToDiagFile("entering CBudaDoc::GetDefaultFileName()");
	// TODO: change the file name here to a default
	CString filename;
	CString timedate;
	int i;

	switch (DefaultFileNameType) {
	default: // default name falls through to case 0
	case 0:  // default name is UserID + Time/Date Stamp
		GetTimeDate(&timedate);
		filename.Format("%s_",PS.UserID);
		filename = DefaultSaveDirectory + filename + "_" + timedate;
		break;
	case 1: // default name is UserID + Sequential number
		i=GetNextFileNumber();
		filename.Format("%s.%d",PS.UserID,i);
		filename = DefaultSaveDirectory + filename;
		break;
	}
	
	DumpToDiagFile("leaving CBudaDoc::GetDefaultFileName()");
	return filename;
}

void CBudaDoc::GetTimeDate(CString *str)
{
	// this file creates a string that shows the time and
	// date in format 12-23-1968_15:35:58
	CTime t = CTime::GetCurrentTime();
	str->Format("%d-%d-%d_%d:%d:%d",t.GetMonth(),t.GetDay(),t.GetYear(),t.GetHour(),t.GetMinute(),t.GetSecond());
}

void CBudaDoc::OnFileSaveAs() 
{
	DumpToDiagFile("entering CBudaDoc::OnFileSaveAs()");
	CString str = GetDefaultFileName();
	CFileDialog dlg(FALSE,NULL,NULL,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,"*.*",NULL);
	dlg.m_ofn.lpstrInitialDir = DefaultSaveDirectory;
	if(dlg.DoModal() == IDCANCEL) return;
	CDocument::OnSaveDocument(dlg.GetPathName());
	DumpToDiagFile("leaving CBudaDoc::OnFileSaveAs()");
}


void CBudaDoc::OnFileSave() 
{
	DumpToDiagFile("entering CBudaDoc::OnFileSave()");
	CString str;
	BOOL Overwrite = FALSE;
	if(PS.MajorVersion != 8) {
		str.Format("Saving version %d.%d file as version 8.0 file...\nDo you wish to overwrite the old file?",PS.MajorVersion,PS.MinorVersion);
		if(pWnd->MessageBox(str,"Overwrite file?",MB_YESNO) != IDYES) return;
		Overwrite = TRUE;
	}

	if(IsModified() && !Overwrite) {
		if(pWnd->MessageBox("Overwrite the old file?","File save?",MB_YESNO) != IDYES) return;
	}

	str = GetPathName();
	if(str.IsEmpty()) {
		str = GetDefaultFileName();
		SetPathName( str, TRUE );
	}

	OnChangeMemo();
	//CDocument::OnSaveDocument(str);
	CDocument::OnFileSave();

	DumpToDiagFile("leaving CBudaDoc::OnFileSave()");
}

int CBudaDoc::GetNextFileNumber()
{
	DumpToDiagFile("entering CBudaDoc::GetNextFileNumber()");
	int i;
	CFileStatus status;
	CFile fl;
	CString str;

	i=0;
	do {
		str.Format("%s.%d",PS.UserID,i);
		str = DefaultSaveDirectory + str;
		i++;
	} while (!fl.GetStatus(str,status));

	DumpToDiagFile("leaving CBudaDoc::GetNextFileNumber()");
	return i;
}


double CBudaDoc::GetMITableMass(long i)
{
	return FreqToMass(FITable[i].F, PS.A, PS.B);
}

BOOL CBudaDoc::SetCurrentSpectrumFocus(int s, int mode)
{
	// this function merely sets the current values of 
	// FreqArray, FITable, and TimeArray to those defined
	// in DS[s] after checking that s is valid.
	//
	// NOTE: calling this function when there is a spectrum view
	// up with ViewAsStickplot == FALSE will require an FFT

	DumpToDiagFile("entering CBudaDoc::SetCurrentSpectrumFocus(..)");

	// verify that s is valid
	if (s == GetCurrentSpectrumFocus()) return TRUE;
	if (s<0 || (int)s>PS.SpectraAcquired-1) {
		// not a valid spectrum number
		MessageBox("Error 23\nSpectrum index out of range","Memory error");
		DumpToDiagFile("leaving CBudaDoc::SetCurrentSpectrumFocus(..) - FALSE");
		return FALSE;
	}

	if(IsDefined) {
		if(PS.Spec[s].IsMIed) FITable = PS.Spec[s].FITable;
		SetIsFFTed(FALSE);
		PS.NoiseThreshold = PS.Spec[s].NoiseLevel;
		PS.TotalIonIntensity = PS.Spec[s].TotalIonInt;
	}

	// get the new time array from file
	if(mode != NO_LOAD_TIMEARRAY) {
		GetTimeArray(s);
	}
	CurrentSpectrumFocus = s;

	DumpToDiagFile("leaving CBudaDoc::SetCurrentSpectrumFocus(..) - TRUE");
	return TRUE;
}

int CBudaDoc::GetCurrentSpectrumFocus()
{
	return CurrentSpectrumFocus;
}

void CBudaDoc::SetStatusBar(CStatusBar *bar)
{
	StatusBar = bar;
}

void CBudaDoc::DeleteContents() 
{
	// first, remove all the temporary files for this document
	// iterate through DS to get all the filenums
	// and then call removetempfile for this filenum
	DumpToDiagFile("entering CBudaDoc::DeleteContents()");
	//long s;

	if(IsDefined) {
		int s;
		// delete all pointers and set them = 0;
		if(FreqArray != 0) {
			CheckMemState("DeleteContents:1");
			delete [] FreqArray;
			FreqArray = 0;
		}
				
		// delete all pointers in PS and set them = 0
		if(PS.Spec != 0) {
			for (s=0; s<PS.SpectraAcquired; s++) {
				SetCurrentSpectrumFocus(s,NO_LOAD_TIMEARRAY);
				if (PS.Spec[s].FITable != 0) {
					CheckMemState("DeleteContents:2");
					delete [] PS.Spec[s].FITable;
					PS.Spec[s].FITable=0;
				}
			}
			//delete [] PS.Spec;
			CheckMemState("DeleteContents:3");
			delete PS.Spec;
			PS.Spec = 0;
		}
		if(TimeArray != 0) {
			delete [] TimeArray;
			TimeArray = 0;
		}
		if(PS.cfp != 0) {
			delete [] PS.cfp;
			PS.cfp = 0;
		}
		if(PS.dac != 0) {
			delete [] PS.dac;
			PS.dac = 0;
		}
		if(PS.arb != 0) {
			delete [] PS.arb;
			PS.arb = 0;
		}
		if(PS.GenPS != 0) {
			delete [] PS.GenPS;
			PS.GenPS = 0;
		}
	}

	IsDefined = FALSE;

	m_bAutoDelete = TRUE;
	CDocument::DeleteContents();
	DumpToDiagFile("leaving CBudaDoc::DeleteContents()");
}

void CBudaDoc::ResetMenus()
{
/*	DumpToDiagFile("entering CBudaDoc::ResetMenus()");
#ifdef INCLUDE_DIAGNOSTICS
	//Diags.SetupDiagnosticsMenu();
#endif
	//Update2DDataMenu();
	//AfxGetApp()->m_pMainWnd->DrawMenuBar();
	DumpToDiagFile("leaving CBudaDoc::ResetMenus()");
*/
}

void CBudaDoc::OnDebugTools() 
{
/*	DumpToDiagFile("entering CBudaDoc::OnDebugTools()");
#ifdef INCLUDE_DIAGNOSTICS
	Diags.SetupDiagnosticsMenu();
#endif
	DumpToDiagFile("leaving CBudaDoc::OnDebugTools()");
*/
}

void CBudaDoc::OnDebugSine() 
{
	DumpToDiagFile("entering CBudaDoc::OnDebugSine()");
	CSineWaveDlg dlg;

	dlg.freq = 0;
	dlg.mass = 0;
	dlg.amp = 1;
	dlg.acqrate = PS.AdcRate;

	dlg.DoModal();

	double freq = dlg.freq;
	double mass = dlg.mass;
	double amp = dlg.amp;
	amp = amp;

	if (mass != 0) freq = MassToFreq(freq, PS.A, PS.B);

	long i;
	for (i=0; i<PS.TransientPoints; i++) {
		//TimeArray[i]=amp*sin(2*pi*freq*i/PS.AdcRate);
	}
	//DS[CurrentSpectrumFocus].IsDirty = TRUE;
	DumpToDiagFile("leaving CBudaDoc::OnDebugSine()");
}

void CBudaDoc::Update2DDataMenu()
{
	DumpToDiagFile("entering CBudaDoc::Update2DDataMenu()");
/*	CMenu* pTopMenu = AfxGetApp()->m_pMainWnd->GetMenu();
	// check to see if the debug menu is already there
	UINT pos, maxpos = pTopMenu->GetMenuItemCount();
	if (maxpos < 0) return;
	CString str, d="2D-Data";
	BOOL TwoDMenuExists = FALSE;
	for(pos=maxpos; pos>0; pos--) {
		pTopMenu->GetMenuString(pos,str,MF_BYPOSITION);
		if (str == d) {
			TwoDMenuExists = TRUE;
			break;
		}
	}
	if(PS.SpectraAcquired<2) {
		if (TwoDMenuExists) pTopMenu->RemoveMenu(pos,MF_BYPOSITION);
		return;
	}
	if(TwoDMenuExists) return;

	// ok, 2DData menu is needed and it doesn't exist, so create it
	//TwoDDataMenu.LoadMenu(IDR_2D_DATA_MENU);
	HMENU mnu = TwoDDataMenu.GetSafeHmenu();
	pTopMenu->InsertMenu(maxpos - 2, MF_POPUP || MF_BYPOSITION,(UINT)mnu,d);
	AfxGetApp()->m_pMainWnd->DrawMenuBar();
	*/
	DumpToDiagFile("leaving CBudaDoc::Update2DDataMenu()");
}

void CBudaDoc::Setup2DDataMenu()
{
//	DumpToDiagFile("entering CBudaDoc::Setup2DDataMenu()");
//	CString d = "testing";
//	CMenu* pTopMenu = AfxGetApp()->m_pMainWnd->GetMenu();
//	TwoDDataMenu.LoadMenu(IDR_2D_DATA_MENU);
//	HMENU mnu = TwoDDataMenu.GetSafeHmenu();
//	pTopMenu->AppendMenu( MF_POPUP || MF_BYPOSITION,(UINT)mnu,d);
//	AfxGetApp()->m_pMainWnd->DrawMenuBar();
//	DumpToDiagFile("leaving CBudaDoc::Setup2DDataMenu()");
}

void CBudaDoc::OnReadTestFftFile() 
{
/*	CString file = "FFTtest.dat";
	CFile fi;
	fi.Open(file,  CFile::modeRead , NULL) ;
	fi.Write(&(PS.SpectrumPoints),4);
	fi.Write(&FreqArray,8*numpts);
	fi.Close();
	PS.TransientPoints = PS.SpectrumPoints*2;
*/
}

void CBudaDoc::OnDebugMakeSineWave() 
{
	PS.TransientPoints = 1024*1024;
	long i=0;
	for(i=0; i<PS.TransientPoints; i++) {
		//TimeArray[i]=sin(double(i)/10);
	}
}


double CBudaDoc::FreqToMass(double freq)
{
	return FreqToMass(freq, PS.A, PS.B);
}

void CBudaDoc::GetCurrentFreqArray(int s, BOOL Do2DFFT)
{
	// turn on the hourglass cursor
	pApp->DoWaitCursor(1);

	// tell the user what the current status is on the statusbar
	UpdateFFTStatus(s);
	SetIsFFTed(FALSE);
	IsPeakPicked=FALSE;
	
	int i;
	if (Do2DFFT) {
		for(i=0; i<PS.SpectraAcquired; i++) {
			AddToFFTQueue(i);
		}
	} else {
		AddToFFTQueue(s);
	}
	
	// turn the cursor back into an arrow
	pApp->DoWaitCursor(-1);

	// set some flags
	CurrentFreqArrayPtr=s;
}

void CBudaDoc::RecalcAllViews()
{ 
	// this sets all the view's ViewLineIsDirty and ViewIsDirty
	// flags then updates them
	POSITION pos = GetFirstViewPosition();
	while(pos != NULL) {
		CBudaView* pView = (CBudaView*)GetNextView( pos );
		pView->ViewIsDirty = TRUE;
		pView->ViewLineIsDirty = TRUE;
	}
	UpdateAllViews(NULL);
}

void CBudaDoc::OnStopfftbtn() 
{
	HaltFFTFlag = TRUE;	
}

void CBudaDoc::OnUpdateStopfftbtn(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(EnableStopFFTBtn);	
}

BOOL CBudaDoc::OnOpenDocument(LPCTSTR lpszPathName) 
{
	if (!CDocument::OnOpenDocument(lpszPathName))
		return FALSE;
	
	//if(PS.MajorVersion!=8) {
	//	SetModifiedFlag(TRUE);
	//}

	if(!FileWasRead) return FALSE;

	return TRUE;
}

BOOL CBudaDoc::CanCloseFrame(CFrameWnd* pFrame) 
{
	// here we check if the file is a version 6 file and
	// don't autosave if it is.
	return CDocument::CanCloseFrame(pFrame);
}

void CBudaDoc::GetEventEdges(long EventIndex, float *start, float *finish)
{
	GetEventEdges(&PS,EventIndex,start,finish,0);
}

void CBudaDoc::GetEventEdges(struct PS* pPS, long EventIndex, float *start, float *finish)
{
	GetEventEdges(pPS,EventIndex,start,finish,0);
}

void CBudaDoc::GetEventEdges(struct PS* pPS, long EventIndex, float *start, float *finish, int EventLoopCtr)
{
	// this function searches the GenPS event array to return the 
	// start and end time of event EventIndex
	// usually, this is trivial, but if the events are linked,
	// then we need to find them

	// note: all times are in milliseconds

	// note: EventLoopCtr is used to keep track of links in the recursive function call. 
	//		since this is a recursive function, detection of cyclic link loops could be
	//		a problem.  This is done here by always calling the full GetEventEdges with
	//		EventLoopCtr from within this function, but outside functions can use the 
	//		polymorphed versions.  Also, the CBudaDoc member variable 
	//		BOOL CyclicLoopCheck[MAX_PS_EVENTS]		is used to be sure we don't link back
	//		up with an event we have already visited

	*start = 0.0;
	*finish = 0.0;

	// return on bad events
	if(EventIndex < 0) return;
	if(EventIndex > pPS->NumEvents) return;
	if(EventIndex > TOTAL_PS_EVENTS) return;
	if(pPS->GenPS[EventIndex].Type == NO_EVENT) return;

	// if not linked, then we are done
	if(pPS->GenPS[EventIndex].Linking == NO_LINK) {
		*start = pPS->GenPS[EventIndex].Delay;
		*finish = *start + pPS->GenPS[EventIndex].Width;
		return;
	}

	if(EventLoopCtr < 0 || EventLoopCtr > pPS->NumEvents) {
		MessageBox("Detected invalid Event number range.  Pulse sequence may be corrupted","Invalid Event number");
		return;
	}
	if(EventLoopCtr == 0) {
		// clear the cyclic loop check array
		int i;
		for(i=0; i < MAX_PS_EVENTS; i++) CyclicLoopCheck[i]=FALSE;
	}

	// if linked, then we need to get the edges of the linked pulse
	// this part is a recursive function call
	float parentstart, parentend;
	// check for the infinite loop situations
	if(CyclicLoopCheck[EventIndex]) {
		MessageBox("Detected a cyclic Link loop in the pulse sequence\nRemoving one of the links\nPlease check the pulse sequence manually","Invalid pulse sequence");
		pPS->GenPS[EventIndex].Link2Event = 0;
		pPS->GenPS[EventIndex].Linking = NO_LINK;
		return;
	} else {
		CyclicLoopCheck[EventIndex] = TRUE;
	}

	EventLoopCtr++;
	GetEventEdges(pPS , pPS->GenPS[EventIndex].Link2Event,&parentstart,&parentend,EventLoopCtr);

	// if linking to the beginning of the link2event
	if(pPS->GenPS[EventIndex].Linking == LINK_BEG) {
		*start = parentstart + pPS->GenPS[EventIndex].Delay;
		*finish = *start + pPS->GenPS[EventIndex].Width;
	}

	// if linking to the end of the lin2event
	if(pPS->GenPS[EventIndex].Linking == LINK_END) {
		*start = parentend + pPS->GenPS[EventIndex].Delay;
		*finish = *start + pPS->GenPS[EventIndex].Width;
	}
}

float CBudaDoc::DacValueAtTime(long dacindex, float time)
{
	int i;
	float volts,start,end;
	volts = 0;
	float LastStepTime;
	LastStepTime = 0;

	for(i=0; i<PS.NumEvents; i++) {
		if(PS.GenPS[i].Channel != dacindex) continue;	// wrong channel
		GetEventEdges(i,&start,&end);

		if(PS.GenPS[i].Type == DAC_INIT) {
			volts = PS.GenPS[i].Value;
			LastStepTime = start;
		}

		if(PS.GenPS[i].Type == DAC_PULSE) {
			if(time<end && time>start) {
				volts = PS.GenPS[i].Value;
			}
		}

		if(PS.GenPS[i].Type == DAC_STEP) {
			if(start>LastStepTime) {
				volts = PS.GenPS[i].Value;
				LastStepTime = start;
			}
		}

		float PrevVolts;
		float m,b;
		if(PS.GenPS[i].Type == DAC_RAMP) {
			if(time<end && time>start) { // ok, gotta interpolate
				// first get the voltage one microsecond before this event
				PrevVolts = DacValueAtTime(dacindex,start-(float)0.001);
				m = (PS.GenPS[i].Value-PrevVolts)/(end-start);
				b = PrevVolts - m*start;
				// now y = m*x + b
				volts = m*time + b;
			} else {
				volts = PS.GenPS[i].Value;
			}
			LastStepTime = end;
		}
	}
	return volts;
}

BOOL CBudaDoc::TestForBigFFTs()
{
	// on fftsizes greater than 2meg, ask the user if he really
	// wants to do this
	long i,j;
	CString str;
	j=1;
	for (i=PS.ZeroFills; i>0; i--) j=j*2;
	PS.FFTSize = j*PS.TransientPoints;
	if(PS.FFTSize>4*1024*1024) {
		str.Format("This will be a %d point FFT\nIt will be slow.\nDo you really want to do this?",PS.FFTSize);
		if(pWnd->MessageBox(str,"Big FFT warning!",MB_YESNO) == IDNO) return TRUE;
	}
	return FALSE;
}

void CBudaDoc::UpdateFFTStatus(int s)
{
	// just keeps us up to date on long fft's
	CString str;
	str.Format("Fourier Transforming spectrum %d of %d",s+1,PS.SpectraAcquired);
	StatusBar->SetPaneText(0, str, TRUE);
}

/*
void CBudaDoc::GetMITable(int s, double* FTData)
{
	// now create and calculate the MI Table
	// note: TotalIonInt is undefined at this point, but it
	// must have a value to get the mass range in EstimateNoiseLevel
	// so define it = 0.
	CString str;
	TotalIonInt = 0;
	DS[s].TotalIonInt = 0;
	
	// get the noise level for the peak pick threshold
	DS[s].NoiseLevel=0;
	EstimateNoiseLevel(FTData, &(DS[s].NoiseLevel));
	DS[s].ScaleFactor = CurrentScaleFactor;

	DS[s].FITable = new struct FITableElement[MAXNUMPEAKS];
	FITable = DS[s].FITable;
	NumPeaks = 0;
	PeakPick();
	PS.PeakCount[s] = NumPeaks;

	// determine charge state
	if(sys->AutoLabelCharge) {
		//ChargeStateDetermination();
		//IsCSed = TRUE;
	}

	// and set the ion intensity used in calibration
	GetTotalIonInt();
	DS[s].TotalIonInt = TotalIonInt;
	str.Format("spectrum #%d: totalionint = %1.2f, noiselevel = %1.2f, numpeaks = %d\n",s,TotalIonInt,DS[s].NoiseLevel,NumPeaks);
	TRACE(str);
}
*/

BOOL CBudaDoc::SetCurrentMIFocus(int s)
{
	DumpToDiagFile("entering CBudaDoc::SetCurrentMIFocus(..)");

	// this function switches the spectrum focus of the MI table 
	// but doesn't update the focus of the transient, spectrum, etc.
	// This is for temporary work dealing with TIC's
	// remember to reset it when done using the function call:
	//		SetCurrentMIFocus(GetCurrentSpectrumFocus());
	// if you don't, the MI Table, TIT's, NoiseLevel, and etc will
	// all be out of sync with the raw data.

	if (s == GetCurrentSpectrumFocus()) return TRUE;
	if (s<0 || (int)s>PS.SpectraAcquired-1) {
		// not a valid spectrum number
		MessageBox("Error 23\nSpectrum index out of range","Memory error");
		DumpToDiagFile("leaving CBudaDoc::SetCurrentSpectrumFocus(..) - FALSE");
		return FALSE;
	}

	if(IsDefined) {
		//FITable = DS[s].FITable;
		//NoiseLevel = DS[s].NoiseLevel;
		//TotalIonInt = DS[s].TotalIonInt;
		//NumPeaks = PS.PeakCount[s];
	}


	DumpToDiagFile("leaving CBudaDoc::SetCurrentMIFocus(..) - TRUE");
	return TRUE;
}

int CBudaDoc::GetV6ChannelNumber(int index)
{
	// this function takes the index and figures out what
	// channel that would correspond to in the version 6
	// GenPS array

	//TRACE("FIRST_EXCITE_DISABLE = %d\n",FIRST_EXCITE_DISABLE);
	//TRACE("LAST_EXCITE_DISABLE = %d\n",LAST_EXCITE_DISABLE);
	//TRACE("FIRST_EXCITE_SOURCEE = %d\n",FIRST_EXCITE_SOURCE);
	//TRACE("LAST_EXCITE_DISABLE = %d\n",LAST_EXCITE_DISABLE);
	if(index>= FIRST_EXCITE_DISABLE && index<=LAST_EXCITE_DISABLE) {
		return EXCITE_DISABLE;
	}
	else if(index>= FIRST_EXCITE_SOURCE && index<=LAST_EXCITE_SOURCE) {
		return EXCITE_SOURCE;
	}
	else if(index>= FIRST_RF_SOURCE && index<=LAST_RF_SOURCE) {
		return RF_SOURCE;
	}
	else if(index>= FIRST_RF_PHASE && index<=LAST_RF_PHASE) {
		return RF_PHASE;
	}
	else if(index>= FIRST_IMP_TRIGGER && index<=LAST_IMP_TRIGGER) {
		return IMP_TRIGGER;
	}
	else if(index>= FIRST_SYN_TRIGGER && index<=LAST_SYN_TRIGGER) {
		return SYN_TRIGGER;
	}
	else if(index>= FIRST_ARB_TRIGGER && index<=LAST_ARB_TRIGGER) {
		return ARB_TRIGGER;
	}
	else if(index>= FIRST_VSP_TRIGGER && index<=LAST_VSP_TRIGGER) {
		return VSP_TRIGGER;
	}
	else if(index>= FIRST_LASER_PULSE && index<=LAST_LASER_PULSE) {
		return LASER_PULSE;
	}
	else if(index>= FIRST_QUADRF_GATE && index<=LAST_QUADRF_GATE) {
		return QUADRF_GATE;
	}
	else if(index>= FIRST_VALVE_PULSE && index<=LAST_VALVE_PULSE) {
		return VALVE_PULSE;
	}
	else if(index>= FIRST_SHUTTER_GATE && index<=LAST_SHUTTER_GATE) {
		return SHUTTER_GATE;
	}
	else if(index>= FIRST_USER5_PULSE && index<=LAST_USER5_PULSE) {
		return USER5_PULSE;
	}
	else if(index>= FIRST_USER6_PULSE && index<=LAST_USER6_PULSE) {
		return USER6_PULSE;
	}
	else if(index>= FIRST_USER7_PULSE && index<=LAST_USER7_PULSE) {
		return USER7_PULSE;
	}
	else if(index>= FIRST_USER8_PULSE && index<=LAST_USER8_PULSE) {
		return USER8_PULSE;
	}
	else if(index>= FIRST_XMIT1 && index<=LAST_XMIT1) {
		return DAC_XMIT1;
	}
	else if(index>= FIRST_XMIT2 && index<=LAST_XMIT2) {
		return DAC_XMIT2;
	}
	else if(index>= FIRST_RCVR1 && index<=LAST_RCVR1) {
		return DAC_RCVR1;
	}
	else if(index>= FIRST_RCVR2 && index<=LAST_RCVR2) {
		return DAC_RCVR2;
	}
	else if(index>= FIRST_TRAP1 && index<=LAST_TRAP1) {
		return DAC_TRAP1;
	}
	else if(index>= FIRST_TRAP2 && index<=LAST_TRAP2) {
		return DAC_TRAP2;
	}
	else if(index>= FIRST_EE && index<=LAST_EE) {
		return DAC_EE;
	}
	else if(index>= FIRST_GAIN && index<=LAST_GAIN) {
		return DAC_GAIN;
	}
	else if(index>= FIRST_USER1_DAC && index<=LAST_USER1_DAC) {
		return DAC_USER1;
	}
	else if(index>= FIRST_USER2_DAC && index<=LAST_USER2_DAC) {
		return DAC_USER2;
	}
	else if(index>= FIRST_USER3_DAC && index<=LAST_USER3_DAC) {
		return DAC_USER3;
	}
	else if(index>= FIRST_USER4_DAC && index<=LAST_USER4_DAC) {
		return DAC_USER4;
	}
	else if(index>= FIRST_USER5_DAC && index<=LAST_USER5_DAC) {
		return DAC_USER5;
	}
	else if(index>= FIRST_USER6_DAC && index<=LAST_USER6_DAC) {
		return DAC_USER6;
	}
	else if(index>= FIRST_USER7_DAC && index<=LAST_USER7_DAC) {
		return DAC_USER7;
	}
	else if(index>= FIRST_USER8_DAC && index<=LAST_USER8_DAC) {
		return DAC_USER8;
	}
	else if(index>= FIRST_USER9_DAC && index<=LAST_USER9_DAC) {
		return DAC_USER9;
	}
	else if(index>= FIRST_USER10_DAC && index<=LAST_USER10_DAC) {
		return DAC_USER10;
	}
	else if(index>= FIRST_USER11_DAC && index<=LAST_USER11_DAC) {
		return DAC_USER11;
	}
	else if(index>= FIRST_USER12_DAC && index<=LAST_USER12_DAC) {
		return DAC_USER12;
	}
	else if(index>= FIRST_USER13_DAC && index<=LAST_USER13_DAC) {
		return DAC_USER13;
	}
	else if(index>= FIRST_USER14_DAC && index<=LAST_USER14_DAC) {
		return DAC_USER14;
	}
	else if(index>= FIRST_USER15_DAC && index<=LAST_USER15_DAC) {
		return DAC_USER15;
	}
	else if(index>= FIRST_USER16_DAC && index<=LAST_USER16_DAC) {
		return DAC_USER16;
	}
	else if(index == LAST_USER16_DAC+1) {
		// the is the temp event
		return NO_EVENT_CHANNEL;
	}
	// if we get here, then we didn't have a valid index
	return -1;
}

void CBudaDoc::UpdateSystemPS()
{
	CopyPS(&(PS),&(Instrument->PS));
	return;

	/*
	if(pWnd->MessageBox("Overwrite the current system pulse sequence\n with the new pulse sequence?","Update Pulse Sequence?",MB_YESNO)==IDNO) {
		return;
	} else {
		Instrument->PS = PS;
	}
	*/
}

float CBudaDoc::EventTime(int index)
{
	return EventTime(&PS,index);
}

float CBudaDoc::EventTime(struct PS* pPS, int index)
{
	// this function returns the absolute time of the
	// start GenPS event given by index
	float start, finish;

	GetEventEdges(pPS, index, &start, &finish);
	return start;
}

int CBudaDoc::SwapEvents(struct PS* pPS, int event1, int event2)
{
	// first, swap event1 and event2 in the GenPS array in PS

	// Then, check GenPS for links to event1 or event2
	//  and if it finds any, shifts them 

	// Then, check if the DetectEvent is linked to 
	// either event1 or event2 and shifts it with them.

	// then if event1 and event2 are synthesizer events,
	// swap their respective cfp's

	struct PSDesc temp;

	temp = pPS->GenPS[event1];
	pPS->GenPS[event1] = pPS->GenPS[event2];
	pPS->GenPS[event2]=temp;

	int i, numlinksfound = 0;
	for (i=0; i<pPS->NumEvents; i++) {
		if(pPS->GenPS[i].Linking != NO_LINK) {
			if(pPS->GenPS[i].Link2Event == event1) {
				pPS->GenPS[i].Link2Event = event2;
				numlinksfound++;
			}
			if(pPS->GenPS[i].Link2Event == event2) {
				pPS->GenPS[i].Link2Event = event1;
				numlinksfound++;
			}
		}
	}
	if(pPS->DetectEvent == event1) {
		pPS->DetectEvent=event2;
		numlinksfound++;
	} else if(pPS->DetectEvent == event2) {
		pPS->DetectEvent=event1;
		numlinksfound++;
	}

	if(pPS->GenPS[event1].Channel == SYN_TRIGGER && pPS->GenPS[event2].Channel == SYN_TRIGGER) {
		// gotta swap cfp's
		// first, find out which cfps go with event1 and event2
		int Event1cfp=-1,Event2cfp=-1;
		int cfpctr=0;
		for(i=0; i<pPS->NumEvents; i++) {
			if(i==event1) Event1cfp=cfpctr;
			if(i==event2) Event2cfp=cfpctr;
			if(pPS->GenPS[i].Channel == SYN_TRIGGER) cfpctr++;
		}
		if(Event1cfp<0 || Event2cfp<0) {
			MessageBox("Error 78: Couldn't identify chirp function packet\nfor synthesizer event","Pulse Sequence Error!!");
			return -1;
		}
		// now swap them;
		struct ChirpPacket temp;
		temp = pPS->cfp[Event1cfp];
		pPS->cfp[Event1cfp]=pPS->cfp[Event2cfp];
		pPS->cfp[Event2cfp]=temp;
	}
	return numlinksfound;
}

void CBudaDoc::SortGenPS()
{
	SortGenPS(&PS);
}

void CBudaDoc::SortGenPS(struct PS* pPS)
{
	// sort genps in channel order removing no_events
	int i,j;
	
	// first move everything that isn't a no_event to the top.
	for(i=0; i<pPS->NumEvents; i++) {
		if(pPS->GenPS[i].Type != NO_EVENT) {
			// for each event that is not a NO_EVENT
			for(j=i; j>0; j--) {
				// bubble that event up until it finds the
				// the first non NO_EVENT event
				if(j==0) break;
				if(pPS->GenPS[j-1].Type != NO_EVENT) break;
				SwapEvents(pPS,j,j-1);
			}
		}
	}

	// before the channel sort, survey the GenPS, and set each NO_EVENT
	// to Channel #40, set each bogus event (channel#'s <0 or >40) = channel#40
	for(i=0; i<pPS->NumEvents; i++) {
		if(pPS->GenPS[i].Channel<0 || pPS->GenPS[i].Channel>39) pPS->GenPS[i].Channel=NO_CHANNEL;
		if(pPS->GenPS[i].Type == NO_EVENT) pPS->GenPS[i].Channel=NO_CHANNEL;
	}

	// second, sort the events in increasing order by channel
	for(i=0; i<pPS->NumEvents; i++) {
		if(pPS->GenPS[i].Channel < pPS->GenPS[i-1].Channel) {
			// if current event channel < last event channel
			for(j=i; j>0; j--) {
				// bubble the current event up until it is 
				// no longer < last event channel
				if(pPS->GenPS[j].Channel >= pPS->GenPS[j-1].Channel) break;
				SwapEvents(pPS,j,j-1);
			}
		}
	}

	// third, for each channel, sort the events by time
	int currentchannel = 0;
	int firstindex = 0;
	for(i=0; i<pPS->NumEvents; i++) {
		if(pPS->GenPS[i].Channel != currentchannel) {
			// if we have come to a channel boundary, 
			// reset current channel and increment i loop
			currentchannel = pPS->GenPS[i].Channel;
			firstindex = i;
			continue;
		}

		if(EventTime(pPS,i) < EventTime(pPS,i-1)) {
			// if current event time < last event event time
			for(j=i; j>firstindex+1; j--) {
				// bubble the current event up until it is 
				// no longer < last event time or until it reaches
				// channel boundary
				if(EventTime(pPS,j) >= EventTime(pPS,j-1)) break;
				SwapEvents(pPS,j,j-1);
			}
		}
	}

	// finally, run through and count the number of events
	// that are not No_Event and reset pPS->NumEvents
	int Events = 0;
	for (i=0; i<MAX_PS_EVENTS; i++) {
		if(pPS->GenPS[i].Type != NO_EVENT) Events++;
	}

	// ok, now, if there is some problem with NumEvents, this
	// should fix it.
	if(Events>pPS->NumEvents) {
		pPS->NumEvents = TOTAL_PS_EVENTS;
		SortGenPS();
	} else {
		pPS->NumEvents = Events;
	}
}


void CBudaDoc::AddToFFTQueue(int s)
{
	// this function creates a new FFTQueueEntry, adds
	// this entry to the FFTQueue, and increments FFTThread's FFTs in queue
	long i;
	struct FFTQueueEntry* qe;

	// create and fill new QueueEntry
	qe = new struct FFTQueueEntry;

	// define event handle and notification hwnd 
	// of view that called this fft
	qe->m_hFFTDone=m_hFFTDone;
	qe->m_hwndNotifyFFTDone = m_hwndNotifyFFTDone;

	qe->pPS = &(PS);
	qe->pDoc = this;

	PS.WindowFunction = sys->WindowFunction;
	PS.ZeroFills = sys->ZeroFills;
	if(PS.ZeroFills < 0) {
		// autocalculate zerofills
		PS.ZeroFills = 2;
		if(PS.TransientPoints > 500000) PS.ZeroFills = 1;
		if(PS.TransientPoints > 1000000) PS.ZeroFills = 0;
	}

	// calculate fftsize so that I can create FreqArray
	PS.FFTSize =  PS.TransientPoints;
	long closestpt = 0;
	if(PS.TruncateTransient) {
		for(i=0; i<30; i++) {
			if(i!=0) closestpt = closestpt*2;
			else closestpt = 2;
			if(closestpt > PS.TruncatePoints) break;
		}
		for(i=0; i<24; i++) {
		}
		// FFTSize is next power of 2 greater than truncatepts
		PS.FFTSize = closestpt;
	}
	for(i=0; i<PS.ZeroFills; i++) PS.FFTSize*=2;
	PS.SpectrumPoints = PS.FFTSize/2;

	// create FreqArray to size for magnitude mode spectrum
	if(FreqArray != 0) delete [] FreqArray;
	FreqArray = new float[PS.SpectrumPoints+1];
	qe->FTData = FreqArray; // this is a pointer to a pointer

	if(PS.Spec[s].FITable!=0) delete [] PS.Spec[s].FITable;
	PS.Spec[s].FITable = new FITableElement[MAXNUMPEAKS];
	
	qe->DSIndex = s;
	qe->HighMassPt = MassToPt(PS.DetectHighMass);
	qe->LowMassPt = MassToPt(PS.DetectLowMass);
	qe->EndofQueue = TRUE;

	// remember: FFTThread only reacts to changes in 
	//			 FFTThrdInfo when FFTsInQueue > 0  and when the 
	//			 StartFFT event is triggered 

	// if the thread is busy add it to the queue
	struct FFTQueueEntry* pQ;
TRACE("AddToFFTQueue: Adding new FFT to queue\n");
	if(pFFT->FFTThrdInfo.FFTsInQueue > 0) {
		pFFT->Lock();
		pQ = pFFT->FFTThrdInfo.CurrentFFT;
		for(i=pFFT->FFTThrdInfo.FFTsInQueue; i>1; i--) {
			pQ=pQ->NextFFT;
		}
		pQ->NextFFT = qe;
		pQ->EndofQueue = FALSE;
		pFFT->FFTThrdInfo.FFTsInQueue++;
		pFFT->Unlock();
		//SetEvent(pFFT->FFTThrdInfo.m_hStartFFT);
	} else { // if not busy get it busy
		pFFT->Lock();
		pFFT->FFTThrdInfo.CurrentFFT = qe;
		pFFT->FFTThrdInfo.FFTsInQueue = 1;
		pFFT->Unlock();
		SetEvent(pFFT->FFTThrdInfo.m_hStartFFT);
	}
	CalculatingFFT = TRUE;

}  // end AddToFFTQueue()

void CBudaDoc::InitializePS()
{
//	CopyPS(&(Instrument->PS),&(PS));

	// this function defines the default PS structure
	// for the system.  It is expected that this structure
	// will be overwritten by ReadDefaultSetupFile(), but 
	// if there is an error on reading, these values will
	// remain

	// the CInstrument::PS structure is the one that will
	// be used for data acquisition.  It will be different
	// from the CBudaDoc::PS's that are around, but the
	// user will be able to copy the CBudaDoc::PS to 
	// this one for data acquisition

	CString str;
	int i,j;
	CString SystemName = "Pete's System";
	CString SystemSerialNumber = "1234";
	CString UserID = "Pete!";

	PS.FileType = SETUP_FILE;
	PS.MajorVersion = -1;
	PS.MinorVersion = 0;

	//// fixed length strings ////
	//fixed length string SystemName
	for(i=0; i<128; i++) {
		if(i>=SystemName.GetLength()) break;;
		PS.SystemName[i]=SystemName.GetAt(i);
	}
	PS.SystemName[i]='\0';
	//fixed length string SystemSerialNumber
	for(i=0; i<32; i++) {
		if(i>=SystemSerialNumber.GetLength()) break;;
		PS.SystemSerialNumber[i]=SystemSerialNumber.GetAt(i);
	}
	PS.SystemSerialNumber[i]='\0';
	//fixed length string UserID
	for(i=0; i<8; i++) {
		if(i>=UserID.GetLength()) break;;
		PS.UserID[i]=UserID.GetAt(i);
	}
	PS.UserID[i]='\0';
	//fixed length string DataFileName
/*	for(i=0; i<64; i++) {
		if(i>=sys->DefaultSetupFile->GetLength()) break;;
		PS.DataFileName[i]=sys->DefaultSetupFile->GetAt(i);
	}
*/
	PS.DataFileName[i]='\0';
	str = "time acquired init'd to none  ";
	//fixed length string TimeAcquired
	for(i=0; i<40; i++) {
		if(i>=str.GetLength()) break;;
		PS.TimeAcquired[i]=str.GetAt(i);
	}
	PS.TimeAcquired[i]='\0';
	str = "time processed init'd to none  ";
	//fixed length string TimeAcquired
	for(i=0; i<32; i++) {
		if(i>=str.GetLength()) break;;
		PS.TimeProcessed[i]=str.GetAt(i);
	}
	PS.TimeAcquired[i]='\0';
	str = "No Experiment Defined - Using setup defaults";
	//fixed length string ExperimentMemo
	for(i=0; i<60; i++) {
		if(i>=str.GetLength()) break;;
		PS.ExperimentMemo[i]=str.GetAt(i);
	}
	PS.ExperimentMemo[i]='\0';
	str = "No Experiment Defined - Using setup defaults";
	//fixed length string ExperimentNotes
	for(i=0; i<600; i++) {
		if(i>=str.GetLength()) break;;
		PS.ExperimentNotes[i]=str.GetAt(i);
	}
	PS.ExperimentNotes[i]='\0';
	//// end fixed length strings ////

	PS.HiResMALDI = 0;
	PS.HiResESI = 0;
	PS.DetectMode = DirectMode;
	PS.ADCGain = 0;
	PS.DetectEvent = 0;
	PS.DetectLowMass = 100;
	PS.DetectHighMass = 10000;
	PS.MonitorMass = 1000;
	PS.MixerFrequency = 100000.0;
	PS.AdcRate = 1000000.0;
	PS.ScanMode = 0;
	PS.TransientPoints = 16*1024;
	PS.SpectraRequested = 1;
	PS.SpectraAcquired = 1;
	PS.CurrentSpectrum = 0;
	PS.ScansPerSpectrum = 1;
	PS.ScansRequested = 1;
	PS.TriggerMode = 0;
	PS.CalibrationVersion = -1;
	PS.A = 100000000;
	PS.B = 1;
	PS.HighMassShift = 0;
	PS.TrapShift = 10;
	PS.IntensityShift = 0.01;
	PS.PeakMatching = 0;
	PS.PeakMatchFactor = 0;
	str = "time acquired init'd to none  ";
	for(i=0; i<MAX_REF_PEAK; i++) {
		PS.RefPeak[i].ActualMass = 1000;
		PS.RefPeak[i].MeasuredFrequency = 100000;
		//fixed length string TimeEntered
		for(j=0; j<40; j++) {
		if(j>=str.GetLength()) break;;
			PS.RefPeak[i].TimeEntered[j]=str.GetAt(j);
		}
		PS.RefPeak[i].TimeEntered[j]='\0';
		PS.RefPeak[i].TotalIntensity = 1;
		PS.RefPeak[i].TrapVoltage = (float)0.46066;
	}
	PS.SpectrometerType = 0;
	PS.DCRemoval = 1;
	PS.WindowFunction = 1;
	PS.ZeroFills = 0;
	PS.TruncateTransient = 0;
	PS.TruncatePoints = 0;
	PS.TruncateOffset = 0;
	PS.FFTSize = PS.TransientPoints;
	PS.SpectrumPoints = PS.FFTSize/2;
	PS.FrequencyDataType = 0; // what is this
	PS.ZoomFactor = 1;
	PS.ZoomLowMass = 1000;
	PS.ZoomHighMass = 1010;
	PS.ZoomFrequencyShift = 1;
	PS.ZoomTransientPoints = 1024*4;
	PS.ZoomPointShift = 1;
	PS.TransientIsStored = 0;
	PS.FrequencyIsStored = 0;
	PS.MITableIsStored = 0;
	PS.NumEvents = 1;

	// fill in ChirpPacket, DacPacket, ArbPacket, GenPS
	// first, create the arrays
	if(PS.cfp != 0) {
		delete [] PS.cfp;
		PS.cfp = 0;
	}
	if(PS.dac != 0) {
		delete [] PS.dac;
		PS.dac = 0;
	}
	if(PS.arb != 0) {
		delete [] PS.arb;
		PS.arb = 0;
	}
	if(PS.GenPS != 0) {
		delete [] PS.GenPS;
		PS.GenPS = 0;
	}
	PS.cfp = new struct ChirpPacket[MAX_SYN_TRIGGERS];
	PS.dac = new struct DacPacket[MAX_DACS];
	PS.arb = new struct ArbPacket[MAX_ARB_TRIGGERS];
	PS.GenPS = new struct PSDesc[MAX_PS_EVENTS];

	for(i=0; i<MAX_SYN_TRIGGERS; i++) {
		PS.cfp[i].Type = CFP_NONE;
		PS.cfp[i].Amplitude = 0;
		PS.cfp[i].StartMass = 1000;
		PS.cfp[i].EndMass = 2000;
	}

	for(i=0; i<MAX_ARB_TRIGGERS; i++) {
		PS.arb[i].WindowLowMass = 400;
		PS.arb[i].WindowHighMass = 1000;
		PS.arb[i].DacFixed = 0;
		PS.arb[i].SamplesFixed = 0;
		PS.arb[i].FilterFixed = 0;
		PS.arb[i].AmplitudeFixed = 0;
		PS.arb[i].DacRate = 100000;
		PS.arb[i].Samples = 16*1024;
		PS.arb[i].Filter = 0;
		PS.arb[i].Amplitude = 0;
		PS.arb[i].Setup = 0;
		PS.arb[i].PacketSize = 0;
		PS.arb[i].PacketSpacing = 1;
		for(j=0; j<MAX_ARB_EVENTS; j++) {
			PS.arb[i].Event[j].Type = ARB_NONE;
			PS.arb[i].Event[j].fit = 0;
			PS.arb[i].Event[j].MassA = PS.A;
			PS.arb[i].Event[j].MassB = PS.B;
			PS.arb[i].Event[j].Amplitude = 0;
			PS.arb[i].Event[j].reserved = 0;
		}
	}

	for(i=0; i<MAX_DACS; i++) {
		PS.dac[i].Units = 0;
		PS.dac[i].MaxValue = 10;
		PS.dac[i].MinValue = -10;
		PS.dac[i].MaxDac = 2048;
		PS.dac[i].MinDac = -2047;
		PS.dac[i].MaxGraph = 10;
		PS.dac[i].MinGraph = -10;
		PS.dac[i].DacAvailable = 0;
	}

	for(i=0; i<MAX_PS_EVENTS; i++) {
		PS.GenPS[i].Channel = NO_CHANNEL; // the NO_CHANNEL channel
		PS.GenPS[i].Type = NO_EVENT;
		PS.GenPS[i].Linking = NO_LINK;
		PS.GenPS[i].Link2Event = NO_CHANNEL;
		PS.GenPS[i].Delay = 0;
		PS.GenPS[i].Width = 0;
		PS.GenPS[i].Value = 0;
	}

	// create struct Spectrum and fill it in
	if(PS.Spec != 0) {
		delete [] PS.Spec;
		PS.Spec = 0;
	}
	PS.Spec = new struct Spectrum[PS.SpectraAcquired];
	for(i=0; i<PS.SpectraAcquired; i++) {
		PS.Spec[i].ElapsedTime = 0;
		PS.Spec[i].IsMIed = FALSE;
		PS.Spec[i].IsFFTed = FALSE;
		PS.Spec[i].NoiseLevel = 0;
		PS.Spec[i].NumPeaks = 0;
		PS.Spec[i].ScaleFactor = 100;
		PS.Spec[i].TotalIonInt = 0;
		PS.Spec[i].FITable = 0;
		PS.Spec[i].TArrayFileOffset = 0;
	}
	//////////////////////////////////////////////////////////
	// finished creating the instrument's default PS structure
	// hopefully, these values will never be seen
}

void CBudaDoc::OnCloseDocument() 
{
	// check to see if we are still fft'ing.  if so, do not close document.
	if(CalculatingFFT) return;

	// if running an experiment, do not close document
	if(pApp->Instrument.Running && pApp->Instrument.pDoc == this) return;
	
	CDocument::OnCloseDocument();
}

void CBudaDoc::ShiftDetectEvent()
{

}

double CBudaDoc::CyclotronFrequency(double ResFreq, float Vtrap, double TotInt, long NumScans, int ADCgain, int direction)
{
	// this function does the frequency corrections for:
	//	- trapping shift
	//  - space charge shift (with TIT corrected for numscans and gain).
	//  - high mass shift
	//
	// direction = 1 means add corrections to frequency (cfreq>freq) 
	//		used in mass calibration
	//		used to get cyclotron frequencies from resonance frequencies
	//
	// direction = -1 means subtract corrections from frequency
	//		not used very often - just in plotting	
	//		used to get resonance frequencies from cyclotron frequencies
	double Trap8=0,Int8,Freq,CyclotronFrequency = 0;
#ifdef DEBUG_VIEWRANGE
	if(0) {
		TRACE("----- cyclotron frequency ---\n");
		TRACE("resonance frequency = %5.5f\n",ResFreq);
		TRACE("Vtrap = %5.5f\n",Vtrap);
		TRACE("TotIonInt = %5.5f\n",TotInt);
		TRACE("NumScans = %d\n",NumScans);
		TRACE("ADCGain = %d\n",ADCgain);
		TRACE("Direction = %d\n",direction);
		TRACE("----- end cyclotron frequency ----\n");
	}
#endif
	Trap8=double(fabs(Vtrap));
	if(NumScans < 1 || ADCgain < 1) Int8 = 0;
	else Int8 = Trap8 * double(TotInt)/double(NumScans)/double(ADCgain);

	if(PS.CalibrationVersion == 6) {
		CyclotronFrequency = ResFreq;
		CyclotronFrequency += direction*PS.TrapShift*Trap8;
		CyclotronFrequency += direction*PS.IntensityShift*Int8;
	}
	else if (PS.CalibrationVersion == 7) {
		if(ResFreq > 0) Freq = ResFreq;
		else Freq = 1.0;

		CyclotronFrequency = ResFreq;
		CyclotronFrequency += direction*PS.TrapShift*Trap8;
		CyclotronFrequency += direction*(PS.IntensityShift*Int8)/sqrt(Freq);
		CyclotronFrequency += direction*PS.HighMassShift/Freq;
	} else {
		if(ResFreq > 0) Freq = ResFreq;
		else Freq = 1;

		CyclotronFrequency = ResFreq;
		CyclotronFrequency += direction*PS.TrapShift*Trap8;
		CyclotronFrequency += direction*PS.IntensityShift*Int8;
		CyclotronFrequency += direction*PS.HighMassShift/Freq;
	}
	return CyclotronFrequency;
}

void CBudaDoc::CopyPS(struct PS* from, struct PS* to)
{
	to->FileType = from->FileType;
	to->MajorVersion = from->MajorVersion;
	to->MinorVersion = from->MinorVersion;

	int i,j;
	CString str;

	for(i=0; i<128; i++) to->SystemName[i] =  from->SystemName[i];
	for(i=0; i<32; i++) to->SystemSerialNumber[i] =  from->SystemSerialNumber[i];
	for(i=0; i<8; i++) to->UserID[i] =  from->UserID[i];
	for(i=0; i<64; i++) to->DataFileName[i] =  from->DataFileName[i];
	for(i=0; i<40; i++) to->TimeAcquired[i] =  from->TimeAcquired[i];
	for(i=0; i<32; i++) to->TimeProcessed[i] =  from->TimeProcessed[i];
	for(i=0; i<60; i++) to->ExperimentMemo[i] =  from->ExperimentMemo[i];
	for(i=0; i<60; i++) to->ExperimentNotes[i] =  from->ExperimentNotes[i];

	to->HiResMALDI = from->HiResMALDI;
	to->HiResESI = from->HiResESI;
	to->DetectMode = from->DetectMode;
	to->ADCGain = from->ADCGain;
	to->DetectEvent = from->DetectEvent;
	to->DetectLowMass = from->DetectLowMass;
	to->DetectHighMass = from->DetectHighMass;
	to->MonitorMass = from->MonitorMass;
	to->MixerFrequency = from->MixerFrequency;
	to->AdcRate = from->AdcRate;
	to->ScanMode = from->ScanMode;
	to->TransientPoints = from->TransientPoints;
	to->SpectraRequested = from->SpectraRequested;
	to->SpectraAcquired = from->SpectraAcquired;
	to->CurrentSpectrum = from->CurrentSpectrum;
	to->ScansPerSpectrum = from->ScansPerSpectrum;
	to->ScansRequested = from->ScansRequested;
	to->TriggerMode = from->TriggerMode;
	to->CalibrationVersion = from->CalibrationVersion;
	to->A = from->A;
	to->B = from->B;
	to->HighMassShift = from->HighMassShift;
	to->TrapShift = from->TrapShift;
	to->IntensityShift = from->IntensityShift;
	to->PeakMatching = from->PeakMatching;
	to->PeakMatchFactor = from->PeakMatchFactor;
	str = "time acquired init'd to none  ";
	for(i=0; i<MAX_REF_PEAK; i++) {
		to->RefPeak[i].ActualMass = from->RefPeak[i].ActualMass;
		to->RefPeak[i].MeasuredFrequency = from->RefPeak[i].MeasuredFrequency;
		for(j=0; j<40; j++) to->RefPeak[i].TimeEntered[j] = from->RefPeak[i].TimeEntered[j];
		to->RefPeak[i].TotalIntensity = from->RefPeak[i].TotalIntensity;
		to->RefPeak[i].TrapVoltage = from->RefPeak[i].TrapVoltage;
	}
	to->SpectrometerType = from->SpectrometerType;
	to->DCRemoval = from->DCRemoval;
	to->WindowFunction = from->WindowFunction;
	to->ZeroFills = from->ZeroFills;
	to->TruncateTransient = from->TruncateTransient;
	to->TruncatePoints = from->TruncatePoints;
	to->TruncateOffset = from->TruncateOffset;
	to->FFTSize = from->FFTSize;
	to->SpectrumPoints = from->SpectrumPoints;
	to->FrequencyDataType = from->FrequencyDataType; // what is this
	to->ZoomFactor = from->ZoomFactor;
	to->ZoomLowMass = from->ZoomLowMass;
	to->ZoomHighMass = from->ZoomHighMass;
	to->ZoomFrequencyShift = from->ZoomFrequencyShift;
	to->ZoomTransientPoints = from->ZoomTransientPoints;
	to->ZoomPointShift = from->ZoomPointShift;
	to->TransientIsStored = from->TransientIsStored;
	to->FrequencyIsStored = from->FrequencyIsStored;
	to->MITableIsStored = from->MITableIsStored;
	to->NumEvents = from->NumEvents;

	// fill in ChirpPacket, DacPacket, ArbPacket, GenPS
	// first, create the arrays
	if(to->cfp != 0) {
		delete [] to->cfp;
		to->cfp = 0;
	}
	if(to->dac != 0) {
		delete [] to->dac;
		to->dac = 0;
	}
	if(to->arb != 0) {
		delete [] to->arb;
		to->arb = 0;
	}
	if(to->GenPS != 0) {
		delete [] to->GenPS;
		to->GenPS = 0;
	}
	if(to->Spec != 0) {
		delete [] to->Spec;
		to->Spec = 0;
	}
	to->Spec = new struct Spectrum[to->SpectraAcquired];
	to->cfp = new struct ChirpPacket[MAX_SYN_TRIGGERS];
	to->dac = new struct DacPacket[MAX_DACS];
	to->arb = new struct ArbPacket[MAX_ARB_TRIGGERS];
	to->GenPS = new struct PSDesc[MAX_PS_EVENTS];

	for(i=0; i<MAX_SYN_TRIGGERS; i++) {
		to->cfp[i] = from->cfp[i];
	}

	for(i=0; i<MAX_ARB_TRIGGERS; i++) {
		to->arb[i] = from->arb[i];
	}

	for(i=0; i<MAX_DACS; i++) {
		to->dac[i] = from->dac[i];
	}

	for(i=0; i<MAX_PS_EVENTS; i++) {
		to->GenPS[i] = from->GenPS[i];
	}

	for(i=0; i<to->SpectraAcquired; i++) {
		to->Spec[i] = from->Spec[i];
	}
}

void CBudaDoc::OnRunExpButton() 
{
	// if already running, ignore input
	if(pApp->Instrument.Running) return;

	// get spectrarequested, scansperspectra.
	CAcquireScansDlg dlg;
	dlg.pDoc = this;
	if(dlg.DoModal() != IDOK) return;
	PS.SpectraRequested = dlg.m_NumSpectra;
	PS.ScansRequested = dlg.m_NumScans;

	// setup info.
	pApp->Instrument.TuneMode = FALSE;
	pApp->Instrument.pDoc = this;

	// clear timearrays
	if(TDisDirty) {
		if(TimeArray != 0) delete [] TimeArray;
		TimeArray = new long[PS.TransientPoints];
	}
	PS.TransientIsStored = 32;
	long i;
	for(i=0; i<PS.TransientPoints; i++) TimeArray[i] = 0;
	

	// start the run.
	pApp->Instrument.ThreadInfo.pMainWnd = pApp->GetMainWnd();
	pApp->Instrument.ThreadInfo.InstrumentCommands = RUN_EXPERIMENT;
	SetEvent(pApp->Instrument.ThreadInfo.m_hProcessCommand);
}

void CBudaDoc::OnUpdateRunExpButton(CCmdUI* pCmdUI) 
{
	if(pApp->Instrument.ThreadInfo.ThreadStatus != IDLE) {
		pCmdUI->Enable(FALSE);
		return;
	}
	if(pApp->Instrument.InstrumentAvailable) pCmdUI->Enable(TRUE);
	else pCmdUI->Enable(FALSE);
}

void CBudaDoc::OnTuneMode() 
{
	pApp->Instrument.TuneMode = TRUE;
	pApp->Instrument.pDoc = this;
	pApp->Instrument.ThreadInfo.InstrumentCommands = RUN_EXPERIMENT;
	SetEvent(pApp->Instrument.ThreadInfo.m_hProcessCommand);
}

void CBudaDoc::OnUpdateTuneMode(CCmdUI* pCmdUI) 
{
	if(pApp->Instrument.ThreadInfo.ThreadStatus != IDLE) {
		pCmdUI->Enable(FALSE);
		return;
	}
	if(pApp->Instrument.InstrumentAvailable) pCmdUI->Enable(TRUE);
	else pCmdUI->Enable(FALSE);
}

void CBudaDoc::OnUpdateSetupButton(CCmdUI* pCmdUI) 
{
	if(SetupViewOpen) pCmdUI->Enable(FALSE);
	else pCmdUI->Enable(TRUE);	
}

void CBudaDoc::ExperimentDone()
{
	// Instrument calls here when it is done running an experiment
	//
	// this function then just updates the views
	UpdateAllViews(NULL,NULL,NULL);
}

void CBudaDoc::RecalcEventTime(struct PS* pPS, int Event, int Linking, int Link2Event)
{
	// this function changes linking status of an event and recalculates the delay value.
	//
	// if Linking == NO_LINK this recalculates the event time for Event
	// then converts the event to an unlinked event, and sets a new absolute event time.
	//		note: this ignores Link2Event

	// if Linking == LINKxxx this function recalculates the event time
	// and then converts the event to a linked event and sets a new relative event time.

	float start1, finish1;
	float start2, finish2;

	GetEventEdges(pPS, Event, &start1, &finish1);

	if(Linking == NO_LINK) {
		pPS->GenPS[Event].Link2Event = 0;
		pPS->GenPS[Event].Linking = NO_LINK;
		pPS->GenPS[Event].Delay = start1;
		return;
	}

	GetEventEdges(pPS, Link2Event, &start2, &finish2);
	if(Linking == LINK_BEG) {
		pPS->GenPS[Event].Delay = start1 - start2;
		pPS->GenPS[Event].Link2Event = Link2Event;
		pPS->GenPS[Event].Linking = LINK_BEG;
		return;
	}

	if(Linking == LINK_END) {
		pPS->GenPS[Event].Delay = start1 - finish2;
		pPS->GenPS[Event].Link2Event = Link2Event;
		pPS->GenPS[Event].Linking = LINK_END;
		return;
	}
}

void CBudaDoc::RecalcEventTime(struct PS* pPS, int Event, int Linking)
{
	RecalcEventTime(pPS, Event,Linking,NULL);
}

CString CBudaDoc::GetLinkedToStr(struct PS* pPS, int Event)
{
	CString returnstr;
	if(Event>PS.NumEvents || Event<0) {
		MessageBox("Error 89: Cannot link to an undefined event","Error: Event invalid");
		return "";
	}

	int i, EventCount=0;
	int Link2Event = pPS->GenPS[Event].Link2Event;
	if(Link2Event>PS.NumEvents || Link2Event<0) {
		MessageBox("Error 88: Cannot link to an undefined event","Error: Event invalid");
		return "";
	}


	for(i=Link2Event; i>0; i--) {
		if(pPS->GenPS[Link2Event].Channel == pPS->GenPS[i].Channel) {
			EventCount++;
			continue;
		} else {
			break;
		}
	}

	returnstr.Format("%s [%d]",sys->ChannelNames[pPS->GenPS[Link2Event].Channel],EventCount);

	return returnstr;
}


BOOL CBudaDoc::ReadSetupFile(CString FileName)
{
	BOOL test;
	CFile setfile;
	CString errorstr,pathname;
	CString filestr = "File Open Error";
	FileWasRead = FALSE;

	// first: get the filename
	if(FileName == "") {
		// use the CFileDialog to get the name of the file
		CFileDialog *dlg;
		dlg = new CFileDialog(TRUE,"*.set","MALDI.set",OFN_HIDEREADONLY | OFN_LONGNAMES,"*.cal",NULL);
		dlg->m_ofn.lpstrFilter = "Setup Files (*.set)\0*.set\0All Files (*.*)\0*.*\0";
		dlg->m_ofn.lpstrInitialDir = pApp->ProgramDirectory;

		if(dlg->DoModal() == IDCANCEL) return FALSE;

		pathname = dlg->GetPathName();
		delete dlg;
	} else {
		pathname = pApp->ProgramDirectory + "\\" + FileName;
	}

	// second: open the file and try to read it
	test = setfile.Open(pathname,  CFile::modeRead, NULL);
	if (!test) {
		errorstr.Format("Couldn't open Setup file \n %s",pathname);
		MessageBox(errorstr,filestr);
		return FALSE;
	}

	CArchive ar(&setfile,CArchive::load,4096,NULL);

	long FileType, MajorVersion, MinorVersion;
	int inpts = 0;
	// first read in version info
	inpts += ar.Read(&FileType,4);
	inpts += ar.Read(&MajorVersion,4);
	inpts += ar.Read(&MinorVersion,4);

	if(inpts<12) FileError("Error Reading Version Number");
	// then reset file to point to beginning of file
	setfile.SeekToBegin();
		
	// third: read in the particular version
	switch(MajorVersion) {
	case 6:
		ReadVersion6File(&setfile,1);
		break;
		// end of loading version 6
	case 8:
		ReadVersion8File(&setfile,1);
		break;
		// end of loading version 8
	default:
		FileError("Cannot Recognize Version Number");
	}
	ar.Close();
	setfile.Close();
	sys = GetSys(); 
	return TRUE;
}

void CBudaDoc::SaveSetupFile()
{
	BOOL test;
	CFile setfile;
	CString errorstr,pathname;
	CString filestr = "File Open Error";

	// first: get the filename
	// use the CFileDialog to get the name of the file
	CFileDialog *dlg;
	dlg = new CFileDialog(TRUE,"*.set","MALDI.set",OFN_HIDEREADONLY | OFN_LONGNAMES,"*.cal",NULL);
	dlg->m_ofn.lpstrFilter = "Setup Files (*.set)\0*.set\0All Files (*.*)\0*.*\0";
	dlg->m_ofn.lpstrInitialDir = pApp->ProgramDirectory;

	if(dlg->DoModal() == IDCANCEL) return;
	pathname = dlg->GetPathName();
	delete dlg;

	// second: open the file and try to read it
	test = setfile.Open(pathname,  CFile::modeCreate | CFile::modeWrite, NULL);
	if (!test) {
		errorstr.Format("Couldn't open Setup file \n %s",pathname);
		MessageBox(errorstr,filestr);
	}

	CArchive ar(&setfile,CArchive::store,4096,NULL);

	StoreVersion8File(&setfile,1);

	ar.Close();
	setfile.Close();
}

void CBudaDoc::DuplicateDoc(CBudaDoc *pDoc)
{
	long i;
	// copies the data elements of pDoc to this document
	ADCGain = pDoc->ADCGain;
	AssignChargStates = pDoc->AssignChargStates;
	AutoDetectThreshold = pDoc->AutoDetectThreshold;
	ThresholdMethod = pDoc->ThresholdMethod;
	AutoSave = pDoc->AutoSave;
	CalculatingFFT = pDoc->CalculatingFFT;
	CurrentFreqArrayPtr = pDoc->CurrentFreqArrayPtr;
	CurrentScaleFactor = pDoc->CurrentScaleFactor;
	CurrentSpectrumFocus = pDoc->CurrentSpectrumFocus;
	for(i=0; i<MAX_PS_EVENTS; i++)
		CyclicLoopCheck[i] = pDoc->CyclicLoopCheck[i];
	DataPointSize = pDoc->DataPointSize;
	DefaultFileNameType = pDoc->DefaultFileNameType;
	DefaultSaveDirectory = pDoc->DefaultSaveDirectory;
	DisplayParms = pDoc->DisplayParms;
	duration = pDoc->duration;
	EnableStopFFTBtn = pDoc->EnableStopFFTBtn;
	FFTIsDirty = pDoc->FFTIsDirty;
	FFTThreadOn = pDoc->FFTThreadOn;
	FileWasRead = pDoc->FileWasRead;
	//if(FreqArray != 0) delete [] FreqArray;
	//FreqArray = new float[PS.SpectrumPoints);
	//for (i=0; i<PS.SpectrumPoints; i++)
	//	FreqArray[i] = pDoc->FreqArray[i];
	//if(FITable != 0) delete [] FITable;
	//FITable = new FITableElement[MAXNUMPEAKS];
	FreqArray = 0;
	FITable = 0;
	HaltFFTFlag = pDoc->HaltFFTFlag;
	IFFTFlag = pDoc->IFFTFlag;
	Instrument = pDoc->Instrument;
	IsCSed = pDoc->IsCSed;
	IsDefined = pDoc->IsDefined;
	//IsFFTed = pDoc->IsFFTed;
	//IsMIed = pDoc->IsMIed;
	IsPeakMatching = pDoc->IsPeakMatching;
	IsPeakPicked = pDoc->IsPeakPicked;
	m_hFFTDone = pDoc->m_hFFTDone;
	m_hwndNotifyFFTDone = pDoc->m_hwndNotifyFFTDone;
	MaxIntensity = pDoc->MaxIntensity;
	order = pDoc->order;
	pApp = pDoc->pApp;
	PeakMatchFactor = pDoc->PeakMatchFactor;
	PeakMatching = pDoc->PeakMatching;
	PeakPickFlag = pDoc->PeakPickFlag;
	pFFT = pDoc->pFFT;
	PS = pDoc->PS;
	//pSetupView = pDoc->pSetupView;
	PSisDirty = pDoc->PSisDirty;
	//pWnd = pDoc->pWnd;
	RefPeak = pDoc->RefPeak;
	SaveTruncatedTransient = pDoc->SaveTruncatedTransient;
	ScaleFactor = pDoc->ScaleFactor;
	SetupOptionsInitPage = pDoc->SetupOptionsInitPage;
	SetupViewOpen = pDoc->SetupViewOpen;
	SICMass = pDoc->SICMass;
	SICmode = pDoc->SICmode;
	SICWidth = pDoc->SICWidth;
	SingleFFTFlag = pDoc->SingleFFTFlag;
	SortIncreasing = pDoc->SortIncreasing;
	SortOnProperty = pDoc->SortOnProperty;
	StatusBar = pDoc->StatusBar;
	sys = pDoc->sys;
	TDisDirty = pDoc->TDisDirty;
	temp = pDoc->temp;
	TransientIsDefined = pDoc->TransientIsDefined;
	TrapVoltsOnDetect = pDoc->TrapVoltsOnDetect;
	//TwoDDataMenu = pDoc->TwoDDataMenu;
	TwoDFFTDone = pDoc->TwoDFFTDone;
	UserSetupFile = pDoc->UserSetupFile;
}

void CBudaDoc::OnExperimentDone()
{
	// this function is called when the instrument has triggered
	// a WM_NEWDATAREADY message.  It has already overwritten this document's 
	// data and we only need to (re)fft, (re)display

	// set IsFFTed and IsMIed FALSE
	if(!IsDefined) return;
	if (!TransientIsDefined) return;
	if(PS.SpectraAcquired <= 0) return;;

	SetIsFFTed(FALSE);
	SetIsMIed(FALSE);


	// if no BudaView open, open one
	BOOL Iview = FALSE;
	POSITION pos = GetFirstViewPosition();   
	while (pos != NULL)   {
      CView* pView = GetNextView(pos);      
	  if(pView->IsKindOf(RUNTIME_CLASS( CBudaView ))) Iview = TRUE;
	}   	

	if(!Iview) {
		CMultiDocTemplate* templ = pApp->m_pBudaViewTemplate;
		CChildFrame* pFrame = (CChildFrame*)templ->CreateNewFrame(this,NULL);
		templ->InitialUpdateFrame(pFrame,this,TRUE);
	}

	// finally, update all views
	UpdateAllViews(NULL,NULL,NULL);
}

long CBudaDoc::ReadTimeArray(int s,CFile* arf)
{
	short* sTArray=0;
	long* lTArray=0;
	long i;
	unsigned long inpts=0;
	double average;

	if(s == CurrentSpectrumFocus) return s;
	// if data stored as shorts, read in shorts
	CheckMemState("ReadTimeArray:1");
	if(TimeArray != 0) delete [] TimeArray;
	TimeArray = new long[PS.TransientPoints];
	CheckMemState("ReadTimeArray:2");

	arf->Seek(PS.Spec[s].TArrayFileOffset,CFile::begin);

	CheckMemState("ReadTimeArray:3");
	if (DataPointSize==16) {
		sTArray = new short[PS.TransientPoints];
		inpts=arf->Read(sTArray,PS.TransientPoints*sizeof(short)); // read the data in as type 'short'
		CheckMemState("ReadTimeArray:3a");
		
		IsDefined = TRUE;
		SetIsFFTed(s,FALSE);

		// subtract off the DC level
		CheckMemState("ReadTimeArray:4");
		average=0;
		if (PS.DCRemoval>0) {
			for (i=0; i<PS.TransientPoints; i++) average+=(double)sTArray[i]/PS.TransientPoints;
		}
		CheckMemState("ReadTimeArray:5");
		for(i=0; i<PS.TransientPoints; i++) TimeArray[i]=long(sTArray[i] - average);
		CheckMemState("ReadTimeArray:6");
		delete [] sTArray;
	} 
	if (DataPointSize==32) {
		lTArray = new long[PS.TransientPoints];
		inpts=arf->Read(lTArray,PS.TransientPoints*sizeof(long)); // read the data in as type 'short'
		IsDefined = TRUE;
		SetIsFFTed(s,FALSE);

		// subtract off the DC level
		average=0;
		if (PS.DCRemoval>0) {
			for (i=0; i<PS.TransientPoints; i++) average+=(double)lTArray[i]/PS.TransientPoints;
		}
		CheckMemState("ReadTimeArray:7");
		for(i=0; i<PS.TransientPoints; i++) TimeArray[i]=long(lTArray[i] - average);
		CheckMemState("ReadTimeArray:8");
		delete [] lTArray;
	} 
	CheckMemState("ReadTimeArray:9");
	return inpts;
}

void CBudaDoc::GetTimeArray(int s)
{	
	CFile file;

	if(s == CurrentSpectrumFocus) return;
	file.Open(GetPathName(),CFile::modeRead);
	ReadTimeArray(s,&file);
	TransientIsDefined = TRUE;
	IsDefined = TRUE;
	SingleFFTFlag = FALSE;
	SetIsFFTed(s,FALSE);
	SetIsMIed(s,FALSE);
	file.Close();
	// check for a spectrum view that has ViewAsStickplot == FALSE
	// and if so set it to a transient plot and set firstplot=true so it
	// will start ffting.
	POSITION pos = GetFirstViewPosition();
	while(pos != NULL) {
		CBudaView* pView = (CBudaView*)GetNextView( pos );
		if(pView->PlotType == MASSSPEC || pView->PlotType == FREQUENCY) {
			pView->PlotType = TRANSIENT;
			pView->FirstPlot = TRUE;
		}
	}
}

BOOL CBudaDoc::SetCurrentSpectrumFocus(int s)
{
	return SetCurrentSpectrumFocus(s,LOAD_TIMEARRAY);
}


void CBudaDoc::SetIsFFTed(int s, BOOL tf)
{
	// only one member of spec can have IsFFTed = TRUE at any time.
	CheckMemState("SetIsFFTed:1");
	if(!IsDefined) return;
	ASSERT(s >= 0);
	ASSERT(s < PS.SpectraAcquired);
	int i;
	if(tf) {
		for(i=0; i<PS.SpectraAcquired; i++) PS.Spec[i].IsFFTed = FALSE;
		PS.Spec[s].IsFFTed = TRUE;
	} else {
		PS.Spec[s].IsFFTed = FALSE;
	}
	CheckMemState("SetIsFFTed:2");
}

void CBudaDoc::SetIsFFTed(BOOL tf)
{
	SetIsFFTed(CurrentSpectrumFocus,tf);
}

BOOL CBudaDoc::IsFFTed()
{
	return PS.Spec[CurrentSpectrumFocus].IsFFTed;
}

void CBudaDoc::SetIsMIed(int s, BOOL tf)
{
	PS.Spec[s].IsMIed = tf;
}

BOOL CBudaDoc::IsMIed()
{
	return PS.Spec[CurrentSpectrumFocus].IsMIed;
}

void CBudaDoc::SetIsMIed(BOOL tf)
{
	SetIsMIed(CurrentSpectrumFocus,tf);
}

void CBudaDoc::CheckMemState(CString str)
{
#ifdef _DEBUG
	//TRACE("****************************\n");
	//TRACE(str+"\n");
	TRACE("****************************\n");
	ASSERT(_CrtCheckMemory());
	//ASSERT(!_CrtDumpMemoryLeaks());
	TRACE("****************************\n");
#endif	// _DEBUG
}

void CBudaDoc::OnChangeMemo()
{
	CChangeMemo dlg;
	int i;

	dlg.m_pMemo = PS.ExperimentMemo;
	dlg.m_pNotes = PS.ExperimentNotes;
	dlg.DoModal();

	for(i=0; i<60 ; i++) {
		if(i<dlg.m_Memo.GetLength()) 
			PS.ExperimentMemo[i] = dlg.m_Memo.GetAt(i);
		else PS.ExperimentMemo[i] = '\0';
	}
	for(i=0; i<600; i++) {
		if(i<dlg.m_Notes.GetLength()) 
			PS.ExperimentNotes[i] = dlg.m_Notes.GetAt(i);
		else PS.ExperimentNotes[i] = '\0';
	}
	UpdateAllViews(NULL,NULL,NULL);
}


void CBudaDoc::GetDefaultDisplayParms()
{
	DisplayParms = pApp->DisplayParms;
}

void CBudaDoc::StoreDefaultDisplayParms()
{
	StoreDefaultDisplayParms(&(DisplayParms));
}

void CBudaDoc::StoreDefaultDisplayParms(struct DispParms* pDp)
{
	pApp->DisplayParms = *(pDp);
	pApp->WriteIniFile();
}

void CBudaDoc::GetMITable()
{
	// now create and calculate the MI Table
	// note: TotalIonInt is undefined at this point, but it
	// must have a value to get the mass range in EstimateNoiseLevel
	// so define it = 0.
	PS.Spec[CurrentSpectrumFocus].TotalIonInt = 0;
	float* FTData = FreqArray;
	long HighMassPt = MassToPt(PS.DetectHighMass);
	long LowMassPt = MassToPt(PS.DetectLowMass);

	if(PS.Spec[CurrentSpectrumFocus].FITable != 0) {
		delete [] PS.Spec[CurrentSpectrumFocus].FITable;
	}
	FITable = PS.Spec[CurrentSpectrumFocus].FITable;
	FITable = new struct FITableElement[MAXNUMPEAKS];
	PS.Spec[CurrentSpectrumFocus].NumPeaks = 0;

	// Peakpick and set the ion intensity used in calibration
	PeakPick(HighMassPt, LowMassPt, FTData);
	GetTotalIonInt();

	// determine charge state
	if(sys->AutoLabelCharge) {
		CChargeStateDetermination CSDet;
		CSDet.pDoc = this;
		CSDet.CSDetermination();
		IsCSed = TRUE;
	}
}

void CBudaDoc::GetTotalIonInt() 
{
	// this adds up the intensity of all peaks for which the
	// program could calculate a resolution - i.e. all big peaks
	// that have more than 3 points above the baseline
	long i;

	float Tint = 0;
	for(i=0; i<PS.Spec[CurrentSpectrumFocus].NumPeaks; i++) {
		if(FITable[i].RP>0) Tint += float(FITable[i].I);
	}
	PS.Spec[CurrentSpectrumFocus].TotalIonInt = Tint;
}

void CBudaDoc::PeakPick(long HighMassPt, long LowMassPt, float* data)
{
	long i;		
	long minpt,maxpt,temp;
	float noise;

	int ptr = 0;			// peakstack pointer
	double PeakStack[100];		// intensity values of peaks
	double PeakXStack[100];		// frequency values of peaks

	noise = float(PS.Spec[CurrentSpectrumFocus].NoiseLevel);
	minpt=HighMassPt;
	maxpt=LowMassPt;
	if(minpt>maxpt) {
		temp = minpt;
		minpt = maxpt;
		maxpt = temp;
	}
	PS.Spec[CurrentSpectrumFocus].NumPeaks = 0;
	// peak requirements:
	//		must have 2 or more points above noise
	//		must have "smooth" surface where smooth is
	//			defined as:
	//				before summit, no negative deviations 
	//				in intensity greater than noise
	//				after summit, no positive deviations 
	//				in intensity greater than noise

	// i = curpoint
	for (i=minpt;i<maxpt;i++) {
		if(data[i]<noise) continue;
		ptr = 0;

		// KeepThisPoint()
		if(ptr<98) {
			PeakStack[ptr]=data[i-1];
			PeakXStack[ptr]=PtToFreq(i-1);
			ptr++;
		}

		// record points until slope change
		for(;i<maxpt-1; i++) {
			if(ptr<98) {
				PeakStack[ptr]=data[i];
				PeakXStack[ptr]=PtToFreq(i);
				ptr++;
			}
			if(data[i] > data[i-1]) continue;

			if(data[i] < noise) break;
			if((data[i-1]-data[i])<(3*noise)) break;
			if(data[i+1] < data[i]) break;
		}

		// we are here because the data indicated a negative slope
		for(i++; i<maxpt-1; i++) {
			// KeepThisPoint()
			if(ptr<98) {
				PeakStack[ptr]=data[i];
				PeakXStack[ptr]=PtToFreq(i);
				ptr++;
			}
			if(data[i]<noise) break;

			if((data[i]-data[i-1])>(3*noise)) {
				i--;
				break;
			}

			if((data[i] > data[i-1]) && (data[i+1] > data[i])){
				i--;
				break;
			}
		}

		PeakFit(PeakStack,PeakXStack,ptr);
		if(PS.Spec[CurrentSpectrumFocus].NumPeaks >= 10000) {
			MessageBox("Peak Pick reached the maximum number of peaks\nbefore completing.  Some peaks may not have been found","Warning! Too many peaks");
			break;
		}
	} //end for loop
	PS.Spec[CurrentSpectrumFocus].IsMIed = TRUE;
//	DumpToDiagFile("leaving CBudaDoc::PeakPick()");

} // end PeakPick()

void CBudaDoc::PeakFit(double *PeakStack, double *PeakXStack, int numpts)
// PeakStack is the array of intensity values
// PeakXStack is the corresponding array of mass values
// numpts is the number of points in these arrays
{
	int i;
	//, PeakFitPts;
	double maxy = 0,maxx = 0,xptmax=0;
	double a[3], *x, *y;
	double deltafreq;
	//double sumxy,sumy;

	BOOL PeakWasFit=FALSE;

	a[0]=a[1]=a[2]=0.0;  
		// initialize parabolic fit constants
		// we want to do the fit with x=the index value and afterwards
		// calulate the matching frequency by interpolation as this 
		// decreases the roundoff errors in the least squares
		//
		// also, we want to set PeakWasFit = FALSE if one of the following
		// conditions is true:
		//		1) if <3 points above noise threshold
		//		2) if <3 points above the half height of the peak
	// first get new y and x arrays that only include the good points
	// good points are those above half height of peak and those above
	// threshold.s

	x=new double [numpts];
	y=new double [numpts];
	maxy = 0;
	maxx = 0;
	for(i=0; i<numpts; i++) {
		if (PeakStack[i]>maxy) 	{
			maxy=double(PeakStack[i]);
			maxx=double(PeakXStack[i]);
		}
	}
	for(i=0; i<numpts; i++) x[i]=i;

#ifdef DEBUG_PEAKFIT
		CString str,crlf = "\r\n";
		TRACE("==========================="+crlf);
		str.Format("New Peak Found : peak number %d",PS.Spec[CurrentSpectrumFocus].NumPeaks);
		TRACE(str+crlf);
		str.Format("numpts = %d",numpts);
		TRACE(str+crlf);
		TRACE("---peak data---"+crlf);
		for(i=0; i<numpts; i++) {
			str.Format("PeakStack[%d] = %1.8f : PeakXStack[%d] = %1.8f",i,PeakStack[i],i,PeakXStack[i]);
			TRACE(str+crlf);
		}
#endif

	PeakWasFit=ParabolicLeastSquares(PeakStack, numpts, a, maxy, sys->PeakPickerLevel/100);
	if (PeakWasFit) {
		// calculate the center pt and intensity and verify that
		//		1) x[center pt] is in x[0]-x[numpt-1] range
		//		2) intensity is not > 2* maxintensity 
		maxx=(-1*a[1])/(2*a[0]);
		maxy=a[0]*maxx*maxx + a[1]*maxx + a[2];
		deltafreq = fabs(PS.AdcRate/double(PS.FFTSize));
		xptmax = maxx;
		if(PS.DetectMode == MixerMode) maxx=PeakXStack[0]-deltafreq*maxx;
		else maxx=PeakXStack[0]+deltafreq*maxx;
#ifdef DEBUG_PEAKFIT
		str.Format("--Peak Was Fit--");
		TRACE(str+crlf);
		str.Format("maxx in points = %1.5f",xptmax);
		TRACE(str+crlf);
		str.Format("deltafreq = %1.8f",deltafreq);
		TRACE(str+crlf);
		str.Format("fitting constant[0] = %1.8f",a[0]);
		TRACE(str+crlf);
		str.Format("fitting constant[1] = %1.8f",a[1]);
		TRACE(str+crlf);
		str.Format("fitting constant[2] = %1.8f",a[2]);
		TRACE(str+crlf);
#endif
		// if maxx outside range of input points, not a good fit
		// remember that the points are plotted backwards for heterodyne 
		if (PS.DetectMode == DirectMode) {
			if(maxx>PeakXStack[numpts-1] || maxx<PeakXStack[0]) PeakWasFit = FALSE;
		} else {
			if(maxx<PeakXStack[numpts-1] || maxx>PeakXStack[0]) PeakWasFit = FALSE;
		}
	} 
	delete [] x;
	delete [] y;

	/*if(!PeakWasFit) {
		// if we couldn't fit a parabola to the peaks, we will do 
		// a center of mass calculation to define the peak
		//
		// x is the intensity weighted average of the x's
		// y is the max intensity
		sumxy=0.0;
		sumy=0.0;
		maxy=0;
		for (i=0; i<numpts; i++) {
			sumxy=sumxy+PeakStack[i]*PeakXStack[i];
			sumy=sumy+PeakStack[i];
			maxy = (maxy<PeakStack[i]) ? PeakStack[i] : maxy;
		}
		maxx=sumxy/sumy;
		PeakWasFit=TRUE;
	}*/
	

	float A = GetPeakArea(PeakStack,PeakXStack,numpts);
	struct FITableElement* FITE = &(FITable[PS.Spec[CurrentSpectrumFocus].NumPeaks]); 

	if (PS.Spec[CurrentSpectrumFocus].NumPeaks<MAXNUMPEAKS) {
		FITE->F=maxx;
		FITE->I=maxy;
		FITE->A = A;
		if(PeakWasFit) FITE->RP = InterpolateForRes(PeakStack,PeakXStack,numpts,xptmax);
		//if(PeakWasFit) FITE->RP = float(CalculateRes(a[0],a[1],a[2],double(maxy),double(maxx),deltafreq));
		else FITE->RP = -1;
		FITE->CS = 1;
		PS.Spec[CurrentSpectrumFocus].NumPeaks++;
#ifdef DEBUG_PEAKFIT
		str.Format("maxx in freq = %1.5f",maxx);
		TRACE(str+crlf);
		str.Format("maxy = %1.5f",maxy);
		TRACE(str+crlf);
		str.Format("area = %1.5f",A);
		TRACE(str+crlf);
		str.Format("RP = %1.5f",FITE->RP);
		TRACE(str+crlf);
#endif		
	}

}//end PeakFit

BOOL CBudaDoc::ParabolicLeastSquares(double *y, int numpts, double *ret, double PeakMaxAmplitude, double PeakPickerLevel)
{  // this function performs a parabolic least squares fit on the x,y arrays each of length numpts
   // that are input.  ret[0], ret[1], ret[3] are the three fit constants

	if (numpts<3) {
		return FALSE; // can't parabolic least squares with less than 3 peaks
	}
	if (numpts == 3) {
		if((y[0]>=y[1]) || (y[2]>=y[1])) return false;
	}

	int i;
	double sumx4, sumx3, sumx2, sumx1, sumx0; // Sum(x^4), Sum(x^3), ..
	double sumy2, sumy1, sumy0;				  // Sum(y*x^2), Sum(y*x), Sum(y)
	double a, b, c;		// parabola coefficients
	double dp;  // denominator in the determinants.
	double f,f2;
	int NumUsablePts;
	double MinFitAmplitude = PeakMaxAmplitude * PeakPickerLevel;
	sumx4=sumx3=sumx2=sumx1=sumx0=0.0;
	sumy2=sumy1=sumy0=0;
	NumUsablePts = 0;

	while(TRUE) {
		sumx4=sumx3=sumx2=sumx1=sumx0=0.0;
		sumy2=sumy1=sumy0=0;
		NumUsablePts = 0;

		for (i=0; i<numpts; i++) {
			if(y[i] > MinFitAmplitude) {
				NumUsablePts++;
				f = i;
				f2 = f*f;

				sumx0++;
				sumx1=sumx1+f;
				sumx2=sumx2+f2;
				sumx3=sumx3+f*f2;
				sumx4=sumx4+f2*f2;

				sumy0=sumy0+y[i];
				sumy1=sumy1+y[i]*f;
				sumy2=sumy2+y[i]*f2;
			}
		}
		if(NumUsablePts >= 3) break;

		PeakPickerLevel -= 0.1;
		if(PeakPickerLevel < 0.33) return FALSE;
		MinFitAmplitude = PeakMaxAmplitude * PeakPickerLevel;		
	}

	// now calculate a,b,c coefficients for the parabola
	dp = GetDeterminant(sumx4, sumx3, sumx2,
						sumx3, sumx2, sumx1,
						sumx2, sumx1, sumx0);
	if (dp == 0) return FALSE;

	a  = GetDeterminant(sumy2, sumx3, sumx2,
						sumy1, sumx2, sumx1,
						sumy0, sumx1, sumx0)/dp;
	b  = GetDeterminant(sumx4, sumy2, sumx2,
						sumx3, sumy1, sumx1,
						sumx2, sumy0, sumx0)/dp;
	c  = GetDeterminant(sumx4, sumx3, sumy2,
						sumx3, sumx2, sumy1,
						sumx2, sumx1, sumy0)/dp;

	// verify that a is negative so that found a peak not a valley 
	if (a>0 || b<0) {
		return FALSE;
	}

	// if we get this far, the peak is good, so return it
	ret[0]=a;
	ret[1]=b;
	ret[2]=c;

	return TRUE;
}

double CBudaDoc::GetDeterminant(double a1, double a2, double a3, double b1, double b2, double b3, double c1, double c2, double c3)
{
	return a1*b2*c3+a2*b3*c1+a3*b1*c2-a1*b3*c2-a2*b1*c3-a3*b2*c1;
}

float CBudaDoc::InterpolateForRes(double *y, double *x, int ptr, double maxx)
{
	double LeftWidth = -1;
	double RightWidth = -1;
	double PeakAmplitude,hh, PeakResolution;
	double PeakFrequency,df;
	int i,CenterIndex;

	PeakAmplitude = y[0];
	CenterIndex = 0;
	for (i=0; i<ptr; i++) {
		if (y[i] > PeakAmplitude) {
			PeakAmplitude =  y[i];
			CenterIndex=i;
		}
	}

	PeakFrequency = double(x[CenterIndex]);
	df = fabs(PS.AdcRate/double(PS.FFTSize));
	
	hh = PeakAmplitude/2.0;

	for(i = int(maxx+0.5); i>0; i--) {
		if(y[i] < hh) {
			LeftWidth = maxx - ((hh-y[i])/(y[i+1]-y[i])) - i;
			break;
		}
	}

	for(i = int(maxx+0.5); i<ptr; i++) {
		if(y[i] < hh) {
			RightWidth = i-((hh-y[i])/(y[i-1]-y[i])) - maxx;
			break;
		}
	}

	PeakResolution = -1;
	if(LeftWidth <= 0) LeftWidth = RightWidth;
	if(RightWidth <= 0) RightWidth = LeftWidth;

	if(LeftWidth > 0) PeakResolution = PeakFrequency/((LeftWidth+RightWidth)*df);
	return float(PeakResolution);
}

float CBudaDoc::GetPeakArea(double *Ypts, double *Xpts, int numpts)
{
	// build in a numerical integration of the peak here.  
	// use trapezoid rule
	//
	// I(1..n) += df*(x[i]+x[i-1])/2
	//
	int i;
	double sum = 0;
	double df = Xpts[1]-Xpts[0];
	for (i=1; i<numpts; i++) {
		sum += df*(Ypts[i]+Ypts[i-1])/2;
	}
	return float(fabs(sum));
}

void CBudaDoc::OnAdjustVoltages() 
{
	CAdjustDacInitDlg dlg;

	dlg.pDoc = this;
	dlg.pPS = &PS;

	// onok and oncancel are handled in the dialog.
	dlg.DoModal();
}

/*
void CBudaDoc::GetChargeState(double minx, double maxx)
{
	// minx and maxx are in mass coordinates.  This function
	// will never be called unless the view is in mass spec mode.

	// this function does a patterson/fourier charge state
	// determination of the data in the minx to maxx window.  
	// parameters can be modified in the options window.

	float* Patt;
	float* Four;
	float* Combo;
	int i,Max_Charge;
	
	// read parms from options
	Max_Charge = 50;
	
	// set up results arrays
	Patt = new float[Max_Charge];
	Four = new float[Max_Charge];
	Combo = new float[Max_Charge];

	// get patterson matching
	CS_Patterson(Patt,minx,maxx,Max_Charge);
	
	// get fourier matching
	CS_Fourier(Patt,minx,maxx,Max_Charge);

	// multiply two techniques
	for(i=0; i<Max_Charge; i++) Combo[i]=Patt[i]*Four[i];

	// pop up a note telling the charge state (and reliability?)

	// get most intense point in combo.
	delete Patt;
	delete Four;
	delete Combo;
}

void CBudaDoc::CS_Patterson(float *Patt, double minx, double maxx, int Max_Charge)
{
	// this function calculates the Patterson charge state determination
	// values for charges from 1 to Max_Charge from minx to maxx in mass domain.

	float p;
	long i,j, minpt, maxpt, tempind, temp;
	float deltam,tempmass,tempminus,tempplus,tempval1,tempval2;

	minpt = MassToPt(maxx);
	maxpt = MassToPt(minx);
	if(minpt>maxpt) {
		temp = minpt;
		minpt = maxpt;
		maxpt = temp;
	}

	// loop through charge states
	for(i=1; i<=Max_Charge; i++) {
		p=0; // zero the temporary value
		deltam = 0.5*(1.0034/i);
		
		for(j=minpt; j<=maxpt; j++) {
			tempval1 = 0;
			tempval2 = 0;

			// get value of data at j-deltam
			tempmass = PtToMass(j);
			tempind = MassToPt(double(tempmass - deltam));
			tempminus = FreqArray[tempind];

			// get value of data at j+deltam
			tempind = MassToPt(tempmass + deltam);
			tempplus = FreqArray[tempind];

			// p(deltam) = sum_lowmass+highmass(f(m(i)-deltam)*f(m(i)+deltam));
			p += tempplus*tempminus;
		}
		Patt[i]=p;
	}
}
*/


void CBudaDoc::OnToolsChefcalculator() 
{
	CChefDlg dlg;
	
	dlg.pDoc = this;
	dlg.DoModal();
}

/*void CBudaDoc::OnFileExport() 
{
	CExportDlg dlg;
	dlg.pDoc = this;
	dlg.DoModal();
}
*/

void CBudaDoc::OnDebugDialogMasstofreq() 
{
	CDebugDialog dlg;
	dlg.pDoc=this;
	dlg.DoModal();	
}

void CBudaDoc::ReadBrukerFIDFile(CFile *arf, int mode)
{
	DumpToDiagFile("entering CBudaDoc::ReadBrukerFIDFile(..)");
	// note : mode 0 is the normal read file mode.  mode 1
	// is the read setupfile mode
	CheckMemState("ReadBrukerFIDFile:1");

	long i;
	int len;
	unsigned long inpts;
	CString pathname, FIDFilename, AcquFilename, SPtypeFilename, AcqusFilename;
	CFile fid,acqu,sptype,acqus;
	long templong;
	double tempdbl;
	CString str;
	
	pathname = arf->GetFilePath();
	FIDFilename=arf->GetFileName();
	len=pathname.Find(FIDFilename,0);
	pathname = pathname.Left(len);
	AcquFilename = pathname+"acqu";
	AcqusFilename = pathname+"acqus";
	SPtypeFilename = pathname+"sptype";
	FIDFilename = pathname + "fid";

	// read in FID for the Bruker file
	PS.FileType = BRUKERFILE;
	strcpy(PS.UserID,"Bruker");
	strcpy(PS.DataFileName,FIDFilename);
	strcpy(PS.ExperimentMemo,"Bruker Data File");
	strcpy(PS.TimeAcquired,"01:23:45");
	strcpy(PS.TimeProcessed,"01:23:45");
	strcpy(PS.ExperimentNotes,"Bruker Data File, no notes");
	PS.MajorVersion = 0;
	PS.MinorVersion = 0;
	inpts=0;

	if(acqu.Open(AcquFilename,CFile::modeRead|CFile::shareDenyWrite)==0) 
	{
		AfxMessageBox("Could not open acqu file", MB_OK);
		return;
	}
	str = "##$NF";
	templong = ExtractLongFromACQUFile(&acqu,&str);
	if(templong>1) {
		// this is time domain data which BUDA doesn't 
		// handle yet.  Sorry
		AfxMessageBox("Sorry, this is 2D data, which BUDA doesn't understand",IDOK);
		return;
	}

	str = "##$SW_h";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	PS.AdcRate=tempdbl;
	if (tempdbl < 20001) PS.DetectMode = MIXER_MODE;
	else PS.DetectMode = DIRECT_MODE;

	str = "##$TD";
	templong = ExtractLongFromACQUFile(&acqu,&str);
	PS.TransientPoints=templong;

	str = "##$MW_high";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	PS.DetectHighMass=tempdbl;
	str = "##$MW_low";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	PS.DetectLowMass=tempdbl;
	str = "##$MW_cent";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	PS.MonitorMass=tempdbl;

	double BrukerCalParm1, BrukerCalParm2, BrukerCalParm3;
	str = "##$ML1";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	BrukerCalParm1=tempdbl;
	str = "##$ML2";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	BrukerCalParm2=tempdbl;
	str="##$ML3";
	tempdbl=ExtractDoubleFromACQUFile(&acqu,&str);
	BrukerCalParm3=tempdbl;

	acqu.Close();

	// convert Bruker Cal Parms to IonSpec Calibration Parms
	PS.A=BrukerCalParm1/2;
	PS.B=BrukerCalParm2;
	PS.TrapShift=0;
	PS.IntensityShift=0;
	PS.HighMassShift=0;
	PS.HiResESI=1;

	// read FID
	long* lTArray=0;
	inpts=0;
	double average;

	if(fid.Open(FIDFilename,CFile::modeRead|CFile::shareDenyWrite)==0) 
	{
		AfxMessageBox("Could not open fid file", MB_OK);
		return;
	}
	CArchive ar(&fid,CArchive::load,4096,NULL);

	if(TimeArray != 0) delete [] TimeArray;
	TimeArray = new long[PS.TransientPoints];
	lTArray = new long[PS.TransientPoints];

	// read the data in as type 'long'
	inpts=ar.Read(lTArray,PS.TransientPoints*sizeof(long)); 
	IsDefined = TRUE;
	SetIsFFTed(0,FALSE);

	// subtract off the DC level
	// convert from Unix to PC format
	for(i=0; i<PS.TransientPoints; i++) TimeArray[i]=ConvertBigEndian(&(lTArray[i]));
	delete [] lTArray;

	average=0;
	if (PS.DCRemoval>0) {
		for (i=0; i<PS.TransientPoints; i++) average+=(double)TimeArray[i]/PS.TransientPoints;
	}
	CheckMemState("ReadTimeArray:7");
	for(i=0; i<PS.TransientPoints; i++) TimeArray[i]=long(TimeArray[i] - average);
	ar.Close();
	fid.Close();
	TransientIsDefined=1;
	PS.TransientIsStored=1;

	PS.PeakMatching=0;
	SetupParmsOnRead();
	return;
}

double CBudaDoc::ExtractDoubleFromACQUFile(CFile *acqu, CString *instr)
{
	double value;
	CString str, getstr,rightstr;
	CArchive ar(acqu,CArchive::load,4096,NULL);
	acqu->SeekToBegin();

	getstr=CompressString(*instr);
	
	while(ar.ReadString(str)) {	// breaks if EOF
		str=CompressString(str);
		if (str.GetLength()==0) continue; // empty lines
		if (str.Left(1)=="%") continue; // comment lines
	
		if (str.Left(getstr.GetLength()) == getstr) {
			rightstr = GetRightString(str);
			sscanf(rightstr,"%lf",&value);
			return value;
		}
	}
	return 0;

}

long CBudaDoc::ExtractLongFromACQUFile(CFile *acqu, CString *instr)
{
	long value;
	CString str, getstr,rightstr;
	CArchive ar(acqu,CArchive::load,4096,NULL);
	acqu->SeekToBegin();

	getstr=CompressString(*instr);
	
	while(ar.ReadString(str)) {	// breaks if EOF
		str=CompressString(str);
		if (str.GetLength()==0) continue; // empty lines
		if (str.Left(1)=="%") continue; // comment lines
	
		if (str.Left(getstr.GetLength()) == getstr) {
			rightstr = GetRightString(str);
			sscanf(rightstr,"%ld",&value);
			return value;
		}
	}
	return 0;	
}

CString CBudaDoc::CompressString(CString str)
{
	CString strleft, strright, outstring;
	outstring = str;
	outstring.MakeUpper();	// make string uppercase
	int i;
	for (i=0; i<outstring.GetLength(); i++) {
		if(outstring.GetAt(i) == ' ' || outstring.GetAt(i) == '\t') {
			// remove spaces and tabs from string
			strleft = outstring.Left(i);
			strright= outstring.Right(outstring.GetLength()-i-1);
			outstring = strleft + strright;
		}
	}
	return outstring;
}

CString CBudaDoc::GetRightString(CString str)
{
	int equalpos;
	equalpos = FindEqualSign(str)+1;
	return str.Right(str.GetLength()-equalpos);
}


int CBudaDoc::FindEqualSign(CString str)
{
	int i=0;
	for(i=0; i<str.GetLength(); i++) {
		if(str.Mid(i,1)=="=") break;
	}
	return i;
}

long CBudaDoc::ConvertBigEndian(long *in)
{
	BYTE* p1, *p2, *p3, *p4;
	BYTE c[4];
	long out;
	long* p;
	short *i;

	c[0]=0;
	c[1]=0;
	c[2]=0;
	c[3]=0;

	out=0;
	p1 = (BYTE*) in;
	p2=p1+1;
	p3=p1+2;
	p4=p1+3;

	if((short)*p4 == 255 && (short)*p3==255) 
	{
		c[0]=*p1;
		c[1]=*p2;
		i = (short *) c;
		out=(long)*i;
	}
	else
	{
		c[0]=*p4;
		c[1]=*p3;
		c[2]=*p2;
		c[3]=*p1;
		p = (long *)c;
		out = *p;
	}

	return out;
}
