// BudaView.cpp : implementation of the CBudaView class
//
/*
The BudaView.cpp file is the code that controls the CBudaView
class.  This class (derived from CView) defines the behavior of the
windows and the plots of the data in them.  It gets the data from the
CBudaDoc class (which it keeps a pointer to), does the decimations
necessary to plot the 1megapoint data arrays in a screen window with
pixel width of maximum 1-2kilopoints.  Controls zooming, the mouse 
cursor data in the status window, and of course plotting.  

The plots use Gigasoft's ProEssentials V2.0 to automate a lot of the
data display and labeling, and most interaction with the plot uses
their subroutines and variables.  

*/
/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "BudaMaster.h"
#include "math.h"

#include "BudaDoc.h"
#include "BudaView.h"
#include "pegrpapi.h" 
#include "pemessag.h"
#include "SetDisplayRngDlg.h"
#include "SetScaleFactorDlg.h"
#include "SetNewThreshold.h"
#include "Deconvolution.h"
#include "DeconOptions.h"
//#include "ChargeStateDetermination.h"
#include "CSMapDlg.h"
#include "ExportDlg.h"

#ifdef _DEBUG
	#define new DEBUG_NEW
	#undef THIS_FILE
	static char THIS_FILE[] = __FILE__;
#endif
#define SCROLL_MAX 32737
#define SCROLL_MIN 0
#define AVG_MAX_DEPTH 32
#define AVERAGE_NOISE 0
#define MIN_MAX 1
#define DECIMATION_METHOD AVERAGE_NOISE

/// Buda debug defines
#ifdef _DEBUG
	//#define DEBUG_CALCMSVL	
	//#define DEBUG_ACCESSVIOLATION
	//#define DEBUG_FFTCRASH
	//#define DEBUG_VIEWRANGE
#endif

/////////////////////////////////////////////////////////////////////////////
// CBudaView

IMPLEMENT_DYNCREATE(CBudaView, CView)

BEGIN_MESSAGE_MAP(CBudaView, CView)
	//{{AFX_MSG_MAP(CBudaView)
	ON_COMMAND(ID_FFTTRANSIENT, GetFreqData)
	ON_WM_SIZE()
	ON_COMMAND(ID_NEWWINDOW, OnNewwindow)
	ON_COMMAND(ID_SETDISPLAYRANGE, OnSetDisplayRange)
	ON_COMMAND(ID_FILE_PRINT, OnFilePrint)
	ON_UPDATE_COMMAND_UI(ID_ZOOM_Y, OnUpdateZoomY)
	ON_COMMAND(ID_ZOOM_Y, OnZoomY)
	ON_COMMAND(ID_ZOOM_OUT, OnZoomOut)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_WM_MOUSEMOVE()
	ON_COMMAND(ID_VIEW_CHROMATOGRAM, OnViewChromatogram)
	ON_COMMAND(ID_2D_SETSPECHERE, On2dSetSpec)
	ON_COMMAND(ID_DIAGS_KEYTEST, OnDiagsKeytest)
	ON_COMMAND(ID_2D_SETSPECNEXT, On2dSetSpecNext)
	ON_COMMAND(ID_2D_SETSPECPREV, On2dSetSpecPrev)
	ON_COMMAND(ID_VIEW_STICKPLOT, OnViewStickplot)
	ON_UPDATE_COMMAND_UI(ID_VIEW_STICKPLOT, OnUpdateViewStickplot)
	ON_UPDATE_COMMAND_UI(ID_VIEW_CHROMATOGRAM, OnUpdateViewChromatogram)
	ON_UPDATE_COMMAND_UI(ID_VIEW_DECON, OnUpdateViewDecon)
	ON_COMMAND(ID_VIEW_DECON, OnViewDecon)
	ON_COMMAND(ID_ANALYZE_SETSCALEFACTOR, OnAnalyzeSetscalefactor)
	ON_COMMAND(ID_ANALYZE_SETPEAKPICKTHRESHOLD, OnAnalyzeSetpeakpickthreshold)
	ON_UPDATE_COMMAND_UI(ID_ANALYZE_SETPEAKPICKTHRESHOLD, OnUpdateAnalyzeSetpeakpickthreshold)
	ON_UPDATE_COMMAND_UI(ID_ChargeStateDeterminationState, OnUpdateCSDet)
	ON_COMMAND(ID_ChargeStateDeterminationState, OnCSDet)
	ON_WM_HSCROLL()
	ON_MESSAGE(WM_FFTDONE, OnFFTDone)
	ON_COMMAND(ID_VIEW_PLOTMASSSPECTRUM, OnPlotMassSpectrum)
	ON_COMMAND(ID_VIEW_PLOTFREQUENCY, OnPlotFrequency)
	ON_COMMAND(ID_VIEW_DECONVOLUTION, OnViewDecon)
	ON_COMMAND(ID_VIEW_PLOTTIME, OnPlotTransient)
	ON_COMMAND(ID_VIEW_COPY_CURSOR_MASS, OnViewCopyCursorXPos)
	ON_COMMAND(ID_FILE_EXPORT, OnFileExport)
	ON_COMMAND(ID_DEBUG_DIALOG_MASSTOFREQ, OnDebugDialogMasstofreq)
	//}}AFX_MSG_MAP

	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP()

struct MIList {
	double M;
	double I;
	int cs;
};

/////////////////////////////////////////////////////////////////////////////
// CBudaView construction/destruction

CBudaView::CBudaView()
{
	int i;

	pApp = (CBudaApp*)AfxGetApp();
	
	Decon = 0;
	FirstPlot = TRUE;
	UpdateScaleFactor = TRUE;
	TwoDFFTStatus = 0;
	TwoDFFT = FALSE;
	TwoDFFTPlotType = TOTAL_ION_CHROMATOGRAM;
	NBMIList = 0;
	NumMI=0;
	NumAnnotations = 0;
	ViewLine=0;
	ViewLineXPts=0;
	ViewLineLength = Buda_VIEWLINELENGTH;
	OldViewLineLength = ViewLineLength;
	ViewLine = new float [ViewLineLength];
	ViewLineXPts = new float [ViewLineLength];

	for (i=0; i<ViewLineLength; i++) {
		ViewLine[i] = float(sin(double(i)/10.0));
		ViewLineXPts[i] = float(0.62830*double(i));
	}

	PostFFTPlotType = MASSSPEC;
	ViewLineIsDirty = TRUE;
	PEIsDefined = FALSE;
	ViewIsDirty = TRUE;
	ViewIsSetup=FALSE;
	IsZoomingY = FALSE;
	ViewLineScaling = 0;
	Printing = FALSE;
	m_hFFTDone = CreateEvent(NULL, TRUE, FALSE, NULL);
}

CBudaView::~CBudaView()
{
	if(Decon) delete Decon;
	delete [] ViewLine;
	delete [] ViewLineXPts;
	PEdestroy(hWndPE);
	if (NBMIList != 0) {
		delete NBMIList;
		NBMIList = 0;
	}
}

BOOL CBudaView::PreCreateWindow(CREATESTRUCT& cs)
{
	cs.lpszClass = AfxRegisterWndClass(CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW,
			AfxGetApp()->LoadStandardCursor(IDC_CROSS), (HBRUSH)(COLOR_WINDOW+1),
			AfxGetApp()->LoadIcon(IDR_MAINFRAME));
	// add a horizontal scroll bar to the view.
	cs.style = cs.style | WS_HSCROLL;

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CBudaView drawing

void CBudaView::OnDraw(CDC* pDC)
{
	CString buff;
	ASSERT_VALID(pDoc);
	if(PlotType == MASSSPEC || PlotType == FREQUENCY) {
		if(ViewAsStickplot) {
			OldPlotType = PlotType;
			//PlotType = STICKPLOT;
		}
	}
	
	PlotViewLine();

	//CRect rect, parmrect;
	//GetClientRect(rect);

	//pDC->SetTextAlign(TA_BASELINE | TA_CENTER);
	//pDC->SetTextColor(::GetSysColor(COLOR_WINDOWTEXT));
	//pDC->SetBkMode(TRANSPARENT);

	//pDC->TextOut(65,20,"Buda",7);
	//pDC->SetTextAlign(TA_RIGHT);
	//buff=CString(pDoc->PS.DataFileName);
	//pDC->TextOut(rect.right-10,5,buff);

	
	//if (pDoc->DisplayParms.allparms) {
	//	parmrect.top = rect.top;
	//	parmrect.bottom = rect.bottom;
	//	parmrect.left = int((rect.right-rect.left)*0.8) + rect.left;
	//	parmrect.right = rect.right;
	//	//WriteParms(pDC, rect);
	//}

	//bmp.LoadBitmap(IDB_SMALL_MONO_LOGO);
	//pt=CPoint(0,0);
	//sz=CSize(64,64);
	//pDC->DrawState(pt,sz,&bmp,DST_BITMAP | DSS_NORMAL);
	pDC = pDC;
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("leaving CBudaView::OnDraw()\n");
#endif
}

/////////////////////////////////////////////////////////////////////////////
// CBudaView printing

BOOL CBudaView::OnPreparePrinting(CPrintInfo* pInfo)
{
	Printing = TRUE;
	CalculateViewLine();
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CBudaView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	Printing = TRUE;
	CalculateViewLine();
	// TODO: add extra initialization before printing
}

void CBudaView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	Printing = FALSE;
	CalculateViewLine();
}

/////////////////////////////////////////////////////////////////////////////
// CBudaView diagnostics

#ifdef _DEBUG
void CBudaView::AssertValid() const
{
	CView::AssertValid();
}

void CBudaView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CBudaDoc* CBudaView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CBudaDoc)));
	return (CBudaDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CBudaView message handlers


void CBudaView::CalculateViewLine()
{
	/*
	  This function takes the CBudaDoc data and calculates a smaller array
	  for plotting on the screen.  Several methods are used.  For transients,
	  the even view points are the maximum point on the data range, and the odd
	  view points are the minima.  For frequency, only the maxima are plotted,
	  and for mass, maxima are plotted except where there is a peak where both
	  max and min are plotted.  These are commented further in the respective
	  functions.

	  CalcViewLine should do nothing more than switch on PlotType to 
	  CalcTransientViewLine, CalcFrequencyViewLine, etc...
	*/

	GetScaleFactor();
	SetViewLineLength();
	switch (PlotType) {
		case TRANSIENT :
			CalcTransientViewLine();
			break;
		case FREQUENCY :
			CalcFrequencyViewLine();
			break;
		case MASSSPEC :
			CalcMassSpecViewLine();
			break;
		case STICKPLOT :
			CalcStickPlotViewLine();
			break;
		case TOTAL_ION_CHROMATOGRAM :
			CalcTICViewLine();
			break;
		case DECONVOLVED_MASS_DATA :
			CalcDeconViewLine();
			break;
		default:
			break;
	}
	ViewLineIsDirty = FALSE;
	ViewIsDirty=TRUE;
	pDoc->UpdateAllViews(NULL,0,NULL);
} // end CFftView::CalculateViewLine(long *data)


void CBudaView::PlotViewLine()
{	
	double miny=0,maxy=0;
	char text[200];

	// if no transient available, plot as mass spectrum
	if(PlotType == TRANSIENT && !pDoc->TransientIsDefined) {
		if(pDoc->IsFFTed()){
			MessageBox("Error 50:\nNo Transient Available\nPlotting Mass Spectrum","Data UnAvailable");
			OldPlotType = PlotType;			
			PlotType = MASSSPEC;
		}
		else if(pDoc->IsMIed()) {
			MessageBox("Error 51:\nNo Transient or Spectrum Available\nPlotting StickPlot","Data UnAvailable");
			OldPlotType = PlotType;			
			PlotType = STICKPLOT;
		}
		else {
			MessageBox("Error 52a:\nData not available\nCan't plot","Data Unavailable");
			OldPlotType = PlotType;			
			PlotType = NO_PLOT;
		}
	}

	// if trying to plot frequency or mass spectrum, and 
	// no spectrum available try plotting as stickplot
	// first ask if we should calculate the mass spectrum.
	if(PlotType == FREQUENCY || PlotType == MASSSPEC) {
		if(!pDoc->IsFFTed()) {
			if(pDoc->IsMIed()) {
				MessageBox("Error 53:\nRaw Spectrum not available\nPlotting Stickplot","Data Unavailable");
				OldPlotType = PlotType;			
				PlotType = STICKPLOT;
			} 
			else if (pDoc->TransientIsDefined) {
				PlotType = TRANSIENT;
			} else {
				MessageBox("Error 52b:\nData not available\nCan't plot","Data Unavailable");
				OldPlotType = PlotType;			
				PlotType = NO_PLOT;
			}			
		}
	}

	// if trying to plot as stickplot and no MI table data
	// available, try to plot as mass spectrum then as transient
	// then as noplot
	if(PlotType == STICKPLOT && !pDoc->IsMIed()) {
		if(pDoc->IsFFTed()) {
			MessageBox("Error 55:\nMI Table not available\nPlotting raw mass spectrum","Data UnAvailable");
			OldPlotType = PlotType;			
			PlotType = MASSSPEC;
		}
		else if(pDoc->TransientIsDefined) {
			MessageBox("Error 56:\nMI table and raw spectrum not available\nPlotting transient","Data UnAvailable");
			OldPlotType = PlotType;
			PlotType = TRANSIENT;
		}
		else {
			MessageBox("Error 52c:\nData not available\nCan't plot","Data Unavailable");
			OldPlotType = PlotType;
			PlotType = NO_PLOT;
		}
	}

	if(PlotType == FREQUENCY || PlotType == MASSSPEC) {
		if(ViewAsStickplot) {
			OldPlotType = PlotType;
			PlotType = STICKPLOT;
		}
	}

	if(PlotType == DECONVOLVED_MASS_DATA) {
		if(!pDoc->IsFFTed()) {
			MessageBox("Error 57:\nData Not available\n Can't Plot","Data Unavailable");
			OldPlotType = PlotType;
			PlotType = MASSSPEC;
		}
	}

	PEreset(hWndPE);
	ViewIsSetup = FALSE;	// testing to see if double lableing of peaks will go away with these two lines
	if(!ViewIsSetup) {
		switch (PlotType) {
			case 0 : // no data loaded
				PEszset (hWndPE, PEP_szMAINTITLE, "No Data Loaded");
				PEszset (hWndPE, PEP_szSUBTITLE,"");
				break;
			case TRANSIENT : // PlotType = Transient
				sprintf(text,"%s    %s\0",pApp->BudaVersion,pDoc->PS.DataFileName);
				PEszset (hWndPE, PEP_szMAINTITLE, text);
				if(pDoc->PS.HiResMALDI && pDoc->PS.HiResESI) sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResMALDI) sprintf(text,"MALDI  :  %s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResESI) sprintf(text,"Electrospray  :  %s\0",pDoc->PS.ExperimentMemo);
				else sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				PEszset (hWndPE, PEP_szSUBTITLE,text);
				PEszset (hWndPE, PEP_szXAXISLABEL, "Time (sec)");
				PEszset (hWndPE, PEP_szYAXISLABEL, "Raw Signal");
				PEnset (hWndPE, PEP_nMANUALSCALECONTROLY, 0);
				break;
			case FREQUENCY : // PlotType = Frequency
				sprintf(text,"%s    %s\0",pApp->BudaVersion,pDoc->PS.DataFileName);
				PEszset (hWndPE, PEP_szMAINTITLE, text);
				if(pDoc->PS.HiResMALDI && pDoc->PS.HiResESI) sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResMALDI) sprintf(text,"MALDI  :  %s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResESI) sprintf(text,"Electrospray  :  %s\0",pDoc->PS.ExperimentMemo);
				else sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				PEszset (hWndPE, PEP_szSUBTITLE,text);
				PEszset (hWndPE, PEP_szXAXISLABEL, "Frequency (Hz)");
				PEszset (hWndPE, PEP_szYAXISLABEL, "Intensity (Raw)");
				break;
			case MASSSPEC : // PlotType = Mass
				sprintf(text,"%s    %s\0",pApp->BudaVersion,pDoc->PS.DataFileName);
				PEszset (hWndPE, PEP_szMAINTITLE, text);
				if(pDoc->PS.HiResMALDI && pDoc->PS.HiResESI) sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResMALDI) sprintf(text,"MALDI  :  %s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResESI) sprintf(text,"Electrospray  :  %s\0",pDoc->PS.ExperimentMemo);
				else sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				PEszset (hWndPE, PEP_szSUBTITLE,text);
				PEszset (hWndPE, PEP_szXAXISLABEL, "Mass/Charge (m/z)");
				PEszset (hWndPE, PEP_szYAXISLABEL, "Intensity (arbitrary units)");
				break;
			case STICKPLOT : // PlotType = stickplot from MI table
				sprintf(text,"%s    %s\0",pApp->BudaVersion,pDoc->PS.DataFileName);
				PEszset (hWndPE, PEP_szMAINTITLE, text);
				if(pDoc->PS.HiResMALDI && pDoc->PS.HiResESI) sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResMALDI) sprintf(text,"MALDI  :  %s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResESI) sprintf(text,"Electrospray  :  %s\0",pDoc->PS.ExperimentMemo);
				else sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				PEszset (hWndPE, PEP_szSUBTITLE,text);
				PEszset (hWndPE, PEP_szXAXISLABEL, "Mass/Charge (m/z)");
				PEszset (hWndPE, PEP_szYAXISLABEL, "Intensity (arbitrary units)");
				break;
			case TOTAL_ION_CHROMATOGRAM :
				sprintf(text,"%s    %s\0",pApp->BudaVersion,pDoc->PS.DataFileName);
				PEszset (hWndPE, PEP_szMAINTITLE, text);
				if(pDoc->PS.HiResMALDI && pDoc->PS.HiResESI) sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResMALDI) sprintf(text,"MALDI  :  %s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResESI) sprintf(text,"Electrospray  :  %s\0",pDoc->PS.ExperimentMemo);
				else sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				PEszset (hWndPE, PEP_szSUBTITLE,text);
				PEszset (hWndPE, PEP_szXAXISLABEL, "Time (min)");
				PEszset (hWndPE, PEP_szYAXISLABEL, "Intensity (arbitrary units)");
				break;
			case DECONVOLVED_MASS_DATA : // PlotType = Deconvolved mass data
				sprintf(text,"%s    %s\0",pApp->BudaVersion,pDoc->PS.DataFileName);
				PEszset (hWndPE, PEP_szMAINTITLE, text);
				if(pDoc->PS.HiResMALDI && pDoc->PS.HiResESI) sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResMALDI) sprintf(text,"MALDI  :  %s\0",pDoc->PS.ExperimentMemo);
				else if(pDoc->PS.HiResESI) sprintf(text,"Deconvoluted Electrospray  :  %s\0",pDoc->PS.ExperimentMemo);
				else sprintf(text,"%s\0",pDoc->PS.ExperimentMemo);
				PEszset (hWndPE, PEP_szSUBTITLE,text);
				PEszset (hWndPE, PEP_szXAXISLABEL, "Mass (Da)");
				PEszset (hWndPE, PEP_szYAXISLABEL, "Intensity (arbitrary units)");
				break;
			default:
				PEszset (hWndPE, PEP_szMAINTITLE, "No Data Loaded");
				PEszset (hWndPE, PEP_szSUBTITLE,"");
				break;
		} // end switch

			// set global properties
		PEnset (hWndPE, PEP_nGRAPHANNOTATIONTEXTSIZE, 100);
		PEnset (hWndPE, PEP_bAUTOSCALEDATA, FALSE);
		PEnset (hWndPE, PEP_bZOOMMODE, TRUE);	// Zooming Allowed	    
		PEnset (hWndPE, PEP_nSUBSETS, 1);	// Object will manage 1 subsets 
		PEnset (hWndPE, PEP_nPOINTS, ViewLineLength);	// Object will manage ViewLineLength points 
		PEnset (hWndPE, PEP_bPREPAREIMAGES, 1);    // Prepare images in memory 
		PEnset (hWndPE, PEP_nDATAPRECISION, 1);    // Table Data has 1 decimal
		PEnset (hWndPE, PEP_nGRIDLINECONTROL, PEGLC_NONE); // no grid
		PEnset (hWndPE, PEP_nXAXISSCALECONTROL, PEAC_NORMAL); // x-axis in linear (not log) mode
		PEnset (hWndPE, PEP_nYAXISSCALECONTROL, PEAC_NORMAL); // y-axis in linear (not log) mode
		PEnset (hWndPE, PEP_nALLOWZOOMING, PEAZ_HORZANDVERT);
		PEnset (hWndPE, PEP_nZOOMINTERFACEONLY, 1);
		PEnset (hWndPE, PEP_bANNOTATIONSINFRONT, TRUE);
		//PEnset (hWndPE, PEP_bALLOWGRAPHHOTSPOTS, TRUE);
		
			/* the PEP_nSUBSETS property is equal to 1 */
		DWORD dwSubsetColors[1];
		dwSubsetColors[0] = RGB(0, 0, 0); // First Subset is black
		PEvset (hWndPE, PEP_dwaSUBSETCOLORS, dwSubsetColors, 1);
		ViewIsSetup = TRUE;
	} // end if setup
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("         CBudaView::PlotViewLine()	- third point\n");
#endif

	if (ViewLineIsDirty) CalculateViewLine();
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("         CBudaView::PlotViewLine()	- fourth point\n");
#endif

	// now, make sure that plot doesn't go below pips
	if(PlotType == FREQUENCY || PlotType == MASSSPEC || PlotType == STICKPLOT || PlotType == DECONVOLVED_MASS_DATA) {
		miny=-4; // plot from -4 to ~100 on scale of 0 to 100
		PEvset (hWndPE, PEP_fMANUALMINY, &miny, 1);
		PEnset (hWndPE, PEP_nMANUALSCALECONTROLY, 1);
		//TRACE("miny = %f, scale = %f\n",miny,scale);
		//TRACE("x = [%f,%f], y = [%f,%f]\n", zoompointer[zoomstack].minx, zoompointer[zoomstack].maxx, zoompointer[zoomstack].miny, zoompointer[zoomstack].maxy);
	}
	
	double minx, maxx;
	minx = zoomstack[zoompointer].minx;
	maxx = zoomstack[zoompointer].maxx;
	miny = zoomstack[zoompointer].miny;
	maxy = zoomstack[zoompointer].maxy;
	PEnset (hWndPE, PEP_nMANUALSCALECONTROLY, PEMSC_MINMAX);
	PEnset (hWndPE, PEP_nMANUALSCALECONTROLX, PEMSC_MINMAX);
	PEvset (hWndPE, PEP_fMANUALMINX, &minx, 1);
	PEvset (hWndPE, PEP_fMANUALMAXX, &maxx, 1);
	PEvset (hWndPE, PEP_fMANUALMINY, &miny, 1);
	PEvset (hWndPE, PEP_fMANUALMAXY, &maxy, 1);

		// link data to graph
	PEvset (hWndPE, PEP_faYDATA, ViewLine, ViewLineLength);
	PEvset (hWndPE, PEP_faXDATA, ViewLineXPts, ViewLineLength);

		// setup annotations as per display options
	/*int i;
	for(i=0; i<NumAnnotations; i++) {
		PEvsetcell (hWndPE, PEP_szaGRAPHANNOTATIONTEXT, i, "");
		PEvsetcell (hWndPE, PEP_naGRAPHANNOTATIONTYPE, i, PEGAT_NOSYMBOL);
		PEvsetcell (hWndPE, PEP_faGRAPHANNOTATIONX, i, 0);
		PEvsetcell (hWndPE, PEP_faGRAPHANNOTATIONY, i, 0);
	}*/

	char szTmp=0;

	NumAnnotations = 0;
	PEreinitializecustoms(hWndPE);
	GetScaleFactor();

	PlotDisplayParameters();

	if(PlotType == MASSSPEC) {
		NumMI = 0;
		if(pDoc->DisplayParms.labelpoints || pDoc->DisplayParms.LabelCharge) {
			CreateNBMIList();
		}

		if(NumMI>0) {
			PlotPointLabels();
			PlotCSLabels();
			PlotPeakPos();
		}
	}

	PEvsetcell (hWndPE, PEP_szaGRAPHANNOTATIONTEXT, NumAnnotations+1, &szTmp);
	PEvsetcell (hWndPE, PEP_naGRAPHANNOTATIONTYPE, NumAnnotations, PEGAT_NOSYMBOL);
	
		// reinitialize and reset image
	PEreinitialize (hWndPE);
	PEresetimage (hWndPE, 0, 0); 
	ViewIsDirty = FALSE;
	ResetMenus();
	 // end of redraw window

	if(FirstPlot) {
		if(pDoc->TransientIsDefined && !pDoc->PS.FrequencyIsStored)
			OnPlotMassSpectrum();
		FirstPlot = FALSE;
	}
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("         CBudaView::PlotViewLine()	- fifth point\n");
#endif

}

void CBudaView::GetFreqData(BOOL Do2DFFT)
{
	if(!pDoc->TransientIsDefined) {
		MessageBox("No Transient Data Available for this file","Plotting Error",MB_OK);
		return;
	}
	pDoc->CalculatingFFT = TRUE;
	pDoc->GetCurrentFreqArray(pDoc->CurrentSpectrumFocus, Do2DFFT);
}

void CBudaView::OnPlotMassSpectrum() 
{
	// note: this is the function that usually is used to trigger the fft
	// as it is the one that calls GetFreqData which calls pDoc->GetCurrentFreqArray
	// after the frequency domain data is retrieved from FFTThrd, the view
	// receives a WM_FFTDONE message which triggers OnFFTDone() which finishes
	// the plotmassspec job.

	if (pDoc->IsFFTed()) {
		if(PlotType != MASSSPEC) {
			PlotType = MASSSPEC;
			InitializeZoomStack();
		}
		ViewLineIsDirty = TRUE;
		ViewIsDirty = TRUE;
		pDoc->UpdateAllViews(NULL,0,NULL);
	} else {
		// set FFTThread to notify this view when FFT is done
		PostFFTPlotType = MASSSPEC;
		PlotType = TRANSIENT;
		InitializeZoomStack();
		ViewLineIsDirty = TRUE;
		ViewIsDirty = TRUE;
		pDoc->m_hwndNotifyFFTDone = this->m_hWnd;
		pDoc->m_hFFTDone = m_hFFTDone;
		if (pDoc->TransientIsDefined) GetFreqData(FALSE);
		else MessageBox("Error 57\nTime domain data undefined\nCannot do FFT","Data Error");
	}
}

void CBudaView::OnPlotFrequency() 
{

	if (pDoc->IsFFTed()) {
		if(PlotType != FREQUENCY) {
			PlotType = FREQUENCY;
			InitializeZoomStack();
		}
		ViewLineIsDirty = TRUE;
		ViewIsDirty = TRUE;
		pDoc->UpdateAllViews(NULL,0,NULL);
	} else {
		// set FFTThread to notify this view when FFT is done
		PostFFTPlotType = FREQUENCY;
		pDoc->m_hwndNotifyFFTDone = this->m_hWnd;
		pDoc->m_hFFTDone = m_hFFTDone;
		if (pDoc->TransientIsDefined) GetFreqData(FALSE);
		else MessageBox("Error 58\nTime domain data undefined\nCannot do FFT","Data Error");
	}
}

void CBudaView::OnPlotTransient() 
{
	if(!pDoc->TransientIsDefined) {
		MessageBox("Error 60\nNo Transient Data Available for this file","Plotting Error",MB_OK);
		return;
	}
	if(PlotType != TRANSIENT) {
		PlotType = TRANSIENT;
		InitializeZoomStack();
	}
	CalculateViewLine();
	pDoc->UpdateAllViews(NULL,0,NULL);
}

void CBudaView::ZoomViewLineIn()
{
	// 1) reset the array extent that CalculateViewLine operates on
	// 2) Calculate new ViewLine
	// 3) reset params to redraw view

	double minx, maxx, miny, maxy;
	long maxpts;

	// get the window's zoom coordinates
	PEvget(hWndPE,PEP_fZOOMMINX,&minx);
	PEvget(hWndPE,PEP_fZOOMMAXX,&maxx);
	PEvget(hWndPE,PEP_fZOOMMINY,&miny);
	PEvget(hWndPE,PEP_fZOOMMAXY,&maxy);

	// if FlagCSDetState is set, don't zoom, call the calculate CSDet
	// routines instead.
	if(FlagCSDetState == 1) {
		if (PlotType != MASSSPEC) {
			FlagCSDetState = 0;
			return;
		}
		//pDoc->GetChargeState(minx,maxx);
		return;
	}

	// otherwize, zoom.
	if (PlotType==TRANSIENT) maxpts=pDoc->PS.TransientPoints;
	else maxpts=pDoc->PS.SpectrumPoints;

	ViewIsDirty = TRUE;
	ViewLineIsDirty = TRUE;
	zoompointer++;
	if (zoompointer>ZOOMSTACKLENGTH-1) {
		MessageBox("Cannot zoom in further \n Please zoom out first",
			"Zoom Error", MB_ICONINFORMATION | MB_OK);
		zoompointer=ZOOMSTACKLENGTH-1;
		return;
	}

	// calculate new points and new viewline
	zoomstack[zoompointer].miny=double(miny*scale);
	zoomstack[zoompointer].maxy=double(maxy*scale);
	zoomstack[zoompointer].minx=(double)minx;
	zoomstack[zoompointer].maxx=(double)maxx;

	SetScrollbarSize();
	SetScrollbarPos();
	
	PEreinitialize(hWndPE);
	PEresetimage (hWndPE, 0, 0); 
	CalculateViewLine();
}

void CBudaView::OnInitialUpdate()
{
	pDoc=GetDocument();
	CView::OnInitialUpdate();

	pFrame = GetParentFrame();
	StatusBar = (CStatusBar*)pFrame->GetMessageBar();

	ViewLineIsDirty = TRUE;
	ViewIsDirty = TRUE;
	FirstPlot = TRUE;
	UpdateScaleFactor = TRUE;
	pDoc->SetStatusBar(StatusBar);

	InitialView = FALSE;
	PlotType = TRANSIENT; // default setting
	if(pDoc->IsFFTed()) {
		PlotType = MASSSPEC;
	} else if(pDoc->TransientIsDefined) {
			PlotType = TRANSIENT;
			//OnPlotMassSpectrum();
			//InitialView = TRUE;
	} else if(pDoc->IsMIed()) PlotType = STICKPLOT;
	else {
		MessageBox("Current document has no data to plot","Error: data error");
	} 

	//if(pDoc->TransientIsDefined) PlotType = TRANSIENT;
	//else if(pDoc->IsFFTed ) PlotType = MASSSPEC;
	//else if(pDoc->IsMIed) PlotType = STICKPLOT;
	OldPlotType = PlotType;
	ViewAsStickplot = FALSE;

	//if(pDoc->PS.SpectraAcquired>1 ) ViewAsStickplot = TRUE;

	InitializeZoomStack();
	CurrentFreqArrayPtr = pDoc->CurrentFreqArrayPtr;

	CreatePlotWindow();
}

BOOL CBudaView::OnCommand(WPARAM wp, LPARAM lp)
{
	//short shiftstate;
#ifdef _WIN32
    switch (HIWORD(wp))
#else    
    switch (HIWORD(lp))
#endif    
    {
        case PEWN_CLICKED: // for 2d spectra, shift spectra if shift-click
			//shiftstate = GetAsyncKeyState(VK_SHIFT); // -32767 = shift button pressed
			//if(pDoc->PS.SpectraAcquired>1 && shiftstate<0) On2dSetSpec();
			break;
		case PEWN_MOUSEMOVE:
			CurrentXYPos();
			break;
		case PEWN_ZOOMIN:
			ZoomViewLineIn();
			break;
		case PEWN_ZOOMOUT:
			ZoomViewLineOut();
			break;
    }
    return CView::OnCommand(wp, lp);
}

void CBudaView::InitializeZoomStack()
{
	UpdateScaleFactor = TRUE;
	ViewIsSetup=FALSE;
	zoompointer = 0;
	//ViewLineLength=Buda_VIEWLINELENGTH;

	switch(PlotType) {
		case TRANSIENT:
			zoomstack[zoompointer].minx=0.0;
			zoomstack[zoompointer].maxx=(double)pDoc->PtToTime(pDoc->PS.TransientPoints-1);
			zoomstack[zoompointer].miny=-1.0;
			zoomstack[zoompointer].maxy=1.0;
			break;
		case FREQUENCY:
			zoomstack[zoompointer].minx=(double)pDoc->MassToFreq(pDoc->PS.DetectHighMass, pDoc->PS.A, pDoc->PS.B);
			zoomstack[zoompointer].maxx=(double)pDoc->MassToFreq(pDoc->PS.DetectLowMass, pDoc->PS.A, pDoc->PS.B);
			zoomstack[zoompointer].miny=0;
			zoomstack[zoompointer].maxy=1.0;
			break;
		case MASSSPEC:
			zoomstack[zoompointer].minx=(double)pDoc->PS.DetectLowMass;
			zoomstack[zoompointer].maxx=(double)pDoc->PS.DetectHighMass;
			zoomstack[zoompointer].miny=0;
			zoomstack[zoompointer].maxy=120.0;
			break;
		case STICKPLOT:
			zoomstack[zoompointer].minx=(double)pDoc->PS.DetectLowMass;
			zoomstack[zoompointer].maxx=(double)pDoc->PS.DetectHighMass;
			zoomstack[zoompointer].miny=0;
			zoomstack[zoompointer].maxy=1.0;
			break;
		case TOTAL_ION_CHROMATOGRAM:
			zoomstack[0].minx=(double)TwoDSpecnumToMin(0);
			zoomstack[0].maxx=(double)TwoDSpecnumToMin(pDoc->PS.SpectraAcquired-1);
			zoomstack[0].miny=0;
			zoomstack[0].maxy=1.0;
			break;
		case DECONVOLVED_MASS_DATA:
			if(!Decon) return;
			zoomstack[0].minx=(double)Decon->DeconLowMassLimit;
			zoomstack[0].maxx=(double)Decon->DeconHighMassLimit;
			zoomstack[0].miny=0;
			zoomstack[0].maxy=1.0;
			break;
	} // end switch(PlotType)
	SetupScrollbar();
}

void CBudaView::ZoomViewLineOut()
{
	ViewIsDirty = TRUE;
	ViewLineIsDirty=TRUE;
	zoompointer--;
	if (zoompointer<0) {
		zoompointer=0;
	//	MessageBox("Cannot zoom out further \n Full data extent displayed",
	//		"Zoom Error", MB_ICONINFORMATION | MB_OK);
	}

	SetScrollbarSize();
	SetScrollbarPos();

	PEreinitialize(hWndPE);
	PEresetimage (hWndPE, 0, 0); 
	CalculateViewLine();
} // end ZoomViewLineOut()

void CBudaView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
	// destroy the old PE window and create a new one to new window size
	if (PEIsDefined) {
		PEdestroy(hWndPE); 
		PEIsDefined = FALSE;
	}
	else return;
	CreatePlotWindow();
	
	// redraw viewline
	OnPlot();
}

void CBudaView::OnNewwindow() 
{
	CDocTemplate *doct = pDoc->GetDocTemplate();
	CFrameWnd *cframe = doct->CreateNewFrame(pDoc,NULL);
	doct->InitialUpdateFrame( cframe, pDoc, TRUE );
	ViewIsDirty = TRUE;
}

void CBudaView::ScaleViewLine()
{
	int i;
	double maxdata, mindata;
	double scalefactor = VScale;

	if (IsZoomingY) {
		for (i=0; i<ViewLineLength; i++) {
			if (ViewLine[i]>zoomstack[zoompointer].maxy) ViewLine[i]=(float)zoomstack[zoompointer].maxy;
			if (ViewLine[i]<zoomstack[zoompointer].miny) ViewLine[i]=(float)zoomstack[zoompointer].miny;
		}
	}

	maxdata = ViewLine[0];
	mindata = ViewLine[0];
	for (i=0; i<ViewLineLength; i++) {
		if (maxdata < ViewLine[i]) maxdata = ViewLine[i];
		if (mindata > ViewLine[i]) mindata = ViewLine[i];
	}
	scale = (maxdata-mindata)/100;
	ViewLineMaxy = maxdata;
	ViewLineMiny = mindata;
	// rescale ViewLine to 100%
	if(scale > 0) {
		for (i=0; i<ViewLineLength; i++) {
			ViewLine[i]=(float)((scalefactor)*((double)ViewLine[i]/double(scale)));
		}
	} else scale = 1;
	
	if(PlotType != TRANSIENT) {
		zoomstack[zoompointer].maxy = double(1.1*ViewLineMaxy/scale);
		zoomstack[zoompointer].miny = double(ViewLineMiny/scale - 0.04*(ViewLineMaxy-ViewLineMiny)/scale);
	} else {
		zoomstack[zoompointer].maxy = double(ViewLineMaxy/scale);
		zoomstack[zoompointer].miny = double(ViewLineMiny/scale);
	}
}

void CBudaView::CalcTransientViewLine()
{
	long *data;
	int i,j;
	double rangepts; // points per sample range
	long beginpt;
	double maxdata, mindata;
	double globalmax, globalmin;
	long minx, maxx, temp, maxpts;
	double miny, maxy;

	
	data = pDoc->TimeArray;
	maxpts=pDoc->PS.TransientPoints;

	minx=pDoc->TimeToPt(zoomstack[zoompointer].minx);
	maxx=pDoc->TimeToPt(zoomstack[zoompointer].maxx);
	miny=zoomstack[zoompointer].miny;
	maxy=zoomstack[zoompointer].maxy;

	if(minx>maxx) {
		temp=minx;
		minx=maxx;
		maxx=temp;
	}

	int mini = 0;
	int maxi = 0;
	// check the initial and final points and set them if necessary
	//ASSERT(beginpt>=0 && beginpt<=maxpts);
	if(minx<0) {
		mini = 2;
		ViewLine[0]=(float)0.01;
		ViewLine[1]=(float)0.01;
		ViewLineXPts[0]=(float)zoomstack[zoompointer].minx;
		ViewLineXPts[1]=(float)(1.0/pDoc->PS.AdcRate);
		minx = 0;
	}
	if((long)minx>pDoc->PS.TransientPoints)
		minx = 0;
	if(maxx<0)
		maxx = pDoc->PS.TransientPoints;
	if((long)maxx>pDoc->PS.TransientPoints) {
		maxi=2;
		maxx = pDoc->PS.TransientPoints;
		ViewLine[ViewLineLength-1]=(float)0.01;
		ViewLine[ViewLineLength-2]=(float)0.01;
		ViewLineXPts[ViewLineLength-1]=(float)zoomstack[zoompointer].maxx;
		ViewLineXPts[ViewLineLength-2]=(float)(pDoc->PS.TransientPoints/pDoc->PS.AdcRate);
	}

	rangepts = (maxx-minx)/double(ViewLineLength-mini-maxi);
	scale = 0;
	
	globalmax = globalmin = data[mini];
	if ((maxx-minx)<ViewLineLength) {  // if zoomed in far enough, plot all the points
		ViewLineLength=maxx-minx;
		for(i=mini;i<ViewLineLength;i++) {
			globalmax = (globalmax > data[minx+i]) ? globalmax : data[minx+i];
			globalmin = (globalmin < data[minx+i]) ? globalmin : data[minx+i];
			ViewLine[i]=(float)(data[minx+i]);
			ViewLineXPts[i]=(float) ((minx+i)/(double)pDoc->PS.AdcRate); 
			if (IsZoomingY) {
				if (ViewLine[i]>maxy) ViewLine[i]=(float)maxy;
				if (ViewLine[i]<miny) ViewLine[i]=(float)miny;
			}
		}
		rangepts=1;
	} else { // plot both minimum and maximum points
		for (i=mini; i<ViewLineLength-maxi; i+=2) {
			beginpt = long(i*rangepts)+minx;
		// now find the data for each individual range of points
			maxdata = data[beginpt];
			mindata = data[beginpt];
			for (j=0; j<rangepts; j++) { // find max and min point on range
				maxdata = (maxdata > data[beginpt+j]) ? maxdata : data[beginpt+j];
				mindata = (mindata < data[beginpt+j]) ? mindata : data[beginpt+j];
				}
			globalmax = (globalmax > maxdata) ? globalmax : maxdata;
			globalmin = (globalmin < mindata) ? globalmin : mindata;

			ViewLine[i] = (float)(maxdata);
			ViewLine[i+1] = (float)(mindata);
			ViewLineXPts[i]=(float)((minx+i*rangepts)/pDoc->PS.AdcRate);
			ViewLineXPts[i+1]=(float)((minx+(i+1)*rangepts)/pDoc->PS.AdcRate);
			if (IsZoomingY) {
				if (ViewLine[i]>maxy) ViewLine[i]=(float)maxy;
				if (ViewLine[i]<miny) ViewLine[i]=(float)miny;
				if (ViewLine[i+1]>maxy) ViewLine[i+1]=(float)maxy;
				if (ViewLine[i+1]<miny) ViewLine[i+1]=(float)miny;
			}
		} // end for
	} // end if
	ScaleViewLine();

} // end CBudaView::CalcTransientViewLine()

void CBudaView::CalcFrequencyViewLine()
{
	float *data;
	int i,j;
	double rangepts; // points per sample range
	long beginpt;
	double maxdata;
	long minx, maxx, temp, maxpts;
	double miny, maxy;
	
	data=pDoc->FreqArray;
	maxpts=pDoc->PS.SpectrumPoints;

	minx=pDoc->FreqToPt(zoomstack[zoompointer].minx);
	maxx=pDoc->FreqToPt(zoomstack[zoompointer].maxx);
	miny=zoomstack[zoompointer].miny;
	maxy=zoomstack[zoompointer].maxy;

	if(minx>maxx) {
		temp=minx;
		minx=maxx;
		maxx=temp;
	}

	// test that the minx and maxx values are good and reset if not
	int mini = 0;
	int maxi = 0;
	// check the initial and final points and set them if necessary
	//ASSERT(beginpt>=0 && beginpt<=maxpts);
	if(minx<=0) { // if negative point numbers or DC
		mini = 2;
		ViewLine[0]=(float)0.01;
		ViewLine[1]=(float)0.01;
		ViewLineXPts[0]=(float)zoomstack[zoompointer].minx;
		ViewLineXPts[1]=(float)pDoc->PtToFreq(1);
		minx = 1;
	}
	// if "lowest" frequency > freq of lowest mass
	// Lowest frequency = freq of highest mass
	if(pDoc->PS.DetectMode != MixerMode) {
		if(minx>pDoc->MassToPt(pDoc->PS.DetectLowMass))
			minx = pDoc->MassToPt(pDoc->PS.DetectHighMass);
		// if highest frequency is negative or DC
		// highest freq = freq of lowest mass
		if(maxx<=0)
			maxx = pDoc->MassToPt(pDoc->PS.DetectLowMass);
		// if highest frequency > freq of lowest mass, 
		if(maxx>pDoc->MassToPt(pDoc->PS.DetectLowMass)) {
			maxi=2;
			maxx = pDoc->MassToPt(pDoc->PS.DetectLowMass);
			ViewLine[ViewLineLength-1]=(float)0.01;
			ViewLine[ViewLineLength-2]=(float)0.01;
			ViewLineXPts[ViewLineLength-1]=(float)zoomstack[zoompointer].maxx;
			ViewLineXPts[ViewLineLength-2]=(float)pDoc->MassToFreq(pDoc->PS.DetectLowMass);
		}
		// finished testing minx and maxx
	}

	rangepts = (maxx-minx)/double(ViewLineLength-mini-maxi);
	ASSERT(rangepts<=(maxpts/ViewLineLength));
	scale=0;

	if ((maxx-minx)<ViewLineLength) {  // if zoomed in far enough, plot all the points
		ViewLineLength=maxx-minx;
		for(i=mini;i<ViewLineLength-maxi;i++) {
			ViewLine[i]=(float)(data[minx+i]);
			ViewLineXPts[i]=(float)pDoc->PtToFreq(minx+i);
			scale= (scale>ViewLine[i]) ? scale : ViewLine[i];
			if (IsZoomingY) {
				if (ViewLine[i]>maxy) ViewLine[i]=(float)maxy;
				if (ViewLine[i]<miny) ViewLine[i]=(float)miny;
			}
		}
		rangepts=1;
	} else { // plot both minimum and maximum points
		for (i=mini; i<ViewLineLength-maxi; i++) {
			beginpt = long(i*rangepts)+minx;
			ASSERT(beginpt>=0 && beginpt<=maxpts);
			maxdata = data[beginpt];
			for (j=0; j<rangepts; j++) { // find max point on range
				maxdata = (maxdata > data[beginpt+j]) ? maxdata : data[beginpt+j];
				}
			ViewLine[i] = (float)(maxdata);
			ViewLineXPts[i]=(float)pDoc->PtToFreq(beginpt+j/2);
			scale= (scale>maxdata) ? scale : maxdata;
			if (IsZoomingY) {
				if (ViewLine[i]>maxy) ViewLine[i]=(float)maxy;
				if (ViewLine[i]<miny) ViewLine[i]=(float)miny;
			}
		} // end for
	}
	ScaleViewLine();
	//CurrentFreqArrayPtr = pDoc->CurrentSpectrumFocus;
}  // end CBudaView::CalcFrequencyViewLine()

void CBudaView::CalcMassSpecViewLine()
{
	float *data;
	double noise;
	int i, avptr=0;
	double rangepts; // points per sample range
	long currentpt=0,beginpt,endpt, oldendpt, k, tempm;
	double maxdata, mindata, avgmax[AVG_MAX_DEPTH];
	double beginmass, endmass, massstep;
	long minx, maxx, temp, maxpts,nullpts;
	double miny, maxy;
	
	for (k=0; k<AVG_MAX_DEPTH; k++) avgmax[k]=0;

	data=pDoc->FreqArray;
	// trace the actual data values here
	CString str;

	// diagnostic check of the data.  Should always be positive due to magnitude
	// mode calculation
#ifdef Debug
	for(temp=0; temp<pDoc->PS.SpectrumPoints; temp++) {
		if(data[temp]<0) {
			str.Format("Buda::CalcMassSpecViewLine data error: data value negative = %e\n", data[temp]);
			TRACE(str);
		}
	}
	// end trace
#endif

	maxpts=pDoc->PS.SpectrumPoints;
	noise=double(pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NoiseLevel);

	beginmass = zoomstack[zoompointer].minx;
	endmass = zoomstack[zoompointer].maxx;
	if (beginmass<pDoc->PS.DetectLowMass) 
		beginmass=pDoc->PS.DetectLowMass;
	if (endmass>100000.0 && pDoc->PS.DetectMode != MixerMode) 
		endmass=100000.0;
	if (endmass>pDoc->PS.DetectHighMass && pDoc->PS.DetectMode == MixerMode) 
		endmass=pDoc->PS.DetectHighMass;
	if (endmass<pDoc->PS.DetectLowMass) endmass = pDoc->PS.DetectHighMass;
	
	minx=pDoc->MassToPt(endmass);
	maxx=pDoc->MassToPt(beginmass);
	miny=zoomstack[zoompointer].miny;
	maxy=zoomstack[zoompointer].maxy;

	if(minx>maxx) {
		temp=minx;
		minx=maxx;
		maxx=temp;
	}

	// test that the minx and maxx values are good and reset if not
	massstep=(endmass-beginmass)/ViewLineLength;
	rangepts = (maxx-minx)/double(ViewLineLength);
	ASSERT(rangepts<=(maxpts/ViewLineLength));
	scale=0;
	nullpts=0;
	oldendpt=pDoc->MassToPt(beginmass);

	if ((maxx-minx)<=ViewLineLength) {  // if zoomed in far enough, plot all the points
		ViewLineLength=maxx-minx;
		for(i=0;i<ViewLineLength;i++) {
			ViewLine[i]=(float)(data[minx+(long)i]);
			ViewLineXPts[i]=(float)pDoc->PtToMass(minx+(long)i);
			scale= (scale>ViewLine[i]) ? scale : ViewLine[i];
		}
		rangepts=1;
	} else { // plot it spaced linearly in mass, not in points
		switch(DECIMATION_METHOD) {
		case AVERAGE_NOISE:
			for (i=0; i<ViewLineLength; i++) {
				if (i==0) {
					endpt=pDoc->MassToPt(beginmass);
					oldendpt = endpt;
				}
				else endpt = oldendpt;
				beginpt = pDoc->MassToPt(beginmass+((i+nullpts+1)*massstep));
				oldendpt = beginpt;
				if(beginpt>endpt) {
					tempm=beginpt;
					beginpt=endpt;
					endpt=tempm;
				}
				if(abs(beginpt-endpt)<2) { // if only one point in the range 
					if (currentpt!=beginpt) {  // if the first time we were at this point
						currentpt=beginpt;
						ViewLine[i] = (float)(data[currentpt]);
						ViewLineXPts[i]=(float)(beginmass+(i*massstep));
						scale= (scale>ViewLine[i]) ? scale : ViewLine[i];
					} else { 
						// if not the first time at this point, decrement ViewLineLength
						// and increment nullpts, and decrement i so that you won't have
						// an undefined point in the ViewLine
						ViewLineLength--;
						nullpts++;
						i--;
					}
				} else {
					maxdata=data[beginpt];
					mindata=data[beginpt];
					for (k=beginpt; k<endpt; k++) { // find max point on range
						maxdata = (maxdata > data[k]) ? maxdata : data[k];
						mindata = (mindata < data[k]) ? mindata : data[k];
					}
					if (maxdata>noise) {
						// if a peak, plot min and max, except:
						//    for min, plot greater of min or 
						//    a rolling array which contains the maxdata
						//    of the last 10 pts which were not peaks
						ViewLine[i] = (float)(maxdata);
						ViewLineXPts[i]=(float)(beginmass+(i*massstep));
						scale= (scale>ViewLine[i]) ? scale : ViewLine[i];	
						i++;
						if (mindata<noise) {
							if (mindata<avgmax[avptr]) {
								mindata = avgmax[avptr];
								avptr++;
								if (avptr>=AVG_MAX_DEPTH) avptr=0;
							}
						}
						ViewLine[i] = (float)(mindata);
						ViewLineXPts[i]=(float)(beginmass+(i*massstep));
					} else {  // not a peak
						// this is not a peak because maxdata<noise
						// so, store maxdata and increment ptr being
						// careful not to increment above 10 pts
						ViewLine[i] = (float)(maxdata);
						ViewLineXPts[i]=(float)(beginmass+(i*massstep));
						scale= (scale>ViewLine[i]) ? scale : ViewLine[i];	
						avgmax[avptr]=maxdata;
						avptr++;
						if (avptr>=AVG_MAX_DEPTH) avptr=0;
					}
				}
			} // end for i
			break;
		case MIN_MAX:
			for (i=0; i<ViewLineLength; i++) {
				if (i==0) {
					endpt=pDoc->MassToPt(beginmass);
					oldendpt = endpt;
				}
				else endpt = oldendpt;
				beginpt = pDoc->MassToPt(beginmass+((i+nullpts+1)*massstep));
				oldendpt = beginpt;
				if(beginpt>endpt) {
					tempm=beginpt;
					beginpt=endpt;
					endpt=tempm;
				}
				if(abs(beginpt-endpt)<2) { // if only one point in the range 
					if (currentpt!=beginpt) {  // if the first time we were at this point
						currentpt=beginpt;
						ViewLine[i] = (float)(data[currentpt]);
						ViewLineXPts[i]=(float)(beginmass+(i*massstep));
						scale= (scale>ViewLine[i]) ? scale : ViewLine[i];
					} else { 
						// if not the first time at this point, decrement ViewLineLength
						// and increment nullpts, and decrement i so that you won't have
						// an undefined point in the ViewLine
						ViewLineLength--;
						nullpts++;
						i--;
					}
				} else {
					maxdata=data[beginpt];
					mindata=data[beginpt];
					for (k=beginpt; k<endpt; k++) { // find max point on range
						maxdata = (maxdata > data[k]) ? maxdata : data[k];
						mindata = (mindata < data[k]) ? mindata : data[k];
					}
					//  plot min and max, 
					ViewLine[i] = (float)(maxdata);
					ViewLineXPts[i]=(float)(beginmass+(i*massstep));
					scale= (scale>ViewLine[i]) ? scale : ViewLine[i];	
					i++;
					ViewLine[i] = (float)(mindata);
					ViewLineXPts[i]=(float)(beginmass+(i*massstep));
				}
			} // end for i
			break;
		}  // end switch (decimation_method)
	} // end if zoomed in far enough

	// now zoom y if necessary
	ScaleViewLine();
	//CurrentFreqArrayPtr = pDoc->CurrentSpectrumFocus;
} // end CBudaView::CalcMassSpecViewLine()

void CBudaView::SetViewLineLength()
{
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("entering CBudaView::SetViewLineLength()\n");
#endif
	long x1=0,x2=0,diff=0;;
	ViewLineLength = Buda_VIEWLINELENGTH;
	if(PlotType != TRANSIENT) ViewLineLength = ViewLineLength/2;
	if(Printing) ViewLineLength = Buda_VIEWLINELENGTH*2;

	switch (PlotType) {
	case TRANSIENT:
		x1=pDoc->TimeToPt(zoomstack[zoompointer].minx);
		x2=pDoc->TimeToPt(zoomstack[zoompointer].maxx);
		break;
	case FREQUENCY:
		x1=pDoc->FreqToPt(zoomstack[zoompointer].minx);
		x2=pDoc->FreqToPt(zoomstack[zoompointer].maxx);
		break;
	case MASSSPEC:
		x1=pDoc->MassToPt(zoomstack[zoompointer].minx);
		x2=pDoc->MassToPt(zoomstack[zoompointer].maxx);
		break;
	case DECONVOLVED_MASS_DATA:
		if(!Decon) return;
		x1=Decon->MassToPt((float)zoomstack[zoompointer].minx);
		x2=Decon->MassToPt((float)zoomstack[zoompointer].maxx);
		break;
	}
	diff = x2-x1;	
	if(diff<0) diff*=-1;
	ViewLineLength = (diff<ViewLineLength) ? diff : ViewLineLength;	
	
	if(ViewLineLength <= OldViewLineLength) return;
	if(ViewLineLength <= 0) {
		MessageBox("Invalid ViewLineLength in SetViewLine()","Error!");
		ViewLineLength = OldViewLineLength;
		return;
	}

#ifdef DEBUG_ACCESSVIOLATION
	TRACE("deleting and reallocating ViewLine and ViewLineXPts");
#endif
	OldViewLineLength = ViewLineLength;

	if(ViewLine != 0) {
		delete [] ViewLine;
		ViewLine = 0;
	}
	if(ViewLineXPts != 0) {
		delete [] ViewLineXPts;
		ViewLineXPts = 0;
	}

	ViewLine = new float [ViewLineLength];
	ViewLineXPts = new float [ViewLineLength];
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("...done\n");
#endif
}

void CBudaView::OnSetDisplayRange() 
{
	long maxpts;
	CSetDisplayRngDlg dlg;

	if (PlotType==TRANSIENT) maxpts=pDoc->PS.TransientPoints;
	else maxpts=pDoc->PS.SpectrumPoints;

	dlg.PlotType=PlotType;
	dlg.m_Minx = zoomstack[zoompointer].minx;
	dlg.m_Maxx = zoomstack[zoompointer].maxx;
	dlg.m_Miny = zoomstack[zoompointer].miny;
	dlg.m_Maxy = zoomstack[zoompointer].maxy;

	switch(PlotType) {
	case TRANSIENT:
		dlg.minxrng=0;
		dlg.maxxrng=pDoc->PtToTime(maxpts-1);
		break;
	case FREQUENCY:
		dlg.minxrng=0;
		dlg.maxxrng=pDoc->PtToFreq(maxpts);
		break;
	case MASSSPEC:
		dlg.minxrng=pDoc->PS.DetectLowMass;
		dlg.maxxrng=pDoc->PS.DetectHighMass;
		break;
	}

	ViewIsDirty = TRUE;
	ViewLineIsDirty = TRUE;
	zoompointer++;
	if (zoompointer>ZOOMSTACKLENGTH-1) {
		MessageBox("Cannot zoom in further \n Please zoom out first",
			"Zoom Error", MB_ICONINFORMATION | MB_OK);
		zoompointer=ZOOMSTACKLENGTH-1;
		return;
	}

	dlg.DoModal();

	zoomstack[zoompointer].minx = (double)dlg.m_Minx;
	zoomstack[zoompointer].maxx = (double)dlg.m_Maxx;
	zoomstack[zoompointer].miny = (double)dlg.m_Miny;
	zoomstack[zoompointer].maxy = (double)dlg.m_Maxy;
	//CalculateViewLine();
	pDoc->UpdateAllViews(NULL);
}

void CBudaView::OnFilePrint() 
{
	Printing = TRUE;
	CalculateViewLine();
	PEreinitialize (hWndPE);
	PEresetimage (hWndPE, 0, 0); 
	OnPEFilePrint();
}

void CBudaView::OnPEFilePrint() 
{
	POINT *pt=0;
	PElaunchprintdialog(hWndPE,TRUE,pt);
	Printing = FALSE;
	CalculateViewLine();
}

void CBudaView::CreatePlotWindow()
{
	RECT rect;
	GetClientRect(&rect); // Get client size of parent 
	//rect.top = 25;
	//rect.right = (int) (0.6 * (rect.right-rect.left)) + rect.left;
	hWndPE = PEcreate (PECONTROL_SGRAPH, WS_VISIBLE, &rect, m_hWnd, 10);
	if (!hWndPE) {
		MessageBox("Cannot Create New Plot \n Error #1001",
		"PECreate Error", MB_ICONINFORMATION | MB_OK);
		return;
		}
	PEIsDefined = TRUE;
}

void CBudaView::WriteParms(CDC* pDC, CRect rect)
{
	CString buff;
	int y=0;
	int height = 10;

	if (pDoc->DisplayParms.acqrate) {
		buff.Format("Acquisition Rate : %d",pDoc->PS.AdcRate);
		y+=height;
		pDC->TextOut(rect.left, rect.top+y,buff,buff.GetLength());
	}
	//pDC->TextOut((int)(rect.right/20),(int)(rect.bottom/20),"Buda",7);
}

void CBudaView::OnUpdateZoomY(CCmdUI* pCmdUI) 
{
	if (IsZoomingY) pCmdUI->SetCheck(1);
	else pCmdUI->SetCheck(0);
}

void CBudaView::OnZoomY() 
{
	if (IsZoomingY) IsZoomingY=FALSE;
	else IsZoomingY=TRUE;
}

void CBudaView::OnZoomOut() 
{
	ZoomViewLineOut();
}

void CBudaView::OnEditCopy() 
{
	CPoint pt;
	pt.x = 10000;
	pt.y = 5000;
	if (PEcopymetatoclipboard(hWndPE,&pt) == 0) {
		MessageBox("Couldn't copy to Clipboard","Error",MB_OK);
	}
}

void CBudaView::GraphClicked()
{
	GRAPHLOC* g;
	g = new GRAPHLOC;
	HOTSPOTDATA HotSpotData;

	PEvget(hWndPE, PEP_structHOTSPOTDATA, &HotSpotData);
	PEvget(hWndPE, PEP_structGRAPHLOC, (void *)g);
		// testing annotations
	PEvset (hWndPE, PEP_naGRAPHANNOTATIONTYPE, PEGAT_NOSYMBOL, 1);
	//PEvset (hWndPE, PEP_faGRAPHANNOTATIONX, &(g.fXval), 1);
	//PEvset (hWndPE, PEP_faGRAPHANNOTATIONY, &(g.fYval), 1);
	PEvset (hWndPE, PEP_szaGRAPHANNOTATIONTEXT, "testing", 7);
	if(g != 0) {
		delete g;
		g=0;
	}
}

void CBudaView::PlotDisplayParameters()
{
	if(PlotType != MASSSPEC && PlotType != FREQUENCY) return;

	struct DispParms dp;
	int i=NumAnnotations;
	double x = 0.45;
	double y, y0,step;

	dp = pDoc->DisplayParms;

	if (!dp.allparms) return;
	PEnset (hWndPE, PEP_bSHOWANNOTATIONS, TRUE);

	x = zoomstack[zoompointer].maxx - 0.20*(zoomstack[zoompointer].maxx - zoomstack[zoompointer].minx);
	y0 = ViewLineMaxy/scale;
	step = (ViewLineMaxy-ViewLineMiny)/(17*scale);

	CString text;
	int j;

	if (dp.timedate) {
		text = pDoc->PS.TimeAcquired;
		for(j=0; j<40; j++) {
			if(pDoc->PS.TimeAcquired[j] == ' ')	break;
		}

		text = text.Left(j);
		text = "|l" + text;
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++; j++;

		text = pDoc->PS.TimeAcquired;
		text = text.Right(text.GetLength()-j);
		text = "|l" + text;
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.acqrate) {
		text.Format("|lAcq. Rate : %1.0f kHz",pDoc->PS.AdcRate/1000);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.apodfunc) {
		switch (pDoc->PS.WindowFunction) {
		case 0 :
			text.Format("|lApod : none");
			break;
		case 1 :
			text.Format("|lApod : Hanning");
			break;
		case 2 :
			text.Format("|lApod : 1/2 Hanning");
			break;
		case 3 :
			text.Format("|lApod : Bartlett");
			break;
		case 4 :
			text.Format("|lApod : Blackman");
			break;
		case 5 :
			text.Format("|lApod : Hamming");
			break;
		}
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.calconst) {
		text.Format("|lCalibration Constants :");
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;

		text.Format("|l    A = %1.1f",pDoc->PS.A);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;

		text.Format("|l    B = %1.4f",pDoc->PS.B);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;

	}
	if (dp.exciteamp) {
		text.Format("|lExcite Voltage : %1.2f", pDoc->GetExciteVoltage());
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.massrange) {
		text.Format("|lMass Range : ");
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;

		text.Format("|l    %4.2f - %4.2f", pDoc->PS.DetectLowMass, pDoc->PS.DetectHighMass);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.numpts) {
		text.Format("|l# Data Pts : %d", pDoc->PS.TransientPoints);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.numscans) {
		text.Format("|l# Scans : %d", pDoc->PS.ScansPerSpectrum);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.numzerofills) {
		text.Format("|l# Zero Fills : %d", pDoc->PS.ZeroFills);
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.trapvolts) {
		text.Format("|lTrap Voltage : %2.2f", pDoc->GetTrapVoltsOnDetect());
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.noisethreshold) {
		text.Format ("|lThreshold : %1.1f%%",pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NoiseLevel*100/(GlobalMax-GlobalMin));
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.PosNeg) {
		if(pDoc->GetTrapVoltsOnDetect()<0) text = "|lNegative Ion Mode";
		else text = "|lPositive Ion Mode";
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}
	if (dp.ScaleFactor) {
		text.Format("|lScaleFactor : %1.2f",VScale*5/scale);
		if((VScale*5/scale) > 1e5) text = "|lScaleFactor : ******";
		y = y0 - (i-NumAnnotations)*step;
		AddTextAnnotation(x,y,i,&text);
		i++;
	}

	NumAnnotations = i;
}

void CBudaView::PlotPointLabels()
{
	if (!pDoc->DisplayParms.labelpoints) return;
	if (PlotType != MASSSPEC) return;

	if (!pDoc->IsMIed()) {
		// no peaks to label... 
		return;
	}

	PEnset (hWndPE, PEP_bSHOWANNOTATIONS, TRUE);

	int OldSortProperty = pDoc->SortOnProperty;
	BOOL OldSortIncreasing = pDoc->SortIncreasing;

	pDoc->SortOnProperty = 2; // abundance
	pDoc->SortIncreasing = FALSE;
	pDoc->SortMITable();

	long i,j,dig=3;
	long len=0;
	double maxx, minx;
	double x,y;
	double textwidth;
	CString text;
	x=0;
	y=0;
	text = "";

	maxx = zoomstack[zoompointer].maxx;
	minx = zoomstack[zoompointer].minx;

	for(i=0; i<NumMI; i++) {
			// set up text string
		text.Format("%1.6f\0",NBMIList[i].M);
			// find the string length;
		if(NBMIList[i].M>=10000.0) len = 6+dig;
		if(NBMIList[i].M>=1000.0 && NBMIList[i].M<10000.0) len = 5+dig;
		if(NBMIList[i].M<1000.0 && NBMIList[i].M>=100.0) len = 4+dig;
		if(NBMIList[i].M<100.0) len = 3+dig;
		text = text.Left(len);
		textwidth = len*(maxx-minx)/100; // 120 characters fit in the screen horizontally

		// set up x,y positions
		// start with 0 shift and if the text doesn't fit, shift right
		// and left in 10% (of textwidth) increments until either it 
		// does fit, or have shifted 50%.  
		for(j=0; j<6; j++) {
			//shift right
			x = NBMIList[i].M + ((double)j/10.0)*textwidth;
			y = NBMIList[i].I/scale;
			if(AddAnnotation(x,y,i,text)) break;
			if(j==0) continue;

			//shift left
			x = NBMIList[i].M - ((double)j/10.0)*textwidth;
			y = NBMIList[i].I/scale;
			if(AddAnnotation(x,y,i,text)) break;
		}		
	}

	pDoc->SortOnProperty = OldSortProperty;
	pDoc->SortIncreasing = OldSortIncreasing;
	pDoc->SortMITable();
}

BOOL CBudaView::PtsInRange(double mass, double intensity, double width)
{
	// return true if there are any line points within the range specified
	// in the parameters otherwise, return false
	int i, j;
	intensity=intensity*1.02;
	
	//first, find the position in ViewLine of the current mass
	for(j=0; ViewLineXPts[j]<mass && j<ViewLineLength; j++);

	// if there are any points in ViewLine within +/- width/2 of 
	// our current mass-intensity point, return true 
	for(i=0; ViewLineXPts[i+j]<mass+width/2; i++) {
		if(j-i<0) break;
		if(j+i > ViewLineLength) break;
		if(ViewLine[j+i] > intensity) return TRUE;
		if(ViewLine[j-i] > intensity) return TRUE;
	}

	/*
	if (direction > 0) { // going up the viewline
		for (i=j+2; ViewLineXPts[i]<pt+width; i++)
			if(ViewLine[i]>height) return TRUE;
	} else {
		for (i=j-2; ViewLineXPts[i]>pt-width; i--)
			if(ViewLine[i]>height) return TRUE;
	}
	*/
	return FALSE;
}

BOOL CBudaView::PeaksInRange(int ptind, double x, double y, double width)
{
	if(ptind < 0 || ptind > NumMI) ptind = NumMI;

	// return TRUE if any textbox overlaps with the textbox given by 
	// the parameters otherwise, return FALSE 
	int i;
	double m,I;

	// sorted on abundance, so don't need to search peaks smaller than current
	// one.  Only returns true if there is a peak in m+/-width that is taller
	// than the current one - that way labels only the tallest peaks
	for(i=0; i<ptind; i++) { 
		if(i==ptind) continue;
		m=NBMIList[i].M;
		I=NBMIList[i].I/scale;
		if(m < x-width) continue;
		if(m > x+width) continue;

		if(I>y) return TRUE;	
	}

	return FALSE;
}

void CBudaView::OnMouseMove(UINT nFlags, CPoint point) 
{
	double x=0, y=0;
	x=x;
	y=y;

	int xp, yp;
	xp = point.x;
	yp = point.y;
	//PEconvpixeltograph(hWndPE, 0, &xp, &yp, &x,&y, FALSE, FALSE, FALSE);

	CView::OnMouseMove(nFlags, point);
}


void CBudaView::DumpToDiagFile(CString str)
{
#ifdef DEBUG_DUMP_DIAGS_TO_FILE
	str=str+"\r\n";
	//(CBudaApp*)AfxGetApp()->diagfile.Write(str,str.GetLength());
	TRACE(str);
#endif
}

void CBudaView::PlotPeakPos()
{
	if (!pDoc->DisplayParms.LabelPeakPos) return;
	if(PlotType != MASSSPEC) return;
	if(!pDoc->IsMIed()) {
		//pDoc->PeakPickFlag = TRUE;
		//pDoc->PeakPick();
		// no peaks to plot
		return;
	}
	PEnset (hWndPE, PEP_bSHOWANNOTATIONS, TRUE);

	// this function places a little plus sign at the point
	// of each element in the MI table. 
	int plus = PEGAT_SMALLPLUS;

	long i;
	double x,y;
	for(i=0; i<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks; i++) {
		PEvsetcell(hWndPE, PEP_naGRAPHANNOTATIONTYPE, i+NumAnnotations, &plus);
		x = GetMITableMass(i);
		y=pDoc->FITable[i].I/scale;
		PEvsetcell(hWndPE, PEP_faGRAPHANNOTATIONX, i+NumAnnotations, &(x));
		PEvsetcell(hWndPE, PEP_faGRAPHANNOTATIONY, i+NumAnnotations, &(y));
	}
	NumAnnotations+=i;
}

double CBudaView::GetMITableMass(struct FITableElement *fi)
{
	return pDoc->FreqToMass(fi->F, pDoc->PS.A, pDoc->PS.B);
}

double CBudaView::GetMITableMass(long i)
{
	return pDoc->GetMITableMass(i);
}

void CBudaView::OnPlot()
{
#ifdef DEBUG_ACCESSVIOLATION
	TRACE("entering CBudaView::OnPlot()\n");
#endif
	switch(PlotType) {
	case TRANSIENT:	
		OnPlotTransient();
		break;
	case FREQUENCY:
		OnPlotFrequency();
		break;
	case MASSSPEC:
		OnPlotMassSpectrum();
		break;
	case TOTAL_ION_CHROMATOGRAM:
		OnViewChromatogram();
		break;
	case STICKPLOT:
		OnPlotStickPlot();
		break;
	case DECONVOLVED_MASS_DATA:
		OnViewDecon();
		break;
	default:
		break;
	}
}

void CBudaView::CurrentXYPos()
{
	POINT pt;
	CString str;

	PEvget(hWndPE,PEP_ptLASTMOUSEMOVE,&pt);

	int i, ptx, pty;// axis number
	double x,y;		// x and y in graph units
	ptx = (int)pt.x;
	pty = (int)pt.y;
	PEconvpixeltograph(hWndPE,&i,&(ptx),&(pty),&x,&y, FALSE, FALSE, FALSE);
	if(pDoc->PS.SpectraAcquired>1) {
		// here if PlotType = TIC, x = min, and plotnum should change
		if (PlotType == TOTAL_ION_CHROMATOGRAM)
			str.Format("scan %d : %3.4f, %2.2f",1+TwoDMinToSpecnum(x),x,y);
		else str.Format("scan %d : %3.4f, %2.2f",1+pDoc->CurrentSpectrumFocus,x,y);
	} else
		str.Format("%3.4f, %2.2f",x,y);
	StatusBar->SetPaneText(0, str, TRUE);
}

void CBudaView::OnViewChromatogram() 
{
	int i;

	BOOL IsMIed = TRUE;
	for (i=0; i<pDoc->PS.SpectraAcquired; i++) {
		if(!pDoc->PS.Spec[i].IsMIed) IsMIed = FALSE;
	}
	if (IsMIed) {
		PlotType = TOTAL_ION_CHROMATOGRAM;
		InitializeZoomStack();
		ViewLineIsDirty = TRUE;
		ViewIsDirty = TRUE;
		pDoc->UpdateAllViews(NULL,0,NULL);
	} else {
		PostFFTPlotType = MASSSPEC; // can't display tic till all spectra are done
		// set FFTThread to notify this view when FFT is done
		pDoc->m_hwndNotifyFFTDone = m_hWnd;
		pDoc->m_hFFTDone = m_hFFTDone;
		TwoDFFTStatus = pDoc->PS.SpectraAcquired;
		TwoDFFT = TRUE;
		TwoDFFTPlotType = TOTAL_ION_CHROMATOGRAM;
		if (pDoc->TransientIsDefined) GetFreqData(TRUE);
		else MessageBox("Error 57\nTime domain data undefined\nCannot do FFT","Data Error");
	}
}

void CBudaView::CalcTICViewLine()
{
	int i;
	int s;
	double tempint;
	double mass, width, peakmass;
	double minx = 0.0;
	double maxx = 0.0;
	double temp = 0.0;

	minx=zoomstack[zoompointer].minx;
	maxx=zoomstack[zoompointer].maxx;
	if(minx>maxx) {
		temp=minx;
		minx=maxx;
		maxx=temp;
	}

	// the line is a plot of the DS[??].TotalIonInts.  The ViewLineXPts
	// is = PS.ElapsedTime[??] * 6000 to change from hundredths of seconds
	// to minutes
	int viewlineptr = 0;

	// first, set the initial zero points
	if(minx < (double)TwoDSpecnumToMin(0)) {
		ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)minx;
		if(ViewLineXPts[viewlineptr]==0.0) ViewLineXPts[viewlineptr]=(float)0.001;
		viewlineptr++;
		ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)TwoDSpecnumToMin(0);
		if(ViewLineXPts[viewlineptr]==0.0) ViewLineXPts[viewlineptr]=(float)0.001;
		viewlineptr++;
	}

	// then set the viewline pts equal to their TIT's
	// note this uses pDoc->SetCurrentMIFocus() - so we have to reset
	// the MIFocus to current spectrum focus at the end of the loop.
	mass = pDoc->SICMass;
	width = pDoc->SICWidth;
	for(i=0; i<pDoc->PS.SpectraAcquired; i++) {
		pDoc->SetCurrentMIFocus(i);
		if( pDoc->SICmode ) {
			tempint = 0;
			for(s=0; s<pDoc->PS.Spec[i].NumPeaks; s++) {
				peakmass = pDoc->FreqToMass(pDoc->FITable[s].F);
				if((peakmass < mass+width) && (peakmass > mass-width)) {
					tempint += (double)pDoc->FITable[s].I;
				}
			}
			ViewLine[viewlineptr]=(float)tempint;
		} else {
			ViewLine[viewlineptr]=(float)pDoc->PS.Spec[i].TotalIonInt;
		}
		if (ViewLine[viewlineptr]==0) ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)TwoDSpecnumToMin(i);
		if(ViewLineXPts[viewlineptr]==0.0) ViewLineXPts[viewlineptr]=(float)0.001;
		viewlineptr++;
	}
	// reset the current MI focus to be in sync with the rest of 
	// the current spectrum focus
	pDoc->SetCurrentMIFocus(pDoc->GetCurrentSpectrumFocus());

	// set the last two zero points.
	if(maxx > (double)TwoDSpecnumToMin(pDoc->PS.SpectraAcquired-1)) {
		ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)TwoDSpecnumToMin(pDoc->PS.SpectraAcquired-1);
		viewlineptr++;
		ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)maxx;
		viewlineptr++;
	}

	ViewLineLength = viewlineptr;

	// scale everything to 100%
	ScaleViewLine();
} // end CalcTICViewLine

void CBudaView::CalcStickPlotViewLine()
{
	// plot this as a zero line from PS.minmass to maxmass
	// with spikes.
	int i;
	double minx, maxx, temp;
	double currmass;

	minx=zoomstack[zoompointer].minx;
	maxx=zoomstack[zoompointer].maxx;
	if(minx>maxx) {
		temp=minx;
		minx=maxx;
		maxx=temp;
	}

	int numpts = pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;
	int viewlineptr=0;
	ViewLine[viewlineptr]=(float)0.01;
	ViewLineXPts[viewlineptr]=(float)minx;
	viewlineptr++;

	for (i=0; i<numpts; i++) {
		currmass = pDoc->FreqToMass(pDoc->FITable[i].F, pDoc->PS.A, pDoc->PS.B);

		if ((double)currmass < minx || (double)currmass > maxx) continue;
		ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)currmass;
		viewlineptr++;
		
		ViewLine[viewlineptr]=(float)pDoc->FITable[i].I;
		ViewLineXPts[viewlineptr]=(float)currmass;
		viewlineptr++;

		ViewLine[viewlineptr]=(float)0.01;
		ViewLineXPts[viewlineptr]=(float)currmass;
		viewlineptr++;
	}

	ViewLine[viewlineptr]=(float)0.01;
	ViewLineXPts[viewlineptr]=(float)maxx;
	viewlineptr++;

	ViewLineLength = viewlineptr;
	ScaleViewLine();
}

void CBudaView::OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView) 
{
	pDoc->ResetMenus();
	CView::OnActivateView(bActivate, pActivateView, pDeactiveView);
}


void CBudaView::ResetMenus()
{
#ifdef INCLUDE_DIAGNOSTICS
	//pDoc->Diags.SetupDiagnosticsMenu();
#endif
	Update2DDataMenu();
	DrawMenuBar();
}

void CBudaView::Update2DDataMenu()
{
/*	CString d = "viewtest";
	CChildFrame* frm = (CChildFrame*)GetParentFrame();
	CMenu* pTopMenu = frm->GetMenu();

	TwoDDataMenu.LoadMenu(IDR_2D_DATA_MENU);
	HMENU mnu = TwoDDataMenu.GetSafeHmenu();
	pTopMenu->AppendMenu( MF_POPUP,(UINT)mnu,d);
*/
}

void CBudaView::On2dSetSpec() 
{
	if(PlotType != TOTAL_ION_CHROMATOGRAM) return;
	POINT pt;
	PEvget(hWndPE,PEP_ptLASTMOUSEMOVE,&pt);

	int i, ptx, pty;// axis number
	double x,y;		// x and y in graph units
	ptx = (int)pt.x;
	pty = (int)pt.y;
	PEconvpixeltograph(hWndPE,&i,&(ptx),&(pty),&x,&y, FALSE, FALSE, FALSE);

	TICLineFocus=TwoDMinToSpecnum(x);
	pDoc->SetCurrentSpectrumFocus(TICLineFocus);

	pDoc->RecalcAllViews();
}

int CBudaView::TwoDMinToSpecnum(double x)
{
	int s,i=0;
	x=x*6000; // change from minutes to hundredths of seconds like ElapsedTime
	for(s=0; s<pDoc->PS.SpectraAcquired; s++) {
		if(fabs(x-pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].ElapsedTime)<fabs(x-pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].ElapsedTime)) i=s;
	}
	return i;
}

void CBudaView::OnDiagsKeytest() 
{
	short i;
	//for (i=0; i<100; i++) {
	//	state = GetAsyncKeyState(VK_SHIFT);
	//}
	
	double *test;
	test = new double[5000];
	PEvget(hWndPE, PEP_faXDATA, test);
	for (i=0; i<5; i++) 
		test[i]=test[i];
	delete test;
}

void CBudaView::OnPlotStickPlot()
{
	PlotType = STICKPLOT;

	if (pDoc->IsFFTed()) {
		// note that not running through InitializeZoomStack() here
		// causes errors on resizing
		InitializeZoomStack();
		ViewLineIsDirty = TRUE;
		ViewIsDirty = TRUE;
		pDoc->UpdateAllViews(NULL,0,NULL);
	} else {
		// set FFTThread to notify this view when FFT is done
		pDoc->m_hwndNotifyFFTDone = m_hWnd;
		pDoc->m_hFFTDone = m_hFFTDone;
		if (pDoc->TransientIsDefined) GetFreqData(FALSE);
		else MessageBox("Error 59\nTime domain data undefined\nCannot do FFT","Data Error");
	}

}

void CBudaView::On2dSetSpecNext() 
{
	int focus;
	focus = pDoc->CurrentSpectrumFocus;
	if(focus < 0 || focus > pDoc->PS.SpectraAcquired) return;
	if(focus+1 >= pDoc->PS.SpectraAcquired) {
		MessageBox("Cannot increment spectrum number\n This is the last spectrum","TIC spectrum focus", MB_OK);
		return;
	}
	pDoc->SetCurrentSpectrumFocus(focus+1);

	pDoc->RecalcAllViews();
}

void CBudaView::On2dSetSpecPrev() 
{
	int focus;
	focus = pDoc->CurrentSpectrumFocus;
	if(focus < 0 || focus > pDoc->PS.SpectraAcquired) return;
	if(focus-1 < 0) {
		MessageBox("Cannot decrement spectrum number\n This is the first spectrum","TIC spectrum focus", MB_OK);
		return;
	}
	pDoc->SetCurrentSpectrumFocus(focus-1);

	pDoc->RecalcAllViews();
}

void CBudaView::OnViewStickplot() 
{
/*	if(ViewAsStickplot == FALSE) {
		OldPlotType = PlotType;
		if(PlotType == MASSSPEC || PlotType == FREQUENCY) PlotType = STICKPLOT;
		ViewAsStickplot = TRUE;
	} else {
		PlotType = OldPlotType;
		ViewAsStickplot = FALSE;
	}
*/
}

void CBudaView::OnUpdateViewStickplot(CCmdUI* pCmdUI) 
{
//	if(ViewAsStickplot == TRUE) pCmdUI->SetCheck(TRUE);
//	else pCmdUI->SetCheck(FALSE);
	pCmdUI->Enable(FALSE);
}

double CBudaView::TwoDSpecnumToMin(int specnum)
{
	double d;

	d=(double)pDoc->PS.Spec[specnum].ElapsedTime;
	d=d/6000; // change from hundredths of seconds like ElapsedTime to minutes 
	return d;
}

void CBudaView::OnUpdateViewChromatogram(CCmdUI* pCmdUI) 
{
	if(pDoc->PS.SpectraAcquired>1) pCmdUI->Enable(TRUE);
	else pCmdUI->Enable(FALSE);
//	pCmdUI->Enable(FALSE);
}

void CBudaView::OnUpdateViewDecon(CCmdUI* pCmdUI) 
{
//	if(pDoc->IsMIed) pCmdUI->Enable(TRUE);
//	else pCmdUI->Enable(FALSE);
	pCmdUI->Enable(FALSE);
}

void CBudaView::OnViewDecon() 
{
	long i;
	// setup Decon class
	if(!Decon) { // if there is no Decon yet defined...
		Decon = new CDeconvolution;
		Decon->pApp = (CBudaApp*)AfxGetApp();
		Decon->pDoc = pDoc;
		Decon->pView = this;
	}
	Decon->AdductMass = pApp->DeconAdductMass;
	Decon->DeconMethod = pApp->DeconMethod;
	Decon->DeconHighChargeLimit = pApp->DeconHighChargeLimit;
	Decon->DeconLowChargeLimit = pApp->DeconLowChargeLimit;
	Decon->DeconHighMassLimit = pApp->DeconHighMassLimit;
	Decon->DeconLowMassLimit = pApp->DeconLowMassLimit;
	Decon->DeconRes = pApp->DeconResolution;
	Decon->DeconIsotopicallyResolved = pApp->DeconIsotopicallyResolved;
	Decon->DeconReportedZ = 0;

	// call up the decon options dialog box to get the current parms.
	CDeconOptions decops;
	int result;
	decops.m_AdductMass = Decon->AdductMass;
	decops.m_DeconRes = Decon->DeconRes;
	decops.m_HighMass = Decon->DeconHighMassLimit;
	decops.m_LowMass = Decon->DeconLowMassLimit;
	decops.m_MaxCharge = Decon->DeconHighChargeLimit;
	decops.m_MinCharge = Decon->DeconLowChargeLimit;
	decops.m_DeconMethod = Decon->DeconMethod;
	decops.m_ReportedZ = Decon->DeconReportedZ;
	result = decops.DoModal();
	if (result == IDCANCEL) {
		delete Decon;
		Decon = 0;	// so that we don't try to delete it later
		return;
	}
	if (result == IDOK) {
		Decon->AdductMass = decops.m_AdductMass;
		Decon->DeconRes = decops.m_DeconRes;
		Decon->DeconHighMassLimit = decops.m_HighMass;
		Decon->DeconLowMassLimit = decops.m_LowMass;
		Decon->DeconHighChargeLimit = decops.m_MaxCharge;
		Decon->DeconLowChargeLimit = decops.m_MinCharge;
		Decon->DeconMethod = decops.m_DeconMethod;
		Decon->DeconIsotopicallyResolved = decops.m_IsotopicallyResolved;
		Decon->DeconReportedZ = decops.m_ReportedZ;
	}

	//first, calculate the deconvolution if necessary
	// call the appropriate viewline function
	HICON oldcursor = GetCursor();
	SetCursor(LoadCursor(NULL, IDC_WAIT));

	switch(Decon->DeconMethod) {
	case DECON_FENN:
		Decon->CalculateFennDecon();
		break;
	case DECON_REINHOLD:
		Decon->CalculateReinholdDecon();
		break;
	case DECON_OCONNOR:
		Decon->CalculateOConnorDecon();
		break;
	//case DECON_ZHANG:
		//Decon->CalculateZhangDecon();
		// create decon array from Zhang Decon Mass/Intensity pairs
//
//		break;
	}

	// note that "zero" values won't plot with ProEssentials stuff.  
	// so, we must make "zero" values in the DeconData almost, but not
	// exactly zero.
	for(i=0; i<Decon->DeconSize; i++) {
		if(Decon->DeconData[i] <= 0) Decon->DeconData[i]=Decon->MaxDeconIntensity/(float)100000.0;
	}

	SetCursor(oldcursor);

	// setup plotting information
	if(PlotType !=DECONVOLVED_MASS_DATA) {
		PlotType = DECONVOLVED_MASS_DATA;
	}
	// note that not running through InitializeZoomStack() here
	// causes errors on resizing
	InitializeZoomStack();
	ViewLineIsDirty = TRUE;
	ViewIsDirty = TRUE;
	pDoc->UpdateAllViews(NULL,0,NULL);
}

void CBudaView::CalcDeconViewLine()
{
	float *data;
	int i,j;
	double rangepts; // points per sample range
	long beginpt;
	double maxdata;
	long minx, maxx, temp, maxpts;
	double miny, maxy;

	if(!Decon) return; // if Decon isn't yet defined, don't do anything.

	data=Decon->DeconData;
	maxpts=Decon->DeconSize;

	minx=Decon->MassToPt((float)zoomstack[zoompointer].minx);
	maxx=Decon->MassToPt((float)zoomstack[zoompointer].maxx);
	miny=zoomstack[zoompointer].miny;
	maxy=zoomstack[zoompointer].maxy;

	// test that the minx and maxx values are good and reset if not
	// check the initial and final points and set them if necessary
	//ASSERT(beginpt>=0 && beginpt<=maxpts);
	if(minx<0) minx=0;
	if(maxx>maxpts) maxx=maxpts;
	if(minx>maxx) {
		temp=minx;
		minx=maxx;
		maxx=temp;
	}


	// check data to see if it's ok.
	if(minx < Decon->MassToPt((float)Decon->DeconLowMassLimit))
		minx = Decon->MassToPt((float)Decon->DeconLowMassLimit);
	if(maxx>Decon->MassToPt((float)Decon->DeconHighMassLimit))
		maxx = Decon->MassToPt((float)Decon->DeconHighMassLimit);
	// finished testing minx and maxx

	rangepts = (double)(maxx-minx)/double(ViewLineLength);
	ASSERT(rangepts<=((double)maxpts/(double)ViewLineLength));
	scale=0;

	if ((maxx-minx)<ViewLineLength) {  // if zoomed in far enough, plot all the points
		ViewLineLength=maxx-minx;
		for(i=0;i<ViewLineLength;i++) {
			ViewLine[i]=(float)(data[minx+i]);
			ViewLineXPts[i]=(float)Decon->PtToMass(minx+i);
			scale= (scale>ViewLine[i]) ? scale : ViewLine[i];
			if (IsZoomingY) {
				if (ViewLine[i]>maxy) ViewLine[i]=(float)maxy;
				if (ViewLine[i]<miny) ViewLine[i]=(float)miny;
			}
		}
		rangepts=1;
	} else { // plot both minimum and maximum points
		for (i=0; i<ViewLineLength; i++) {
			beginpt = long(i*rangepts)+minx;
			ASSERT(beginpt>=0 && beginpt<=maxpts);
			maxdata = data[beginpt];
			for (j=0; j<rangepts; j++) { // find max point on range
				maxdata = (maxdata > data[beginpt+j]) ? maxdata : data[beginpt+j];
				}
			ViewLine[i] = (float)(maxdata);
			ViewLineXPts[i]=(float)Decon->PtToMass(beginpt+j/2);
			scale= (scale>maxdata) ? scale : maxdata;
			if (IsZoomingY) {
				if (ViewLine[i]>maxy) ViewLine[i]=(float)maxy;
				if (ViewLine[i]<miny) ViewLine[i]=(float)miny;
			}
		} // end for
	}
	ScaleViewLine();
}

void CBudaView::OnFFTDone()
{
	// when an FFT finishes, the message w_hNotifyFFTDone is posted
	// and BudaView's message map sends program flow here.  
	PlotType = PostFFTPlotType;
	UpdateScaleFactor = TRUE;
	pDoc->CalculatingFFT = FALSE;
	TRACE("CBudaView::OnFFTDone  TwoDFFTStatus = %d\n",TwoDFFTStatus);
	if(TwoDFFTStatus > 0) {
		TwoDFFTStatus--;
		pDoc->CalculatingFFT = TRUE;
	}
	if(TwoDFFT && TwoDFFTStatus <= 0) PlotType = TwoDFFTPlotType; 

	pDoc->FFTIsDirty = FALSE;
	InitializeZoomStack();
	OnPlot();
}

void CBudaView::CreateNBMIList()
{
	if(!pDoc->IsMIed()) return;
	double maxx, minx;
	maxx = zoomstack[zoompointer].maxx;
	minx = zoomstack[zoompointer].minx;

	if (NBMIList != 0) {
		delete NBMIList;
		NBMIList = 0;
	}
	NumMI = 0;
	if(pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks < 0) return;
	NBMIList = new struct MIList[pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks];

	int i,j;
	double M;
	for(i=0; i<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks; i++) {
		M = GetMITableMass(i);
		if(M<minx || M>maxx) continue;
		NBMIList[NumMI].M = M;
		NBMIList[NumMI].I = pDoc->FITable[i].I;
		NBMIList[NumMI].cs = pDoc->FITable[i].CS;
		NumMI++;
	}

	// now sort MIList in decreasing order of abundance
	struct MIList temp;
	for(i=0; i<NumMI; i++) {
		if(NBMIList[i].I > NBMIList[i-1].I) {
			j=i;
			while(j>0 && NBMIList[j].I > NBMIList[j-1].I) {
				temp = NBMIList[j];
				NBMIList[j]=NBMIList[j-1];
				NBMIList[j-1]=temp;
				j--;
			}
		}
	}		
}

void CBudaView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	if(pDoc->FFTIsDirty) {
		if(PlotType == MASSSPEC) OnPlotMassSpectrum();
		if(PlotType == FREQUENCY) OnPlotFrequency();
	} else
		CView::OnUpdate(pSender,lHint,pHint);		
}

void CBudaView::AddTextAnnotation(double x, double y, long Num, CString* text)
{
		PEvsetcell (hWndPE, PEP_szaGRAPHANNOTATIONTEXT, Num, (void *)LPCTSTR(*text));
		PEvsetcell (hWndPE, PEP_naGRAPHANNOTATIONTYPE, Num, PEGAT_NOSYMBOL);
		PEvsetcell (hWndPE, PEP_faGRAPHANNOTATIONX, Num, &x);
		PEvsetcell (hWndPE, PEP_faGRAPHANNOTATIONY, Num, &y);
}

void CBudaView::PlotCSLabels()
{
	DumpToDiagFile("entering CBudaView::PlotCSLabels()");
	// if we do this, we can't turn off the charge labeling.
	//if(pDoc->sys->AutoLabelCharge) pDoc->DisplayParms.LabelCharge = TRUE;
	if(!pDoc->DisplayParms.LabelCharge) return;

	// first, find the cs groups in NBMI and count them
	int i,j;
	int group = 0;
	int *groupMIindex = new int[NumMI];
	double *groupmass = new double[NumMI];
	for(i=0; i<NumMI; i++) {	// clear the arrays
		groupMIindex[i]=-1;
		groupmass[i]=0;
	}

	for(i=0; i<NumMI; i++) {
		if(groupMIindex[i]<0) {	// undefined group
			// search through and find all peaks +/- 2 da from this peak
			// and assign them to this group 
			groupMIindex[i] = group;
			for(j=i; j<NumMI; j++) {
				if(fabs(NBMIList[j].M - NBMIList[i].M) <2 && NBMIList[j].cs == NBMIList[i].cs) {
					groupMIindex[j] = group;
				}
			}	// for j
			group++;
		}	// if 
	}	// for i

	double x,y, summi, sumi;
	CString cs;
	int chargestate=0;

	for(i=0; i<group; i++) {
		// second, calculate the center-of-mass of the group.
		summi = 0;
		sumi = 0;
		y=0;
		for(j = 0; j<NumMI; j++) {
			if(groupMIindex[j] == i) {
				chargestate = NBMIList[j].cs;
				sumi += NBMIList[j].I;
				summi += NBMIList[j].I*NBMIList[j].M;
				if(y<NBMIList[j].I) y=NBMIList[j].I;
			}
		}

		x = summi/sumi;
		ASSERT(scale>0);
		y = y/scale;
		y = y+5;

		// don't label charge 0, check before labeling charge 1
		if(chargestate == 0) continue;
		if(chargestate == 1 && !pDoc->sys->LabelCharge1) continue;

		// third put the label there. 
		cs.Format("%d",chargestate);
		if(NBMIList[i].cs>0) cs = "+" + cs;
		AddAnnotation(x,y,-1,cs);
	}

	delete [] groupMIindex;
	delete [] groupmass;
	DumpToDiagFile("leaving CBudaView::PlotCSLabels()");
}

BOOL CBudaView::AddAnnotation(double mass, double intensity, int MIindex, CString text)
{
	double textwidth,textheight;
	int no = PEGAT_NOSYMBOL;
	if(intensity==0) return FALSE;
	if(text == "") return FALSE;

	// first, find the lower left and upper right limits of the string.
	//double x,y;
	//double width;
	double maxx,minx;
	maxx = zoomstack[zoompointer].maxx;
	minx = zoomstack[zoompointer].minx;
	if(mass > maxx) return FALSE;
	if(mass < minx) return FALSE;

	// find the string length;
	textwidth = text.GetLength()*(maxx-minx)/80;	// 120 characters fit in the screen horizontally
	textheight = 5;									// 20 lines per page, scaled to 100%

	// then check to see if they are within bounds if not, return FALSE
	if((mass+textwidth) > maxx) return FALSE;
	if((mass-textwidth) < minx) return FALSE;
	if((intensity+textheight) > 100/scale) return FALSE;

	// check for data points within textbox
	if(PtsInRange(mass, intensity, textwidth)) return FALSE;
	// check for peaks (including peaks' text) within textbox
	if(PeaksInRange(MIindex, mass, intensity, textwidth)) return FALSE;
	// check for other annotations within textbox
	if(AnnotationsInRange(mass,intensity,textwidth)) return FALSE;

	// everything is ok, so add the annotation
	text = "|c"+text;
		
	PEvsetcell (hWndPE, PEP_szaGRAPHANNOTATIONTEXT, NumAnnotations, (void *)LPCTSTR(text));
	PEvsetcell (hWndPE, PEP_naGRAPHANNOTATIONTYPE, NumAnnotations, &no);
	PEvsetcell (hWndPE, PEP_faGRAPHANNOTATIONX, NumAnnotations, &mass);
	PEvsetcell (hWndPE, PEP_faGRAPHANNOTATIONY, NumAnnotations, &intensity);
	NumAnnotations++;
	
	// if everything ok, return TRUE.
	return TRUE;
}

void CBudaView::GetScaleFactor()
{
	if(!UpdateScaleFactor) return;
	long i,temp;
	VScale = 1;
	if(!pDoc->IsDefined) return;

	GlobalMax = 0;
	long lowpt = pDoc->MassToPt(pDoc->PS.DetectHighMass);
	long highpt = pDoc->MassToPt(pDoc->PS.DetectLowMass);
	if(PlotType == DECONVOLVED_MASS_DATA) {
		if(!Decon) return;
		lowpt = Decon->MassToPt((float)Decon->DeconLowMassLimit);
		highpt = Decon->MassToPt((float)Decon->DeconHighMassLimit);
	}
	if(PlotType == TRANSIENT) {
		lowpt = 0;
		highpt = pDoc->PS.TransientPoints-1;
	}

	if(lowpt>highpt) {
		temp = lowpt;
		lowpt = highpt;
		highpt = temp;
	}

	if(PlotType == MASSSPEC || PlotType == FREQUENCY) {
		GlobalMax = GlobalMin = pDoc->FreqArray[lowpt];
		for(i=lowpt; i<highpt; i++){
			if(pDoc->FreqArray[i]>GlobalMax) GlobalMax = pDoc->FreqArray[i];
			if(pDoc->FreqArray[i]<GlobalMin) GlobalMin = pDoc->FreqArray[i];
		}
	}
	if(PlotType == TRANSIENT) {
		GlobalMax = GlobalMin = double(pDoc->TimeArray[lowpt]);
		for(i=0; i<pDoc->PS.TransientPoints; i++){
			if(double(pDoc->TimeArray[i])>GlobalMax) GlobalMax = double(pDoc->TimeArray[i]);
			if(double(pDoc->TimeArray[i])<GlobalMin) GlobalMin = double(pDoc->TimeArray[i]);
		}
	}
	if(PlotType == DECONVOLVED_MASS_DATA) {
		GlobalMax = GlobalMin = (double) Decon->DeconData[lowpt];
		for(i=0; i<Decon->DeconSize; i++){
			if((double) Decon->DeconData[i]>GlobalMax) GlobalMax = (double) Decon->DeconData[i];
			if((double) Decon->DeconData[i]<GlobalMin) GlobalMin = (double) Decon->DeconData[i];
		}
	}

	//if(GlobalMax > 0) VScale = 10000/GlobalMax;
	UpdateScaleFactor = FALSE;
}

void CBudaView::OnAnalyzeSetscalefactor() 
{
	CSetScaleFactorDlg dlg;
	dlg.VScale = (float)(500.0/(GlobalMax-GlobalMin));
	dlg.ScaleFactor = (float)(VScale*5.0/scale);
	double oldscale = scale;

	if(dlg.DoModal() == IDOK) {
		scale = 5/dlg.ScaleFactor;
		VScale = double(oldscale/scale);

		ViewIsDirty = TRUE;
		ViewLineIsDirty = TRUE;
		PEreinitialize(hWndPE);
		PEresetimage (hWndPE, 0, 0); 
		CalculateViewLine();
	}
}

void CBudaView::OnAnalyzeSetpeakpickthreshold() 
{
	CSetNewThreshold dlg;
	dlg.m_Thresh = pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NoiseLevel*100/(GlobalMax-GlobalMin);
	if(dlg.DoModal() == IDOK) {
			pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NoiseLevel = dlg.m_Thresh*(GlobalMax-GlobalMin)/100;
			pDoc->GetMITable();
			pDoc->UpdateAllViews(NULL,NULL,NULL);
	}
}

void CBudaView::OnUpdateAnalyzeSetpeakpickthreshold(CCmdUI* pCmdUI) 
{
	if(pDoc->IsFFTed()) pCmdUI->Enable(TRUE);
	else pCmdUI->Enable(FALSE);
}

BOOL CBudaView::AnnotationsInRange(double mass, double intensity, double textwidth)
{
	// return true if there are any other annotations within the range 
	// in the parameters otherwise, return false
	int i;
	double bottom = intensity;
	double top = intensity+5;
	double left = mass - textwidth/2;
	double right = mass + textwidth/2;
	char buffer[100];
	for(i=0; i<100; i++) buffer[i]='\0';
	CString text;
	double maxx = zoomstack[zoompointer].maxx;
	double minx = zoomstack[zoompointer].minx;
	double width;

	double x,y;
	for(i=0; i<NumAnnotations; i++) {
		PEvgetcell (hWndPE, PEP_faGRAPHANNOTATIONX, i, &x);
		PEvgetcell (hWndPE, PEP_faGRAPHANNOTATIONY, i, &y);
		PEvgetcell (hWndPE, PEP_szaGRAPHANNOTATIONTEXT, i, (void *)buffer);
		text = CString(buffer);
		width = text.GetLength()*(maxx-minx)/100;
		if((x + width/2) > left && (x-width/2) < right && y > bottom && y < top) return TRUE;
	}
	return FALSE;
}

void CBudaView::OnUpdateCSDet(CCmdUI* pCmdUI) 
{
	pCmdUI=pCmdUI;
//	if (FlagCSDetState) pCmdUI->SetCheck(1);
//	else pCmdUI->SetCheck(0);
}

void CBudaView::OnCSDet() 
{
	// if > 2 Da in view window return
	double maxx,minx;
	maxx = zoomstack[zoompointer].maxx;
	minx = zoomstack[zoompointer].minx;

	if((maxx - minx) > (double)10.0) {
		MessageBox("Please zoom in to < 10.0 Da to do charge state determination","Error!");
		return;
	}

	// otherwise, find maximum point in window, call CSDet with that point
	long i,pt0,pt1,maxi;
	double maxv;
	pt0 = pDoc->MassToPt(maxx);
	pt1 = pDoc->MassToPt(minx);
	maxv = pDoc->FreqArray[pt0];
	maxi = pt0;
	for(i=pt0; i<pt1; i++) {
		if(pDoc->FreqArray[i] > maxv) {
			maxi=i;
			maxv = pDoc->FreqArray[i];
		}
	}

	CCSMapDlg dlg;
	dlg.pDoc = pDoc;
	dlg.m_MinMass = (float)minx;
	dlg.m_MaxMass = (float)maxx;
	dlg.m_MaxPeak = maxi;

	dlg.DoModal();
/*	CChargeStateDetermination CSDet;
	CSDet.pDoc = pDoc;
	CSDet.pSys = pDoc->sys;
	z = CSDet.GetChargeState(maxi);
	CString str;
	str.Format("Charge state = %d",z);
	MessageBox(str,"Charge state value");
*/
}

BOOL CBudaView::OnScroll(UINT nScrollCode, UINT nPos, BOOL bDoScroll) 
{
	// TODO: Add your specialized code here and/or call the base class
	
	return CView::OnScroll(nScrollCode, nPos, bDoScroll);
}

void CBudaView::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	double ScrollMax, ScrollMin;
	double minx,maxx;

	// this calculates the scrollbar size from the ratio
	// of minx and maxx to the total spectrum size
	minx = zoomstack[zoompointer].minx;
	maxx = zoomstack[zoompointer].maxx;
	ScrollMax = 100;
	ScrollMin = 0;

	switch (PlotType) {
		case TRANSIENT:
			ScrollMax = (double)pDoc->PS.TransientPoints*(double)pDoc->PS.AdcRate;
			ScrollMin = 0;
			break;
		case FREQUENCY : // PlotType = Frequency
			ScrollMax = (double)pDoc->PS.AdcRate/(double)2.0;
			ScrollMin = (double)pDoc->MassToFreq(pDoc->PS.DetectHighMass);
			break;
		case MASSSPEC : // PlotType = Mass
			ScrollMax = (double)pDoc->PS.DetectHighMass;
			ScrollMin = (double)pDoc->PS.DetectLowMass;
			break;
		case DECONVOLVED_MASS_DATA:
			if(!Decon) return;
			ScrollMax = (double)Decon->DeconHighMassLimit;
			ScrollMin = (double)Decon->DeconLowMassLimit;
			break;
		default:
			break;
	}// end switch

	// Here we handle the scrolling functions.
	// first adjust the current zoomstack elements
	switch(nSBCode) {
	case SB_LEFT:	// scroll to far left
		zoomstack[zoompointer].minx = ScrollMax-(maxx-minx);
		zoomstack[zoompointer].maxx = ScrollMax;
		break;
	case SB_RIGHT:	// scroll to far right
		zoomstack[zoompointer].maxx = ScrollMin+(maxx-minx);
		zoomstack[zoompointer].minx = ScrollMin;
		break;
	case SB_LINELEFT:	// scroll left
		zoomstack[zoompointer].minx = minx-(double)0.10*(maxx-minx);
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx+(maxx-minx);
		break;
	case SB_LINERIGHT:	// scroll right
		zoomstack[zoompointer].minx = minx+(double)0.10*(maxx-minx);
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx+(maxx-minx);
		break;
	case SB_PAGELEFT:	// scroll one page left
		zoomstack[zoompointer].minx = minx-(double)0.90*(maxx-minx);
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx+(maxx-minx);
		break;
	case SB_PAGERIGHT:	// scroll one page right
		zoomstack[zoompointer].minx = minx+(double)0.90*(maxx-minx);
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx+(maxx-minx);
		break;
	case SB_ENDSCROLL:	// end scroll??
		// recalculate viewline and redisplay it.
		PEreinitialize(hWndPE);
		PEresetimage (hWndPE, 0, 0); 
		CalculateViewLine();
		break;
	case SB_THUMBPOSITION:	// scroll to abolute position nPos ??
		// calculate minx from npos and reset.
		zoomstack[zoompointer].minx = (ScrollMax-ScrollMin)*((double)nPos/(double)(SCROLL_MAX-SCROLL_MIN)) + ScrollMin;
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx + (maxx-minx);
		PEreinitialize(hWndPE);
		PEresetimage (hWndPE, 0, 0); 
		CalculateViewLine();
		break;
	case SB_THUMBTRACK:	// drag stroll box to nPos ??
		// calculate minx from npos and reset.
		zoomstack[zoompointer].minx = (ScrollMax-ScrollMin)*((double)nPos/(double)(SCROLL_MAX-SCROLL_MIN)) + ScrollMin;
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx + (maxx-minx);
		PEreinitialize(hWndPE);
		PEresetimage (hWndPE, 0, 0); 
		CalculateViewLine();
		break;
	} // end switch
	// then call SetScrollbarPos();

	// so that we can't scroll past the ends...
	if(zoomstack[zoompointer].minx<ScrollMin) {
		zoomstack[zoompointer].minx = ScrollMin;
		zoomstack[zoompointer].maxx = zoomstack[zoompointer].minx + (maxx-minx);
	}
	if(zoomstack[zoompointer].maxx>ScrollMax) {
		zoomstack[zoompointer].maxx = ScrollMax;
		zoomstack[zoompointer].minx = zoomstack[zoompointer].maxx - (maxx-minx);
	}

	SetScrollbarPos();

	CView::OnHScroll(nSBCode, nPos, pScrollBar);
}

void CBudaView::SetupScrollbar()
{
	GetScrollInfo(SB_HORZ,&Scroll_Info,SIF_ALL);
	Scroll_Info.nMax = SCROLL_MAX;
	Scroll_Info.nMin = SCROLL_MIN;
	Scroll_Info.nPage = SCROLL_MAX;
	Scroll_Info.nPos = SCROLL_MIN;
	Scroll_Info.nTrackPos = SCROLL_MIN;
	SetScrollInfo(SB_HORZ,&Scroll_Info,TRUE);
}

void CBudaView::SetScrollbarSize()
{
	double ScrollMax, ScrollMin;
	double minx,maxx;
	int scrollsize;

	// this calculates the scrollbar size from the ratio
	// of minx and maxx to the total spectrum size
	minx = zoomstack[zoompointer].minx;
	maxx = zoomstack[zoompointer].maxx;
	ScrollMax = 100;
	ScrollMin = 0;

	switch (PlotType) {
		case TRANSIENT:
			ScrollMax = (double)pDoc->PS.TransientPoints/(double)pDoc->PS.AdcRate;
			ScrollMin = 0;
			break;
		case FREQUENCY : // PlotType = Frequency
			ScrollMax = (double)pDoc->PS.AdcRate/(double)2.0;
			ScrollMin = (double)pDoc->MassToFreq(pDoc->PS.DetectHighMass);
			break;
		case MASSSPEC : // PlotType = Mass
			ScrollMax = (double)pDoc->PS.DetectHighMass;
			ScrollMin = (double)pDoc->PS.DetectLowMass;
			break;
		case DECONVOLVED_MASS_DATA:
			if(!Decon) return;
			ScrollMax = (double)Decon->DeconHighMassLimit;
			ScrollMin = (double)Decon->DeconLowMassLimit;
			break;
		default:
			break;
	}// end switch

	// now size of scroll bar = 327687(maxx-minx)/(scrollmax-scrollmin);
	scrollsize = (int)(((maxx-minx)/(ScrollMax-ScrollMin))*(SCROLL_MAX - SCROLL_MIN));
	Scroll_Info.nPage = scrollsize;
	SetScrollInfo(SB_HORZ,&Scroll_Info,TRUE);
}

void CBudaView::SetScrollbarPos()
{
	double ScrollMax, ScrollMin;
	double minx,maxx;
	int scrollpos;

	// this calculates the scrollbar size from the ratio
	// of minx and maxx to the total spectrum size
	minx = zoomstack[zoompointer].minx;
	maxx = zoomstack[zoompointer].maxx;
	ScrollMax = 100;
	ScrollMin = 0;

	switch (PlotType) {
		case TRANSIENT:
			ScrollMax = (double)pDoc->PS.TransientPoints/(double)pDoc->PS.AdcRate;
			ScrollMin = 0;
			break;
		case FREQUENCY : // PlotType = Frequency
			ScrollMax = (double)pDoc->PS.AdcRate/(double)2.0;
			ScrollMin = (double)pDoc->MassToFreq(pDoc->PS.DetectHighMass);
			break;
		case MASSSPEC : // PlotType = Mass
			ScrollMax = (double)pDoc->PS.DetectHighMass;
			ScrollMin = (double)pDoc->PS.DetectLowMass;
			break;
		case DECONVOLVED_MASS_DATA:
			if(!Decon) return;
			ScrollMax = (double)Decon->DeconHighMassLimit;
			ScrollMin = (double)Decon->DeconLowMassLimit;
			break;
		default:
			break;
	}// end switch

	// now pos of scroll bar = 32767*minx/(scrollmax-scrollmin);
	scrollpos = (int)(((minx-ScrollMin)/(ScrollMax-ScrollMin))*(SCROLL_MAX - SCROLL_MIN));
	Scroll_Info.nPos = scrollpos;
	SetScrollInfo(SB_HORZ,&Scroll_Info,TRUE);
}

void CBudaView::OnViewCopyCursorXPos() 
{
	HANDLE hText = NULL;
	LPSTR lpszText;

//  this function copies the current X-Axis position of
//	the cursor to the windows clipboard as text.
	if ( !OpenClipboard() )
	{
		MessageBox( "Cannot open the Clipboard" );
		return;
	}
	// Remove the current Clipboard contents
	if( !EmptyClipboard() )
	{
		MessageBox( "Cannot empty the Clipboard" );
		return;
	}
	// ...
	// Get the current cursor point
	POINT pt;
	CString str;

	PEvget(hWndPE,PEP_ptLASTMOUSEMOVE,&pt);

	int i, ptx, pty;// axis number
	double x,y;		// x and y in graph units
	ptx = (int)pt.x;
	pty = (int)pt.y;
	PEconvpixeltograph(hWndPE,&i,&(ptx),&(pty),&x,&y, FALSE, FALSE, FALSE);
	str.Format("%3.4f\r\n",x);

	hText = GlobalAlloc(GMEM_DDESHARE,20);
	if (!hText) {
	    MessageBox( "Unable to Allocate clipboard memory" );
		return;
	}
	lpszText=(char *)GlobalLock(hText);
	if (!lpszText) {
	    MessageBox( "Unable to Lock Clipboard data" );
		return;
	}
    strcpy(lpszText, str);
    GlobalUnlock (hText);
    if (SetClipboardData(CF_TEXT, hText) == NULL) 
	{
		i=GetLastError();
	    MessageBox( "Unable to set Clipboard data" );
	}
    CloseClipboard();
	return;

}

void CBudaView::OnDebugDialogMasstofreq()
{
	pDoc->OnDebugDialogMasstofreq();
}

void CBudaView::OnFileExport() 
{
	CExportDlg dlg;
	dlg.pDoc = pDoc;
	dlg.pView = this;
	dlg.DoModal();
}

