// ChargeStateDetermination.cpp : implementation file
//
/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "stdafx.h"
#include "Buda.h"
#include "ChargeStateDetermination.h"
#include "BudaMaster.h"
#include "BudaDoc.h"

#include <stdlib.h>
#include <string.h>
#include <math.h>

#define CHARGE_STATE_MAP_SIZE 1024

// intel's digital signalling processing library
//#define nsp_UsesAll
#define nsp_UsesTransform
#define nsp_UsesVector
#define nsp_UsesWin
#define nsp_UsesConversion
#include "nsp.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CChargeStateDetermination

CChargeStateDetermination::CChargeStateDetermination()
{
	maxgroup = 0;
	ChargeLimit = 30;
}

CChargeStateDetermination::~CChargeStateDetermination()
{
}


BEGIN_MESSAGE_MAP(CChargeStateDetermination, CWnd)
	//{{AFX_MSG_MAP(CChargeStateDetermination)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CChargeStateDetermination message handlers

void CChargeStateDetermination::InitializeCSDet()
{
	ClearGroupStruct();
}

void CChargeStateDetermination::RoughGroupAssignment()
{
	int i,j,numAssigned, group, rmaxi;
	double* mz, avg;
	float * rel,rmax;
	int* grp;
	double delta, mindelta, leftdelta, rightdelta,ratio,temp;
	long NumPeaks = pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;

	// first create and fill mz, rel, grp arrays;
	mz = new double[NumPeaks];
	rel = new float[NumPeaks];
	grp = new int[NumPeaks];
	scale = 0;
	for(i=0; i<NumPeaks; i++) {
		mz[i]= pDoc->FreqToMass(pDoc->FITable[i].F);
		rel[i]=float(pDoc->FreqToMass(pDoc->FITable[i].I));
		if(rel[i]>scale) scale = rel[i];
		grp[i]=0;
	}
	group = 0;

	// main loop - assign each peak to a group
	for(numAssigned=0; numAssigned<NumPeaks;) {
		rmax = 0;
		rmaxi = 0;
		// find first unassigned peak.
		for(i=0; i<NumPeaks; i++) {
			if(grp[i] == 0) {
				rmax = rel[i];
				rmaxi = i;
				break;
			}
		}
		// now find largest unassigned peak
		for(;i<NumPeaks; i++) {
			if(grp[i]==0 && rel[i]>rmax) {
				rmax = rel[i];
				rmaxi = i;
			}				
		}

		// start a new group with this peak
		group++;
		grp[rmaxi]=group;
		numAssigned++;
		AddToCSGroup(group, rmaxi);

		// determine if this peak is near other isotope peaks 
		// and estimate spacing right and left;
		delta = leftdelta = rightdelta = 0.0;
		ChargeLimit = pDoc->sys->ChargeLimit;
		mindelta = 1.0/double(ChargeLimit);

		// check lower m/z
		// (get first unassigned peak lower in m/z that has amplitude
		// at least 40% of rmax and is within 0.75 m/z from rmaxi
		// to exclude cs's lower than 2)
		for(j=rmaxi-1; j>0; j--) {
			if(grp[j]==0 && rel[j] >= 0.4*rel[rmaxi]) {
				delta = fabs(mz[rmaxi]-mz[j]);
				if(delta<mindelta) continue;
				if(delta <= 1.1) leftdelta = delta;
				break;
			}
		}

		// check higher m/z
		for(j=rmaxi+1; j<NumPeaks; j++) {
			if(grp[j]==0 && rel[j] >= 0.4*rel[rmaxi]) {
				delta = fabs(mz[j]-mz[rmaxi]);
				if(delta<mindelta) continue;
				if(delta <= 1.1) rightdelta = delta;
				break;
			}
		}

		// lone peak
		if(leftdelta == 0 && rightdelta == 0) continue;
		
		if(leftdelta == 0) avg = rightdelta;
		else if(rightdelta == 0) avg = leftdelta;
		else { 
			// peaks on both sides, so if match within 10%, just average them
			if(leftdelta<rightdelta) {
				// make sure that right delta is always the lower of the two
				temp = rightdelta;
				rightdelta = leftdelta;
				leftdelta = temp;
			}
			ratio=leftdelta/rightdelta;
			if(ratio < 1.5) avg = (leftdelta+rightdelta)/2;
			// for cases where right = 2xleft
			else if(ratio < 2.5) avg = (leftdelta/2 + rightdelta)/2;
			else avg = rightdelta;
		}

		// now search at higher m/z for all peaks that belong to this group
		// to belong their delta must be within 12% of the previous avg
		j = rmaxi;
		for(i=rmaxi+1; i<NumPeaks; i++) {
			if( grp[i] == 0) {
				delta = fabs(mz[i]-mz[j]);
				if(delta > 1.12*avg) break;
				if(delta >= 0.88*avg) {
					grp[i]=group;
					numAssigned++;
					AddToCSGroup(group,i);
					j=i;
				}
			}
		}

		// ditto for lower m/z
		j = rmaxi;
		for(i=rmaxi+1; i>0; i--) {
			if( grp[i] == 0) {
				delta = fabs(mz[j]-mz[i]);
				if(delta > 1.12*avg) break;
				if(delta >= 0.88*avg) {
					grp[i]=group;
					numAssigned++;
					AddToCSGroup(group,i);
					j=i;
				}
			}
		}
	}	// end main (numAssigned) loop

	// now, convert grp[i], avg[i], mz[i] avg
	/* cs vestiges
	// first peak goes in new group
	AddToCSGroup(0,0);
		currmass = pDoc->FreqToMass(pDoc->FITable[i].F);
		if (!fit) {  // if not start a new group
			AddToCSGroup(maxgroup,i);
		}
		for(j=0;j<maxgroup;j++) {
			if (fabs(currmass-groups[j].mass)<2) { // is it within 2 Da of 1st peak in grp j?
				AddToCSGroup(j,i);
				fit = TRUE;
				break; // break out of j loop
			} 
		}
	*/
	// clean up by deleting internal arrays
	delete [] rel;
	delete [] mz;
	delete [] grp;
}	// end RoughGroupAssignment()

void CChargeStateDetermination::ExtractOverlappingCSDists()
{

}

void CChargeStateDetermination::ComputeDeltaMassCS()
{
	int i,j,k;
	//int deltapeaks[MAX_PEAKS_PER_GROUP][MAX_PEAKS_PER_GROUP];
	double deltamass[MAX_PEAKS_PER_GROUP][MAX_PEAKS_PER_GROUP];
	int numpeaks;
	double mass1=0,mass2=0;
	double mass=0,weights=0,cs;
	int ics;

	RoughGroupAssignment();

	for(i=0; i<maxgroup; i++) {
		if(groups[i].numpeaks < 1) continue;
		if(groups[i].numpeaks == 1) {
			// only one peak, so give it cs = 1
			groups[i].average = 1;
			groups[i].chargestate = 1;
			continue;
		}

		// for each group, find all mass differences and 
		// number of peaks between masses
		numpeaks = groups[i].numpeaks;
		if (numpeaks>MAX_PEAKS_PER_GROUP) numpeaks = MAX_PEAKS_PER_GROUP;
		for(j=0; j<numpeaks; j++) {
			deltamass[j][j]=0;
			//deltapeaks[j][j]=0;
			mass1 = Mass(groups[i].MITableIndex[j]);
			for(k=j+1; k<numpeaks; k++) {
				mass2 = Mass(groups[i].MITableIndex[k]);
				deltamass[j][k] = fabs(mass1-mass2);
				deltamass[k][j] = deltamass[j][k];
			} // end for k
		} // end for j

		// now, do a weighted average of the deltamasses/deltapeaks.
		// where the weight is = deltapeaks.
		//
		//	i.e. dm = (sum((k-j)*deltamass[j][k]))/(sum(j-k))
		//
		mass = 0;
		weights = 0;
		for(j=0; j<numpeaks; j++) {
			for(k=0; k<numpeaks; k++) {
				//mass += abs(j-k)*deltamass[j][k]/abs(j-k);
				mass += deltamass[j][k];
				weights += abs(j-k);
			}
		}
		cs = weights/mass;
		mass = mass/weights;

		// now round cs and make sure it != 0
		ics=int(cs);
		if(cs-double(ics) >= 0.5) ics++;
		if(ics == 0) ics = 1;
		if(ics>ChargeLimit) ics = 1;

		// note: for z>2, we must find z/2 or 5 peaks per group (whichever is less)
		// or we cannot define the cs so set cs = 1;
		if(groups[i].numpeaks < ics/2) {
			if(groups[i].numpeaks < 5) {
				ics = 1;
			}
		}
		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) ics = ics * -1;

		groups[i].average = mass;
		groups[i].chargestate = ics;

		// finally dump cs info back into the mi table
		for(j=0; j<groups[i].numpeaks; j++) {
			pDoc->FITable[groups[i].MITableIndex[j]].CS = ics;
		}
	} // end for i

	// note: for speed, may be able to use
	// maxgroup = 0; 
	// instead of full cleargroupstruct;
	ClearGroupStruct();
}

void CChargeStateDetermination::AddToCSGroup(int groupnum, int MINum)
{
	// if grp[groupnum] exists, add peak to it.  Otherwise,
	// create it.
	if(groupnum > maxgroup) {
		//MessageBox("Error 34 : invalid charge state peak grouping number","Charge state determination error");
		return;
	}
	if(groupnum == maxgroup) {
		// new group - create it
		groups[groupnum].mass = pDoc->FreqToMass(pDoc->FITable[MINum].F);
		groups[groupnum].average = 1;
		groups[groupnum].chargestate = 1;
		groups[groupnum].numpeaks = 1;
		groups[groupnum].stdev = 100; // effectively infinite
		groups[groupnum].MITableIndex[0]=MINum;
		maxgroup++;
	} else {
		// old group - add to it
		groups[groupnum].MITableIndex[groups[groupnum].numpeaks]=MINum;
		groups[groupnum].numpeaks++; 
	}
}

void CChargeStateDetermination::CSDetermination()
{
	ChargeLimit = pDoc->sys->ChargeLimit;

	pDoc->SortIncreasing = TRUE;
	pDoc->SortFITableOnMass();
	maxgroup = 1;

	switch(pDoc->CSDetMethod) {
	case CSDET_DELTAMASS:
		ComputeDeltaMassCS();
		break;
	case CSDET_MULTIPLICATIVE:
		ComputeMultiplicativeCS();
		break;
	case CSDET_FOURIER:
		ComputeFourierCS();
		break;
	case CSDET_COMBO:
		ComputeComboCS();
		break;
	case CSDET_AUTOCORR:
		ComputeAutoCorrCS();
		break;
	case CSDET_OCONNOR:
		ComputeOConnorCS();
		break;
	case CSDET_ENTROPY:
		ComputeComboCS();
		break;
	case CSDET_SPARE1:
		ComputeComboCS();
		break;
	}	

	// making sure that all peaks have cs the same polarity as trap volts
	int i;
	for(i=0; i<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks; i++) {
		if(pDoc->TrapVoltsOnDetect < 0) 
			if(pDoc->FITable[i].CS > 0) pDoc->FITable[i].CS *= -1;
		if(pDoc->TrapVoltsOnDetect > 0) 
			if(pDoc->FITable[i].CS < 0) pDoc->FITable[i].CS *= -1;
	}

}	// end charge state determination

double CChargeStateDetermination::Mass(int n)
{
	return pDoc->FreqToMass(pDoc->FITable[n].F);
}

void CChargeStateDetermination::ClearGroupStruct()
{
	int i;
	struct groupstruct NullGrp;
	NullGrp.average=0;
	NullGrp.chargestate=0;
	NullGrp.mass=0;
	NullGrp.numpeaks=0;
	NullGrp.stdev=0;
	for(i=0; i<MAX_PEAKS_PER_GROUP; i++) NullGrp.MITableIndex[i]=0;
	for(i=0; i<MAX_NUMBER_OF_GROUPS; i++) groups[i]=NullGrp;
	maxgroup=0;
}

int CChargeStateDetermination::CountPeaks(double m1, double m2)
{
	// counts up all peaks in the MITable between masses m1 and m2
	// does not include peaks m1 and m2
	int i,count = 0;
	long numpeaks = pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;

	double mass=0;
	if(m1>m2) {
		mass = m1;
		m1=m2;
		m2=mass;
	}
	mass = 0;

	for(i=0; i<numpeaks; i++) {
		mass = Mass(i);
		if(mass>m1 && mass<m2) count++;
	}
	return count;
}

//#define _DEBUG_CS2

void CChargeStateDetermination::ComputeXCorrCS()
{
#ifdef _DEBUG_CS2
	CString str;
	int k;
	CFile fi;
#endif
	float deltamass = 0,max=0, centermass=0, abund=0;
	int i,j,MaxXDistIndex;
	float MaxXDist;
	int cs,cs2;
	// this function uses crosscorrelation functions in the Intel spl
	// library for calculating charge states.

	float lowmass, himass, xcorr, peakwidth,mass;
	long lowpoint,hipoint,numpts,centerpt;
	float *model=0, *data=0, *xdist=0;
	if(maxgroup == 0) return;
	for(i=0; i<maxgroup; i++) {
		if(groups[i].numpeaks<=1) continue;
		max=0;
		// first get an average charge state
		deltamass = (float)fabs(Mass(groups[i].MITableIndex[0]) - Mass(groups[i].MITableIndex[1]));

		lowmass = 0; himass = 100000; peakwidth = 0;
		for(j=0; j<groups[i].numpeaks-1; j++) {
			mass =  (float)Mass(groups[i].MITableIndex[j]);
			if(j==0) lowmass = himass = mass;
			abund = (float)pDoc->FITable[groups[i].MITableIndex[j]].I;
			if(max<abund){
				max = abund;
				centermass = mass;
			}
			if(mass<lowmass) lowmass = mass;
			if(mass>himass) himass = mass;
			peakwidth += (float)pDoc->FITable[groups[i].MITableIndex[j]].RP;
		}

		// convert average resolving power to average peak width
		peakwidth = peakwidth/groups[i].numpeaks-1;
		peakwidth = centermass/peakwidth;

		cs = int(1/deltamass);
		if((1/deltamass - cs) >0.5) cs++;
		cs2 = int(cs/2)+1;		
		
		himass += 2*deltamass;	lowpoint = pDoc->MassToPt(himass);
		lowmass -= 2*deltamass; hipoint = pDoc->MassToPt(lowmass);
		numpts = abs(hipoint-lowpoint);
		centerpt = pDoc->MassToPt(centermass) - lowpoint;
		if(data!=0) delete data;
		if(model!=0) delete model;
		if(xdist!=0) delete xdist;
		data = new float[numpts];
		model = new float[numpts];
		xdist = new float[cs];

		// now, do crosscorrelation over a number of charge states near 1/deltamass
		for(j=cs2; j<cs+cs2 ; j++) { // for these cs's
			// generate a filtered raw data distribution
			GenerateFilteredRawData(lowpoint, data, numpts);
			
			// generate a model chargestate distribution
			//GetModelIsotopicDistribution(centermass, j, lowpoint, peakwidth, model, numpts);

			// xcorrelate and find the maximum xcorr value
			xcorr=0;
			nspsCrossCorr(data,numpts,model,numpts,&xcorr,0,0);
			xdist[j-cs2] = xcorr;

			// for debugging
			#ifdef _DEBUG_CS2
				str.Format("pkmtch%d-%d.txt",i,j);
				fi.Open(str,CFile::modeCreate | CFile::modeWrite);
				str += "\r\n";
				fi.Write(str,str.GetLength());
				str.Format("mass = %1.5f\r\n",mass);
				fi.Write(str,str.GetLength());
				str.Format("cs = %d       xcorr = %1.3f\r\n",j,xcorr);
				fi.Write(str,str.GetLength());
				str.Format("mass   , data[...] , model[...]\r\n");
				fi.Write(str,str.GetLength());
				str.Format("-------------------------------\r\n");
				fi.Write(str,str.GetLength());
				for(k=0; k<numpts; k++) {
					str.Format("%1.4f , %1.4f , %1.4f\r\n",pDoc->PtToMass(k+lowpoint),data[k],model[k]);
					fi.Write(str,str.GetLength());
				}
				fi.Close();
			#endif
		}
		// find maximum point in xdist and that is the chargestate.

		MaxXDist = 0;
		MaxXDistIndex = 0;
		for(j=0; j<cs; j++) {
			if(xdist[j]>MaxXDist) {
				MaxXDistIndex = j;
				MaxXDist = xdist[j];
			}
		}
		// finally dump cs info back into the mi table
		for(j=0; j<groups[i].numpeaks; j++) {
			pDoc->FITable[groups[i].MITableIndex[j]].CS = MaxXDistIndex+cs2;
		}

		delete data; data = 0;
		delete model; model = 0;
		delete xdist; xdist = 0;
	}

	/*
	questions:

	what mass range covered?  lowpoint-2*deltamass .. highpoint + 2*deltamass
	how many points?  
	what point spacing?
	same for model and data?

	how filter data?
		idea: use fitable.RP to calculate a peak width.  
		take raw data and zero out all regions not MITable+/- 1 peakwidth
	how generate model distribution? 
		idea:	use binomial distribution assuming m*cs/12 carbons
		and force base peak intensities to align.

	*/
}

void CChargeStateDetermination::GenerateFilteredRawData(long lowpoint, float *data, int numpts)
{
	// for now, no filtering at all.
	int i;
	for(i=0; i<numpts; i++) {
		data[i] = pDoc->FreqArray[i+lowpoint];
	}
}

void CChargeStateDetermination::GetModelIsotopicDistribution(double mass, float cs, float *prob, int numpts)
{
	/*************
	problem: how to generate model distribution? 

	solution: calculate a probability distribution using the standard
	binomial distribution equation:
			B(i) = combinations(n,i) * p^i * (1-p)^(n-i)
			where combinations(n,i) = n!/(i!*(n-i)!)

	Nothing unusual there.  The trick is getting the probability(p)
	and the number n.

	for these, we use p = 0.011 (c13 probability)
	and n = CF * (mass*cs)/12.000

	CF is the adjustable parameter of weight percent of carbon
	in the model molecule (we start with 0.53 for Ubiquitin)

    This model gives us a good estimation for the intensities
	of an isotopic distribution.
	
	------------
	other notes:
	remember to scale this distribution so that the most intense
	peak has intensity = most intense peak in data distribution

    remember to shift the mass of this distribution so that the most
	intense peak has the same mass as the base peak in the data

	remember that peak spacings will be spaced by (i*1.0034)/cs

	**************/

	int i,n,m;
	double p,CF;
	float maxp;

	// clear the model array
	if(numpts < 32) return;
	for(i=0; i<numpts; i++) {
		prob[i]=0;
	}

	// get the probability based intensity distribution
	CF = 0.53;
	p = 0.011;
	n = int(CF * (mass*cs)/12.000);
	m=n;
	if(n>32) m=32;
	for(i=0; i<32; i++) prob[i]=0;
	int maxpi=0;
	maxp=0;

	CF = 0;
	for(i=0; i<m; i++) {
		prob[i] = float(bico(n,i)*pow(p,i)*pow((1-p),(n-i)));
		CF += prob[i];
		if(prob[i] > maxp) {
			maxp = prob[i];
			maxpi = i;
		}
		if(prob[i]<(maxp*1e-2)) break;
	}
	// normalize prob
	for(i=0; i<m; i++) prob[i]=(float)((double)prob[i]/CF);
	return;
}

double CChargeStateDetermination::bico(int n, int k)
{
	if(n<0) return 0;
	if(k<0) return 0;
	if((n-k)<0) return 0;
	return floor(0.5+exp(factln(n)-factln(k)-factln(n-k)));

} // from numerical recipes

double CChargeStateDetermination::factln(int n)
{
	static double a[101];

	if (n <= 1) return 0.0;
	if (n <= 100) return a[n] ? a[n] : (a[n]=gammln(n+1.0));
	else return gammln(n+1.0);
} // from numerical recipes

double CChargeStateDetermination::gammln(double xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
} // from numerical recipes

void CChargeStateDetermination::ComputeMultiplicativeCS()
{
	int i;
	int z;
	double mass;
	float ZMap[1024];
	float Zvalues[1024];

	// for each peak in the mi table
	for(i=0; i<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;i++) {
		mass = pDoc->FreqToMass(pDoc->FITable[i].F);
		z = GetMultiplicativeMap(mass, ZMap, Zvalues);

		// now make sure it != 0
		if(z == 0) z = 1;
		if(z>ChargeLimit) z = 1;

		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) z = z * -1;

		// finally dump cs info back into the mi table
		pDoc->FITable[i].CS = z;
	} // end for i

}

void CChargeStateDetermination::ComputeFourierCS()
{
	int pk;
	float FourMap[CHARGE_STATE_MAP_SIZE];
	float Zvalue[CHARGE_STATE_MAP_SIZE];
	float ZMap[CHARGE_STATE_MAP_SIZE];
	double mass;
	int z;

	// for each peak in mi table
	for(pk=0; pk<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;pk++) {
		mass = pDoc->FreqToMass(pDoc->FITable[pk].F);

		GetFourierMap(mass,FourMap,Zvalue);
		z = GetFourierCSValue(FourMap,Zvalue,ZMap);

		// now make sure it != 0
		if(z == 0) z = 1;
		if(z>ChargeLimit) z = 1;

		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) z = z * -1;

		// finally dump cs info back into the mi table
		pDoc->FITable[pk].CS = z;
	}	// end for pk (mitable loop)

}

void CChargeStateDetermination::ComputeComboCS()
{
	double mass;
	long pk;
	float ZMap[1024], Zvalue[1024];
	int ics;

	// for each peak in the mi table
	for(pk=0; pk<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;pk++) {


		mass = pDoc->FreqToMass(pDoc->FITable[pk].F);
		ics = GetComboCSMap(mass, ZMap, Zvalue);


		// now make sure it != 0
		if(ics == 0) ics = 1;
		if(ics>ChargeLimit) ics = 1;

		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) ics = ics * -1;

		// finally dump cs info back into the mi table
		pDoc->FITable[pk].CS = ics;
		
		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) pDoc->FITable[pk].CS = pDoc->FITable[pk].CS * -1;

	} // end for i
}

double CChargeStateDetermination::GetIntensityatMass(double mass)
{
	long pt0,pt1,pt2;
	float int1,int2,intensity,slope;
	double tempmass;

	if(mass > pDoc->PS.DetectHighMass) return 0.0;
	if(mass < pDoc->PS.DetectLowMass) return 0.0;

	pt0 = pDoc->MassToPt(mass);
	// check if mass of pt1 is above or below mass
	tempmass = pDoc->PtToMass(pt0);
	if (mass > tempmass) {
		pt1 = pt0;
		pt2 = pt1+1;
	} else {
		pt2=pt0;
		pt1=pt2-1;
	}
	int1 = pDoc->FreqArray[pt1];
	int2 = pDoc->FreqArray[pt2];

	// interpolate
	tempmass = pDoc->PtToMass(pt1);
	slope = float(double(int2-int1)/(pDoc->PtToMass(pt2)-tempmass));
	intensity = int1+(slope*float(mass - tempmass));
	return (float)fabs((float)intensity);
}

void CChargeStateDetermination::GetFourierMap(double Mass, float *ZMap,float* Zvalue)
{
	// this function takes the Frequency domain data in the mass
	// range from lowmass to highmass, baseline corrects it, zerofills
	// to the next power of 2, FFT's it, and returns this data as well
	// as the associated zvalue array in the arrays CSMap and ZValues
	//
	// CSMap has a maximum of maxpts data elements.  
	// ZValues is constrained by the FT of yeilding maxpts/2+1 data elements,
	//		but it is initialized to the same size as CSMap for simplicity

	//double lowmass,highmass;
	long centerpt;
	double deltamass;
	int i,j;
	long pt0,pt2,numpts,maxpts;
	float avg,scale;
	int num,order;
	SCplx *FTData;

	maxpts = CHARGE_STATE_MAP_SIZE;
	centerpt = pDoc->MassToPt(Mass);
	deltamass = fabs(pDoc->PtToMass(centerpt) - pDoc->PtToMass(centerpt-1));
	pt0 = pDoc->MassToPt(Mass-1.5);
	pt2 = pDoc->MassToPt(Mass+1.5);

	if(abs(pt0-centerpt) > maxpts/2 || abs(pt2-centerpt) > maxpts/2) {
		pt0 = centerpt+maxpts/2-1;
		pt2 = centerpt-maxpts/2;
	}
		
	if(pt0 < pt2) {
		// swap using numpts as the temp variable.
		numpts = pt2;
		pt2=pt0;
		pt0=numpts;
	}

	numpts = pt0-pt2;
	avg = 0;
	for(j=0; j < numpts; j++) {
		ZMap[j] = pDoc->FreqArray[pt2+j];
		avg += ZMap[j];
	}

	// baseline correct it
	avg = avg / float(numpts);
	for(j=0; j < numpts; j++) {
		ZMap[j] = ZMap[j] - avg;
	}
		
	// pad with zeros to next order of 2
	num = 1;
	for(i=0; i<10; i++) {
		if(num > numpts) break;
		num =num*2;
	}
	order = i;
	for(i=numpts; i<num; i++) ZMap[i]=0;

	// power spectrum fft
	if(num > 1 && order > 1){
		FTData = new SCplx[maxpts/2+1];
		nspsRealFftNip(ZMap,FTData,order,NSP_Forw);
		scale=float(100.0)/float(num);
		nspcbPowerSpectr(FTData,ZMap,num/2);
		nspsbMpy1(scale, ZMap, num/2);	// 4/n scaling
		delete FTData;
	}

	// get the associated z values and zero those values of ZMap that 
	// could be below z = 1 to remove the dc offset;
	for(i=0; i<num/2+1; i++) {
		Zvalue[i]=float(i)/(float(num)*float(deltamass));
		if(Zvalue[i] < 0.95) ZMap[i] = 0;
		if(i>maxpts/2+1) break;
	}
}

int CChargeStateDetermination::GetFourierCSValue(float *FourMap, float *Zvalue, float *ZMap)
{
	int i,k;
	float z,zmax,zret;
	float slope;

	z = 1;
	zmax = 0;
	zret = 1;

	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		for(i = 0; i<CHARGE_STATE_MAP_SIZE/2; i++) {
			if(z < Zvalue[i]) break;
		}
		if(Zvalue[i-1] == Zvalue[i]) {
			ZMap[k] = FourMap[i-1];
		} else {
			slope = (FourMap[i-1]-FourMap[i])/(Zvalue[i-1]-Zvalue[i]);
			ZMap[k] = (float) fabs((double)(FourMap[i-1] + slope*(z-Zvalue[i-1])));
		}
	}

	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		Zvalue[k] = z;
		if(ZMap[k] > zmax) {
			zmax = ZMap[(int)(z*10)];
			zret = z;
		}
	}

	i=(int)floor(zret);

	if(zret - (float)i > 0.5) return i+1;
	else return i;	
}

int CChargeStateDetermination::GetMultiplicativeMap(double mass, float *MpyMap, float* Zvalue )
{
	float ics,zmax;
	int i,j,k,numisos,pt,pt1,pt2;
	float z;
	double mass0,mass2;
	double intensity0,intensity2, intensity;

	ics = 1;
	numisos = 1;

	pt = pDoc->MassToPt(mass);
	intensity = pDoc->FreqArray[pt];

	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		Zvalue[k] = z;
		if(z<0.85) {
			continue;
		}
		if(z<2){
			pt1 = pDoc->MassToPt(mass - 2);
			pt2 = pDoc->MassToPt(mass + 2);
		} else {
			pt1 = pDoc->MassToPt(mass - 1);
			pt2 = pDoc->MassToPt(mass + 1);
		}
		//if(mass*z > 2500) numisos = 2;
		//if(mass*z > 5000) numisos = 3;
		//if(mass*z > 10000) numisos = 4;
		//if(mass*z > 20000) numisos = 5;
		MpyMap[k] = (float)fabs(intensity);
		for(j=1; j<= numisos; j++) {
			mass2 = mass + 1.0034/double(z);
			intensity2 = fabs(GetIntensityatMass(mass2));
			if(intensity2 == 0.0) intensity2 = 1.0;

			mass0 = mass - 1.0034/double(z);
			intensity0 = fabs(GetIntensityatMass(mass0));
			if(intensity0 == 0.0) intensity0 = 1.0;
			
			// get Multiplicative value
			MpyMap[k] *= float(intensity0 * intensity2 / (double)(numisos*2+1));
		}
	}

	zmax = 0;
	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		if(MpyMap[k] > zmax) {
			zmax = MpyMap[(int)(z*10)];
			ics = z;
		}
	}

	i=(int)floor(ics);

	if(ics - (float)i > 0.5) return i+1;
	else return i;
}

int CChargeStateDetermination::GetChargeState(long index)
{
	// This is a function for outside usage.  It returns the 
	// ComboCS charge state for the peak at pDoc->FreqArray[index]
	double mass;
	long i,z;
	float FourMap[1024];
	float Zvalue[1024];
	float ZMap[1024];

	for(i=0; i<1024; i++) {
		CSValue[i]=0;
		CSMap[i]=0;
		FourMap[i]=0;
		Zvalue[i]=0;
		ZMap[i]=0;
	}

	mass = pDoc->PtToMass(index);
	z = 1;

	switch(pDoc->CSDetMethod) {
	case CSDET_DELTAMASS:
		z = GetComboCSMap(mass,ZMap,Zvalue);
		break;
	case CSDET_MULTIPLICATIVE:
		z = GetMultiplicativeMap(mass, ZMap, Zvalue);
		break;
	case CSDET_FOURIER:
		GetFourierMap(mass,FourMap,Zvalue);
		z = GetFourierCSValue(FourMap,Zvalue,ZMap);
		break;
	case CSDET_COMBO:
		z = GetComboCSMap(mass, ZMap, Zvalue);
		break;
	case CSDET_AUTOCORR:
		z = GetAutoCorrMap(mass,ZMap,Zvalue);
		break;
	case CSDET_OCONNOR:
		GetMultiplicativeMap(mass,ZMap,Zvalue);
		GetFourierMap(mass,FourMap,Zvalue);
		z = GetOConnorCSValue(FourMap,Zvalue,ZMap);
		break;
	case CSDET_ENTROPY:
		z = GetEntropyMap(mass, ZMap, Zvalue);
		break;
	case CSDET_SPARE1:
		z = GetComboCSMap(mass,ZMap,Zvalue);
		break;
	}

	for(i=0; i<ChargeLimit*10; i++) {
		CSMap[i] = ZMap[i];
		CSValue[i] = (float)((double)(i/10.0));
	}

	CSCalcResult = z;
	return z;
}


void CChargeStateDetermination::ComputeAutoCorrCS()
{
	int pk;
	float Zvalue[CHARGE_STATE_MAP_SIZE];
	float ZMap[CHARGE_STATE_MAP_SIZE];
	double mass1;
	int z;

	// for each peak in mi table
	for(pk=0; pk<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;pk++) {
		mass1 = pDoc->FreqToMass(pDoc->FITable[pk].F);

		z = GetAutoCorrMap(mass1,ZMap,Zvalue);

		// now make sure it != 0
		if(z == 0) z = 1;
		if(z>ChargeLimit) z = 1;

		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) z = z * -1;

		// finally dump cs info back into the mi table
		pDoc->FITable[pk].CS = z;
	}	// end for pk (mitable loop)

}

int CChargeStateDetermination::GetAutoCorrMap(double mass, float *ZMap, float *Zvalue)
{
	long pt, pt1, pt2,i,k;
	double deltamass;
	float *src;
	float *dst;
	int lags;
	int z;
	z=0;
	float cs,zmax,slope;
	zmax = 0;
	cs = 1;

	pt = pDoc->MassToPt(mass);
	deltamass = fabs(pDoc->PtToMass(pt) - pDoc->PtToMass(pt-1));
	pt1 = pDoc->MassToPt(mass - 1.5);
	pt2 = pDoc->MassToPt(mass + 1.5);
	if(pt2 < pt1) {
		// swap using i as the temp variable.
		i = pt2;
		pt2=pt1;
		pt1=i;
	}

	src = new float [pt2-pt1];

	for(i=pt1; i<pt2; i++) src[i-pt1] = pDoc->FreqArray[i];
	lags = (int)(1.1/deltamass);
	dst = new float [lags];

	nspsAutoCorr(src,pt2-pt1,dst,lags,NSP_Normal);
	for(i=0; i<lags; i++) {
		Zvalue[lags-i] = (float)(1.0/((double)i*deltamass));
		if(Zvalue[lags-i] > ChargeLimit) dst[i] = 0;
		ZMap[lags-i] = dst[i];
		if(dst[i] > zmax) {
			zmax = dst[i];
			cs = Zvalue[lags-i];
		}
	}

	delete [] dst;
	delete [] src;

	for(k=0; k<ChargeLimit*10; k++) {
		zmax = (float) k/(float)10.0;
		for(i = 0; i<CHARGE_STATE_MAP_SIZE; i++) {
			if(zmax < Zvalue[i]) break;
		}
		slope = (ZMap[i-1]-ZMap[i])/(Zvalue[i-1]-Zvalue[i]);
		ZMap[k] = (float) fabs((double)(ZMap[i-1] + slope*(zmax-Zvalue[i-1])));
	}
	for(k=0; k<ChargeLimit*10; k++) {
		zmax = (float) k/(float)10.0;
		Zvalue[k] = zmax;
	}

	z = (int)floor(cs);
	if(cs - (float)z > 0.5) return z+1;
	else return z;
}

int CChargeStateDetermination::GetAutoCorrFourierCSValue(float *FourMap, float *Zvalue, float *ZMap)
{
	float deltaz;
	int lags,i;
	int z;
	z=0;
	float cs,zmax;
	zmax = 0;
	cs = 1;

	deltaz = (float)fabs((double)(Zvalue[5] - Zvalue[6]));
	lags = (int)(1.1*(double)ChargeLimit/deltaz);

	nspsAutoCorr(FourMap,CHARGE_STATE_MAP_SIZE,ZMap,lags,NSP_Normal);
	for(i=0; i<lags; i++) {
		Zvalue[i] = (float)((double)i*deltaz);
		if(Zvalue[i] > ChargeLimit) ZMap[i] = 0;
		if(ZMap[i] > zmax) {
			zmax = ZMap[i];
			cs = Zvalue[i];
		}
	}

	z = (int)floor(cs);
	if(cs - (float)z > 0.5) return z+1;
	else return z;
}

void CChargeStateDetermination::ComputeOConnorCS()
{
	int pk;
	float FourMap[CHARGE_STATE_MAP_SIZE];
	float Zvalue[CHARGE_STATE_MAP_SIZE];
	float ZMap[CHARGE_STATE_MAP_SIZE];
	double mass;
	int z;

	// for each peak in mi table
	for(pk=0; pk<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;pk++) {
		mass = pDoc->FreqToMass(pDoc->FITable[pk].F);
		GetFourierMap(mass,FourMap,Zvalue);

		// Multiplicative map the fourier space
		z = GetOConnorCSValue(FourMap,Zvalue,ZMap);

		// now make sure it != 0
		if(z == 0) z = 1;
		if(z>ChargeLimit) z = 1;

		// pos or neg?
		if(pDoc->TrapVoltsOnDetect < 0) z = z * -1;

		// finally dump cs info back into the mi table
		pDoc->FITable[pk].CS = z;
	}	// end for pk (mitable loop)
}


int CChargeStateDetermination::GetOConnorCSValue(float *FourMap, float *Zvalue, float *ZMap)
{
	float MpyMap[1024];
	double fourval,slope;
	int j,i,k,m;
	float z;
	// int imax;
	double cs,zmax;
	zmax = 0;
	cs = 1;

	for(i=0; i<1024; i++) {
		MpyMap[i]=ZMap[i];
	}

#define USE_LOG
//#define USE_MPY
#define NUM_FOURIER_PEAKS 5

#ifdef USE_LOG
	double modelval;
	double model[CHARGE_STATE_MAP_SIZE/2];
	double fmap[CHARGE_STATE_MAP_SIZE/2];
	double dz;
	double z0,z1;
	double entval;

	for(i=0; i<CHARGE_STATE_MAP_SIZE/2; i++) fmap[i] = 0;
	fourval = 0;

	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		if(z<0.85) {
			continue;
		}

		// get fourval here.
		for(j = 0; j<CHARGE_STATE_MAP_SIZE/2; j++) {
			if(z < Zvalue[j]) break;
		}
		if(Zvalue[j-1] == Zvalue[j]) {
			fourval = (double) FourMap[j];
		}
		else {
			slope = ((double)FourMap[j-1]-(double)FourMap[j])/((double)Zvalue[j-1]-(double)Zvalue[j]);
			fourval = fabs(((double)FourMap[j-1] + slope*(z-(double)Zvalue[j-1])));
		}
		fmap[k] = fourval;
	}

	//normalize fmap
	zmax = 0;
	for(k=0; k<ChargeLimit*10; k++) {
		zmax += fmap[k];
	}
	for(k=0; k<ChargeLimit*10; k++) {
		fmap[k] = fmap[k]/zmax;
	}

	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		if(z<0.85) {
			ZMap[k] = 0;
			continue;
		}

		// create model;	
		for(i=0; i<ChargeLimit*10; i++) model[i] = 0;
		for(i=1; i<NUM_FOURIER_PEAKS; i++) {

			// get fourier value for charge x*i
			z0 = (double)z*(double)i;
			if(z0 > ChargeLimit*2) continue;
			modelval = 1.0/(double)i;

			for(j=0; j<20; j++) {
				z1 = z0+((double)j-10)/10.0;
				dz = (z1 - z0)/0.5;
				dz = exp(-1*dz*dz);
				m = (int)(z1*10.0);
				if(m < CHARGE_STATE_MAP_SIZE/2 && m >= 0) 
					model[(int)(z1*10.0)] += modelval*dz;
			}
		}
		// normalize model;
		dz = 0;
		for(j = 0; j<ChargeLimit*10; j++) dz += model[j];
		for(j = 0; j<ChargeLimit*10; j++) model[j] = model[j]/dz;		
		
		entval = 0;
		for(i=0; i<ChargeLimit*10; i++) {
			z = (float) i/(float)10.0;
			if(z<0.85) continue;

			// the entropy distance equation
			if(model[i] > 1.0e-5) entval += model[i]*log(model[i]/fmap[i]); 
		}
		ZMap[k] = (float)(exp(-1.0*entval))*MpyMap[k];
		//ZMap[k] = (float)(exp(-1.0*entval));
	}

/*	rescale the zmap
	zmax = 0;
	z0 = 0;
	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/10;
		if (z>9.3 && z < 9.5) {
			z=z;
		}
		if(ZMap[k] > zmax) zmax = ZMap[k];
		if(ZMap[k] < z0) z0 = ZMap[k];
	}
	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/10;
		if (z>9.3 && z < 9.5) {
			z=z;
		}
		ZMap[k] = 100.0*(ZMap[k]-z0)/(zmax-z0);
	}
	*/
#endif

#ifdef USE_MPY
	for(z=0; z<ChargeLimit; z+=(float)0.1) {
		// get maximum in fourmap within +/- 0.25 Z of z
		zmax = 0;
		imax = (int)z;
		for(i=0; i<CHARGE_STATE_MAP_SIZE/2; i++) {
			if(Zvalue[i] < (z - (float)0.25)) continue;
			if(Zvalue[i] > (z + (float)0.25)) break;
			if(FourMap[i] > zmax) {
				zmax = FourMap[i]; 
				imax = i;
			}
		}
		fourval = zmax;
		for(i=1; i<NUM_FOURIER_PEAKS; i++) {
			// get fourier value for charge x*i
			z0 = Zvalue[imax] + z*(float)i;
			if(z0 > ChargeLimit*2) continue;
			for(j = 0; j<CHARGE_STATE_MAP_SIZE/2; j++) {
				if(z0 < Zvalue[j]) break;
			}
			if(Zvalue[j-1] == Zvalue[j]) {
				fourval *= FourMap[j];
			}
			else {
				slope = (FourMap[j-1]-FourMap[j])/(Zvalue[j-1]-Zvalue[j]);

				// multiply it by fourval
				fourval *= (float) fabs((double)(FourMap[j-1] + slope*(z0-Zvalue[j-1])));
			}
		}
		ZMap[(int)(z*10)] = fourval*MpyMap[(int)(z*10)]; 

	}

#endif


	zmax = 0;
	for(z=0; z<ChargeLimit; z+=(float)0.1) {
		if (z>9.3 && z < 9.5) {
			z=z;
		}
		Zvalue[(int)(z*10)] = z;
		if(ZMap[(int)(z*10)] > zmax) {
			zmax = ZMap[(int)(z*10)];
			cs = z;
		}
	}



	i = (int)floor(cs);
	if(cs - (float)i > 0.5) return i+1;
	else return i;
}

int CChargeStateDetermination::GetComboCSMap(double mass, float *ComboMap, float* Zvalue )
{
	float ics;
	int z;
	float zmax;
	ics = 1;
	float AutoCorrMap[1024], FourMap[1024];
	for(z=0; z<1024; z++) {
		AutoCorrMap[z]=0;
		FourMap[z]=0;
	}


	// Multiplicative part
	GetAutoCorrMap(mass, AutoCorrMap, Zvalue);

	// fourier part
	GetFourierMap(mass,FourMap,Zvalue);
	GetFourierCSValue(FourMap,Zvalue,ComboMap);

	// now combomap = fourier map * Multiplicative map.  ics is z of max pt.
	zmax=0;
	ics = 1.0;
	for(z=0; z<ChargeLimit*10; z++) {
		ComboMap[z]=AutoCorrMap[z]*ComboMap[z];
		if(zmax<ComboMap[z]) {
			zmax=ComboMap[z];
			ics = (float)z/(float)10.0;
		}
	}

	z=(int)floor(ics);

	if(ics - (float)z > 0.5) return z+1;
	else return z;
}

int CChargeStateDetermination::GetEntropyMap(double mass, float *EMap, float *Zvalue)
{
	float ics,zmax;
	int i,j,k,maxprob;
	long pt,pt1,pt2,pt0,temp,numpts;
	float z;
	double mass2;
	double value;
	float prob[32];
	float model[10000];
	float TotalData;
	long PtWidth;

	ics = 1;
	numpts = 0;
	PtWidth = 0;

	pt = pDoc->MassToPt(mass);
#define a pDoc->FreqArray
	for(PtWidth = 1; PtWidth < 20; PtWidth++) {
		if(a[pt+PtWidth] > a[pt+PtWidth-1]) break;
	}
	if(PtWidth < 2) PtWidth = 2;
	PtWidth = PtWidth /2;

	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		Zvalue[k] = z;
		if(z<0.85) {
			EMap[k] = 0;
			continue;
		}
		if(z<2){
			pt1 = pDoc->MassToPt(mass - 2);
			pt2 = pDoc->MassToPt(mass + 2);
		} else {
			pt1 = pDoc->MassToPt(mass - 1);
			pt2 = pDoc->MassToPt(mass + 1);
		}
		if(pt2<pt1) {
			temp = pt1;
			pt1=pt2;
			pt2=temp;
		}
		TotalData = 0;
		maxprob = 0;
		numpts = (int) fabs((double)pt2-(double)pt1);
		for(j=pt1; j<pt2; j++) {
			TotalData += pDoc->FreqArray[j];
		}

		GetModelIsotopicDistribution(mass,z,prob,32);
		for(j=0; j<32; j++) {
			if(prob[j] > prob[maxprob]) maxprob = j;
		}
		
		// now make model
		for(j=0; j<numpts; j++) model[j]=0;
		for(j=0; j<32; j++) {
			if((maxprob+j) >= 32) continue;
			if(prob[maxprob+j] < 0.1*prob[maxprob]) continue;
			mass2 = mass + (double)j*1.0034/double(z);
			pt0 = pDoc->MassToPt(mass2);
			if(pt0 < pt1 || pt0 > pt2) continue;
			for(i = -PtWidth*3; i<PtWidth*3+1; i++) {
				if((pt0+i)<=pt1) continue;
				if((pt0+i)>=pt2) continue;
				value = ((double)i/(double)PtWidth)*((double)i/(double)PtWidth);
				value = (double)prob[maxprob+j] * exp(-1.0*value);
				model[(pt0-pt1)+i] += (float)value;
			}
			if(j==0) continue;
			
			if((maxprob-j) < 0) continue;
			if(prob[maxprob-j] < 0.1*prob[maxprob]) continue;;
			mass2 = mass - (double)j*1.0034/double(z);
			pt0 = pDoc->MassToPt(mass2);
			if(pt0 < pt1 || pt0 > pt2) continue;
			for(i = -PtWidth*3; i<PtWidth*3+1; i++) {
				if((pt0+i)<=pt1) continue;
				if((pt0+i)>=pt2) continue;
				value = (i/PtWidth)*(i/PtWidth);
				value = prob[maxprob-j] * exp(-1*value);
				if((pt0-pt1)+i > numpts) {
					i=i;
				}
				model[(pt0-pt1)+i] += (float)value;
			}
		}
		// normalize model
		value = 0;
		for(i=0; i<numpts; i++) value += (double) model[i];
		for(i=0; i<numpts; i++) model[i] = model[i]/(float)value;

		value = 0;
		for(i=pt1; i<pt2; i++) {
			// the entropy distance equation
			if((i-pt1) > numpts) {
				break;
			}
			if(model[i-pt1] > 1.0e-5) value += model[i-pt1]*log(model[i-pt1]/(pDoc->FreqArray[i]/TotalData)); 
		}
		EMap[k] = (float)(exp(-1.0*value));
		EMap[k] *= EMap[k];
	}

	zmax = 0;
	for(k=0; k<ChargeLimit*10; k++) {
		z = (float) k/(float)10.0;
		if(EMap[k] > zmax) {
			zmax = EMap[(int)(z*10)];
			ics = z;
		}
	}

	i=(int)floor(ics);

	if(ics - (float)i > 0.5) return i+1;
	else return i;

}
