// Deconvolution.cpp: implementation of the CDeconvolution class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "buda.h"
#include "BudaDoc.h"
#include "BudaView.h"
#include <math.h>
#include <stdlib.h>
#include "Deconvolution.h"
#include "ChargeStateDetermination.h"

//#define USE_SECOND_ZHANG_DECON
#define MAX_PEAK_SIZE 256
#define SYSTEM_HIGH_MZ_LIMIT 10000.0
#define MAX_MI_LENGTH 8192

//#define ZHANG_DECON_AVAILABLE
#ifdef ZHANG_DECON_AVAILABLE

#include "ZhangDecon.h"
#define SPECTRUM_DECONV_LOWZ 0
#define SPECTRUM_DECONV_HIGHZ 1
#define SPECTRUM_DECONV_MIXEDZ 2

#endif

#define DELTA_C13 1.00335
#define MAX_PEAK_SIZE 256

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDeconvolution::CDeconvolution()
{
	long i;
	DeconSize = MAX_DECON_SIZE;
	
	for(i=0; i<MAX_DECON_SIZE; i++) DeconData[i] = (float)0.0;
	DeconHighChargeLimit = 30;
	DeconLowChargeLimit = 1;
	DeconHighMassLimit = 30000.0;
	DeconLowMassLimit = 1000.0;
	DeconMethod = DECON_FENN;
	DeconRes = (float)0.1;
	MaxDeconIndex = 0;
	MaxDeconIntensity = 0;
	DeconIsotopicallyResolved = 1;
}

CDeconvolution::~CDeconvolution()
{
}

void CDeconvolution::CalculateFennDecon()
{
	long i,z;
	double CurrentMass;
	double CurrentMZ;
	double CurrentIntensity;
	long lowcharge,highcharge;
	lowcharge = (long)DeconLowChargeLimit;
	highcharge = (long)DeconHighChargeLimit;

	MaxDeconIntensity = 0;
	CheckDeconSize();

	for(i=0; i<DeconSize; i++) {
		// this is the main loop.
		CurrentMass = (double)i*DeconRes + DeconLowMassLimit;
		CurrentIntensity=0;
		// Fenn deconvolution is the sum of the data points at mass/z 
		// where z is the charges expected .... regardless of noise problems
		for(z=highcharge; z>lowcharge; z--) {
			CurrentMZ = (CurrentMass + (double)z*AdductMass)/(double)z;
			CurrentIntensity += GetIntensityatMass(CurrentMZ);
		}
		DeconData[i] = (float)CurrentIntensity;
		if(MaxDeconIntensity < DeconData[i]) {
			MaxDeconIntensity = DeconData[i];
			MaxDeconIndex = i;
		}
	} // end for i
}

void CDeconvolution::CalculateReinholdDecon()
{
	/*****************************************
	This is a modified form of the original Reinhold Deconvolution.
	as defined in:
	(1)	"Electrospray ionization mass spectrometry: Deconvolution 
	by an entropy-based algorithm" Reinhold, B. B.; Reinhold, V. N. 
	Journal of the American Society for Mass Spectrometry 1992, 3, 207-215.

	In this form, the model distribution is assumed to be unit abundance
	peaks at the appropriate m/z places.  This allows a much faster 
	calculation, but with less specificity.

	The form of the reinhold eqn. is 
	decon(mass) = sum_z(model(m/z)*log(model(m/z)/data(m/z)))
	when model(m/z) = 1, this is simply sum_z(log(1/data(m/z))

    since log(A) + log(B) = log(A*B), this form becomes 
	log(prod_z(data(m/z))).

	todo:
	Implement a better model.  What should it be?

	*****************************************/
	long i,z;
	double CurrentMass;
	double CurrentMZ;
	double CurrentIntensity;
	double CurrentVal;
	double Threshold;
	long lowcharge,highcharge;
	lowcharge = (long)DeconLowChargeLimit;
	highcharge = (long)DeconHighChargeLimit;


	MaxDeconIntensity = 0;
	CheckDeconSize();
	Threshold = GetNoiseThreshold(); 
	// set threshold to the average noise and don't include 
	// intensities below the average noise value + 2*sigma.  
	Threshold += 2*NoiseSigma;

	for(i=0; i<DeconSize; i++) {
		// this is the main loop.
		CurrentMass = (double)i*DeconRes + DeconLowMassLimit;
		CurrentIntensity=1;

		for(z=highcharge; z>lowcharge; z--) {
			CurrentMZ = (CurrentMass + (double)z*AdductMass)/(double)z;
			CurrentVal = GetIntensityatMass(CurrentMZ);
			if(CurrentVal !=0 ) {
				CurrentIntensity *= CurrentVal/Threshold;
			} else {
				CurrentIntensity *= Threshold/1000;
			}

		}
		//DeconData[i] = (float)log(CurrentIntensity);
		DeconData[i] = (float)CurrentIntensity;
		if(MaxDeconIntensity < DeconData[i]) {
			MaxDeconIntensity = DeconData[i];
			MaxDeconIndex = i;
		}
	} // end for i

}

void CDeconvolution::CalculateOConnorDecon()
{
	CalculateOConnor2Decon();
	return;
	/*
	long i,z;
	double CurrentMass;
	double CurrentMZ;
	double CurrentIntensity;
	float* ReinDecon;
	long lowcharge,highcharge;
	lowcharge = (long)DeconLowChargeLimit;
	highcharge = (long)DeconHighChargeLimit;

	/*****************************************
	This is the deconvolution method defined in:
	(1)	"Oligomer Characterizationof 4-22 kDa Polymers by 
	Electrospray Fourier-Transform Mass Spectrometry" 
	O'Connor, P. B.; McLafferty, F. W. Journal of the American 
	Chemical Society 1996, 117, 12826-31.

	In this deconvolution, the high discrimination of the Reinhold
	method is used to filter the highly accurate abundances from 
	the Fenn method.

	*****************************************/

	/*CalculateReinholdDecon();
	GetDeconThreshold();

	// duplicate the ReinholdDeconData and clear the DeconData Array;
	ReinDecon = new float[DeconSize];
	for(i=0; i<DeconSize; i++) {
		ReinDecon[i]=DeconData[i];
		DeconData[i]=0;
	}

	for(i=0; i<DeconSize; i++) {
		// this is the main loop.
		CurrentMass = (double)i*DeconRes + DeconLowMassLimit;
		CurrentIntensity=0;
		// Fenn deconvolution is the sum of the data points at mass/z 
		// where z is the charges expected .... regardless of noise problems
		for(z=highcharge; z>lowcharge; z--) {
			CurrentMZ = (CurrentMass + (double)z*AdductMass)/(double)z;
			CurrentIntensity += GetIntensityatMass(CurrentMZ);
		}
		if(ReinDecon[i] > (float)(DeconThreshold + 3*DeconSigma)) 
			DeconData[i] = (float)CurrentIntensity;
		else DeconData[i] = (float)0.0;
	} // end for i

	MaxDeconIntensity = DeconData[MaxDeconIndex];
	delete [] ReinDecon;
	*/
}

long CDeconvolution::MassToPt(float mass)
{
	double m;
	// this converts an absolute mass in decon space to 
	// a point number in the decondata array.

	// point = (highpoint-lowpoint)/(highmass-lowmass)*(mass-lowmass) + lowpoint
	// mass = (highmass-lowmass)/(highpoint-lowpoint)*(point-lowpoint) + lowmass
	// lowpoint = 0
	m = (double)DeconSize/(DeconHighMassLimit-DeconLowMassLimit);
	return (long)(((double)mass-DeconLowMassLimit)*m);
}

float CDeconvolution::PtToMass(long point)
{
	double m;
	// this converts a point number in the decondata array
	// to an absolute mass in decon space.

	// point = (highpoint-lowpoint)/(highmass-lowmass)*(mass-lowmass) + lowpoint
	// mass = (highmass-lowmass)/(highpoint-lowpoint)*(point-lowpoint) + lowmass
	// lowpoint = 0
	m = (DeconHighMassLimit-DeconLowMassLimit)/(double)DeconSize;
	return (float)(m*(double)point + DeconLowMassLimit);
}

double CDeconvolution::GetIntensityatMass(double mass)
{
	// this function returns the FTData value at a particular
	// mass to charge ratio (mz).  If this falls between to data
	// points in the FT domain, it interpolates linearly.
	long pt0,pt1,pt2;
	double int1,int2,intensity,slope;

	if(mass < pDoc->PS.DetectLowMass) return 0;
	if(mass > pDoc->PS.DetectHighMass) return 0;

	pt0 = pDoc->MassToPt(mass);
	double tempmass;

	// check if mass of pt1 is above or below mass
	tempmass = pDoc->PtToMass(pt0);
	if (mass > tempmass) {
		pt1 = pt0;
		pt2 = pt1+1;
	} else {
		pt2=pt0;
		pt1=pt2-1;
	}
	int1 = pDoc->FreqArray[pt1];
	int2 = pDoc->FreqArray[pt2];

	// interpolate
	tempmass = pDoc->PtToMass(pt1);
	slope = double(double(int2-int1)/(pDoc->PtToMass(pt2)-tempmass));
	intensity = int1+(slope*double(mass - tempmass));
	return intensity;
}

void CDeconvolution::CheckDeconSize()
{
	// if the number of data points requested by the mass range and
	// resolution exceeds the maximum, adjust resolution till it 
	// is ok.

	double TempDeconRes;
	long TotalPoints;

	TempDeconRes = DeconRes;
	TotalPoints = (long)((DeconHighMassLimit - DeconLowMassLimit) / TempDeconRes);
	while(TotalPoints > MAX_DECON_SIZE) {
		TempDeconRes*=2;
		TotalPoints = (long)((DeconHighMassLimit - DeconLowMassLimit) / TempDeconRes);
	}
	if(TempDeconRes != DeconRes) {
		pView->MessageBox("Deconvolution requested too many data points.  Adjusting resolution to compensate","Memory Error",MB_OK);
		DeconRes = TempDeconRes;
		pDoc->DeconResolution = DeconRes;
		pApp->DeconResolution = DeconRes;
	}
	DeconSize=TotalPoints;
}

double CDeconvolution::GetNoiseThreshold()
{
	long i, firstpt, lastpt, numpts;
	double avg, threshold, error;
	double sigma, oldthreshold, oldsigma;
	int passnum;

	// this function will automatically determine the 2*sigma noise threshold

	// first choose a random region in the spectrum of 1000 points or so.
	// then, find the average value for the noise in this region rejecting
	// those points outside average+3*sigma (they are peaks).
	//
	// repeat this until threshold stabilizes.
	// if >90% of the points in this region are "noise", accept this value
	// otherwise, try another random region.
	
	numpts = 0;
	threshold = 0;
	sigma = 0;

	while (numpts < 900) {
		firstpt = (long)((double)pDoc->PS.SpectrumPoints*(double)rand()/(double)RAND_MAX);
		if(firstpt > pDoc->PS.SpectrumPoints - 1000) firstpt -= 1000;
		lastpt = firstpt+1000;
		numpts = 0;

		threshold = 0.0;	// set initial value for threshold = maximum value in range
		for(i=firstpt; i<lastpt; i++) 
			if(pDoc->FreqArray[i]>threshold) threshold = pDoc->FreqArray[i];
		oldthreshold = 0.0;
		error = 0.001*threshold; // within 0.1% of maximum peak.
		sigma = threshold;	// very big initial sigma

		passnum=0;
		while(fabs(threshold-oldthreshold) > error || passnum < 3) {
			oldthreshold = threshold;
			oldsigma=sigma;
			sigma=0;
			avg = 0.0;
			numpts=0;
			for(i=firstpt; i<lastpt; i++) {
				if(pDoc->FreqArray[i] < oldthreshold+3*oldsigma) {
					avg+=pDoc->FreqArray[i];
					sigma += (pDoc->FreqArray[i]-oldthreshold)*(pDoc->FreqArray[i]-oldthreshold);
					numpts++;
				}
			} // for i
			avg=avg/numpts;
			sigma=sqrt(sigma/double(numpts-1.0));
			threshold = avg;
			passnum++;
		} // while threshold > error
	} // while numpts

	NoiseThreshold = threshold;
	NoiseSigma = sigma;
	return threshold;
}


double CDeconvolution::GetDeconThreshold()
{
	long i, firstpt, lastpt, numpts;
	double avg, threshold, error;
	double sigma, oldthreshold, oldsigma;
	int passnum;

	// this function will automatically determine the 2*sigma noise threshold

	// first choose a random region in the spectrum of 1000 points or so.
	// then, find the average value for the noise in this region rejecting
	// those points outside average+3*sigma (they are peaks).
	//
	// repeat this until threshold stabilizes.
	// if >90% of the points in this region are "noise", accept this value
	// otherwise, try another random region.
	
	firstpt = MaxDeconIndex-1500;
	if(firstpt < 0 ) firstpt = MaxDeconIndex+500;
	if(firstpt > DeconSize - 1000) firstpt -= 1000;
	lastpt = firstpt+1000;
	numpts = 0;

	threshold = 0.0;	// set initial value for threshold = maximum value in range
	for(i=firstpt; i<lastpt; i++) 
		if(DeconData[i]>threshold) threshold = DeconData[i];
	oldthreshold = 0.0;
	error = 0.001*threshold; // within 0.1% of maximum peak.
	sigma = threshold;	// very big initial sigma

	passnum=0;
	while(fabs(threshold-oldthreshold) > error || passnum < 3) {
		oldthreshold = threshold;
		oldsigma=sigma;
		sigma=0;
		avg = 0.0;
		numpts=0;
		for(i=firstpt; i<lastpt; i++) {
			if(DeconData[i] < oldthreshold+3*oldsigma) {
				avg+=DeconData[i];
				sigma += (DeconData[i]-oldthreshold)*(DeconData[i]-oldthreshold);
				numpts++;
			}
		} // for i
		avg=avg/numpts;
		sigma=sqrt(sigma/double(numpts-1.0));
		threshold = avg;
		passnum++;
	} // while threshold > error

	DeconThreshold = threshold;
	DeconSigma = sigma;
	return threshold;
}

void CDeconvolution::CalculateOConnor2Decon()
{
	/* this decon method will try to use the superior charge state
	 determination routines already available in cs_det to get a 
	 deconvolved mass spectrum using a modified peaklist algorithm.

	The idea is to iterate through the raw data doing so:
		1) find highest unprocessed peak from peaklist.
		2) find it's charge state
		3) find all it's neighboring peaks of the same MASS 
			isotopic distribution(other charge states )
		4) add them to the DeconData array
		5) mark these peaks "processed"
		6) goto 1
	*/
	long i;
	double CurrentIntensity;
	double Threshold;
	long NumPeaks, nextpeak;
	int CurrentCharge;
	long* indexx;
	CurrentIntensity = 0.0;

	MaxDeconIntensity = 0;
	CheckDeconSize();
	for(i=0; i<DeconSize; i++) DeconData[i] = (float)0.0;

	// first, get the noise threshold so that we can avoid lengthy calcs on noise.
	Threshold = GetNoiseThreshold(); 
	// set threshold to the average noise and don't include 
	// intensities below the average noise value + 2*sigma.  
	Threshold += 2*NoiseSigma;

	// second, mark all the raw spectrum points unprocessed
	// this will be done by creating an indextable of the mitable.
	// if a peak is processed, it will be given the index value of -1
	// otherwize, it will be >=0.
	NumPeaks = pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;
	indexx = new long[NumPeaks];
	// create index table of the mi table.
	CreateIndexTable(NumPeaks, indexx);

	// do the main loop.
	for(nextpeak = 0; nextpeak < NumPeaks; nextpeak++) {
		if(indexx[nextpeak]<0) continue;
		CurrentCharge = pDoc->FITable[indexx[nextpeak]].CS;
		FindNeighbors(CurrentCharge,nextpeak,indexx);
	}

	MaxDeconIntensity = 0;
	for(i=0; i<DeconSize; i++) {
		if(DeconData[i]>MaxDeconIntensity) {
			MaxDeconIntensity = DeconData[i];
			MaxDeconIndex = i;
		}
	}
}

void CDeconvolution::FindNeighbors(int CurrentCharge, long CurrentIndex, long* indexx)
{
	/* now, we need to add this peak to DeconData array, but
	   we also need to find all of it's neighbors (isotopes and charge states)
	   and add them as well.  

	   So, 1) find all the isotopes and note their masses.
		   2) for each of these masses, check all the charge states to see
				if there is intensity above threshold and to see if the charge
				state at that peak is what you think it is.
		   3) if ok, add to DeconData.
	*/
	double CurrentMass;
	double CurrentMZ;
	long i, j, isomassindex,pt,NumPeaks;
	int lowcharge, highcharge;
	double IsotopeMasses[128];
	double tempint;
	double tempmass,tempMZ;
	double plusminus;
	double deltafreq;

	NumPeaks = pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks;
	CurrentMZ = pDoc->FreqToMass(pDoc->FITable[indexx[CurrentIndex]].F);
	CurrentMass = CurrentCharge*(CurrentMZ - AdductMass);
	tempint=pDoc->FITable[indexx[CurrentIndex]].I;
	
	// 1) find all the isotopes and note their masses.
	isomassindex = 1;
	IsotopeMasses[0] = CurrentMZ;

	// calculate isotope masses (plus side first)
	pt = pDoc->FreqToPt(pDoc->FITable[indexx[CurrentIndex]].F);
	deltafreq = pDoc->PtToFreq(pt) - pDoc->PtToFreq(pt-1);
	indexx[CurrentIndex] = -1;

	plusminus = +1.0;
	while(1) {
		// this loop is executed twice so that plusminus can be first
		// = +1.0, then = -1.0 to get both sides of the isotope distribution
		i=1;
		while(1) {
			// this loop is executed one peak at a time until the isotopes are 
			// no longer above noisethreshold
			tempmass=CurrentMass+DELTA_C13*(double)i*plusminus;
			tempMZ = tempmass/(double)CurrentCharge+AdductMass;
			tempint = 0;

			// is there a peak there?  What is it's intensity?
			// first check mi table, but also check raw data to go lower
			// than the mi table does into the noise
			pt = GetMITableInt(tempMZ,deltafreq);
			if(pt > 0) {
				tempint = pDoc->FITable[pt].I;
				for(j=0; j<NumPeaks; j++) {
					if(indexx[j]==pt) {
						indexx[j] = -1;
						break;
					}
				}
			} else {
				tempint = GetIntensityatMass(tempMZ);
			}
			i++;

			if(tempint > NoiseThreshold+NoiseSigma) {
				IsotopeMasses[isomassindex] = tempMZ;
				isomassindex++;
			} else break;
		}  // end while isotopes
		if(plusminus == +1.0) 
			plusminus = -1.0;
		else break;
	}	// end while plusminus


//  2) for each of these masses, check all the charge states to see
//	if there is intensity above threshold and to see if the charge
//	state at that peak is what you think it is.
//  3) if ok, add to DeconData.

	lowcharge = DeconLowChargeLimit;
	highcharge = (int)(DeconHighMassLimit/CurrentMZ) + 1;
	if((unsigned int)highcharge > DeconHighChargeLimit) highcharge = DeconHighChargeLimit;

	for(j=0; j<isomassindex; j++) {
		AddToDeconData(IsotopeMasses[j],CurrentCharge);
	}

	for(j=0; j<isomassindex; j++) {
		for(i=lowcharge; i<highcharge; i++) {
			if (i==CurrentCharge) continue;	// don't bother to check this one again.  already done.
			tempmass = IsotopeMasses[j]/(double)i;
			tempint = GetIntensityatMass(tempmass);
			if(tempint > NoiseThreshold) {
				AddToDeconData(tempmass,i);
			}
		}
	}
}

void CDeconvolution::AddToDeconData(double MZ, int charge)
{
	// this function adds this peak to the decondata array 

	long i,pt;
	long massindex, peakstart, peakstop;
	long DeconNum0, DeconNum1;
	double DeconMass;
	double value;
	long n;
	double x[MAX_PEAK_SIZE];
	double y[MAX_PEAK_SIZE];
	double y2[MAX_PEAK_SIZE];
	double yp1,ypn;

	// 1) find the true peakstart and stop.

	// for code readability, use this define
	#define A pDoc->FreqArray
	#define T NoiseThreshold

	massindex = pDoc->MassToPt(MZ);
	// find peakstart
	for(i=massindex; TRUE ; i++) {
		// break out of here only if nextpoint and peak after are both
		// greater than thispoint+NoiseThreshold "T".
		if((A[i] < A[i+1]+T) && (A[i] < A[i+2]+T)) break;
	}
	peakstop = i;
	// find peakstop
	for(i=massindex; TRUE ; i--) {
		// break out of here only if nextpoint and peak after are both
		// greater than thispoint+NoiseThreshold "T".
		if((A[i] < A[i-1]+T) && (A[i] < A[i-2]+T)) break;
	}
	peakstart = i;

	if(((pDoc->PtToMass(peakstart)-AdductMass)*charge)>DeconHighMassLimit) return;
	if(((pDoc->PtToMass(peakstart)-AdductMass)*charge)<DeconLowMassLimit) return;
	if(((pDoc->PtToMass(peakstop)-AdductMass)*charge)>DeconHighMassLimit) return;
	if(((pDoc->PtToMass(peakstop)-AdductMass)*charge)<DeconLowMassLimit) return;

	// 2) now, turn all these m/z points into mass points and add to decondata

	// get the spline that correctly estimates these points then sample
	// that spline on the decon points.
	n=peakstop-peakstart;
	if(n<5) return;

	if(n>MAX_PEAK_SIZE) pApp->MessageBox("DeconError!","Peak too big");
	for(i=0; i<MAX_PEAK_SIZE; i++) {
		x[i]=0.0;
		y[i]=0.0;
		y2[i]=0.0;
	}

	for(i=n; i>=0; i--) {
		pt=n-i + peakstart;
		x[i] = (pDoc->PtToMass(pt)-AdductMass)*(double)charge;
		y[i] = pDoc->FreqArray[pt];
		y2[i]=0;
	}
	yp1 = (y[2]-y[1])/(x[2]-x[1]);
	ypn = (y[n]-y[n-1])/(x[n]-x[n-1]);

#ifdef _DEBUG
/*	if(x[0] < 2000) {
		for(i=n; i>=0; i--) {
			TRACE("x[%d] = %1.4f, y%d = %1.4f\n",i,x[i],i,y[i]);
		}
	}
*/
#endif

	spline(x,y,n,yp1,ypn,y2);

	DeconNum0 = (long)((x[1] - DeconLowMassLimit)/DeconRes);
	DeconNum1 = (long)((x[n] - DeconLowMassLimit)/DeconRes);
	ASSERT(DeconNum0 >= 0);
	ASSERT(DeconNum0 <= DeconSize);
	ASSERT(DeconNum1 >= 0);
	ASSERT(DeconNum1 <= DeconSize);

	for(i=DeconNum0; i<DeconNum1+1; i++) {
		DeconMass = (double)i*DeconRes + DeconLowMassLimit;
		splint(x,y,y2,n,DeconMass,&value);
		DeconData[i]+=(float)value;
	}
}

long CDeconvolution::GetMITableInt(double mass, double deltafreq)
{
	// this looks through the MI table for a peak at this 
	// mass +/- deltafreq.

	long i;
	double lowfreq, highfreq;
	lowfreq = pDoc->MassToFreq(mass);
	highfreq = lowfreq + deltafreq;
	lowfreq -= deltafreq;
	for(i=0; i<pDoc->PS.Spec[pDoc->CurrentSpectrumFocus].NumPeaks; i++) {
		if(pDoc->FITable[i].F < highfreq && pDoc->FITable[i].F > lowfreq) {
			return i;
		}
	}
	return -1;
}

long CDeconvolution::FindHighestUnprocessedPeak(long *indexx, double *CurrentIntensity)
{
	// search through pDoc->FreqArray and return the highest remaining
	// unprocessed point.
	long i, index;
	double maxpt;
	index = -1;

	maxpt = 0;
	for(i=0;i<pDoc->PS.SpectrumPoints; i++) {
		if(maxpt < pDoc->FreqArray[i]) {
			if(!indexx[i]) {
				maxpt = pDoc->FreqArray[i];
				index = i;
			} // end if processed
		} // end if maxpt < pt
	} // end for i
	*CurrentIntensity = maxpt;
	return index;
}

int CDeconvolution::GetChargeState(long index)
{
	CChargeStateDetermination CSDet;
	CSDet.pDoc = pDoc;

	return CSDet.GetChargeState(index);
}


void CDeconvolution::CreateIndexTable(long n, long *indexx)
{
	// this is a simple sort using the straight insertion method
	// on pg 243 of Numerical Recipes in C also called piksrt
	// it isn't particularly fast, but should be fine for this.
	long i,j,l;
	double temp;

	// initialize indexx array;
	for(j=0; j<n; j++) indexx[j] = j;
	if(n<=1) return;

	for(j=1; j<n; j++) {
		l = indexx[j];
		temp = pDoc->FITable[indexx[j]].I;
		i=j-1;
		while(i>=0 && pDoc->FITable[indexx[i]].I < temp) {
			indexx[i+1] = indexx[i];
			i--;
		}
		indexx[i+1]=l;
	}
} //CreateIndexTable

void CDeconvolution::spline(double* x,double* y,long n,double yp1,double ypn,double* y2)
{
        long i,k;
        double p,qn,sig,un,u[MAX_PEAK_SIZE];

		for(i=0; i<MAX_PEAK_SIZE; i++) u[i]=0.0;

        if (yp1 > 0.99e30)
                y2[1]=u[1]=0.0;
        else {
                y2[1] = -0.5;
                u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
        }
        for (i=2;i<=n-1;i++) {
                sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
                p=sig*y2[i-1]+2.0;
                y2[i]=(sig-1.0)/p;
                u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
                u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
        }
        if (ypn > 0.99e30)
                qn=un=0.0;
        else {
                qn=0.5;
                un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
        }
        y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
        for (k=n-1;k>=1;k--)
                y2[k]=y2[k]*y2[k+1]+u[k];
}

void CDeconvolution::splint(double* xa,double* ya,double* y2a,long n,double x,double* y)
{
        long klo,khi,k;
        double h,b,a;

        klo=1;
        khi=n;
        while (khi-klo > 1) {
                k=(khi+klo) >> 1;
                if (xa[k] > x) khi=k;
                else klo=k;
        }
        h=xa[khi]-xa[klo];
        if (h == 0.0) {
			pApp->MessageBox("SplintError!","Bad XA input to routine SPLlong");
		}
        a=(xa[khi]-x)/h;
        b=(x-xa[klo])/h;
        *y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}

/*  vestigial code
	// 2) now, turn all these m/z points into mass points and add to decondata
	for(i=peakstart; i<peakstop-1; i++) {
		tempmass0 = ((pDoc->PtToMass(i+1)-AdductMass) * (double)charge);
		tempmass1 = ((pDoc->PtToMass(i)-AdductMass) * (double)charge);
		if(DeconRes <= fabs(tempmass1-tempmass0)) {
			// deconres < deltamass (i.e. deconvolution has morepoint
			// resolution than the data)
			//
			// now, tempMZ will usually land on the DeconData space between two
			// DeconData bins.  Also, depending on DeconRes and frequency resolution
			// there will often be several bins between adjacent points and the
			// mass spacing will be very nonlinear.  So, for eacy peak, I will
			// linearly interpolate to get values for the deconbins
			DeconNum0 = (long)((tempmass0 - DeconLowMassLimit)/DeconRes);
			DeconMass0 = (double)DeconNum0*DeconRes + DeconLowMassLimit;
			DeconNum1 = (long)((tempmass1 - DeconLowMassLimit)/DeconRes);
			DeconMass1 = (double)DeconNum1*DeconRes + DeconLowMassLimit;

			ASSERT(DeconNum0 >= 0);
			ASSERT(DeconNum0 <= DeconSize);
			ASSERT(DeconNum1 >= 0);
			ASSERT(DeconNum1 <= DeconSize);

			if(DeconNum0 > DeconNum1) {
				j=DeconNum1;
				DeconNum1=DeconNum0;
				DeconNum0 = j;

				temp = DeconMass1;
				DeconMass1 = DeconMass0;
				DeconMass0 = temp;

				temp = tempmass1;
				tempmass1=tempmass0;
				tempmass0=temp;
			}

			deltay = (double)(pDoc->FreqArray[i] - pDoc->FreqArray[i+1]);
			deltax = tempmass0-tempmass1;
			dydx = deltay/deltax;

			for(j=DeconNum0+1; j<DeconNum1; j++) {
				DeconMass = DeconRes*j + DeconLowMassLimit;
				value = (float)fabs((DeconMass-DeconMass0)*dydx + (double)pDoc->FreqArray[i]);
				DeconData[j] += value;
				#ifdef _DEBUG
				//TRACE("%d: DeconData[%d] += %1.5f\n",j-DeconNum0,j,value);
				#endif
			}
			DeconData[DeconNum0] += (float)((double)pDoc->FreqArray[i]*fabs(DeconMass0-tempmass0)/DeconRes);
			DeconData[DeconNum1] += (float)((double)pDoc->FreqArray[i+1]*fabs(DeconMass1-tempmass1)/DeconRes);
		} else {
			// deltamass < deconres (i.e. deconvolution has less point
			// resolution than the data)
			//
			// in this case, you need to add this point to both neighboring
			// deconpoints in a weighted fashion.

			DeconNum0 = (long)((tempmass0 - DeconLowMassLimit)/DeconRes);
			DeconMass0 = (double)DeconNum0*DeconRes + DeconLowMassLimit;

			DeconData[DeconNum0] += (float)((double)pDoc->FreqArray[i]*fabs(DeconMass0-tempmass0)/DeconRes);
			DeconData[DeconNum0+1] += (float)((double)pDoc->FreqArray[i]*fabs(DeconMass0+DeconRes-tempmass0)/DeconRes);
		}


*/

long CDeconvolution::CreateMITable(float* Mass, float* Intensity)
{
	long i;		
	long minpt,maxpt,temp;
	float noise;
	double Threshold;
	float* data;
	long NumPeaks;

	long PeakMax = 0;

	// Need to define these for various systems
	data = pDoc->FreqArray;
	minpt=pDoc->MassToPt(SYSTEM_HIGH_MZ_LIMIT);
	maxpt=pDoc->MassToPt(pDoc->PS.DetectLowMass);

	// set threshold to the average noise and don't include 
	// intensities below the average noise value + 2*sigma.  
	Threshold = NoiseThreshold + 3*NoiseSigma;
	noise = float(Threshold);

	// create empty array for peaks, and clear counter
	NumPeaks = 0;

	// find peaks
	if(minpt>maxpt) {
		temp = minpt;
		minpt = maxpt;
		maxpt = temp;
	}
	// peak requirements:
	//		must have 2 or more points above noise
	//		must have "smooth" surface where smooth is
	//			defined as:
	//				before summit, no negative deviations 
	//				in intensity greater than noise
	//				after summit, no positive deviations 
	//				in intensity greater than noise

	// i = curpoint
	for (i=minpt;i<maxpt;i++) {
		if(data[i]<noise) continue;

		// record points until slope change
		for(;i<maxpt-1; i++) {
			if(data[i] > data[i-1]) continue;

			if(data[i] < noise) break;
			if((data[i-1]-data[i])<noise) break;
			if(data[i+1] < data[i]) break;
		}

		PeakMax = i;
		//found a peak, add an entry to the MI tables
		Mass[NumPeaks] = (float)pDoc->PtToMass(PeakMax);
		Intensity[NumPeaks] = data[PeakMax];
		NumPeaks++;

		// we are here because the data indicated a negative slope
		// need to find end of peak before continuing.
		for(i++; i<maxpt-1; i++) {
			if(data[i]<noise) break;

			if((data[i]-data[i-1])>noise) {
				i--;
				break;
			}

			if((data[i] > data[i-1]) && (data[i+1] > data[i])){
				i--;
				break;
			}
		}
		
		// verify that we won't overflow the MI buffers
		if(NumPeaks >= MAX_MI_LENGTH) {
			pDoc->MessageBox("Peak Pick reached the maximum number of peaks\nbefore completing.  Some peaks may not have been found","Warning! Too many peaks");
			break;
		}
	} //end for loop
	return NumPeaks;
} 

/*CfReal CDeconvolution::fpMassToPt(CfReal mass)
{ 
	long pt;
	CBudaDoc* pD;

	pD = GetpDoc();

	pt = pD->MassToPt((double) mass);

	return (float) pt;
}

CfReal CDeconvolution::fpPtToMass(CfReal point)
{
	return (float) pDoc->PtToMass((long) point);
}

CBudaDoc* CDeconvolution::GetpDoc()
{
	return pDoc;
}*/
