// FFTWorkerThread1.cpp: implementation of the CFFTWorkerThread class.
//
//////////////////////////////////////////////////////////////////////
// FFTWorker.cpp : Defines the FFT thread function.
//
/*
File Description: FFTWorkerThread.cpp defines the main FFT worker
thread function.  This function executes the Fast Fourier Transform
and all zerofilling, apodization, and peakpicking as defined in the 
structure FFTThreadInfo (defined in FFTWorkerThread.h).  

******************************************************************/
/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "stdafx.h"
#include "Buda.h"
#include "BudaMaster.h"
#include "BudaDoc.h"
#include "Math.h"
#include "ChargeStateDetermination.h"
#include "FFTWorkerThread1.h"

//#define SINGLE_PRECISION_FFT
#define float_PRECISION_FFT
//#define ABUNDANCE_PEAKPICKER
#define FREQ_PEAKPICKER

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]=__FILE__;
	#define new DEBUG_NEW
#endif

// Buda debugging definitions
#ifdef _DEBUG
	//#define DEBUG_FFTCRASH	// used for debugging crash on close during fft bug
	//#define DEBUG_PEAKFIT
	//#define DEBUG_NSP_SQRT
#endif

// intel's digital signalling processing library
//#define nsp_UsesAll
#define nsp_UsesTransform
#define nsp_UsesVector
#define nsp_UsesWin
#define nsp_UsesConversion
#include "nsp.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFFTWorkerThread::CFFTWorkerThread()
{
	pMainApp = (CBudaApp*)AfxGetApp();
	Access = new CCriticalSection();
	DataLock = new CSingleLock(Access,FALSE);

	// FFTThrdInfo
	FFTThrdInfo.FFTsInQueue=0;
	FFTThrdInfo.m_hStartThread = CreateEvent(NULL, TRUE, FALSE, NULL	);	// event initially reset, manual reset
	FFTThrdInfo.m_hKillThread = CreateEvent(NULL, TRUE, FALSE, NULL	);	// event initially reset, manual reset
	FFTThrdInfo.m_hStartFFT = CreateEvent(NULL, TRUE, FALSE, NULL	);	// event initially reset, manual reset
	FFTThrdInfo.m_hThreadDead = CreateEvent(NULL, TRUE, FALSE, NULL	);	// event initially reset, manual reset
	// thread progress messages will go to mainframe window for updating the status bar
	//FFTThrdInfo.m_hwndNotifyFFTProgress = pWnd->m_hWnd;
	//FFTThrdInfo.pMainWnd = pMainApp->m_pMainWnd;

	// create FFTThread
	pFFT = this;
	FFTThrd = AfxBeginThread(FFTThreadProc, this);
	FITable = 0;
}

CFFTWorkerThread::~CFFTWorkerThread()
{
	if(Access != 0) {
		delete Access;
		Access = 0;
	}
	if(DataLock != 0) {
		delete DataLock;
		DataLock = 0;
	}
	if(FITable != 0) {
		delete [] FITable;
		FITable = 0;
	}
}

//////////////////////////////////////////////////////////////////////
// MainThread function
//////////////////////////////////////////////////////////////////////
UINT FFTThreadProc(LPVOID pFFTThrdClass) 
{
/* This thread runs in an infinite loop.  It is started in CBudaDoc constructor
in BudaApp.cpp and kills itself when m_hEventFFTThreadKilled is set which
is done in the CBudaDoc destructor.

The thread waits until event m_hEventStartFFT is set, calculates the 
FFTs (including apodization, zerofilling, and peakpicking as necessary)
until it runs out of FFT's to do and then resets m_hEventStartFFT.

*/

	struct PS* pPS;
	CFFTWorkerThread* pFFT = (CFFTWorkerThread*)pFFTThrdClass;
	int s;
	struct FFTThreadInfo* pInfo;	// pointer to the current FFT info
	pInfo = &(pFFT->FFTThrdInfo);
	struct FFTQueueEntry* CurrFFT;
	long i=0, j=0, order=0, pts=0;;
	SCplx* FTData=0;
	float* tarray = 0;
	float avg;


	WaitForSingleObject(pInfo->m_hStartThread, INFINITE);
	pInfo->pMainWnd = pFFT->pMainApp->m_pMainWnd;

	while (TRUE) // infinite loop
	{
		if(pFFT->UpdateProgress(0) == -1) break;	// initialize thread idle

		pFFT->Lock();
		pInfo->ThreadStatus = 0; // fft in stage 0 (idle)
		pFFT->Unlock();
		pInfo->pMainWnd->PostMessage(WM_FFTPROGRESS,0,0);

		// next check if QueueIsEmpty if so,  
		// wait for new StartFFT event then cycle loop.
		if(pInfo->FFTsInQueue == 0) {
			// reset StartFFTEvent then idle and wait for new StartFFTevent
			ResetEvent(pInfo->m_hStartFFT);
			// first wait for new StartFFT event so that this 
			// loop doesn't use up all the clock cycles just spinning
			pFFT->Lock();
			pInfo->ThreadStatus=0;	// thread idling
			pFFT->Unlock();
			WaitForSingleObject(pInfo->m_hStartFFT, INFINITE);	//*** wait spot
			ResetEvent(pInfo->m_hStartFFT);
			continue;
		}

		// start of FFT
		if(pFFT->UpdateProgress(1) == -1) break; // fft in stage 1 (setup, truncate, zerofill, apodize)
		CurrFFT = pInfo->CurrentFFT;

		pPS = CurrFFT->pPS;
		pFFT->pDoc = CurrFFT->pDoc;
		s = CurrFFT->DSIndex;

		// *** get current spectrum focus ***
			// get fftsize so that nspdRealFFT can calculate
			// its twiddle tables and such
		j = pPS->FFTSize;
		for(i=0; j>1; i++) j=j/2; // i = order = base 2 log(PS.FFTSize)

		// *** initialize & Zerofill ***
		// now initialize twiddle tables and such (if fft size has changed
		if(i!=order) {	//i.e. if the size of the FFT has changed
			order = i;
			nspsRealFftNip(tarray, FTData, order, NSP_Init);

			// create time domain data array
			if(tarray != 0) {
				delete [] tarray;
				tarray = 0;
			}
			tarray = new float [pPS->FFTSize];

			// create complex ft array
			if(FTData != 0) {
				delete [] FTData;
				FTData = 0;
			}
			FTData = new SCplx [pPS->FFTSize/2 + 1];
		}
		nspsbZero(tarray, pPS->FFTSize);

		// *** truncate ***
		// copy data into tarray. and truncate simultaneously
		pFFT->Lock();
		pFFT->pDoc->GetTimeArray(CurrFFT->DSIndex);
		if(!pPS->TruncateTransient) {
			pPS->TruncatePoints = pPS->TransientPoints;
			pPS->TruncateOffset = 0;
		}
		for(i=pPS->TruncateOffset; i<pPS->TruncatePoints+pPS->TruncateOffset && i < pPS->TransientPoints; i++) {
			tarray[i-pPS->TruncateOffset]=(float)pFFT->pDoc->TimeArray[i];
		}
		pFFT->Unlock();

		// *** filter ***
		// filter data in time domain
		// tested this idea.  it doesn't work.  It simply gives a rolling
		// baseline that smooths out high frequency peaks (and noise) but 
		// does nothing at the lower frequencies - pete
		if(pFFT->pDoc->BoxcarAvg) {
			pFFT->Lock();
			pts = pFFT->pDoc->BoxcarPoints;
			for(j=0; j<pPS->TruncatePoints; j++) {
				avg=0.0;
				for(i=j-pts/2; i<j+pts/2; i++) {
					if(i<0 || i>=pPS->TruncatePoints) continue;
					avg+=tarray[i]/pts;
				}
				tarray[j]=avg;
			}
			pFFT->Unlock();
		}

		// *** apodize ***
		pFFT->Apodize(pPS->WindowFunction,pPS->TruncatePoints,tarray);

		// *** fft *** 
		if(pFFT->UpdateProgress(2) == -1) break; // fft in stage 2 (fft)
		// the actual fft function call
		nspsRealFftNip(tarray, FTData, order, NSP_Forw);


		// *** magnitude mode ***
		if(pFFT->UpdateProgress(3) == -1) break;	// fft in stage 3 (magnitude mode)

		// data copied back to doc 
		// note send in TransientPoints here for scaling not FFTSize so that amplitudes
		// remain roughly constant if you change the number of zerofills
		pFFT->Magnitude(pPS->TransientPoints,pPS->FFTSize,FTData,float(pPS->Spec[s].ScaleFactor));
		pFFT->pDoc->SetIsFFTed(s,TRUE);

		// *** peakpick and charge state determination*** 
		if(pFFT->UpdateProgress(4) == -1) break;	// FFT in stage 4 (GetMITable, cleanup)

		pFFT->GetMITable(CurrFFT->HighMassPt, CurrFFT->LowMassPt, CurrFFT->pDoc->FreqArray);
		pFFT->pDoc->SetIsMIed(s,TRUE);
		// end of FFT
		
		// notify fftdone and fftprogress
		if(pFFT->UpdateProgress(5) == -1) break;
		::PostMessage(CurrFFT->m_hwndNotifyFFTDone,WM_FFTDONE,0,0);
		SetEvent(CurrFFT->m_hFFTDone);
		//::PostMessage(pInfo->m_hwndNotifyFFTProgress, WM_FFTPROGRESS,0,0);
		pInfo->pMainWnd->PostMessage(WM_FFTPROGRESS,0,0);

		// if not end of queue, update CurrentFFT = NextFFT and
		// delete current FFTQueueEntry (CurrentFFT)
		pFFT->Lock();
		if(pInfo->FFTsInQueue>1) {
			pInfo->CurrentFFT = CurrFFT->NextFFT;
		}
		delete CurrFFT;

		// decrement FFTsInQueue
		pInfo->FFTsInQueue--;
		pFFT->Unlock();
		//FTData = 0;
	}	// end FFTThread's infinite loop

	if(FTData != 0) {
		delete [] FTData;
		FTData = 0;
	}
	if(tarray != 0) {
		delete [] tarray;
		tarray = 0;
	}

	pFFT->Unlock();
	SetEvent(pInfo->m_hThreadDead);

	TRACE("FFTThreadProc: exiting normally\n");
	return 0; // function exited normally
}	// end FFTThreadProc();

void CFFTWorkerThread::Apodize(long WindowFunc, long n, float* FTData)
{
	long i=0;

	float factor;

	switch (WindowFunc) {
	case 0:
		// no apod (or rectangular window function)
		break;
	case 1:	// hann window
		nspsWinHann(FTData,n);
		break; // end case 1
	case 2:
		// partial Hanning apod
		// Hanning apod as above but just on first and last
		// 25% of spectrum
		for(i=0;i<n/4;i++) {
			factor=(float(1.0)-float(cos(pi*i/n)))/2;
			FTData[i]=FTData[i]*factor;
			FTData[n-i]=FTData[n-i]*factor;
		} // end for
		break;
	case 3: //Bartlet window
		nspsWinBartlett(FTData,n);
		break;
	case 4: //Blackman window
		nspsWinBlackmanStd(FTData,n);
		break;
	case 5: // Hamming Window
		nspsWinHamming(FTData,n);
		break;
	} // end switch


} // end Apodize()

void CFFTWorkerThread::Magnitude(long TransientPoints, long FFTSize, SCplx* FTData, float ScaleFactor)
{
	float *temp;
	float scale;

	pFFT->Lock();
	temp = pDoc->FreqArray;	

	scale=float(0.5/TransientPoints);

	nspcbMag(FTData, temp, FFTSize/2);	// magnitude mode calculation
	nspsbMpy1(scale, temp, FFTSize/2);	// 4/n scaling

	ScaleFactor=100/ScaleFactor;  // Scalefactor = 100/max point
	pFFT->Unlock();
}

void CFFTWorkerThread::GetMITable(long HighMassPt, long LowMassPt, float* FTData)
{
	// now create and calculate the MI Table
	// note: TotalIonInt is undefined at this point, but it
	// must have a value to get the mass range in EstimateNoiseLevel
	// so define it = 0.
	CChargeStateDetermination CSDet;

#ifdef _DEBUG
	CBudaApp* pApp = ((CBudaApp*)AfxGetApp());
	pApp->DebugCMassCounter = 0;
#endif
	TotalIonInt = 0;

	// get the noise level for the peak pick threshold
	NoiseLevel=0;
	EstimateNoiseLevel(HighMassPt, LowMassPt, FTData);

	FITable = new struct FITableElement[MAXNUMPEAKS];
	NumPeaks = 0;
#ifdef FREQ_PEAKPICKER
	PeakPick(HighMassPt, LowMassPt, FTData);
#endif
#ifdef ABUNDANCE_PEAKPICKER
	NewPeakPick(HighMassPt, LowMassPt, FTData);
#endif
	// and set the ion intensity used in calibration
	GetTotalIonInt();
	CopyFITable();

	// determine charge state
	if(pDoc->sys->AutoLabelCharge) {
		CSDet.pDoc = pDoc;
		CSDet.CSDetermination();
		pDoc->IsCSed = TRUE;
	}
	if(FITable != 0) {
		delete [] FITable;
		FITable = 0;
	}
}

void CFFTWorkerThread::GetTotalIonInt() 
{
	// this adds up the intensity of all peaks for which the
	// program could calculate a resolution - i.e. all big peaks
	// that have more than 3 points above the baseline
	long i;

	float Tint = 0;
	for(i=0; i<NumPeaks; i++) {
		if(FITable[i].RP>0) Tint += float(FITable[i].I);
	}
	TotalIonInt = Tint;
}

void CFFTWorkerThread::EstimateNoiseLevel(long HighMassPt, long LowMassPt, float* data )
{
	switch(pDoc->ThresholdMethod) {
	case ITERATIVE_AVERAGING_METHOD:
		CalculateNoiseThresholdbyIterativeAveraging(HighMassPt, LowMassPt, data);
		break;
	case HISTOGRAM_METHOD:
		CalculateNoiseThresholdbyHistogram(HighMassPt, LowMassPt, data);
		break;
	}
}

void CFFTWorkerThread::PeakPick(long HighMassPt, long LowMassPt, float* data)
{
	long i;		
	long minpt,maxpt,temp;
	float noise;

	int ptr = 0;			// peakstack pointer
	double PeakStack[100];		// intensity values of peaks
	double PeakXStack[100];		// frequency values of peaks

	noise = float(NoiseLevel);
	minpt=HighMassPt;
	maxpt=LowMassPt;
	if(minpt>maxpt) {
		temp = minpt;
		minpt = maxpt;
		maxpt = temp;
	}
	NumPeaks = 0;
	// peak requirements:
	//		must have 2 or more points above noise
	//		must have "smooth" surface where smooth is
	//			defined as:
	//				before summit, no negative deviations 
	//				in intensity greater than noise
	//				after summit, no positive deviations 
	//				in intensity greater than noise

	// i = curpoint
	for (i=minpt;i<maxpt;i++) {
		if(data[i]<noise) continue;
		ptr = 0;

		// KeepThisPoint()
		if(ptr<98) {
			PeakStack[ptr]=data[i-1];
			PeakXStack[ptr]=PtToFreq(i-1);
			ptr++;
		}

		// record points until slope change
		for(;i<maxpt-1; i++) {
			if(ptr<98) {
				PeakStack[ptr]=data[i];
				PeakXStack[ptr]=PtToFreq(i);
				ptr++;
			}
			if(data[i] > data[i-1]) continue;

			if(data[i] < noise) break;
			if((data[i-1]-data[i])<(3*noise)) break;
			if(data[i+1] < data[i]) break;
		}

		// we are here because the data indicated a negative slope
		for(i++; i<maxpt-1; i++) {
			// KeepThisPoint()
			if(ptr<98) {
				PeakStack[ptr]=data[i];
				PeakXStack[ptr]=PtToFreq(i);
				ptr++;
			}
			if(data[i]<noise) break;

			if((data[i]-data[i-1])>(3*noise)) {
				i--;
				break;
			}

			if((data[i] > data[i-1]) && (data[i+1] > data[i])){
				i--;
				break;
			}
		}

		PeakFit(PeakStack,PeakXStack,ptr);
		if(NumPeaks >= 10000) {
			pDoc->MessageBox("Peak Pick reached the maximum number of peaks\nbefore completing.  Some peaks may not have been found","Warning! Too many peaks");
			break;
		}
	} //end for loop
	IsMIed = TRUE;
//	DumpToDiagFile("leaving CBudaDoc::PeakPick()");

} // end PeakPick()

double CFFTWorkerThread::PtToFreq(long pt)
{
	return pDoc->PtToFreq(pt);
}

void CFFTWorkerThread::PeakFit(double *PeakStack, double *PeakXStack, int numpts)
// PeakStack is the array of intensity values
// PeakXStack is the corresponding array of mass values
// numpts is the number of points in these arrays
{
	int i;
	//, PeakFitPts;
	double maxy = 0,maxx = 0,xptmax=0;
	double a[3], *x, *y;
	double deltafreq;
	//double sumxy,sumy;

	BOOL PeakWasFit=FALSE;

	a[0]=a[1]=a[2]=0.0;  
		// initialize parabolic fit constants
		// we want to do the fit with x=the index value and afterwards
		// calulate the matching frequency by interpolation as this 
		// decreases the roundoff errors in the least squares
		//
		// also, we want to set PeakWasFit = FALSE if one of the following
		// conditions is true:
		//		1) if <3 points above noise threshold
		//		2) if <3 points above the half height of the peak
	// first get new y and x arrays that only include the good points
	// good points are those above half height of peak and those above
	// threshold.s

	x=new double [numpts];
	y=new double [numpts];
	maxy = 0;
	maxx = 0;
	for(i=0; i<numpts; i++) {
		if (PeakStack[i]>maxy) 	{
			maxy=double(PeakStack[i]);
			maxx=double(PeakXStack[i]);
		}
	}
	for(i=0; i<numpts; i++) x[i]=i;

#ifdef DEBUG_PEAKFIT
		CString str,crlf = "\r\n";
		TRACE("==========================="+crlf);
		str.Format("New Peak Found : peak number %d",NumPeaks);
		TRACE(str+crlf);
		str.Format("numpts = %d",numpts);
		TRACE(str+crlf);
		TRACE("---peak data---"+crlf);
		for(i=0; i<numpts; i++) {
			str.Format("PeakStack[%d] = %1.8f : PeakXStack[%d] = %1.8f",i,PeakStack[i],i,PeakXStack[i]);
			TRACE(str+crlf);
		}
#endif

	PeakWasFit=ParabolicLeastSquares(PeakStack, numpts, a, maxy, pDoc->sys->PeakPickerLevel/100);
	if (PeakWasFit) {
		// calculate the center pt and intensity and verify that
		//		1) x[center pt] is in x[0]-x[numpt-1] range
		//		2) intensity is not > 2* maxintensity 
		maxx=(-1*a[1])/(2*a[0]);
		maxy=a[0]*maxx*maxx + a[1]*maxx + a[2];
		deltafreq = fabs(pDoc->PS.AdcRate/double(pDoc->PS.FFTSize));
		xptmax = maxx;
		if(pDoc->PS.DetectMode == MixerMode) maxx=PeakXStack[0]-deltafreq*maxx;
		else maxx=PeakXStack[0]+deltafreq*maxx;
#ifdef DEBUG_PEAKFIT
		str.Format("--Peak Was Fit--");
		TRACE(str+crlf);
		str.Format("maxx in points = %1.5f",xptmax);
		TRACE(str+crlf);
		str.Format("deltafreq = %1.8f",deltafreq);
		TRACE(str+crlf);
		str.Format("fitting constant[0] = %1.8f",a[0]);
		TRACE(str+crlf);
		str.Format("fitting constant[1] = %1.8f",a[1]);
		TRACE(str+crlf);
		str.Format("fitting constant[2] = %1.8f",a[2]);
		TRACE(str+crlf);
#endif
		// if maxx outside range of input points, not a good fit
		// remember that the points are plotted backwards for heterodyne 
		if (pDoc->PS.DetectMode == DirectMode) {
			if(maxx>PeakXStack[numpts-1] || maxx<PeakXStack[0]) PeakWasFit = FALSE;
		} else {
			if(maxx<PeakXStack[numpts-1] || maxx>PeakXStack[0]) PeakWasFit = FALSE;
		}
	} 
	delete [] x;
	delete [] y;

	/*if(!PeakWasFit) {
		// if we couldn't fit a parabola to the peaks, we will do 
		// a center of mass calculation to define the peak
		//
		// x is the intensity weighted average of the x's
		// y is the max intensity
		sumxy=0.0;
		sumy=0.0;
		maxy=0;
		for (i=0; i<numpts; i++) {
			sumxy=sumxy+PeakStack[i]*PeakXStack[i];
			sumy=sumy+PeakStack[i];
			maxy = (maxy<PeakStack[i]) ? PeakStack[i] : maxy;
		}
		maxx=sumxy/sumy;
		PeakWasFit=TRUE;
	}*/
	

	float A = GetPeakArea(PeakStack,PeakXStack,numpts);
	struct FITableElement* FITE = &(FITable[NumPeaks]); 

	if (NumPeaks<MAXNUMPEAKS) {
		FITE->F=maxx;
		FITE->I=maxy;
		FITE->A = A;
		if(PeakWasFit) FITE->RP = InterpolateForRes(PeakStack,PeakXStack,numpts,xptmax);
		//if(PeakWasFit) FITE->RP = float(CalculateRes(a[0],a[1],a[2],double(maxy),double(maxx),deltafreq));
		else FITE->RP = -1;
		FITE->CS = 1;
		NumPeaks++;
#ifdef DEBUG_PEAKFIT
		str.Format("maxx in freq = %1.5f",maxx);
		TRACE(str+crlf);
		str.Format("maxy = %1.5f",maxy);
		TRACE(str+crlf);
		str.Format("area = %1.5f",A);
		TRACE(str+crlf);
		str.Format("RP = %1.5f",FITE->RP);
		TRACE(str+crlf);
#endif		
	}

}//end PeakFit

BOOL CFFTWorkerThread::ParabolicLeastSquares(double *y, int numpts, double *ret, double PeakMaxAmplitude, double PeakPickerLevel)
{  // this function performs a parabolic least squares fit on the x,y arrays each of length numpts
   // that are input.  ret[0], ret[1], ret[3] are the three fit constants

	if (numpts<3) {
		return FALSE; // can't parabolic least squares with less than 3 peaks
	}
	if (numpts == 3) {
		if((y[0]>=y[1]) || (y[2]>=y[1])) return false;
	}

	int i;
	double sumx4, sumx3, sumx2, sumx1, sumx0; // Sum(x^4), Sum(x^3), ...
	double sumy2, sumy1, sumy0;				  // Sum(y*x^2), Sum(y*x), Sum(y)
	double a, b, c;		// parabola coefficients
	double dp;  // denominator in the determinants.
	double f,f2;
	int NumUsablePts;
	double MinFitAmplitude = PeakMaxAmplitude * PeakPickerLevel;
	sumx4=sumx3=sumx2=sumx1=sumx0=0.0;
	sumy2=sumy1=sumy0=0;
	NumUsablePts = 0;

	while(TRUE) {
		sumx4=sumx3=sumx2=sumx1=sumx0=0.0;
		sumy2=sumy1=sumy0=0;
		NumUsablePts = 0;

		for (i=0; i<numpts; i++) {
			if(y[i] > MinFitAmplitude) {
				NumUsablePts++;
				f = i;
				f2 = f*f;

				sumx0++;
				sumx1=sumx1+f;
				sumx2=sumx2+f2;
				sumx3=sumx3+f*f2;
				sumx4=sumx4+f2*f2;

				sumy0=sumy0+y[i];
				sumy1=sumy1+y[i]*f;
				sumy2=sumy2+y[i]*f2;
			}
		}
		if(NumUsablePts >= 3) break;

		PeakPickerLevel -= 0.1;
		if(PeakPickerLevel < 0.33) return FALSE;
		MinFitAmplitude = PeakMaxAmplitude * PeakPickerLevel;		
	}

	// now calculate a,b,c coefficients for the parabola
	dp = GetDeterminant(sumx4, sumx3, sumx2,
						sumx3, sumx2, sumx1,
						sumx2, sumx1, sumx0);
	if (dp == 0) return FALSE;

	a  = GetDeterminant(sumy2, sumx3, sumx2,
						sumy1, sumx2, sumx1,
						sumy0, sumx1, sumx0)/dp;
	b  = GetDeterminant(sumx4, sumy2, sumx2,
						sumx3, sumy1, sumx1,
						sumx2, sumy0, sumx0)/dp;
	c  = GetDeterminant(sumx4, sumx3, sumy2,
						sumx3, sumx2, sumy1,
						sumx2, sumx1, sumy0)/dp;

	// verify that a is negative so that found a peak not a valley 
	if (a>0 || b<0) {
		return FALSE;
	}

	// if we get this far, the peak is good, so return it
	ret[0]=a;
	ret[1]=b;
	ret[2]=c;

	return TRUE;
}

double CFFTWorkerThread::GetDeterminant(double a1, double a2, double a3, double b1, double b2, double b3, double c1, double c2, double c3)
{
	return a1*b2*c3+a2*b3*c1+a3*b1*c2-a1*b3*c2-a2*b1*c3-a3*b2*c1;
}

float CFFTWorkerThread::InterpolateForRes(double *y, double *x, int ptr, double maxx)
{
	double LeftWidth = -1;
	double RightWidth = -1;
	double PeakAmplitude,hh, PeakResolution;
	double PeakFrequency,df;
	int i,CenterIndex;

	PeakAmplitude = y[0];
	CenterIndex = 0;
	for (i=0; i<ptr; i++) {
		if (y[i] > PeakAmplitude) {
			PeakAmplitude =  y[i];
			CenterIndex=i;
		}
	}

	PeakFrequency = double(x[CenterIndex]);
	df = fabs(pDoc->PS.AdcRate/double(pDoc->PS.FFTSize));
	
	hh = PeakAmplitude/2.0;

	for(i = int(maxx+0.5); i>0; i--) {
		if(y[i] < hh) {
			LeftWidth = maxx - ((hh-y[i])/(y[i+1]-y[i])) - i;
			break;
		}
	}

	for(i = int(maxx+0.5); i<ptr; i++) {
		if(y[i] < hh) {
			RightWidth = i-((hh-y[i])/(y[i-1]-y[i])) - maxx;
			break;
		}
	}

	PeakResolution = -1;
	if(LeftWidth <= 0) LeftWidth = RightWidth;
	if(RightWidth <= 0) RightWidth = LeftWidth;

	if(LeftWidth > 0) PeakResolution = PeakFrequency/((LeftWidth+RightWidth)*df);
	return float(PeakResolution);

/*	// find maximum point, then find two peaks closest to the FWHM points
	// on a model peak, interpolate to find the approximate x values of these
	// points and then calculate f/delta(f) and return it.
	int i,maxx=0;
	double maxy,b,m,retval;
	double fwhmx1, fwhmx2;
	int pt1,pt2,count;

	// find maximum point and count number of points above NoiseLevel
	maxy=0;
	count=0;
	for (i=0; i<ptr; i++) {
		if (y[i] > maxy) {
			maxy =  y[i];
			maxx=i;
		}
		if (y[i]>NoiseLevel) count++;
	}
	//if (count<5) return -1;

	// now find left side two points straddling the half height
	pt1=pt2=0;
	for (i=0; i<ptr; i++) {
		pt1=i;
		pt2=i+1;
		if (y[pt1]>maxy/2) break;
	}
	if (pt2>ptr-1) {
		pt2=ptr-1;
		pt1=pt2-1;
	}
	// for y=mx+b, m=y2-y1/x2-x1, b=y2-mx2
	m=(y[pt2]-y[pt1])/(x[pt2]-x[pt1]);
	b=y[pt2]-m*x[pt2];
	// now, x=(y-b)/m;
	fwhmx1=(maxy/2-b)/m;
	// now find right side two points straddling the half height
	pt1=pt2=0;
	for (i=maxx; i<ptr ; i++) {
		pt1=i;
		pt2=i+1;
		if (y[pt2]<maxy/2) break;
	}
	if (pt2>ptr-1) {
		pt2=ptr-1;
		pt1=pt2-1;
	}
	// for y=mx+b, m=y2-y1/x2-x1, b=y2-mx2
	m=(y[pt2]-y[pt1])/(x[pt2]-x[pt1]);
	b=y[pt2]-m*x[pt2];
	// now, x=(y-b)/m;
	fwhmx2=(maxy/2-b)/m;

	retval= x[maxx]/(fwhmx2-fwhmx1);
	if(pDoc->PS.DetectMode == MixerMode) retval = retval*-1;

	// now round resolving power to nearest 100
	if(retval<0) {
		return float(-1);
	}

	b = retval/100;
	m = double(int(b));
	if((b-m)>0.5) m+=1;;
	if(m==0) {
		return float(-1);
	}
	retval = m*100;

	return float(retval);
	*/
}


BOOL CFFTWorkerThread::Lock()
{
	int i=0;
	while(DataLock->IsLocked()) {
		i++;
	}
	return DataLock->Lock(INFINITE);
}

BOOL CFFTWorkerThread::Unlock()
{
	return DataLock->Unlock();
}

int CFFTWorkerThread::UpdateProgress(int level)
{
		pFFT->Lock();
		if( WaitForSingleObject(pFFT->FFTThrdInfo.m_hKillThread, 0) == WAIT_OBJECT_0 ) {
			// if here, then m_hKillThread was signalled, so we must break out of
			// the infinite loop and exit/terminate the thread
			ResetEvent(pFFT->FFTThrdInfo.m_hKillThread);
			return -1;
		}

		pFFT->FFTThrdInfo.ThreadStatus = level; // fft in stage 1 (setup, truncate, zerofill, apodize)
		pFFT->FFTThrdInfo.pMainWnd->PostMessage(WM_FFTPROGRESS,0,0);
		pFFT->Unlock();
		if( WaitForSingleObject(pFFT->FFTThrdInfo.m_hStopFFT, 0) == WAIT_OBJECT_0 ) {
			// if here, then m_hStopFFT was signalled, so we must 
			// delete the FFT queue, set queue counter to zero, and
			// cycle the loop to dump thread back into idle mode

			ResetEvent(pFFT->FFTThrdInfo.m_hStopFFT);
			pFFT->Lock();
				// delete all currently used objects
			pFFT->Unlock();

			return 1;
		}
		return 0;
}

void CFFTWorkerThread::CopyFITable()
{
	pFFT->Lock();
	int s = FFTThrdInfo.CurrentFFT->DSIndex;
	struct Spectrum *Spec = &(pDoc->PS.Spec[s]);
	long i;
	struct FITableElement* pFI = Spec->FITable;

	Spec->TotalIonInt = TotalIonInt;
	pDoc->SetIsMIed(s,TRUE);
	Spec->NoiseLevel = NoiseLevel;
	Spec->NumPeaks = NumPeaks;
	Spec->ScaleFactor = ScaleFactor;

	for (i=0; i<NumPeaks; i++) {
		pFI[i]=FITable[i];
	}
	pDoc->FITable = pFI;
	pFFT->Unlock();
}


float CFFTWorkerThread::GetPeakArea(double *Ypts, double *Xpts, int numpts)
{
	// build in a numerical integration of the peak here.  
	// use trapezoid rule
	//
	// I(1..n) += df*(x[i]+x[i-1])/2
	//
	int i;
	double sum = 0;
	double df = Xpts[1]-Xpts[0];
	for (i=1; i<numpts; i++) {
		sum += df*(Ypts[i]+Ypts[i-1])/2;
	}
	return float(fabs(sum));
}


void CFFTWorkerThread::NewPeakPick(long highpt, long lowpt, float *FTdata)
{
	// this function finds peaks and submits them to peakfit for addition to 
	// the MITable.
	//
	// This algorithm is different than in peakpick.  The pseudocode follows
	//
	//	duplicate data array
	//	FindLargestPoint
	//  track both right and left from this point to get full peak
	//		stop tracking if get a significant slope change or reach noiselevel
	// submit to peakfit
	// find next largest point

	float* data,max;
	long temp,numpt,i,maxi,ptr;
	double PeakStack[100];		// intensity values of peaks
	double PeakXStack[100];		// frequency values of peaks
	double last;

	if(highpt==lowpt) return;
	if(highpt<lowpt){
		temp = highpt;
		highpt = lowpt;
		lowpt = temp;
	}
	numpt = highpt-lowpt;

		
	data = new float[numpt];
	nspsbCopy(&(FTdata[lowpt]),data,numpt);
	nspsbThresh1(data,numpt,float(NoiseLevel),NSP_LT);
	maxi = lowpt;
	while(1) {
		max = float(NoiseLevel);
		for(i=0; i<numpt; i++) {
			if(data[i]>max) {
				maxi=i;
				max=data[i];
			}
		}

		if(NumPeaks>MAXNUMPEAKS) break;
		if(max == float(NoiseLevel)) break;
		ptr=0;
		PeakStack[ptr] = max;
		PeakXStack[ptr] = PtToFreq(maxi+lowpt);
		ptr++;
		data[maxi]=0;

		// go right first
		last = max;
		for(i=maxi+1; ;i++) {
			if(data[i] <= NoiseLevel) break;
			if(ptr>99) break;
			if(i>numpt) break;
			// slope wrong and significant
			if((last - data[i]) < (-1*NoiseLevel)) break;
			last = double(data[i]);
			PeakStack[ptr] = last;
			PeakXStack[ptr] = PtToFreq(i+lowpt);
			ptr++;
			data[i]=0;
		}
		
		// now go left
		last = max;
		for(i=maxi-1; ;i--) {
			if(data[i] <= NoiseLevel) break;
			if(ptr>99) break;
			if(i<0) break;
			// slope wrong and significant
			if((last - data[i]) < (-1*NoiseLevel)) break;
			last = double(data[i]);
			PeakStack[ptr] = last;
			PeakXStack[ptr] = PtToFreq(i+lowpt);
			ptr++;
			last = double(data[i]);
			data[i]=0;
		}

		// now fit the peak to a parabola and add it to the mitable
		PeakFit(PeakStack, PeakXStack, ptr);
	}
	delete [] data;
}


/*
void CFFTWorkerThread::FitPeak(double *PeakY, double *PeakF, numpts)
{
	long j;
	long PeakMaxIndex;
	long NumUsablePts;
	float PeakMaxAmplitude;
	double df;
	float PeakResolution;
	float PeakAmplitude;
	double PeakFrequency;
	BOOL PeakWasFit;

	PeakResolution = -1.0;
	PeakMaxAmplitude = PeakY[0];
	PeakMaxIndex = 0;
	for(j=1; j<numpts; j++) {
		if(PeakY[j]>PeakMaxAmplitude) {
			PeakMaxAmplitude = PeakY[j];
			PeakMaxIndex = j;
		}
	}
	NumUsablePts = numpts;

	// if only 3 pts, they must be triangular else ignore
	if(NumUsablePts = 3) {
		if((PeakY[0] >= PeakY[1]) || (PeakY[3] >= PeakY[1])) {
			return;
		}
	}

}*/

double CFFTWorkerThread::CalculateRes(double a, double b, double c, double maxy, double maxx, double df)
{
	// calculates a peak width from the half-height points of the parabola defined by
	// maxy/2 = a*x*x + b*x + c 
	
	// using the quadratic formula the peak width (x+ - x-) derives to 
	//	sqrt(b*b - 4*a*(c-maxy/2))/a

	double width = fabs(sqrt(b*b - 4*a*(c - maxy/2.0))/a);
	width = width * df;

	return maxx/width;
}

void CFFTWorkerThread::CalculateNoiseThresholdbyIterativeAveraging(long HighMassPt, long LowMassPt, float* data )
{
	long i,j;
	float sum=0.0;
	long highpt,lowpt,numpts;
	float scale=0, noise=0;

	lowpt=HighMassPt;
	highpt=LowMassPt;
	if (highpt<lowpt) { // swap them
		numpts = highpt; // use numpts as a temp variable
		highpt = lowpt;
		lowpt = numpts;
	}

	if ((highpt-lowpt)<6) {
		noise=0.0;
		return;
	}
	for (i=lowpt; i<highpt; i++) {
		sum+=data[i];
		if(data[i]>scale) scale = data[i]; // get highest point for scale
	}

	// if no data points, very simple
	if (scale == 0) {
		NoiseLevel = 0;
		ScaleFactor = 0;
		pDoc->ScaleFactor = 0;
		return;
	}
	ScaleFactor = 100.0/scale;

	// if setting the noise threshold manually, very simple
	if(pDoc->pApp->AutoDetectThreshold == 0) {
		noise = float(pDoc->pApp->MinimumAutoThreshold/ScaleFactor);
		NoiseLevel = noise;
		pDoc->ScaleFactor=ScaleFactor;
		return;
	}
		
	numpts=highpt-lowpt;

	//noise=10*sum/numpts; // first estimate is 10* average
	noise = scale;

	for (j=0; j<pDoc->sys->PeakPickerLoopCount; j++) {
		sum=0.0;
		for (i=lowpt; i<highpt; i++) {
			if(data[i]<noise) {
				sum+=data[i];
				//sum+=data[i]*data[i];
			} else { 
				sum+=noise;
				//sum+=noise*noise;
				//sum+=noise*noise*float(NOISE_LEVEL_OVERCORRECTION_FACTOR);
			}
		}
		noise=float(pDoc->sys->PeakPickerMeanFactor*sum/(numpts+1)); // include all pts < 2*rms
		//noise=float(pDoc->sys->PeakPickerMeanFactor*sqrt(sum/numpts)); // include all pts < 2*rms
	}
	if(noise>scale) noise = scale;
	if(noise <  pDoc->pApp->MinimumAutoThreshold/ScaleFactor) noise = float(pDoc->sys->MinimumAutoThreshold/ScaleFactor);
	NoiseLevel = noise;
	pDoc->ScaleFactor = ScaleFactor;

	// ok, store this calculation in the correct data section of 
	// the PS structure.  It belongs in PS.Spec[s].---
}

void CFFTWorkerThread::CalculateNoiseThresholdbyHistogram(long HighMassPt, long LowMassPt, float* data )
{
// please note that this function has not yet been built.  
}
