/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#if !defined(AFX_INSTRUMENT_H__0CDD3C21_06AD_11D2_9106_44455354616F__INCLUDED_)
#define AFX_INSTRUMENT_H__0CDD3C21_06AD_11D2_9106_44455354616F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000
// Instrument.h : header file
//

#include "stdafx.h"
#include "MainFrm.h"
#include "BudaMaster.h"
#include <sys\timeb.h>


// declare classes that will be needed so that I
// can use interlinked includes
class CBudaApp;
//class CBudaDoc;
class CMainFrame;

//---------------------------------------------------------------
//	Instrument Thread Structures
//---------------------------------------------------------------
struct InstrumentThreadInfo
{
	int ThreadStatus;	// used for thread to communicate with BudaApp
						// -2 = TD error
						// -1 = Instrument Unavailable
						// 0 = idle
						// 1 = Kill this thread
						// 2 = acquire data
						// 3 = run diagnostics
						// 4 = reset
						// 5 = running experiment
						// 6 = abort experiment
	char InstrumentStatus;

	int InstrumentCommands;	// used to tell the thread what functions to go into
							// commands defined in compiler directives below

	// define FFT control events
	HANDLE m_hStartThread;	// event used to start the thread from its initial stopped state
	HANDLE m_hKillThread;	// event used to kill Instrument thread from ExitInstance
	HANDLE m_hProcessCommand;		// event used to process a command
	HANDLE m_hAbortExperiment;		// event used to immediately abort the experiment
	HANDLE m_hStopAcquisition;		// event used to stop the experiment at the end of the current run
	HANDLE m_hThreadDead;	// event used by thread to notify BudaApp that it is dead
	HANDLE m_hDiagDataUpdated;	// event used by thread to notify BudaApp that Diagnostic data has changed

	HWND m_hwndNotifyDiagDataUpdated;	// hwnds for postmessages - will be set to appropriate windows
	HWND m_hwndNotifyRunPSDone;	// hwnds for postmessages - will be set to appropriate windows

	CWnd* pMainWnd;
};

//Instrument Commands
#define	IDLE				0
#define	KILL_THREAD			1
#define	RUN_EXPERIMENT		2
#define ABORT_EXPERIMENT	3
#define RUN_DIAGNOSTICS		4
#define RESET_INSTRUMENT	5
#define	UPDATE_STATUS		6
#define CHANGE_SYSTEM		7

// Thread Status constants
#define THREAD_IDLE					0
#define THREAD_RUNNING_EXPERIMENT	1
#define THREAD_ABORT_EXPERIMENT		2
#define	THREAD_KILLTHREAD			4
#define THREAD_INSTRUMENT_RESET		5

// Instrument status bitmasks
#define INSTRUMENT_DATA_READY	0x0004
#define INSTRUMENT_IDLE			0x0001

// define triplet structure
struct triplet {
	long timeon;
	long timeoff;
	long bitpattern;
};

/////////////////////////////////////////////////////////////////////////////
// CInstrument document

class CInstrument : public CDocument
{
public:
	CInstrument();           // protected constructor used by dynamic creation
	DECLARE_DYNCREATE(CInstrument)

// Attributes
public:
	struct sys sys;
	struct PS PS;

	#ifdef DEBUG_DUMP_DIAGS_TO_FILE
		CFile* diagfile;
	#endif

// Operations
public:
	CString CommaReturn(CString);
	int FindComma(CString);
	CString CommaParse(CString);
	CString GetRightString(CString);
	void InitializeSys();
	CString CompressString(CString);
	int FindEqualSign(CString);
	BOOL ReadConfigurationFile();

	#ifdef ENCODE_SERIAL_NUMBER
		void SetSerialNumber();
	#endif

	int ScrambleFactor;
	CString SerialNumber;
	void DumpToDiagFile(CString);
	CString GetSerialNumber();
	BOOL CheckSerialNumber(CString);
	BOOL ReadCalibrationFile();
	void MessageBox(CString, CString);
	CMainFrame* pWnd;
	BOOL Setup();
	struct sys* GetSys();
	CString	EncryptedSerialNumber;
	void OnSetupPulsesequence();


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CInstrument)
	public:
	virtual void Serialize(CArchive& ar);   // overridden for document i/o
	protected:
	virtual BOOL OnNewDocument();
	//}}AFX_VIRTUAL

// Implementation
public:
	void GetEventEdges(long, float*, float*);
	int ShiftLinks(int, int);
	float EventTime(int);
	void SortGenPS();
	virtual ~CInstrument();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

	// Generated message map functions
public:
	void bcd32(long* BinWord, long* BcdWord);
	void ProgramFSPfrequency(long* Address, double f, BOOL ReleaseZeroPhase);
	void ProgramFSPdelaytrigger(long* Address);
	void ProgramFSPend(long* Address);
	void ProgramFSPzerophase(long* Address);
	void ProgramFSPword(long* Address, double f, BOOL ZeroPhase, BOOL FreqValid, BOOL EndSeq, BOOL Trigger);
	struct FlagPacket flags;
	float GainDacVoltage(float Vout, float Vin, float FixedGain, float Divisor);
	void PrepDac(long DacNumber, long Units, float vMax, long codeMax, float vMin, long codeMin, float dispMax, float dispMin, float Vinit);
	void ProgramVSPword(long* Address, long Channel, long DACvalue);
	long ValueToDac(float Value, float Scale, float Offset);
	void ProgramPSG(long* pulsetable, long* ticktable);
	void ProgramFSP3();
	void ProgramFSP2();
	void ProgramARB();
	void ProgramFSP();
	void ProgramVSP();
	//void ProgramPSG(long* PulseTable, long* TickTable);
	//void ProgramFSP2();
	//void ProgramFSP();
	//void ProgramARB();
	//void ProgramVSP();
	void BuildPS();
	void FitCFP(long* pEvent, long* packet, double* StartMass, double* EndMass, double* StartFreq, double* EndFreq, long* IncrFreq, long* NumSteps);
	void InterpretDACString(CString str, int dacnum);
	BOOL acqStop;
	long ADCRateCode(double rate);
	void StartTimer();
	double scanduration;
	clock_t now;
	clock_t start;
	long* TimeArray;
	long acqScansInTD;
	long acqScansInMemory;
	long acqScansCompleted;
	long acqGroupSize;
	long RateCode;
	long TriggerCode;
	BOOL acqAbort;
	BOOL acqTRASH;
	long TDAstat;
	long FSPstat;
	long PSGstat;
	long ARBstat;
	long VSPstat;
	void SetUserID(CString ID);
	void SetActiveSystem(int index);
	int ActiveSystem;
	BOOL InterpretConfigFileString(CString str);
	CString SourceName;
//	int NumSources;
	void ReadTD();
	BOOL Running;
	void UpdateStatus();
	void AbortExperiment();
	int SetBudaPath(CString pathname);
	float QuadRFNeeded(double Mass);
	float FlightTime(double Mass, double QuadOffset);
	int CurrentInstrument;
	CBudaDoc* pDoc;
	void CopyPS(struct PS* from, struct PS* to);
	BOOL TuneMode;
	BOOL RunPSGDiagnostic();
	BOOL RunVSPDiagnostic();
	BOOL RunFSPDiagnostic();
	BOOL RunARBDiagnostic();
	BOOL RunTDDiagnostic();
	BOOL PSG_Tested;
	BOOL VSP_Tested;
	BOOL FSP_Tested;
	BOOL ARB_Tested;
	BOOL TD_Tested;
	BOOL PSG_Passed;
	BOOL VSP_Passed;
	BOOL FSP_Passed;
	BOOL ARB_Passed;
	BOOL TD_Passed;
	BOOL DiagnosticsComplete;

	BOOL Unlock();
	BOOL Lock();
	BOOL InstrumentAvailable;
	void ResetInstrument();
	void RunExperiment();
	CWinThread* InstrumentThread;
	CSingleLock* DataLock;
	CCriticalSection* Access;
	CBudaApp* pApp;
	void InitializeInstrument();
	struct InstrumentThreadInfo ThreadInfo;
	void KillInstrumentThread();
	void StartInstrumentThread();
	void CopySys(/* int from, int to */);
//	CString ConfigFileName;
	CFile ConfigFile;
	BOOL ReadCNFFileESI();
	BOOL ReadCNFFileMALDI();
	CString RemoveSpaces(CString);
	int GetV6ChannelNumber(int);
	void ReadVersion8SetupFile(CFile*);
	void ReadVersion6SetupFile(CFile*);
	BOOL StartupDiagnostics();
	BOOL ReadDefaultSetupFile();
	void InitializePS();
	//{{AFX_MSG(CInstrument)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_INSTRUMENT_H__0CDD3C21_06AD_11D2_9106_44455354616F__INCLUDED_)
