// SetupFFT.cpp : implementation file
//
/*
file description: SetupFFT is the code for the CSetupFFT
class which is derived from CDialog and merely provides a GUI for 
changing the FFT options such as number of zero fills, type of apod,
whether or not to peakpick, and etc.

*/
/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "BudaMaster.h"
#include "BudaDoc.h"
#include "SetupFFT.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSetupFFT property page

IMPLEMENT_DYNCREATE(CSetupFFT, CPropertyPage)

CSetupFFT::CSetupFFT() : CPropertyPage(CSetupFFT::IDD)
{
	//{{AFX_DATA_INIT(CSetupFFT)
	m_apodfunc = -1;
	m_IsPeakPicking = FALSE;
	m_NumZeroFills = 0;
	m_TruncatePoints = 0;
	m_TruncateOffset = 0;
	m_FFTSize = _T("");
	m_NDP = _T("");
	m_SpecPts = _T("");
	//}}AFX_DATA_INIT
}

CSetupFFT::~CSetupFFT()
{
}

void CSetupFFT::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CSetupFFT)
	DDX_Control(pDX, IDC_NUMZEROFILLS, m_NZFEdit);
	DDX_Control(pDX, IDC_TRUNCATE, m_TruncBtn);
	DDX_Control(pDX, IDC_TRUNCATE_POINTS, m_TruncPtsBox);
	DDX_Control(pDX, IDC_TRUNCATE_OFFSET, m_TruncOffBox);
	DDX_Control(pDX, IDC_APODCOMBO, m_ApodComboBox);
	DDX_CBIndex(pDX, IDC_APODCOMBO, m_apodfunc);
	DDX_Check(pDX, IDC_PEAKPICK, m_IsPeakPicking);
	DDX_Check(pDX, IDC_AUTOCALCCS, m_AutoCalcCS);
	DDX_Text(pDX, IDC_NUMZEROFILLS, m_NumZeroFills);
	DDX_Text(pDX, IDC_TRUNCATE_POINTS, m_TruncatePoints);
	DDX_Text(pDX, IDC_TRUNCATE_OFFSET, m_TruncateOffset);
	DDX_Text(pDX, IDC_FFTOPT_FFTSIZE, m_FFTSize);
	DDX_Text(pDX, IDC_FFTOPT_NDP, m_NDP);
	DDX_Text(pDX, IDC_FFTOPT_SPECPTS, m_SpecPts);
	//}}AFX_DATA_MAP
	ValidateTruncatePoints();
	ValidateTruncateOffset();
}


BEGIN_MESSAGE_MAP(CSetupFFT, CPropertyPage)
	//{{AFX_MSG_MAP(CSetupFFT)
	ON_EN_UPDATE(IDC_NUMZEROFILLS, OnUpdateNumzerofills)
	ON_CBN_SELCHANGE(IDC_APODCOMBO, OnSelchangeApodcombo)
	ON_BN_CLICKED(IDC_PEAKPICK, OnPeakpick)
	ON_BN_CLICKED(IDC_AUTOCALCCS, OnAutoCalcCS)
	ON_BN_CLICKED(IDC_TRUNCATE, OnTruncate)
	ON_EN_UPDATE(IDC_TRUNCATE_OFFSET, OnUpdateTruncateOffset)
	ON_EN_UPDATE(IDC_TRUNCATE_POINTS, OnUpdateTruncatePoints)
	ON_EN_CHANGE(IDC_NUMZEROFILLS, OnChangeNumzerofills)
	ON_EN_KILLFOCUS(IDC_NUMZEROFILLS, OnKillfocusNumzerofills)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSetupFFT message handlers

BOOL CSetupFFT::OnKillActive() 
{
	m_apodfunc=m_ApodComboBox.GetCurSel();
	return CPropertyPage::OnKillActive();
}

BOOL CSetupFFT::OnSetActive() 
{
	m_ApodComboBox.ResetContent();
	m_ApodComboBox.InsertString(0,"None");
	m_ApodComboBox.InsertString(1,"Full Hanning");
	m_ApodComboBox.InsertString(2,"Half Hanning");
	m_ApodComboBox.InsertString(3,"Bartlett");
	m_ApodComboBox.InsertString(4,"Blackman");
	m_ApodComboBox.InsertString(5,"Hamming");
	m_ApodComboBox.SetCurSel(m_apodfunc);
	
	OnUpdate();
	return CPropertyPage::OnSetActive();
}

void CSetupFFT::OnOK() 
{
	if(!CheckBigFFTs()) m_NumZeroFills = 0;
	CPropertyPage::OnOK();
}

BOOL CSetupFFT::OnApply() 
{
	CheckBigFFTs();
	DataDirty = TRUE;
	
	return CPropertyPage::OnApply();
}

void CSetupFFT::OnUpdateNumzerofills() 
{
	CString str;
	int oldnzf = m_NumZeroFills;
	m_NZFEdit.GetWindowText(str);
	if(sscanf(str,"%d",&m_NumZeroFills) <= 0) m_NumZeroFills = 1;
	SetModified(TRUE);
	if(!CheckBigFFTs()) m_NumZeroFills = oldnzf;
	OnUpdate();
}

void CSetupFFT::OnSelchangeApodcombo() 
{
	m_apodfunc = m_ApodComboBox.GetCurSel();
	SetModified(TRUE);	
	OnUpdate();
}

void CSetupFFT::OnPeakpick() 
{
	SetModified(TRUE);
	OnUpdate();
}

void CSetupFFT::OnAutoCalcCS() 
{
	SetModified(TRUE);
	OnUpdate();
}

void CSetupFFT::OnTruncate() 
{
	SetModified(TRUE);
	if(m_TruncBtn.GetCheck()) {
		TruncateTransient=1;
		m_TruncPtsBox.EnableWindow(TRUE);
		m_TruncOffBox.EnableWindow(TRUE);
	} else {
		TruncateTransient=0;
		m_TruncPtsBox.EnableWindow(FALSE);
		m_TruncOffBox.EnableWindow(FALSE);
	}
}

void CSetupFFT::OnUpdateTruncateOffset() 
{
	SetModified(TRUE);
}

void CSetupFFT::OnUpdateTruncatePoints() 
{
	SetModified(TRUE);
}

void CSetupFFT::ValidateTruncatePoints()
{
	CString str;

	if (m_TruncatePoints < 0 || m_TruncatePoints > long(PDoc->PS.TransientPoints)) {
		str.Format("Please enter a value between 0 and %d",(PDoc->PS.TransientPoints));
		MessageBox(str, "Invalid Entry", MB_OK);
		m_TruncatePoints = (long)PDoc->PS.TransientPoints/2;
	}
}

BOOL CSetupFFT::OnQueryCancel() 
{
	DataDirty = FALSE;	
	return CPropertyPage::OnQueryCancel();
}

void CSetupFFT::ValidateTruncateOffset()
{	
	CString str;

	if (m_TruncateOffset < 0 || m_TruncateOffset > long(PDoc->PS.TransientPoints - m_TruncatePoints)) {
		str.Format("Please enter a value between 0 and %d",(PDoc->PS.TransientPoints - m_TruncatePoints));
		MessageBox(str, "Invalid Entry", MB_OK);
		m_TruncateOffset = 0;
	}
}

void CSetupFFT::OnUpdate()
{
	// activate/deactivate the truncation info
	if(TruncateTransient == 0) {
		m_TruncBtn.SetCheck(0);
		m_TruncPtsBox.EnableWindow(FALSE);
		m_TruncOffBox.EnableWindow(FALSE);
	} else {
		m_TruncBtn.SetCheck(1);
		m_TruncPtsBox.EnableWindow(TRUE);
		m_TruncOffBox.EnableWindow(TRUE);
	}

	
	// set up the three data point strings.
	m_NDP.Format("%ld",PDoc->PS.TransientPoints);
	if(m_NDP.GetLength() > 7) m_NDP.Format("%ld Mpts",PDoc->PS.TransientPoints/1048576);
	if(m_NDP.GetLength() > 7) m_NDP.Format("%ld Mp",PDoc->PS.TransientPoints/1048576);
	if(m_NDP.GetLength() > 7) m_NDP.Format("******");

	// find fft size.
	long j=PDoc->PS.TransientPoints;
	if(TruncateTransient) j = m_TruncatePoints;
	int i;
	for(i=0; i<m_NumZeroFills; i++) {
		j*=2;
		if (j>=1073741824) {
			j = 1073741824;
			m_NumZeroFills = i;
			break;
		}
		if(j<=0) {
			m_NumZeroFills = 0;
			j = PDoc->PS.TransientPoints;
			if(TruncateTransient) j = m_TruncatePoints;
			break;
		}
	}
	m_FFTSize.Format("%ld",j);
	if(m_FFTSize.GetLength() > 7) m_FFTSize.Format("%ld Mpts",j/1048576);
	if(m_FFTSize.GetLength() > 7) m_FFTSize.Format("%ld Mp",j/1048576);
	if(m_FFTSize.GetLength() > 7) m_FFTSize.Format("******");

	m_SpecPts.Format("%ld",j/2);
	if(m_SpecPts.GetLength() > 7) m_SpecPts.Format("%ld Mpts",j/2097152);
	if(m_SpecPts.GetLength() > 7) m_SpecPts.Format("%ld Mp",j/2097152);
	if(m_SpecPts.GetLength() > 7) m_SpecPts.Format("******");

	UpdateData(FALSE);
}

BOOL CSetupFFT::CheckBigFFTs()
{
	// find fft size.
	long j=PDoc->PS.TransientPoints;
	if(TruncateTransient) j = m_TruncatePoints;
	int i;
	for(i=0; i<m_NumZeroFills; i++) j*=2;

	CString text;
	if(j>=4194304) {
		text.Format("This is a %d Megapoint FFT\nIt will be slow.\n\nDo you really want to do this?",j/1048576);
		if(MessageBox(text,"FFTSize Warning.",MB_YESNO) == IDYES) return TRUE;
		else {
			OnUpdate();
			return FALSE;
		}
	}
	return TRUE;
}

void CSetupFFT::OnChangeNumzerofills() 
{
//	OnUpdate();	
}

void CSetupFFT::OnKillfocusNumzerofills() 
{
//	OnUpdate();	
}
