// SetupView.cpp : implementation file
//
/****************************************************************

Boston University data analysis (BUDA) package 

Description: Buda is a data analysis program for Fourier 
Transform Mass Spectrometry data.  

by Peter B. O'Connor

    Copyright (C) 2000  Peter B. O'Connor

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "stdafx.h"
#include "Buda.h"
#include "BudaMaster.h"
#include "SetupView.h"
#include <math.h>
#include "DetectDlg.h"
#include "SetupSYNDlg.h"
#include "SetupARBDlg.h"
#include "SetupQuadDlg.h"
#include "SetupImpDlg.h"
#include "DACSetupDlg.h"
#include "SetupTTLDlg.h"
#include "DACChannelList.h"
#include "TTLChannelList.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSetupView

IMPLEMENT_DYNCREATE(CSetupView, CView)

CSetupView::CSetupView()
{
	pInstrument = &(((CBudaApp*)AfxGetApp())->Instrument);
	pSys = &(pInstrument->sys);

	PlotChArea = new CRect[40];
	ChannelIndices = new int[40];

	UpdateMenu();
}

CSetupView::~CSetupView()
{
	pDoc->SetupViewOpen = FALSE;
	if(PlotChArea != 0) {
		delete [] PlotChArea;
		PlotChArea = 0;
	}
	if(ChannelIndices != 0) {
		delete [] ChannelIndices;
		ChannelIndices = 0;
	}
}


BEGIN_MESSAGE_MAP(CSetupView, CView)
	//{{AFX_MSG_MAP(CSetupView)
	ON_WM_LBUTTONDBLCLK()
	ON_COMMAND(IDC_ADDDAC, OnAdddac)
	ON_COMMAND(IDC_ADDTTL, OnAddttl)
	ON_COMMAND(ID_EDITQUADRF, OnEditQuadRF)
	ON_COMMAND(ID_EDITEXCITATION, OnEditExcite)
	ON_COMMAND(ID_EDITARB, OnEditArb)
	ON_COMMAND(ID_EDITIMPULSE, OnEditImpulse)
	ON_COMMAND(ID_SETUPDETECT, OnEditDetect)
	ON_COMMAND(ID_EDITDAC, OnEditDAC)
	ON_COMMAND(ID_EDITTTL, OnEditTTL)
	
	ON_WM_RBUTTONDOWN()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSetupView drawing

void CSetupView::OnDraw(CDC* pDeviceContext)
{
	pDC = pDeviceContext;
	pDoc = (CBudaDoc*)GetDocument();
	pDoc->SetupViewOpen = TRUE;
	pDoc->pSetupView = this;

	// Define drawing limits and constants
	AxisLineYpos = 4500;
	PlotArea.top = 32768-2500;
	PlotArea.right = 32768 - 1000;
	PlotArea.bottom = 4500;
	PlotArea.left = 7500;

	pDC->GetWindow()->GetClientRect(&ViewSize);

	// place page title
	CString str = "Pulse and DAC Sequences";
	CPoint pt;
	pt.x = 16384;
	pt.y = 32700;
	Text(str,pt,CENTER_JUSTIFIED);

	/*	// test the ranging by drawing an x in the screen
	MoveTo(0,0);			// lower right
	LineTo(16384,16384);	// center
	LineTo(32768,0);		// upper left
	LineTo(16384,16384);	// center
	LineTo(0,32768);		// lower right
	LineTo(16384,16384);	// center
	LineTo(32768,32768);	// upper right
	*/

	DrawTimeAxis();
	DrawChannels();
	LabelChannels();
}

/////////////////////////////////////////////////////////////////////////////
// CSetupView diagnostics

#ifdef _DEBUG
void CSetupView::AssertValid() const
{
	CView::AssertValid();
}

void CSetupView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CSetupView message handlers

CPoint CSetupView::PTtoViewPT(CPoint pt)
{
	// this function maps the window specified by ViewSize
	// into a logical window from lower-left = 0,0 to 
	// upper right = 32767,32767

	// the input variable pt is assumed to be in the 0-32767
	// window, while the output is in the viewsize window.

	if (pt.x>32767) pt.x = 32767;
	if (pt.y>32767) pt.y = 32767;
	if (pt.x<0) pt.x = 0;
	if (pt.y<0) pt.y = 0;

	CPoint outpt;
	double x,y,mx,my,by;
	my = ((double)ViewSize.bottom-(double)ViewSize.top)/32768.0;
	mx = ((double)ViewSize.right-(double)ViewSize.left)/32768.0;
	by = (double)ViewSize.bottom-(double)ViewSize.top;
	y = by - my*(double)pt.y;
	x = mx*(double)pt.x;

	outpt.y = (short)y;
	outpt.x = (short)x;
	return outpt;
}


CPoint CSetupView::PTtoViewPT(int x, int y)
{
	CPoint pt;
	pt.x=x;
	pt.y=y;
	return PTtoViewPT(pt);
}	

void CSetupView::DrawTimeAxis()
{
	// this function draws a time axis out to the last event
	// in the pulse sequence
	int AxisLineXstart = PlotArea.left;
	int AxisLineXend = PlotArea.right;

	pGenPS = pDoc->PS.GenPS;
	long NumEvents = pDoc->PS.NumEvents;

	// first find the first and last events in the sequence
	long i, j, FirstEvent=0, LastEvent=0;

	// initialize the plot channel info
	NumPlotChannels=0;
	for(i=0; i<40; i++) PlotChannels[i]=FALSE;

	float start=0, finish=0, FirstEventStart=0, LastEventEnd=0;
	BOOL StartStopUndefined = TRUE;

	for(i=0; i<NumEvents; i++) {
		//	First, skip all the events that are in channels
		// that will not be plotted
		if(pGenPS[i].Channel == EXCITE_DISABLE) continue;
		if(pGenPS[i].Channel == EXCITE_SOURCE) continue;
		if(pGenPS[i].Channel == RF_SOURCE) continue;
		if(pGenPS[i].Channel == ULTIMA_SWITCH) continue;
		if(pGenPS[i].Channel == RF_PHASE) continue;
		//if(pGenPS[i].Channel == IMP_TRIGGER) continue;
		//if(pGenPS[i].Channel == SYN_TRIGGER) continue;
		//if(pGenPS[i].Channel == ARB_TRIGGER) continue;
		//if(pGenPS[i].Channel == QUADRF_GATE) continue;
		if(pGenPS[i].Channel == VSP_TRIGGER) continue;
		if(pGenPS[i].Channel == DAC_GAIN) continue;

		if(pGenPS[i].Type != DAC_INIT)
			// for DACS, don't plot channels that do nothing but init.
			PlotChannels[pGenPS[i].Channel] = TRUE;
		if(pGenPS[i].Channel == QUADRF_GATE) {
			// for quads, set FALSE unless at least one cfp event is type CFP_QUAD
			PlotChannels[QUADRF_GATE] = FALSE;
			for(j=0; j<pDoc->PS.NumCFP; j++) {
				if(pDoc->PS.cfp[j].Type == CFP_QUAD) 
					PlotChannels[QUADRF_GATE] = TRUE;
			}
		}
		if(pGenPS[i].Channel == SYN_TRIGGER) {
			// not true if only quadrf events
			PlotChannels[SYN_TRIGGER] = FALSE;
			for(j=0; j<pDoc->PS.NumCFP; j++) {
				if(pDoc->PS.cfp[j].Type != CFP_QUAD && pDoc->PS.cfp[j].Type != CFP_NONE) 
					PlotChannels[SYN_TRIGGER] = TRUE;
			}
		}

		// get first event time and last event time, ignoring DAC_INITs
		pDoc->GetEventEdges(i,&start,&finish);
		if(StartStopUndefined) {
			StartStopUndefined = FALSE;
			FirstEventStart = start;
			FirstEvent = i;
			LastEventEnd = finish;
			LastEvent = i;
		}
		if(start<FirstEventStart && pDoc->PS.GenPS[i].Type != DAC_INIT) {
			FirstEventStart = start;
			FirstEvent = i;
		}
		if(finish>LastEventEnd && pDoc->PS.GenPS[i].Type != DAC_INIT) {
			LastEventEnd = finish;
			LastEvent = i;
		}
	}
	// check to see if the detection is actually the last thing
	// in time, if so, correct LastEventEnd for this
	pDoc->GetEventEdges(pDoc->PS.DetectEvent,&start,&finish);
	float DetectEnd = finish + (float)(1000*(double)pDoc->PS.TransientPoints/pDoc->PS.AdcRate); 
	if(LastEventEnd<DetectEnd) LastEventEnd = DetectEnd;

	// count up the channels that will be plotted
	NumPlotChannels=0;
	for(i=0;i<40; i++) {
		if(PlotChannels[i]) NumPlotChannels++;
	}

	TimeStart = (double) FirstEventStart;
	TimeEnd = (double) LastEventEnd;
	double timebase = (double)LastEventEnd - (double)FirstEventStart;

	// second draw the axis line 
	MoveTo(AxisLineXstart,AxisLineYpos);
	LineTo(AxisLineXend,AxisLineYpos);

	// place tics
	// there will be <=10 tics
	int ticpos[10];
	double tictime[10];
	for(i=0; i<10; i++) {
		ticpos[i] = 0;
		tictime[i] = 0;
	}

	// there will be 5-10 tics  
	// ticsteps can be (order of magnitude of timebase) / 1, 2, 5 
	double a = log10(timebase);
	double b = floor(a);	// b = order of magnitude of timebase
	double ticstep = pow(10,b);
	double nt;	
	int numtics;
	nt = timebase/ ticstep;

	if(nt>=5) numtics = (int)nt;
	if(nt>2 && nt<5) {
		ticstep=pow(10,b)/2;
		numtics = (int)(timebase/ticstep);
	}
	if(nt<=2) {
		ticstep = pow(10,b)/5;
		numtics = (int)(timebase/ticstep);
	}

	// find the position of the first tic.
	// if zero is in the range, I want zero to be one of the ticx
	double NumTixb4Zero = (double)FirstEventStart/ticstep;
	if(NumTixb4Zero>0) {// GenPS starts after zero
		tictime[0]=(floor(NumTixb4Zero)+1) * ticstep;
	}
	if (NumTixb4Zero == 0) tictime[0] = 0;
	if (NumTixb4Zero < 0) { // GenPS starts before zero
		tictime[0] = -1 * floor(-1*NumTixb4Zero) * ticstep;
	}

	// find the position of each tic relative to tictime[0]
	for(i=1; i<10; i++) {
		tictime[i]=ticstep*i+tictime[0];
		if(tictime[i]>LastEventEnd) break;
	}
	numtics = i-1;

	// find tic x positions
	// remember that position 1000 = FirstEventStart
	// and 31768 = LastEventEnd;
				// create a linear eqn y = a*x + b
				// where y's are the plotspace #'s and 
				// x's are the timebase space #'s
	a = (AxisLineXend-AxisLineXstart)/(LastEventEnd - FirstEventStart);
	b = AxisLineXstart - a*FirstEventStart;
	for(i=0; i<numtics; i++) {
		ticpos[i] = int(a*tictime[i] + b);
	}

	// now actually plot the tics
	for(i=0; i<numtics; i++) {
		MoveTo(ticpos[i],AxisLineYpos);
		LineTo(ticpos[i],AxisLineYpos-500);
	}

	// lable tics
	CString str;
	CPoint pt;
	CSize sz;

	str.Format("Milliseconds");
	pt.x = 16384;
	pt.y = 2000;
	Text(str,pt,CENTER_JUSTIFIED);

	for(i=0; i<numtics; i++) {
		str.Format("%1.1f",tictime[i]);
		pt.x = ticpos[i];
		pt.y = 4000;
		Text(str,pt,CENTER_JUSTIFIED);
	}

	// end of draw axis
}

void CSetupView::DrawChannels()
{
	// this function plots a line for each channel that has 
	// events in the pulse sequence
	
	// first get plot areas on 0-32k grid
	int PlotHeight;
	PlotHeight = (PlotArea.top-PlotArea.bottom)/(NumPlotChannels+1); 
	CRect ChannelArea = PlotArea, box;
	//PlotChArea = new CRect[40];
	//ChannelIndices = new int[40];

	long i,plotnum=0;

	// then plot each channel in its region
	// first draw the quadrupole rf pulse(s)
	int channel=0;
	if(PlotChannels[QUADRF_GATE]) {
		ChannelArea.top = PlotArea.top;
		ChannelArea.bottom = ChannelArea.top-PlotHeight;
		ChannelArea.right = PlotArea.right;
		ChannelArea.left = PlotArea.left;
		PlotChArea[0]=ChannelArea;
		ChannelIndices[0]=QUADRF_GATE;
		DrawChannel(QUADRF_GATE,ChannelArea);
		channel++;
	}

	// then draw the synthesizer pulses
	if(PlotChannels[SYN_TRIGGER]) {
		ChannelArea.top = PlotArea.top-channel*PlotHeight;
		ChannelArea.bottom = ChannelArea.top-PlotHeight;
		ChannelArea.right = PlotArea.right;
		ChannelArea.left = PlotArea.left;
		PlotChArea[channel]=ChannelArea;
		ChannelIndices[channel]=SYN_TRIGGER;
		DrawChannel(SYN_TRIGGER,ChannelArea);
		channel++;
	}
	// then draw the Arbitrary waveform generator pulses
	if(PlotChannels[ARB_TRIGGER]) {
		ChannelArea.top = PlotArea.top-channel*PlotHeight;
		ChannelArea.bottom = ChannelArea.top-PlotHeight;
		ChannelArea.right = PlotArea.right;
		ChannelArea.left = PlotArea.left;
		PlotChArea[channel]=ChannelArea;
		ChannelIndices[channel]=ARB_TRIGGER;
		DrawChannel(ARB_TRIGGER,ChannelArea);
		channel++;
	}

	// then draw the Impulse pulses
	if(PlotChannels[IMP_TRIGGER]) {
		ChannelArea.top = PlotArea.top-channel*PlotHeight;
		ChannelArea.bottom = ChannelArea.top-PlotHeight;
		ChannelArea.right = PlotArea.right;
		ChannelArea.left = PlotArea.left;
		PlotChArea[channel]=ChannelArea;
		ChannelIndices[channel]=IMP_TRIGGER;
		DrawChannel(IMP_TRIGGER,ChannelArea);
		channel++;
	}

		// then draw the stylized transient
	ChannelArea.top = PlotArea.top-channel*PlotHeight;
	ChannelArea.bottom = ChannelArea.top-PlotHeight;
	ChannelArea.right = PlotArea.right;
	ChannelArea.left = PlotArea.left;
	PlotChArea[channel]=ChannelArea;
	ChannelIndices[channel]=0;
	PlotTransient(ChannelArea);
	channel++;
	
	// finally draw all other channels (except the reserved
	// internal channels) note: this part plots them counting up
	// from the bottom of the page.
	plotnum=0;
	for(i = 0; i<40; i++) {
		// if not plotting this channel, cycle
		if(!PlotChannels[i]) continue;
		if(i==SYN_TRIGGER) continue;
		if(i==IMP_TRIGGER) continue;
		if(i==ARB_TRIGGER) continue;
		if(i==QUADRF_GATE) continue;
		// plot this channel
		ChannelArea.bottom = plotnum*PlotHeight+PlotArea.bottom;
		ChannelArea.top = ChannelArea.bottom + PlotHeight;
		ChannelArea.right = PlotArea.right;
		ChannelArea.left = PlotArea.left;
		PlotChArea[channel]=ChannelArea;
		ChannelIndices[channel]=i;
		channel++;
		DrawChannel(i,ChannelArea);
		plotnum++;
	}

	// finally store the corrected ChannelAreas in PlotChArea
	CRect a;
	CPoint p;
	NumPlotChannels++; // for signal
	for(i=0; i<NumPlotChannels; i++) {
		p.y=PlotChArea[i].top;
		p.x=ViewSize.left;
		p=PTtoViewPT(p);
		PlotChArea[i].top=p.y;
		PlotChArea[i].left=ViewSize.left;
		p.y=PlotChArea[i].bottom;
		p=PTtoViewPT(p);
		PlotChArea[i].bottom=p.y;
		PlotChArea[i].right=ViewSize.right;
	}
}

void CSetupView::LabelChannels()
{
	// this function labels the channels
	int i;
	CPoint pt;
	CString str;
	CSize sz;

	for(i=0; i<NumPlotChannels; i++) {
		pt.y = (PlotChArea[i].bottom+PlotChArea[i].top)/2;
		pt.x = 0;
		if(ChannelIndices[i] == QUADRF_GATE) str.Format("Quadrupole RF");
		else if(ChannelIndices[i] == SYN_TRIGGER) str.Format("Synthesizer");
		else if(ChannelIndices[i] == 0) str.Format("Ion Signal");
		else if(ChannelIndices[i] == IMP_TRIGGER) str.Format("Impulse Excite");
		else if(ChannelIndices[i] == ARB_TRIGGER) str.Format("ARB Excite");
		else str = pDoc->sys->ChannelNames[ChannelIndices[i]];
		str = str.Left(14);

		sz = pDC->GetTextExtent(str);

		pt.y = pt.y - sz.cy/2;
		pDC->TextOut(pt.x,pt.y,str);	
	}
}

// the 4 moveto and lineto functions should have
// inputs on the 0-32768 range
void CSetupView::MoveTo(int x, int y)
{
	pDC->MoveTo(PTtoViewPT(x,y));
}

void CSetupView::MoveTo(CPoint pt)
{
	pDC->MoveTo(PTtoViewPT(pt));
}

void CSetupView::LineTo(int x, int y)
{
	pDC->LineTo(PTtoViewPT(x,y));
}

void CSetupView::LineTo(CPoint pt)
{
	pDC->LineTo(PTtoViewPT(pt));
}

void CSetupView::Text(CString str, CPoint pt, int just)
{
	CSize sz = pDC->GetTextExtent(str);

	CPoint pt1,pt2;
	pt1 = PTtoViewPT(pt);
	if(just == CENTER_JUSTIFIED) pt1.x = pt1.x-sz.cx/2;
	if(just == LEFT_JUSTIFIED) pt1.y = pt1.y - sz.cy/2;
	if(just == CENTER_JUSTIFIED2) {
		pt1.x = pt1.x-sz.cx/2;
		pt1.y = pt1.y - sz.cy/2;
	}

	pDC->TextOut(pt1.x,pt1.y,str);
}

int CSetupView::TimeToPos(double time)
{
	int left = PlotArea.left;
	int right = PlotArea.right;

	// time = da*(time - t0)/dt + a0
	double temp;
	temp = (right-left)*(time - TimeStart)/(TimeEnd-TimeStart) + left;
	return (int)temp;
}

void CSetupView::DrawChannel(int Channel, CRect ChannelArea)
{
	// draw an individual channel, Channel, within rect ChannelArea

	BOOL Quad = FALSE;
	if (Channel == QUADRF_GATE) {
		Channel = SYN_TRIGGER;
		Quad = TRUE;
	}

	// find all the events for this channel
	int* events = new int[pDoc->PS.NumEvents];
	long i,j;
	int NumChannelEvents=0;
		// first for non-excite events
	if(Channel!=SYN_TRIGGER) {
		for(i=0; i<pDoc->PS.NumEvents; i++) {
			if(pGenPS[i].Channel==Channel) {
				events[NumChannelEvents] = i;
				NumChannelEvents++;
			}
		} 
	}

	// now for excite events
	int cfppointer=-1;
	if(Channel== SYN_TRIGGER) {
		for(i=0; i<pDoc->PS.NumEvents; i++) {
			if(pGenPS[i].Channel!=SYN_TRIGGER) continue;
			cfppointer++;
			if(cfppointer>pDoc->PS.NumCFP) {
				// error, more syn events than cfp's
			}
			// all synthesizer events are either quad or not quad for plotting purposes
			// so we need to store the events for these sets separately so we don't double count
			if(Quad && pDoc->PS.cfp[cfppointer].Type == CFP_QUAD) {
				events[NumChannelEvents] = i;
				NumChannelEvents++;
			}
			if(!Quad && pDoc->PS.cfp[cfppointer].Type != CFP_QUAD) {
				events[NumChannelEvents] = i;
				NumChannelEvents++;
			}
		}
	}
	
	// EventPts array will be on a vertical scale of 0 to 100
	// and will be converted to ChannelArea just before plotting
	CPoint* EventPts=0;
	int NumEventPts=0;
	float start,end,dt;

	// calculate the EventPts for QUADRF_GATE : the quadrupole rf voltage
	if(Channel == QUADRF_GATE) {
		EventPts = new CPoint[200*NumChannelEvents+2];
		float MaxQuadRF=0;
		// first find the maximum quadrf voltage
		for(i=0; i<pDoc->PS.NumCFP; i++) {
			if(pDoc->PS.cfp[i].Type == CFP_QUAD) {
				if(pDoc->PS.cfp[i].Amplitude > MaxQuadRF) 
					MaxQuadRF=pDoc->PS.cfp[i].Amplitude;
			}
		}

		// then make an EventPts Array for all the QuadRF_gate'ed ones
		for(i=0; i<NumChannelEvents; i++) {
			if(i==0) {
				EventPts[NumEventPts].x = TimeToPos(TimeStart);
				EventPts[NumEventPts].y = 50;
				NumEventPts++;
			}
			pDoc->GetEventEdges(events[i],&start,&end);

			EventPts[NumEventPts].x = TimeToPos(start);
			EventPts[NumEventPts].y = 50;
			NumEventPts++;
			EventPts[NumEventPts].x = TimeToPos(start);
			EventPts[NumEventPts].y = 100;
			NumEventPts++;
			dt = (end-start)/100;
			for(j=0; j<100; j++) {
				if(j % 2) { // go low on odd numbers
					EventPts[NumEventPts].x = TimeToPos(j*dt+start);
					EventPts[NumEventPts].y = 0;
					NumEventPts++;
				} else {	// go high on even numbers
					EventPts[NumEventPts].x = TimeToPos(j*dt+start);
					EventPts[NumEventPts].y = 100;
					NumEventPts++;
				}
			}
			EventPts[NumEventPts].x = TimeToPos(end);
			EventPts[NumEventPts].y = 50;
			NumEventPts++;

			if(i==NumChannelEvents-1) {
				EventPts[NumEventPts].x = TimeToPos(TimeEnd);
				EventPts[NumEventPts].y = 50;
				NumEventPts++;
			}
		}
	}

	// calculate the EventPts for Syn_trigger
	if (Channel == SYN_TRIGGER || Channel==ARB_TRIGGER) {
		PlotSynthesizer(events,NumChannelEvents, ChannelArea);
		if(EventPts != 0) {
			delete [] EventPts;
			EventPts = 0;
		}
		if(events != 0) {
			delete [] events;
			events = 0;
		}
		return;
	}

	// calculate the EventPts for TTL's/Pulses
	if(Channel>=0 && Channel < 16 && Channel!=SYN_TRIGGER && Channel!=QUADRF_GATE) { // ttl pulse channel
		EventPts = new CPoint[4*NumChannelEvents+2];
		for(i=0; i<NumChannelEvents; i++) {
			if(pGenPS[events[i]].Type == NO_EVENT) continue;
			if(pGenPS[events[i]].Type != TTL_PULSE) {
				// error! invalid event type for TTL
			}
			// start channel plot low
			if(i==0) {	
				EventPts[NumEventPts].x = TimeToPos(TimeStart);
				EventPts[NumEventPts].y = 0;
				NumEventPts++;
			}
			pDoc->GetEventEdges(events[i],&start,&end);
			// put plot line for channel high
			EventPts[NumEventPts].x = TimeToPos(start);
			EventPts[NumEventPts].y = 0;
			NumEventPts++;
			EventPts[NumEventPts].x = TimeToPos(start);
			EventPts[NumEventPts].y = 100;
			NumEventPts++;
			// bring plot line for channel back low again
			EventPts[NumEventPts].x = TimeToPos(end);
			EventPts[NumEventPts].y = 100;
			NumEventPts++;
			EventPts[NumEventPts].x = TimeToPos(end);
			EventPts[NumEventPts].y = 0;
			NumEventPts++;
			// end channel plot low
			if(i==NumChannelEvents-1) {	
				EventPts[NumEventPts].x = TimeToPos(TimeEnd);
				EventPts[NumEventPts].y = 0;
				NumEventPts++;
			}
		}
	}

	// calculate the EventPts for DAC's
	float MaxValue=0,MinValue=0;
	int PreviousValue=0;

	if(Channel>15 && Channel < 40) { // DAC channel
		EventPts = new CPoint[4*NumChannelEvents+2];
		MaxValue = pDoc->sys->dac[Channel-16].MaxGraph;
		MinValue = pDoc->sys->dac[Channel-16].MinGraph;

		for(i=0; i<NumChannelEvents; i++) {
			if(pGenPS[events[i]].Type == NO_EVENT) continue;
			if(pGenPS[events[i]].Type == TTL_PULSE) {
				// error! invalid event type for DAC
			}
			pDoc->GetEventEdges(events[i],&start,&end);

			switch (pGenPS[events[i]].Type) {
			case DAC_INIT:	// event is DAC-INIT
				EventPts[NumEventPts].x = TimeToPos(TimeStart);
				EventPts[NumEventPts].y = (int)(100*(pGenPS[events[i]].Value-MinValue)/(MaxValue-MinValue));
				PreviousValue=EventPts[NumEventPts].y;
				NumEventPts++;
				break;
			case DAC_STEP:	// event is DAC-STEP
				EventPts[NumEventPts].x = TimeToPos(start);
				EventPts[NumEventPts].y = PreviousValue;
				NumEventPts++;
				EventPts[NumEventPts].x = TimeToPos(start);
				EventPts[NumEventPts].y = (int)(100*(pGenPS[events[i]].Value-MinValue)/(MaxValue-MinValue));
				PreviousValue=EventPts[NumEventPts].y;
				NumEventPts++;
				break;
			case DAC_PULSE:	// event is DAC-PULSE
				EventPts[NumEventPts].x = TimeToPos(start);
				EventPts[NumEventPts].y = PreviousValue;
				NumEventPts++;
				EventPts[NumEventPts].x = TimeToPos(start);
				EventPts[NumEventPts].y = (int)(100*(pGenPS[events[i]].Value-MinValue)/(MaxValue-MinValue));
				NumEventPts++;
				EventPts[NumEventPts].x = TimeToPos(end);
				EventPts[NumEventPts].y = (int)(100*(pGenPS[events[i]].Value-MinValue)/(MaxValue-MinValue));
				NumEventPts++;
				EventPts[NumEventPts].x = TimeToPos(end);
				EventPts[NumEventPts].y = PreviousValue;
				PreviousValue=EventPts[NumEventPts].y;
				NumEventPts++;
				break;
			case DAC_RAMP:	// event is DAC-RAMP
				EventPts[NumEventPts].x = TimeToPos(start);
				EventPts[NumEventPts].y = PreviousValue;
				NumEventPts++;
				EventPts[NumEventPts].x = TimeToPos(end);
				EventPts[NumEventPts].y = (int)(100*(pGenPS[events[i]].Value-MinValue)/(MaxValue-MinValue));
				PreviousValue=EventPts[NumEventPts].y;
				NumEventPts++;
				break;
			}
			if(i==NumChannelEvents-1) {	
				EventPts[NumEventPts].x = TimeToPos(TimeEnd);
				EventPts[NumEventPts].y = PreviousValue;
				NumEventPts++;
			}

		}
	}

	// Plot all the EventPts on ChannelArea
	// plot them only from 10% to 90% of the height of channel area
	// so that they don't run into each other
	double tempy,dx = ChannelArea.top - ChannelArea.bottom;
	for(i=0; i<NumEventPts; i++) {
		EventPts[i].y = (int)(((double)EventPts[i].y*0.8)+10.0);
		tempy = dx*(double)EventPts[i].y/100.0 + (double)ChannelArea.bottom;
		EventPts[i].y = (int)tempy;
		if(i==0) MoveTo(EventPts[i]);
		LineTo(EventPts[i]);
	}

	if(EventPts != 0) {
		delete [] EventPts;
		EventPts = 0;
	}
	if(events != 0) {
		delete [] events;
		events = 0;
	}
}

void CSetupView::Box(CPoint topleft, CPoint bottomright)
{
	// this function draws a 100 cycle oscillating line 
	// that simulates an excite pulse in a box from upperleft to 
	// lower right

	int top = topleft.y;
	int left = topleft.x;
	int bottom = bottomright.y;
	int right = bottomright.x;
	int center = (top+bottom)/2;
	int i,y;

	// then draw the synthesizer pulses
	MoveTo(left,center);
	LineTo(left,top);
	float dt = (float)(right-left)/100;
	for(i=0; i<100; i++) {
		if(i%2) y = top;
		else y= bottom;
		LineTo((int)(i*dt)+left,y);
	}
	LineTo(right,center);
}

void CSetupView::PlotSynthesizer(int *events, int NumEvents, CRect Area)
{
	int i;
	int top = Area.bottom+(int)(0.9*((float)(Area.top-Area.bottom)));
	int bottom = Area.bottom+(int)(0.1*((float)(Area.top-Area.bottom)));
	int right = Area.right;
	int left = Area.left;
	int center = (top+bottom)/2;
	CPoint tl,br;
	float start,end;

	MoveTo(left,center);

	for(i=0; i<NumEvents; i++) {
		pDoc->GetEventEdges(events[i],&start,&end);

		LineTo(TimeToPos(start),center);
		tl.y = top;
		br.y = bottom;
		tl.x = TimeToPos(start);
		br.x = TimeToPos(end);
		Box(tl,br);
		MoveTo(TimeToPos(end),center);
	}
	LineTo(right,center);
}

void CSetupView::PlotTransient(CRect Area)
{
	// plots the stylized transient on the setup screen
	int i,y;
	int top = Area.bottom+(int)(0.9*((float)(Area.top-Area.bottom)));
	int bottom = Area.bottom+(int)(0.1*((float)(Area.top-Area.bottom)));
	int right = Area.right;
	int left = Area.left;
	int center = (top+bottom)/2;
	float start,end,x;
	CPoint pt;

	MoveTo(left,center);

	if(pDoc->PS.DetectEvent < 0 || pDoc->PS.DetectEvent >= pDoc->PS.NumEvents) {
		CString str = "No detection event defined";
		pt.x = (right+left)/2;
		pt.y = center;
		Text(str,pt,CENTER_JUSTIFIED2);
		return;
	}
	// the transient starts at the end of the excite pointed to 
	// by DetectEvent and ends after all the data points have been
	// used up
	pDoc->GetEventEdges(pDoc->PS.DetectEvent,&start,&end);
	start=end;
	end = start + (float)(1000*(double)pDoc->PS.TransientPoints/pDoc->PS.AdcRate);

	LineTo(TimeToPos(start),center);
	LineTo(TimeToPos(start),top);
	LineTo(TimeToPos(start),bottom);
	float dt = (end-start)/200;
	float dy = (float)(top-bottom)/400;
	for(i=0; i<200; i++) {
		x=(i*dt+(float)start);
		if(i%2) y = center+(int)((200-i)*dy);
		else y = center-(int)((200-i)*dy);
		LineTo(TimeToPos(x),y);
	}
		
	MoveTo(TimeToPos(end),center);
	LineTo(right,center);
}

void CSetupView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	CView::OnLButtonDblClk(nFlags, point);

	// first, find out which channel we are pointing to.
	int i,channel=0;
	for(i=0; i<NumPlotChannels; i++) {
		if(point.y > PlotChArea[i].top && point.y < PlotChArea[i].bottom) {
			if (channel!=0) {
				// error, more than one area for this point
				// should never happen.
			}
			channel = i;
		}
	}

	// now, if ttl, call up the ttl setup dialog
	//		if dac, call up the dac setup dialog
	//		if detect, call up the detect dialog
	//		if synthesizer or quadrf, call up the synthesizer dialog
	channel = ChannelIndices[channel];

	if (channel<0 || channel>39) {
		// error invalid channel number
		MessageBox("Error 76 :  Invalid Channel Number","Error!!");
		return;
	}

	if (channel == 0) {	// this is the detection channel number
		// use dialog IDD_SETUP_DETECTION_DLG
		OnEditDetect();
	}

	if (channel == QUADRF_GATE ) { // quad channel
		// use dialog IDD_SETUP_QUADRF_EVENTS
		OnEditQuad();
	}

	if (channel == SYN_TRIGGER ) { // synthesizer channel
		// use dialog IDD_SETUP_SYNTHESIZER_DLG
		OnEditExcite();
	}

	if (channel == ARB_TRIGGER ) { // arb channel
		// use dialog IDD_SETUP_ARB_EVENTS
		OnEditArb();
	}

	if (channel == IMP_TRIGGER ) { // impulse excite channel
		// use dialog IDD_SETUP_IMPULSE_EVENTS
		OnEditImpulse();
	}

	if (channel >0 && channel < 16 && channel != SYN_TRIGGER && channel != QUADRF_GATE && channel != ARB_TRIGGER && channel != IMP_TRIGGER) {
		OnEditTTL(channel);
	}

	if (channel > 15 && channel < 40) { // DAC channel
		OnEditDAC(channel);
	}

	pDoc->UpdateAllViews(NULL,NULL,NULL);
}

void CSetupView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// create a floating popup menu here
	// from menu IDR_SETUP_MENU
	CMenu menu;
	VERIFY(menu.LoadMenu(IDR_SETUP_MENU));
	CMenu* pPopup = menu.GetSubMenu(0);
	ASSERT(pPopup != NULL);
	ClientToScreen(&point);
 	pPopup->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, FromHandle(m_hWnd));

	CView::OnRButtonDown(nFlags, point);
}

void CSetupView::OnAdddac() 
{
	// open a dialog box to select one of the dacs
	// then create a CDACSetupDlg and open it with a new BLANK_EVENT
	// added
	int Channel=0;
	CDACChannelList dlg;
	dlg.InFlags = SELECT_UNUSED;
	dlg.Channel = Channel;
	dlg.pPS = &(pDoc->PS);
	dlg.pDoc = pDoc;
	dlg.pSys = pSys;
	dlg.OutFlag = NORMAL;

	dlg.DoModal();

	Channel = dlg.Channel;
	if (dlg.OutFlag == CANCEL) return;

	//OnEditDAC(Channel);

	// use dialog IDD_SETUP_DAC_DLG
	CDACSetupDlg dlg3;

	dlg3.Channel = Channel;
	dlg3.m_DACName = pDoc->sys->ChannelNames[Channel];
	dlg3.pDoc = pDoc;
	dlg3.pSys = pSys;
	dlg3.pPS = &(pDoc->PS);
	dlg3.StartFlags = ADD_NEW_EVENT;

	dlg3.DoModal();
}

void CSetupView::OnAddttl() 
{
	int Channel=0;
	CTTLChannelList dlg;
	dlg.InFlags = SELECT_UNUSED;
	dlg.Channel = Channel;
	dlg.pPS = &(pDoc->PS);
	dlg.pSys = pSys;
	dlg.pDoc = pDoc;
	dlg.OutFlag = NORMAL;

	dlg.DoModal();

	Channel = dlg.Channel;
	if (dlg.OutFlag == CANCEL) return;

	// use dialog IDD_SETUP_DAC_DLG
	CSetupTTLDlg dlg3;

	dlg3.Channel = Channel;
	dlg3.m_TTLName = pDoc->sys->ChannelNames[Channel];
	dlg3.pDoc = pDoc;
	dlg3.pPS = &(pDoc->PS);
	dlg3.StartFlags = ADD_NEW_EVENT;

	dlg3.DoModal();	
}

void CSetupView::OnEditQuadRF()
{
	OnEditQuad();
}

void CSetupView::OnEditTTL(int channel)
{
	// use dialog IDD_SETUP_TTL_DLG
	CSetupTTLDlg dlg2;

	dlg2.Channel = channel;
	dlg2.m_TTLName = pDoc->sys->ChannelNames[channel];
	dlg2.pDoc = pDoc;
	dlg2.pPS = &(pDoc->PS);

	dlg2.DoModal();
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}

void CSetupView::OnEditDAC(int channel)
{
	// use dialog IDD_SETUP_DAC_DLG
	CDACSetupDlg dlg3;

	dlg3.Channel = channel;
	dlg3.m_DACName = pDoc->sys->ChannelNames[channel];
	dlg3.pDoc = pDoc;
	dlg3.pPS = &(pDoc->PS);
	dlg3.pSys = pSys;
	dlg3.StartFlags = STANDARD_STARTUP;

	dlg3.DoModal();
	pDoc->UpdateAllViews(NULL,NULL,NULL);
}

void CSetupView::OnEditDAC()
{
	// first get channel
	int Channel=0;
	CDACChannelList dlg;
	dlg.InFlags = SELECT_CURRENT;
	dlg.Channel = Channel;
	dlg.pPS = &(pDoc->PS);
	dlg.pDoc = pDoc;
	dlg.pSys = pSys;
	dlg.OutFlag = NORMAL;

	dlg.DoModal();

	Channel = dlg.Channel;
	if (dlg.OutFlag == CANCEL) return;
	if(Channel<16 || Channel>40) return;
	
	// then call the other OnEditDac()
	OnEditDAC(Channel);
}

void CSetupView::OnEditTTL()
{
	// first get channel
	int Channel=0;
	CTTLChannelList dlg;
	dlg.InFlags = SELECT_CURRENT;
	dlg.Channel = Channel;
	dlg.pPS = &(pDoc->PS);
	dlg.pSys = pSys;
	dlg.pDoc = pDoc;
	dlg.OutFlag = NORMAL;

	dlg.DoModal();

	Channel = dlg.Channel;
	if (dlg.OutFlag == CANCEL) return;
	if(Channel<0 || Channel>16) return;

	// then call the other OnEditTTL()
	OnEditTTL(Channel);
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}

void CSetupView::OnEditExcite()
{
	CSetupSYNDlg dlg1;
		
	dlg1.pDoc = pDoc;
	dlg1.PS = pDoc->PS;

	dlg1.DoModal();
	// do all the copy-back inside the dialog's 
	// OnOK() function
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}

void CSetupView::OnEditDetect()
{
	CDetectDlg dlg0;

	if(pDoc->PS.SpectraRequested>1) dlg0.AcqMode = Chromatography;
	else dlg0.AcqMode = Single;

	dlg0.pPS = &(pDoc->PS);
	dlg0.pSys = pSys;
	dlg0.pDoc = pDoc;
	dlg0.DetectMode = pDoc->PS.DetectMode;
	dlg0.m_HighMass = pDoc->PS.DetectHighMass;
	dlg0.m_LowMass = pDoc->PS.DetectLowMass;
	dlg0.m_MonitorMass = pDoc->PS.MonitorMass;
	dlg0.NumDataPoints = pDoc->PS.TransientPoints;
	dlg0.AdcRate = pDoc->PS.AdcRate;
	dlg0.m_NumScans = pDoc->PS.ScansPerSpectrum;
	dlg0.m_NumSpectra = pDoc->PS.SpectraAcquired;

	if (dlg0.DoModal() == IDOK) {
		pDoc->PS.AdcRate = dlg0.AdcRate;
		pDoc->PS.TransientPoints = dlg0.NumDataPoints;
		pDoc->PS.DetectMode = (short)dlg0.DetectMode;
		pDoc->PS.DetectHighMass = dlg0.m_HighMass;
		pDoc->PS.DetectLowMass = dlg0.m_LowMass;
		pDoc->PS.MonitorMass = dlg0.m_MonitorMass;
		pDoc->PS.ScansRequested = dlg0.m_NumScans;
		pDoc->PS.SpectraRequested = dlg0.m_NumSpectra;
	}
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}

void CSetupView::UpdateMenu()
{
	//CMenu* MainMenu = AfxGetMainWnd()->GetMenu();
}

void CSetupView::OnEditQuad()
{
	CSetupQuadDlg dlg1;
		
	dlg1.pDoc = pDoc;
	dlg1.PS = pDoc->PS;

	dlg1.DoModal();
	// do all the copy-back inside the dialog's 
	// OnOK() function
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}

void CSetupView::OnEditArb()
{
	CSetupARBDlg dlg5;
		
	dlg5.pDoc = pDoc;
	dlg5.PS = pDoc->PS;

	dlg5.DoModal();
	// do all the copy-back inside the dialog's 
	// OnOK() function
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}

void CSetupView::OnEditImpulse()
{
	CSetupImpDlg dlg6;
		
	dlg6.pDoc = pDoc;
	dlg6.PS = pDoc->PS;

	dlg6.DoModal();
	// do all the copy-back inside the dialog's 
	// OnOK() function
	InvalidateRect( ViewSize, TRUE );
	UpdateWindow();
}
